/* Copyright (C) 2000 MySQL AB, portions Copyright (C) 2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
General Public License version 2.0 for more details.

You should have received a copy of the GNU General Public License
version 2.0 along with this program; if not, write to the Free
Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA */


#ifdef __WIN__
#define __PRETTY_FUNCTION__   __FUNCTION__
#endif


#include "shared/handler/BrighthouseHandler.h"
#include "shared/core/RCTableImpl.h"

char infobright_version[] = "Infobright Community Edition 3.3.1";

handler *rcbase_create_handler(handlerton *hton, TABLE_SHARE *table, MEM_ROOT *mem_root);
int rcbase_panic_func(handlerton *hton, enum ha_panic_function flag);
int rcbase_close_connection(handlerton *hton, THD* thd);
int rcbase_commit(handlerton *hton, THD *thd, mysql_bool all);
int rcbase_rollback(handlerton *hton, THD *thd, mysql_bool all);
uchar* rcbase_get_key(RCBASE_SHARE *share, size_t *length, my_bool not_used __attribute__((unused)));

extern const char *ha_rcbase_exts[];
extern my_bool infobright_bootstrap;
extern HASH rcbase_open_tables; // Hash used to track open tables

/*
 ::info() is used to return information to the optimizer.
 see my_base.h for the complete description

 Currently this table handler doesn't implement most of the fields
 really needed. SHOW also makes use of this data
 Another note, you will probably want to have the following in your
 code:
 if (records < 2)
 records = 2;
 The reason is that the server will optimize for cases of only a single
 record. If in a table scan you don't know the number of records
 it will probably be better to set records to two so you can return
 as many records as you need.
 Along with records a few more variables you may wish to set are:
 records
 deleted
 data_file_length
 index_file_length
 delete_length
 check_time
 Take a look at the public variables in handler.h for more information.

 Called in:
 filesort.cc
 ha_heap.cc
 item_sum.cc
 opt_sum.cc
 sql_delete.cc
 sql_delete.cc
 sql_derived.cc
 sql_select.cc
 sql_select.cc
 sql_select.cc
 sql_select.cc
 sql_select.cc
 sql_show.cc
 sql_show.cc
 sql_show.cc
 sql_show.cc
 sql_table.cc
 sql_union.cc
 sql_update.cc

 */

int rcbase_init_func(void *p)
{
	DBUG_ENTER(__PRETTY_FUNCTION__);
	rcbase_hton= (handlerton *)p;
	//VOID(pthread_mutex_init(&rcbase_mutex, MY_MUTEX_INIT_FAST));
	(void) hash_init(&rcbase_open_tables, system_charset_info, 32, 0, 0, (hash_get_key) rcbase_get_key, 0, 0);

	rcbase_hton->state = SHOW_OPTION_YES;
	rcbase_hton->db_type = DB_TYPE_BRIGHTHOUSE_DB;
	rcbase_hton->create = rcbase_create_handler;
	rcbase_hton->flags = HTON_NO_FLAGS;
	rcbase_hton->flags |= HTON_ALTER_NOT_SUPPORTED;
	rcbase_hton->panic = rcbase_panic_func;
	rcbase_hton->close_connection = rcbase_close_connection;
	rcbase_hton->commit = rcbase_commit;
	rcbase_hton->rollback = rcbase_rollback;

	// When mysqld runs as bootstrap mode, we do not need to initialize memmanager.
	if (infobright_bootstrap)
		DBUG_RETURN(0);

	int ret = 1;
	rceng = NULL;

	try {
		rceng = new RCEngine();
		ret = rceng->Init(ha_rcbase_exts[0], total_ha);
	} catch (std::exception& e) {
		my_message(BHERROR_UNKNOWN, e.what(), MYF(0));
		rclog << lock << "Error: " << e.what() << unlock;
	} catch (...) {
		my_message(BHERROR_UNKNOWN, "An unknown system exception error caught.", MYF(0));
		rclog << lock << "Error: An unknown system exception error caught." << unlock;
	}

	//FIX: Can not be deleted right now. Some internal variables need to
	// deleted properly before deleting rceng.
	//if (ret ==1 && rceng != NULL)
	//	delete rceng;

	DBUG_RETURN(ret);
}



int BrighthouseHandler::info(uint flag)
{
	DBUG_ENTER(__PRETTY_FUNCTION__);

	int ret = 1;
	try {
		if(m_table_path) {
			if(flag & HA_STATUS_VARIABLE) {
				if(!table_lock_manager.HasAnyLockForTable(*table->in_use, m_table_path)
					&& table_share->tmp_table!=TRANSACTIONAL_TMP_TABLE) {
					std::auto_ptr<RCTable> tab(rceng->GetCurrentTableState(m_table_path, table_share));
					stats.records = tab->NoValues();
					stats.data_file_length = 0;
					stats.mean_rec_length = 0;
					if(stats.records > 0) {
						AttrInfo* attr_info = 0;
						uint no_attrs = 0;
						rceng->GetTableAttributes(*tab, attr_info, no_attrs);
						for(uint j=0; j < no_attrs; j++)
							stats.data_file_length += attr_info[j].comp_size; // compressed size
						delete [] attr_info;
						stats.mean_rec_length = stats.data_file_length / stats.records;
					}
				} else {

					stats.records = (ha_rows)rceng->GetTableNoObj(m_table_path, table_share);
					stats.data_file_length = 0;
					stats.mean_rec_length = 0;
					if(stats.records > 0) {
						AttrInfo* attr_info;
						uint no_attrs;
						rceng->GetTableAttributes(m_table_path, table_share, attr_info, no_attrs);
						for(uint j=0; j < no_attrs; j++)
							stats.data_file_length += attr_info[j].comp_size; // compressed size
						delete[] attr_info;
						stats.mean_rec_length = stats.data_file_length / stats.records;
					}
				}
			}
			if(flag & HA_STATUS_CONST)
				stats.create_time = rceng->GetTableCreateTime(m_table_path, table_share);
			if(flag & HA_STATUS_TIME)
				stats.update_time = rceng->GetTableUpdateTime(m_table_path, table_share);
		}
		ret = 0;
	} catch (std::exception& e) {
		my_message(BHERROR_UNKNOWN, e.what(), MYF(0));
		rclog << lock << "Error: " << e.what() << unlock;
	} catch (...) {
		my_message(BHERROR_UNKNOWN, "An unknown system exception error caught.", MYF(0));
		rclog << lock << "Error: An unknown system exception error caught." << unlock;
	}

	DBUG_RETURN(ret);
}

/*
 write_row() inserts a row. No extra() hint is given currently if a bulk load
 is happeneding. buf() is a byte array of data. You can use the field
 information to extract the data from the native byte array type.
 Example of this would be:
 for (Field **field=table->field ; *field ; field++)
 {
 ...
 }

 See ha_tina.cc for an example of extracting all of the data as strings.
 ha_berekly.cc has an example of how to store it intact by "packing" it
 for ha_berkeley's own native storage type.

 See the note for update_row() on auto_increments and timestamps. This
 case also applied to write_row().

 Called from item_sum.cc, item_sum.cc, sql_acl.cc, sql_insert.cc,
 sql_insert.cc, sql_select.cc, sql_table.cc, sql_udf.cc, and sql_update.cc.
 */
int BrighthouseHandler::write_row(uchar* buf)
{
	DBUG_ENTER(__PRETTY_FUNCTION__);
	DBUG_RETURN(HA_ERR_WRONG_COMMAND);
}

/*
 Yes, update_row() does what you expect, it updates a row. old_data will have
 the previous row record in it, while new_data will have the newest data in
 it.
 Keep in mind that the server can do updates based on ordering if an ORDER BY
 clause was used. Consecutive ordering is not guarenteed.
 Currently new_data will not have an updated auto_increament record, or
 and updated timestamp field. You can do these for example by doing these:
 if (table->timestamp_field_type & TIMESTAMP_AUTO_SET_ON_UPDATE)
 table->timestamp_field->set_time();
 if (table->next_number_field && record == table->record[0])
 update_auto_increment();

 Called from sql_select.cc, sql_acl.cc, sql_update.cc, and sql_insert.cc.
 */
int BrighthouseHandler::update_row(const uchar* old_data, uchar* new_data)
{
	DBUG_ENTER(__PRETTY_FUNCTION__);
	DBUG_RETURN(HA_ERR_WRONG_COMMAND);
}

/*
 This will delete a row. buf will contain a copy of the row to be deleted.
 The server will call this right after the current row has been called (from
 either a previous rnd_nexT() or index call).
 If you keep a pointer to the last row or can access a primary key it will
 make doing the deletion quite a bit easier.
 Keep in mind that the server does no guarentee consecutive deletions. ORDER BY
 clauses can be used.

 Called in sql_acl.cc and sql_udf.cc to manage internal table information.
 Called in sql_delete.cc, sql_insert.cc, and sql_select.cc. In sql_select it is
 used for removing duplicates while in insert it is used for REPLACE calls.
 */
int BrighthouseHandler::delete_row(const uchar * buf)
{
	DBUG_ENTER(__PRETTY_FUNCTION__);
	DBUG_RETURN(HA_ERR_WRONG_COMMAND);
}

/*
 Used to delete all rows in a table. Both for cases of truncate and
 for cases where the optimizer realizes that all rows will be
 removed as a result of a SQL statement.

 Called from item_sum.cc by Item_func_group_concat::clear(),
 Item_sum_count_distinct::clear(), and Item_func_group_concat::clear().
 Called from sql_delete.cc by mysql_delete().
 Called from sql_select.cc by JOIN::rein*it.
 Called from sql_union.cc by st_select_lex_unit::exec().
 */
int BrighthouseHandler::delete_all_rows()
{
	DBUG_ENTER(__PRETTY_FUNCTION__);
	DBUG_RETURN(HA_ERR_WRONG_COMMAND);
}

/*
 Renames a table from one name to another from alter table call.

 If you do not implement this, the default rename_table() is called from
 handler.cc and it will delete all files with the file extentions returned
 by bas_ext().

 Called from sql_table.cc by mysql_rename_table().
 */
int BrighthouseHandler::rename_table(const char * from, const char * to)
{
	return 0;
}

/*
  Sets brighthouse alter table flag.

  Called from mysql_alter_table (sql_table.cc)
*/
void BrighthouseHandler::set_brighthouse_alter_table(int& brighthouse_alter_table)
{
}

/*
  Alters table except for rename table.

  Called from mysql_alter_table (sql_table.cc)
*/
int BrighthouseHandler::brighthouse_alter_table(int drop_last_column)
{
	return FALSE;
}
