/* Copyright (C)  2005-2009 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef AGGREGATOR_ADVANCED_H_
#define AGGREGATOR_ADVANCED_H_

#include "Aggregator.h"

/*! \brief A generalization of aggregation algorithms (counters) to be used in GROUP BY etc.
 * See "Aggregator.h" for more details.
 *
 * \note The aim of this class hierarchy is to make aggregators more pluggable,
 * as well as to replace multiple ifs and switch/cases by virtual methods.
 * Therefore it is suggested to implement all variants as separate subclasses,
 * e.g. to distinguish 32- and 64-bit counters.
 */

///////////////////////////////////////////////////////////////////////////////////////////////
//! Abstract class for all "second central moment"-like statistical operations,
//  i.e. VAR_POP, VAR_SAMP, STD_POP, STD_SAMP
class AggregatorStat : public Aggregator
{
public:
	// Buffer contents for all functions, 8+8+8 bytes:
	//	<no_obj> <sum> <sum_squares>  - int64, double, double
	virtual _int64 BufferByteSize() 				{ return 24; }
	virtual void Reset(unsigned char *buf)			{ *((_int64*)buf) = 0; *((double*)(buf + 8)) = 0; *((double*)(buf + 16)) = 0; }

	///////////// Optimization part /////////////////
	// we need sum of squares to aggregate the whole pack

protected:
	// Note: it is a subclass task to make sure NoObj > 0 (or 1 for sample version), otherwise NULL
	_int64 NoObj(unsigned char *buf)				{ return *((_int64*)buf); }
	double Sum(unsigned char *buf)					{ return *((double*)(buf + 8)); }
	double SumSq(unsigned char *buf)				{ return *((double*)(buf + 16)); }
	double VarPop(unsigned char *buf)
	{
		double res = SumSq(buf) / NoObj(buf) - Sum(buf) * Sum(buf) / NoObj(buf) / double(NoObj(buf));
		if(res <= DBL_MIN || NoObj(buf) == 1)
			return 0;
		return res;
	}
	double VarSamp(unsigned char *buf)
	{
		double res = SumSq(buf) / (NoObj(buf) - 1) - Sum(buf) * Sum(buf) / NoObj(buf) / double(NoObj(buf) - 1);
		if(res <= DBL_MIN)
			return 0;
		return res;
	}
};

///////////////////////////////////////////////////////////////////////////////////////////////

class AggregatorVarPop64 : public AggregatorStat
{
public:
	AggregatorVarPop64(int precision)
	{ prec_factor = PowOfTen(precision); }
	virtual void PutAggregatedValue(unsigned char *buf, _int64 v, _int64 factor);
	virtual _int64 GetValue64(unsigned char *buf);

private:
	double prec_factor;			// precision factor: the values must be divided by it
};

///////////////////////////////////////////////////////////////////////////////////////////////

class AggregatorVarSamp64 : public AggregatorStat
{
public:
	AggregatorVarSamp64(int precision)
	{ prec_factor = PowOfTen(precision); }
	virtual void PutAggregatedValue(unsigned char *buf, _int64 v, _int64 factor);
	virtual _int64 GetValue64(unsigned char *buf);

private:
	double prec_factor;			// precision factor: the values must be divided by it
};
///////////////////////////////////////////////////////////////////////////////////////////////

class AggregatorStdPop64 : public AggregatorStat
{
public:
	AggregatorStdPop64(int precision)
	{ prec_factor = PowOfTen(precision); }
	virtual void PutAggregatedValue(unsigned char *buf, _int64 v, _int64 factor);
	virtual _int64 GetValue64(unsigned char *buf);

private:
	double prec_factor;			// precision factor: the values must be divided by it
};
///////////////////////////////////////////////////////////////////////////////////////////////

class AggregatorStdSamp64 : public AggregatorStat
{
public:
	AggregatorStdSamp64(int precision)
	{ prec_factor = PowOfTen(precision); }
	virtual void PutAggregatedValue(unsigned char *buf, _int64 v, _int64 factor);
	virtual _int64 GetValue64(unsigned char *buf);

private:
	double prec_factor;			// precision factor: the values must be divided by it
};

///////////////////////////////////////////////////////////////////////////////////////////////

class AggregatorVarPopD : public AggregatorStat
{
public:
	virtual void PutAggregatedValue(unsigned char *buf, _int64 v, _int64 factor);
	virtual void PutAggregatedValue(unsigned char *buf, const RCBString& v, _int64 factor);
	virtual _int64 GetValue64(unsigned char *buf);
};

///////////////////////////////////////////////////////////////////////////////////////////////

class AggregatorVarSampD : public AggregatorStat
{
public:
	virtual void PutAggregatedValue(unsigned char *buf, _int64 v, _int64 factor);
	virtual void PutAggregatedValue(unsigned char *buf, const RCBString& v, _int64 factor);
	virtual _int64 GetValue64(unsigned char *buf);
};

///////////////////////////////////////////////////////////////////////////////////////////////

class AggregatorStdPopD : public AggregatorStat
{
public:
	virtual void PutAggregatedValue(unsigned char *buf, _int64 v, _int64 factor);
	virtual void PutAggregatedValue(unsigned char *buf, const RCBString& v, _int64 factor);
	virtual _int64 GetValue64(unsigned char *buf);
};

///////////////////////////////////////////////////////////////////////////////////////////////

class AggregatorStdSampD : public AggregatorStat
{
public:
	virtual void PutAggregatedValue(unsigned char *buf, _int64 v, _int64 factor);
	virtual void PutAggregatedValue(unsigned char *buf, const RCBString& v, _int64 factor);
	virtual _int64 GetValue64(unsigned char *buf);
};

//////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////

class AggregatorBitAnd : public Aggregator
{
public:
	virtual _int64 BufferByteSize() 				{ return 8; }
	virtual void Reset(unsigned char *buf)			{ *((_int64*)buf) = 0xFFFFFFFFFFFFFFFFULL; }
	virtual void PutAggregatedValue(unsigned char *buf, _int64 v, _int64 factor)
	{ *((_int64*)buf) = (*((_int64*)buf) & v); }
	virtual void PutAggregatedValue(unsigned char *buf, const RCBString& v, _int64 factor);
	virtual _int64 GetValue64(unsigned char *buf)	{ return *((_int64*)buf); }

	virtual bool FactorNeeded()						{ return false; }
	virtual bool IgnoreDistinct()					{ return true; }
};

class AggregatorBitOr : public Aggregator
{
public:
	virtual _int64 BufferByteSize() 				{ return 8; }
	virtual void Reset(unsigned char *buf)			{ *((_int64*)buf) = 0; }
	virtual void PutAggregatedValue(unsigned char *buf, _int64 v, _int64 factor)
	{ *((_int64*)buf) = (*((_int64*)buf) | v); }
	virtual void PutAggregatedValue(unsigned char *buf, const RCBString& v, _int64 factor);
	virtual _int64 GetValue64(unsigned char *buf)	{ return *((_int64*)buf); }

	virtual bool FactorNeeded()						{ return false; }
	virtual bool IgnoreDistinct()					{ return true; }
};

class AggregatorBitXor : public Aggregator
{
public:
	virtual _int64 BufferByteSize() 				{ return 8; }
	virtual void Reset(unsigned char *buf)			{ *((_int64*)buf) = 0; }
	virtual void PutAggregatedValue(unsigned char *buf, _int64 v, _int64 factor)
	{ if(factor%2 == 1) *((_int64*)buf) = (*((_int64*)buf) ^ v); }
	virtual void PutAggregatedValue(unsigned char *buf, const RCBString& v, _int64 factor);
	virtual _int64 GetValue64(unsigned char *buf)	{ return *((_int64*)buf); }
};

#endif
