/* Copyright (C)  2005-2009 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "ColumnBinEncoder.h"
#include "vc/VirtualColumn.h"
#include "bintools.h"

ColumnBinEncoder::ColumnBinEncoder(int flags)
{
	ignore_nulls		= ((flags & ENCODER_IGNORE_NULLS) != 0);
	monotonic_encoding	= ((flags & ENCODER_MONOTONIC) != 0);
	descending			= ((flags & ENCODER_DESCENDING) != 0);
	decodable			= ((flags & ENCODER_DECODABLE) != 0);

	disabled = false;

	vc = NULL;
	val_offset = 0;
	val_sec_offset = 0;
	val_size = 0;
	val_sec_size = 0;

	my_encoder.reset();
	my_sec_encoder.reset();
}

ColumnBinEncoder::ColumnBinEncoder(ColumnBinEncoder const &sec)
{
	ignore_nulls = sec.ignore_nulls;
	monotonic_encoding = sec.monotonic_encoding;
	descending = sec.descending;
	decodable = sec.decodable;
	disabled = sec.disabled;
	vc = sec.vc;
	val_offset = sec.val_offset;
	val_sec_offset = sec.val_sec_offset;
	val_size = sec.val_size;
	val_sec_size = sec.val_sec_size;
	my_encoder = sec.my_encoder;
	my_sec_encoder = sec.my_sec_encoder;
}

ColumnBinEncoder& ColumnBinEncoder::operator = (ColumnBinEncoder const &sec)
{
	if(&sec != this) {
		ignore_nulls = sec.ignore_nulls;
		monotonic_encoding = sec.monotonic_encoding;
		descending = sec.descending;
		decodable = sec.decodable;
		disabled = sec.disabled;
		vc = sec.vc;
		val_offset = sec.val_offset;
		val_sec_offset = sec.val_sec_offset;
		val_size = sec.val_size;
		val_sec_size = sec.val_sec_size;
		my_encoder = sec.my_encoder;
		my_sec_encoder = sec.my_sec_encoder;
	}
	return *this;
}

ColumnBinEncoder::~ColumnBinEncoder()
{
}

void ColumnBinEncoder::Clear()
{
	my_encoder.reset();
	my_sec_encoder.reset();
	disabled = false;
}

bool ColumnBinEncoder::PrepareEncoder(VirtualColumn* _vc, VirtualColumn* _vc2)
{
	if(_vc == NULL) 
		return false;
	bool nulls_possible = false;
	if(!ignore_nulls)
		nulls_possible = _vc->NullsPossible() || (_vc2 != NULL && _vc2->NullsPossible());
	vc = _vc;
	ColumnType vct = vc->Type();
	ColumnType vct2 = _vc2 ? _vc2->Type() : ColumnType();
	if(vct.IsFixed() && (!_vc2 || (vct2.IsFixed() && vct.GetScale() == vct2.GetScale()))) {		// int/dec of the same scale
		my_encoder = ColumnValueEncoder_ptr_t(new ColumnBinEncoder::EncoderInt(vc, decodable, nulls_possible, descending));
	} else if(vct.IsFloat() || (_vc2 && vct2.IsFloat())) {
		my_encoder = ColumnValueEncoder_ptr_t(new ColumnBinEncoder::EncoderDouble(vc, decodable, nulls_possible, descending));
	} else if(vct.IsFixed()) {	// Decimals for different scale (the same scale is done by EncoderInt)
		my_encoder = ColumnValueEncoder_ptr_t(new ColumnBinEncoder::EncoderDecimal(vc, decodable, nulls_possible, descending));
	} else if(vct.GetTypeName() == RC_DATE) {
		my_encoder = ColumnValueEncoder_ptr_t(new ColumnBinEncoder::EncoderDate(vc, decodable, nulls_possible, descending));
	} else if(vct.GetTypeName() == RC_YEAR) {
		my_encoder = ColumnValueEncoder_ptr_t(new ColumnBinEncoder::EncoderYear(vc, decodable, nulls_possible, descending));
	} else if(vct.IsString()) {
		if((monotonic_encoding || !decodable)	// (!mon. && dec.) => non-sorted cols in sorter, don't UTF-encode. (!mon. && !dec.) => used in join/distinct => UTF-encode
			&& RequiresUTFConversions(vc->GetCollation()))	
			my_encoder = ColumnValueEncoder_ptr_t(new ColumnBinEncoder::EncoderText_UTF(vc, decodable, nulls_possible, descending));
		else
			my_encoder = ColumnValueEncoder_ptr_t(new ColumnBinEncoder::EncoderText(vc, decodable, nulls_possible, descending));
	// TODO: Lookup
	} else if(vct.IsDateTime()) {		// Date/time types except special cases (above)
			my_encoder = ColumnValueEncoder_ptr_t(new ColumnBinEncoder::EncoderInt(vc, decodable, nulls_possible, descending));
	} else {
		BHASSERT_WITH_NO_PERFORMANCE_IMPACT(!"wrong combination of encoded columns");	// Other types not implemented yet
		my_encoder = ColumnValueEncoder_ptr_t(new ColumnBinEncoder::EncoderText(vc, decodable, nulls_possible, descending));
	}
	if(_vc2 != NULL) {		// multiple column encoding?
		bool encoding_possible = my_encoder->SecondColumn(_vc2);
		if(!encoding_possible)
			return false;
	}
	val_size = my_encoder->ValueSize();
	val_sec_size = my_encoder->ValueSizeSec();
	return true;
}

void ColumnBinEncoder::Encode(uchar* buf, MIIterator& mit, VirtualColumn* alternative_vc, bool update_stats)
{
	my_encoder->Encode(buf + val_offset, buf + val_sec_offset, (alternative_vc ? alternative_vc : vc), mit, update_stats);
}

void ColumnBinEncoder::PutValue64(uchar* buf, _int64 v, bool sec_column, bool update_stats)
{
	my_encoder->Encode(buf + val_offset, buf + val_sec_offset, v, sec_column, update_stats);
}

_int64 ColumnBinEncoder::GetValue64(uchar* buf, bool &is_null)
{
	is_null = false;
	if(my_encoder->IsNull(buf + val_offset, buf + val_sec_offset)) {
		is_null = true;
		return NULL_VALUE_64;
	}
	return my_encoder->GetValue64(buf + val_offset, buf + val_sec_offset);
}

RCBString ColumnBinEncoder::GetValueT(uchar* buf)
{
	if(my_encoder->IsNull(buf + val_offset, buf + val_sec_offset))
		return RCBString();
	return my_encoder->GetValueT(buf + val_offset, buf + val_sec_offset);
}

bool ColumnBinEncoder::ImpossibleValues(_int64 pack_min, _int64 pack_max)		// return true if the current contents of the encoder is out of scope
{
	return my_encoder->ImpossibleValues(pack_min, pack_max);
}

void ColumnBinEncoder::ClearStatistics()
{
	my_encoder->ClearStatistics();
}

////////////////////////////////////////////////////////////////////////

ColumnBinEncoder::EncoderInt::EncoderInt(VirtualColumn* vc, bool decodable, bool nulls_possible, bool _descending)
					: ColumnValueEncoder(vc, decodable, nulls_possible, _descending)
{
	min_val = vc->RoughMin();
	max_code = _uint64(vc->RoughMax() - min_val);
	if(nulls_possible) {
		if(max_code < 0xFFFFFFFFFFFFFFFFULL) {
			null_status = 1;	// 0 is null
			max_code++;
		} else
			null_status = 2;	// separate byte
	} else
		null_status = 0;
	size = CalculateByteSize(max_code) + (null_status == 2 ? 1 : 0);
	size_sec = 0;
	min_found = PLUS_INF_64;
	max_found = MINUS_INF_64;
}

bool ColumnBinEncoder::EncoderInt::SecondColumn(VirtualColumn* vc)
{
	if(!vc->Type().IsFixed() && !(this->vc_type.IsDateTime() && vc->Type().IsDateTime()))
		return false;
	// Easy case: integers/decimals with the same precision
	_int64 new_min_val = vc->RoughMin();
	_int64 max_val = max_code + min_val - (null_status == 1 ? 1 : 0);
	_int64 new_max_val = vc->RoughMax();
	if(min_val > new_min_val)
		min_val = new_min_val;
	if(max_val < new_max_val)
		max_val = new_max_val;
	max_code = _uint64(max_val - min_val);
	if(null_status == 1 && max_code == 0xFFFFFFFFFFFFFFFFULL)
			null_status = 2;	// separate byte
	size = CalculateByteSize(max_code) + (null_status == 2 ? 1 : 0);
	return true;
}

void ColumnBinEncoder::EncoderInt::Encode(uchar* buf, uchar* buf_sec, VirtualColumn* vc, MIIterator& mit, bool update_stats)
{
	if(null_status > 0 && vc->IsNull(mit))
		SetNull(buf, buf_sec);
	else
		Encode(buf, buf_sec, vc->GetValueInt64(mit), false, update_stats);
}

void ColumnBinEncoder::EncoderInt::Encode(uchar* buf, uchar* buf_sec, _int64 v, bool sec_column, bool update_stats)
{
	if(null_status > 0 && v == NULL_VALUE_64) {
		SetNull(buf, buf_sec);
		return;
	}
	_uint64 coded_val;
	int loc_size = size;
	if(update_stats) {
		if(v > max_found)
			max_found = v;
		if(v < min_found)
			min_found = v;
	}
	coded_val = _uint64(v - min_val) + (null_status == 1 ? 1 : 0);
	if(descending)
		coded_val = max_code - coded_val;
	if(null_status == 2) {
		*buf = (descending ? '\0' : '\1');			// not null
		buf++;
		loc_size--;
	}
	uchar* val_ptr = (uchar *)(&coded_val) + loc_size; 	// the end of meaningful bytes
	for(int i = 0; i < loc_size; i++)
		*(buf++) = *(--val_ptr);			// change endianess - to make numbers comparable by memcmp()
}

void ColumnBinEncoder::EncoderInt::SetNull(uchar* buf, uchar* buf_sec)
{
	// Assume three bytes of int
	// mode              |   null    |   not null  |
	// ---------------------------------------------
	// status=1, ascend. |    000    |    xxx      |
	// status=2, ascend. |   0000    |   1xxx      |
	// status=1, descend.|    max    |    xxx      |
	// status=2, descend.|   1000    |   0xxx      |
	// ---------------------------------------------

	assert(null_status > 0);
	memset(buf, 0, size);		// zero on the first and all other bytes (to be changed below)
	if(descending) {
		if(null_status == 1) {
			uchar* val_ptr = (uchar *)(&max_code) + size; 	// the end of meaningful bytes
			for(int i = 0; i < size; i++)
				*(buf++) = *(--val_ptr);			// change endianess - to make numbers comparable by memcmp()
		} else {
			*buf = '\1';
		}
	}
}

bool ColumnBinEncoder::EncoderInt::IsNull(uchar* buf, uchar* buf_sec)
{
	if(null_status == 0)
		return false;
	_uint64 zero = 0;
	if(descending) {
		if(null_status == 1) {
			_uint64 coded_val = 0;
			uchar* val_ptr = (uchar *)(&coded_val) + size; 	// the end of meaningful bytes
			for(int i = 0; i < size; i++)
				*(--val_ptr) = *(buf++);			// change endianess
			return (coded_val == max_code);
		} else
			return (*buf == '\1');
	} else {
		if(null_status == 1)
			return (memcmp(buf, &zero, size) == 0);
		else
			return (*buf == '\0');
	}
}

_int64 ColumnBinEncoder::EncoderInt::GetValue64(uchar* buf, uchar* buf_sec)
{
	int loc_size = size;
	if(null_status == 2) {
		if((!descending && *buf == '\0') ||
			(descending && *buf == '\1'))
			return NULL_VALUE_64;
		buf++;
		loc_size--;
	}
	_uint64 coded_val = 0;
	uchar* val_ptr = (uchar *)(&coded_val) + loc_size; 	// the end of meaningful bytes
	for(int i = 0; i < loc_size; i++)
		*(--val_ptr) = *(buf++);			// change endianess

	if(descending)
		coded_val = max_code - coded_val;
	return coded_val - (null_status == 1 ? 1 : 0) + min_val;
}

bool ColumnBinEncoder::EncoderInt::ImpossibleValues(_int64 pack_min, _int64 pack_max)
{
	if(pack_min == NULL_VALUE_64 || pack_max == NULL_VALUE_64)
		return false;
	if(pack_min > max_found || pack_max < min_found)
		return true;
	return false;
}

void ColumnBinEncoder::EncoderInt::ClearStatistics()
{
	min_found = PLUS_INF_64;
	max_found = MINUS_INF_64;
}

////////////////////////////////////////////////////////////////////////

ColumnBinEncoder::EncoderDecimal::EncoderDecimal(VirtualColumn* vc, bool decodable, bool nulls_possible, bool _descending)
: EncoderInt(vc, decodable, nulls_possible, _descending)
{
	multiplier = 1.0;
	sec_multiplier = 1.0;
	scale = vc->Type().GetScale();
	_int64 pmin = vc->RoughMin();
	_int64 pmax = vc->RoughMax();
	min_val = pmin;
	max_code = _uint64(pmax - pmin);
	if(nulls_possible) {
		if(max_code < 0xFFFFFFFFFFFFFFFFULL) {
			null_status = 1;	// 0 is null
			max_code++;
		} else
			null_status = 2;	// separate byte
	} else
		null_status = 0;
	size = CalculateByteSize(max_code);
	size_sec = 0;
}

bool ColumnBinEncoder::EncoderDecimal::SecondColumn(VirtualColumn* vc)
{
	_int64 max_val = max_code + min_val - (null_status == 1 ? 1 : 0);
	_int64 new_min_val = vc->RoughMin();
	_int64 new_max_val = vc->RoughMax();
	int new_scale = vc->Type().GetScale();
	if(new_scale > scale) {
		multiplier = PowOfTen(new_scale - scale);
		min_val *= multiplier;
		max_val *= multiplier;
	} else if(new_scale < scale) {
		sec_multiplier = PowOfTen(scale - new_scale);
		new_min_val *= sec_multiplier;
		new_max_val *= sec_multiplier;
	}
	scale = max(new_scale, scale);
	if(min_val > new_min_val)
		min_val = new_min_val;
	if(max_val < new_max_val)
		max_val = new_max_val;
	max_code = _uint64(max_val - min_val);
	if(null_status == 1 && max_code == 0xFFFFFFFFFFFFFFFFULL)
		null_status = 2;	// separate byte
	size = CalculateByteSize(max_code) + (null_status == 2 ? 1 : 0);
	return true;
}

void ColumnBinEncoder::EncoderDecimal::Encode(uchar* buf, uchar* buf_sec, VirtualColumn* vc, MIIterator &mit, bool update_stats /*= false*/)
{
	if(null_status > 0 && vc->IsNull(mit))
		SetNull(buf, buf_sec);
	else {
		_int64 v = vc->GetValueInt64(mit);
		if(vc->Type().GetScale() < scale) 
			v *= PowOfTen(scale - vc->Type().GetScale());
		EncoderInt::Encode(buf, buf_sec, v, false, update_stats);
	}
}

void ColumnBinEncoder::EncoderDecimal::Encode(uchar* buf, uchar* buf_sec, _int64 v, bool sec_column, bool update_stats)
{
	if(null_status > 0 && v == NULL_VALUE_64)
		SetNull(buf, buf_sec);
	else {
		if(sec_column)
			v *= sec_multiplier;
		else
			v *= multiplier;
		EncoderInt::Encode(buf, buf_sec, v, false, update_stats);
	}
}

bool ColumnBinEncoder::EncoderDecimal::ImpossibleValues(_int64 pack_min, _int64 pack_max)
{
	if(pack_min == NULL_VALUE_64 || pack_max == NULL_VALUE_64)
		return false;
	pack_min *= sec_multiplier;			// assuming that pack_min, pack_max always belong to the second column!
	pack_max *= sec_multiplier;
	if(pack_min > max_found || pack_max < min_found)
		return true;
	return false;
}

////////////////////////////////////////////////////////////////////////

ColumnBinEncoder::EncoderDate::EncoderDate(VirtualColumn* vc, bool decodable, bool nulls_possible, bool _descending)
						: EncoderInt(vc, decodable, nulls_possible, _descending)
{
	_int64 pmin = DateSortEncoding(vc->RoughMin());
	_int64 pmax = DateSortEncoding(vc->RoughMax());
	min_val = pmin;
	max_code = _uint64(pmax - pmin);
	if(nulls_possible) {
		null_status = 1;	// 0 is null - because dates never reach max. int.
		max_code++;
	} else
		null_status = 0;
	size = CalculateByteSize(max_code);
	size_sec = 0;
}

bool ColumnBinEncoder::EncoderDate::SecondColumn(VirtualColumn* vc)
{
	// Possible conversions: only dates.
	if(vc->Type().GetTypeName() != RC_DATE)
		return false;
	_int64 new_min_val = DateSortEncoding(vc->RoughMin());
	_int64 max_val = max_code + min_val - (null_status == 1 ? 1 : 0);
	_int64 new_max_val = DateSortEncoding(vc->RoughMax());
	if(min_val > new_min_val)
		min_val = new_min_val;
	if(max_val < new_max_val)
		max_val = new_max_val;
	max_code = _uint64(max_val - min_val) + (null_status == 1 ? 1 : 0);;
	size = CalculateByteSize(max_code);
	return true;
}

void ColumnBinEncoder::EncoderDate::Encode(uchar* buf, uchar* buf_sec, VirtualColumn* vc, MIIterator &mit, bool update_stats)
{
	if(null_status > 0 && vc->IsNull(mit))
		SetNull(buf, buf_sec);
	else
		EncoderInt::Encode(buf, buf_sec, DateSortEncoding(vc->GetValueInt64(mit)), false, update_stats);
}

void ColumnBinEncoder::EncoderDate::Encode(uchar* buf, uchar* buf_sec, _int64 v, bool sec_column, bool update_stats)
{
	if(null_status > 0 && v == NULL_VALUE_64)
		SetNull(buf, buf_sec);
	else
		EncoderInt::Encode(buf, buf_sec, DateSortEncoding(v), false, update_stats);
}

_int64 ColumnBinEncoder::EncoderDate::GetValue64(uchar* buf, uchar* buf_sec)
{
	if(IsNull(buf, buf_sec))
		return NULL_VALUE_64;
	return DateSortDecoding(EncoderInt::GetValue64(buf, buf_sec));
}

bool ColumnBinEncoder::EncoderDate::ImpossibleValues(_int64 pack_min, _int64 pack_max)
{
	if(pack_min == NULL_VALUE_64 || pack_max == NULL_VALUE_64)
		return false;
	if(DateSortEncoding(pack_min) > max_found || DateSortEncoding(pack_max) < min_found)
		return true;
	return false;
}

void ColumnBinEncoder::EncoderDate::ClearStatistics()
{
	min_found = PLUS_INF_64;
	max_found = MINUS_INF_64;
}

////////////////////////////////////////////////////////////////////////

ColumnBinEncoder::EncoderYear::EncoderYear(VirtualColumn* vc, bool decodable, bool nulls_possible, bool _descending)
						: EncoderInt(vc, decodable, nulls_possible, _descending)
{
	_int64 pmin = YearSortEncoding(vc->RoughMin());
	_int64 pmax = YearSortEncoding(vc->RoughMax());
	min_val = pmin;
	max_code = _uint64(pmax - pmin);
	if(nulls_possible) {
		null_status = 1;	// 0 is null - because years never reach max. int.
		max_code++;
	} else
		null_status = 0;
	size = CalculateByteSize(max_code);
	size_sec = 0;
}

bool ColumnBinEncoder::EncoderYear::SecondColumn(VirtualColumn* vc)
{
	// Possible conversions: only years.
	if(vc->Type().GetTypeName() != RC_YEAR)
		return false;

	_int64 new_min_val = YearSortEncoding(vc->RoughMin());
	_int64 max_val = max_code + min_val - (null_status == 1 ? 1 : 0);
	_int64 new_max_val = YearSortEncoding(vc->RoughMax());
	if(min_val > new_min_val)
		min_val = new_min_val;
	if(max_val < new_max_val)
		max_val = new_max_val;
	max_code = _uint64(max_val - min_val) + (null_status == 1 ? 1 : 0);;
	size = CalculateByteSize(max_code);
	return true;
}

void ColumnBinEncoder::EncoderYear::Encode(uchar* buf, uchar* buf_sec, VirtualColumn* vc, MIIterator &mit, bool update_stats)
{
	if(null_status > 0 && vc->IsNull(mit))
		SetNull(buf, buf_sec);
	else
		EncoderInt::Encode(buf, buf_sec, YearSortEncoding(vc->GetValueInt64(mit)), false, update_stats);
}

void ColumnBinEncoder::EncoderYear::Encode(uchar* buf, uchar* buf_sec, _int64 v, bool sec_column, bool update_stats)
{
	if(null_status > 0 && v == NULL_VALUE_64)
		SetNull(buf, buf_sec);
	else
		EncoderInt::Encode(buf, buf_sec, YearSortEncoding(v), false, update_stats);
}

_int64 ColumnBinEncoder::EncoderYear::GetValue64(uchar* buf, uchar* buf_sec)
{
	if(IsNull(buf, buf_sec))
		return NULL_VALUE_64;
	return YearSortDecoding(EncoderInt::GetValue64(buf, buf_sec));
}

bool ColumnBinEncoder::EncoderYear::ImpossibleValues(_int64 pack_min, _int64 pack_max)
{
	if(pack_min == NULL_VALUE_64 || pack_max == NULL_VALUE_64)
		return false;
	if(YearSortEncoding(pack_min) > max_found || YearSortEncoding(pack_max) < min_found)
		return true;
	return false;
}

void ColumnBinEncoder::EncoderYear::ClearStatistics()
{
	min_found = PLUS_INF_64;
	max_found = MINUS_INF_64;
}

////////////////////////////////////////////////////////////////////////

ColumnBinEncoder::EncoderDouble::EncoderDouble(VirtualColumn* vc, bool decodable, bool nulls_possible, bool _descending)
							: ColumnValueEncoder(vc, decodable, nulls_possible, _descending)
{
	multiplier_vc1 = 0;	// 0 if the first vc is double, otherwise it is regarded as int (decimal) which must be multiplied
	multiplier_vc2 = 0;	// 0 if the second vc is double, otherwise it is regarded as int (decimal) which must be multiplied
	if(vc->Type().IsFixed())
		multiplier_vc1 = PowOfTen(vc->Type().GetScale());		// encode int/dec as double
	if(nulls_possible) {
		null_status = 2;
		size = 9;
	} else {
		null_status = 0;
		size = 8;
	}
}

bool ColumnBinEncoder::EncoderDouble::SecondColumn(VirtualColumn* vc)
{
	// Possible conversions: all numericals.
	if(!vc->Type().IsFixed() && !vc->Type().IsFloat())
		return false;
	if(vc->Type().IsFixed())
		multiplier_vc2 = PowOfTen(vc->Type().GetScale());		// encode int/dec as double
	return true;
}

void ColumnBinEncoder::EncoderDouble::Encode(uchar* buf, uchar* buf_sec, VirtualColumn* vc, MIIterator& mit, bool update_stats)
{
	if(null_status == 2) {
		if(vc->IsNull(mit)) {
			SetNull(buf, buf_sec);
			return;
		}
		*buf = (descending ? '\0' : '\1');			// not null
		buf++;
	}
	double val = vc->GetValueDouble(mit);			// note that non-float columns will be properly converted here
	_int64 coded_val = MonotonicDouble2Int64(*(_int64*)&val);
	if(descending)
		Negate((uchar *)&coded_val, 8);
	uchar* val_ptr = (uchar *)(&coded_val) + 8; 	// the end of meaningful bytes
	for(int i = 0; i < 8; i++)
		*(buf++) = *(--val_ptr);			// change endianess - to make numbers comparable by memcmp()
}

void ColumnBinEncoder::EncoderDouble::Encode(uchar* buf, uchar* buf_sec, _int64 v, bool sec_column, bool update_stats)
{
	if(null_status == 2) {
		if(v == NULL_VALUE_64) {
			SetNull(buf, buf_sec);
			return;
		}
		*buf = (descending ? '\0' : '\1');			// not null
		buf++;
	}
	_int64 local_mult = (sec_column ? multiplier_vc2 : multiplier_vc1);
	_int64 coded_val;
	if(local_mult == 0)
		coded_val = MonotonicDouble2Int64(v);
	else {
		double d = double(v) / local_mult;			// decimal encoded as double
		coded_val = MonotonicDouble2Int64(*((_int64*)(&d)));
	}
	if(descending)
		Negate((uchar *)&coded_val, 8);
	uchar* val_ptr = (uchar *)(&coded_val) + 8; 	// the end of meaningful bytes
	for(int i = 0; i < 8; i++)
		*(buf++) = *(--val_ptr);			// change endianess - to make numbers comparable by memcmp()
}

void ColumnBinEncoder::EncoderDouble::SetNull(uchar* buf, uchar* buf_sec)
{
	// mode              |   null    |   not null  |
	// ---------------------------------------------
	// status=2, ascend. |   00..00  |   1<64bit>  |
	// status=2, descend.|   10..00  |   0<64bit>  |
	// ---------------------------------------------
	assert(null_status == 2);
	memset(buf, 0, 9);
	if(descending)
		*buf = '\1';
}

bool ColumnBinEncoder::EncoderDouble::IsNull(uchar* buf, uchar* buf_sec)
{
	if(null_status != 2)
		return false;
	if(descending)
		return (*buf == '\1');
	return (*buf == '\0');
}

_int64 ColumnBinEncoder::EncoderDouble::GetValue64(uchar* buf, uchar* buf_sec)
{
	if(null_status == 2) {
		if((!descending && *buf == '\0') ||
			(descending && *buf == '\1'))
			return NULL_VALUE_64;
		buf++;
	}
	_uint64 coded_val = 0;
	uchar* val_ptr = (uchar *)(&coded_val) + 8; 	// the end of meaningful bytes
	for(int i = 0; i < 8; i++)
		*(--val_ptr) = *(buf++);			// change endianess
	if(descending)
		Negate((uchar *)&coded_val, 8);
	return MonotonicInt642Double(coded_val);
}

////////////////////////////////////////////////////////////////////////

ColumnBinEncoder::EncoderText::EncoderText(VirtualColumn* vc, bool decodable, bool nulls_possible, bool _descending)
						: ColumnValueEncoder(vc, decodable, nulls_possible, _descending)
{
	size = vc->MaxStringSize() + 2;						// 2 bytes for len
	size_sec = 0;
}

ColumnBinEncoder::EncoderText::~EncoderText()
{}

bool ColumnBinEncoder::EncoderText::SecondColumn(VirtualColumn* vc)
{
	size = max(size, vc->MaxStringSize() + 2);						// 2 bytes for len
	return true;
}

void ColumnBinEncoder::EncoderText::Encode(uchar* buf, uchar* buf_sec, VirtualColumn* vc, MIIterator& mit, bool update_stats)
{
	if(vc->IsNull(mit)) {
		SetNull(buf, buf_sec);
		return;
	}
	memset(buf, 0, size);
	RCBString s;
	vc->GetValueString(s, mit, false);
	if(s.len > 0)
		memcpy(buf, s.GetDataBytesPointer(), s.len);
	buf[size - 2] = (s.len + 1) / 256;
	buf[size - 1] = (s.len + 1) % 256;
	if(descending)
		Negate(buf, size);
}

void ColumnBinEncoder::EncoderText::SetNull(uchar* buf, uchar* buf_sec)
{
	if(descending)
		memset(buf, 255, size);
	else
		memset(buf, 0, size);
}

bool ColumnBinEncoder::EncoderText::IsNull(uchar* buf, uchar* buf_sec)
{
	if(descending)
		return (buf[size - 2] == 255 && buf[size - 1] == 255);
	return (buf[size - 2] == 0 && buf[size - 1] == 0);
}

RCBString ColumnBinEncoder::EncoderText::GetValueT(uchar* buf, uchar* buf_sec)
{
	if(IsNull(buf, buf_sec))
		return RCBString();
	if(descending)
		Negate(buf, size);
	int len = int(buf[size - 2]) * 256 + int(buf[size - 1]) - 1;
	return RCBString((char *)buf, len, false, (len == 0));			// the RCBString is generated as temporary
}

////////////////////////////////////////////////////////////////////////

ColumnBinEncoder::EncoderText_UTF::EncoderText_UTF(VirtualColumn* vc, bool decodable, bool nulls_possible, bool _descending)
: ColumnValueEncoder(vc, decodable, nulls_possible, _descending)
{
	collation = vc->GetCollation();
	int coded_len = CollationBufLen(collation,  vc->MaxStringSize());
	size = coded_len + 2;						// 2 bytes for len
	size_sec = 0;
	if(decodable) {
		size_sec = vc->MaxStringSize() + 2;		// just a raw data plus len
	}
}

ColumnBinEncoder::EncoderText_UTF::~EncoderText_UTF()
{}

bool ColumnBinEncoder::EncoderText_UTF::SecondColumn(VirtualColumn* vc)
{
	if(collation.collation != (vc->GetCollation()).collation)
		return false;
	int coded_len = CollationBufLen(collation, vc->MaxStringSize());
	size = max(size, coded_len + 2);						// 2 bytes for len
	if(size_sec > 0)
		size_sec = max(size_sec, vc->MaxStringSize() + 2);						// 2 bytes for len
	return true;
}

void ColumnBinEncoder::EncoderText_UTF::Encode(uchar* buf, uchar* buf_sec, VirtualColumn* vc, MIIterator& mit, bool update_stats)
{
	if(vc->IsNull(mit)) {
		SetNull(buf, buf_sec);
		return;
	}
	memset(buf, 0, size);
	RCBString s;
	vc->GetValueString(s, mit, false);
	collation.collation->coll->strnxfrm(collation.collation, buf, size - 2, (uchar*)s.Value(), s.len);
	int coded_len = CollationBufLen(collation, s.len);
	buf[size - 2] = (coded_len + 1) / 256;
	buf[size - 1] = (coded_len + 1) % 256;
	if(descending)
		Negate(buf, size);
	if(size_sec > 0) {
		memset(buf_sec, 0, size_sec);
		buf_sec[size_sec - 2] = (s.len + 1) / 256;
		buf_sec[size_sec - 1] = (s.len + 1) % 256;
		if(s.len > 0)
			memcpy(buf_sec, s.GetDataBytesPointer(), s.len);
	}
}

void ColumnBinEncoder::EncoderText_UTF::SetNull(uchar* buf, uchar* buf_sec)
{
	if(descending)
		memset(buf, 255, size);
	else
		memset(buf, 0, size);
	if(size_sec > 0)
		memset(buf_sec, 0, size_sec);
}

bool ColumnBinEncoder::EncoderText_UTF::IsNull(uchar* buf, uchar* buf_sec)
{
	assert(size_sec > 0);
	return (buf_sec[size_sec - 2] == 0 && buf_sec[size_sec - 1] == 0);
}

RCBString ColumnBinEncoder::EncoderText_UTF::GetValueT(uchar* buf, uchar* buf_sec)
{
	assert(size_sec > 0);
	if(IsNull(buf, buf_sec))
		return RCBString();
	int len = int(buf_sec[size_sec - 2]) * 256 + int(buf_sec[size_sec - 1]) - 1;
	return RCBString((char *)buf_sec, len, false, (len == 0));			// the RCBString is generated as temporary
}
