/* Copyright (C)  2005-2009 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */


#ifndef COLUMNBINENCODER_H_
#define COLUMNBINENCODER_H_


#include "vc/VirtualColumn.h"

class MIIterator;


class ColumnBinEncoder
{
public:
	static const int ENCODER_IGNORE_NULLS	= 0x01;
	static const int ENCODER_MONOTONIC		= 0x02;
	static const int ENCODER_DESCENDING		= 0x04;
	static const int ENCODER_DECODABLE		= 0x08;

	class ColumnValueEncoder;
	// Special cases:
	class EncoderInt;
	class EncoderDecimal;
	class EncoderDate;
	class EncoderYear;
	class EncoderDouble;
	class EncoderText;
	class EncoderText_UTF;

	/////////////////////////////////////////////////
	ColumnBinEncoder(int flags = 0);
	ColumnBinEncoder(ColumnBinEncoder const &sec);
	ColumnBinEncoder& operator=(ColumnBinEncoder const&);
	~ColumnBinEncoder();

	void Clear();								// clear all encoding data, keep flags
	bool PrepareEncoder(VirtualColumn *vc1, VirtualColumn *vc2 = NULL);		// encoder for one column, or a common encoder for two of them
	// return false if encoding of a second column is not possible (incompatible)

	void Disable()								{ disabled = true; }
	bool IsEnabled()							{ return !disabled; }	// disabled if set manually or trivial value for sorting
	bool IsNontrivial()							{ return (val_size > 0); }

	// Buffer descriptions
	int GetPrimarySize()						// no. of bytes in the primary buffer
	{ return val_size; }
	int GetSecondarySize()						// no. of bytes in the secondary buffer (or 0 if not needed)
	{ return val_sec_size; }

	void SetPrimaryOffset(int _offset)			// offset of the primary storage in buffer
	{ val_offset = _offset; }
	void SetSecondaryOffset(int _offset)			// offset of the primary storage in buffer
	{ val_sec_offset = _offset; }

	// Set / retrieve values
	void LockSourcePacks(MIIterator &mit)
	{ if(vc) vc->LockSourcePacks(mit); }

	void Encode(unsigned char *buf, MIIterator& mit, VirtualColumn *alternative_vc = NULL, bool update_stats = false);
	void PutValue64(unsigned char *buf, _int64 v, bool sec_column, bool update_stats = false);

	_int64 GetValue64(unsigned char *buf, bool &is_null);
	RCBString GetValueT(unsigned char *buf);

	bool ImpossibleValues(_int64 pack_min, _int64 pack_max);		// return true if the current contents of the encoder is out of scope
	void ClearStatistics();

private:
	VirtualColumn *vc;

	bool ignore_nulls;
	bool monotonic_encoding;
	bool descending;
	bool decodable;

	bool disabled;				// if true, then the column will not be encoded

	// Address part:
	int	val_offset;		// buffer offset of the value - externally set
	int	val_sec_offset;	// secondary buffer, if needed - externally set
	int	val_size;		// number of bytes in the main part
	int	val_sec_size;	// number of bytes in the secondary part, if needed

	typedef std::auto_ptr<ColumnValueEncoder> ColumnValueEncoder_ptr_t;
	mutable ColumnValueEncoder_ptr_t my_encoder;
	mutable ColumnValueEncoder_ptr_t my_sec_encoder;
};

/////////////////////////////////////////////////////////////////////////

class ColumnBinEncoder::ColumnValueEncoder			// base class for detailed encoders
{
public:
	ColumnValueEncoder(VirtualColumn *vc, bool decodable, bool nulls_possible, bool _descending)
						: descending(_descending), null_status(0), vc_type(vc->Type()), size(0), size_sec(0) {}
	virtual ~ColumnValueEncoder() {}

	// Encode the second column to make common encoding
	virtual bool SecondColumn(VirtualColumn *vc) 			{ return false; }

	virtual void Encode(unsigned char *buf, unsigned char *buf_sec, VirtualColumn *vc, MIIterator& mit, bool update_stats = false) = 0;
	virtual void SetNull(unsigned char *buf, unsigned char *buf_sec) = 0;
	virtual void Encode(unsigned char *buf, unsigned char *buf_sec, _int64 v, bool sec_column, bool update_stats)		{ assert(0); }	// used in special cases only

	virtual bool IsNull(unsigned char *buf, unsigned char *buf_sec) = 0;
	virtual _int64 GetValue64(unsigned char *buf, unsigned char *buf_sec)			{ assert(0); return NULL_VALUE_64; }
	virtual RCBString GetValueT(unsigned char *buf, unsigned char *buf_sec)			{ assert(0); return RCBString(); }
	virtual bool ImpossibleValues(_int64 pack_min, _int64 pack_max)					{ return false; }		// default: all values are possible
	virtual void ClearStatistics()													{}

	int ValueSize()											{ return size; }
	int ValueSizeSec()										{ return size_sec; }

protected:
	bool descending;		// true if descending
	int	null_status;		// 0 - no nulls, 1 - integer shifted (encoded as 0), 2 - separate byte
	ColumnType vc_type;		// a type of the first encoded virtual column
	int size;				// total size in bytes
	int size_sec;			// total size in bytes for a secondary location, or 0

	void Negate(unsigned char *buf, int loc_size)
	{ for(int i = 0; i < loc_size; i++) buf[i] = ~(buf[i]); }
};
////////////////////////////////////////////////////////////////////////

class ColumnBinEncoder::EncoderInt : public ColumnBinEncoder::ColumnValueEncoder
{
public:
	EncoderInt(VirtualColumn *vc, bool decodable, bool nulls_possible, bool _descending);

	virtual bool SecondColumn(VirtualColumn *vc);

	virtual void Encode(unsigned char *buf, unsigned char *buf_sec, VirtualColumn *vc, MIIterator& mit, bool update_stats = false);
	virtual void Encode(unsigned char *buf, unsigned char *buf_sec, _int64 v, bool sec_column, bool update_stats);
	virtual void SetNull(unsigned char *buf, unsigned char *buf_sec);
	virtual bool IsNull(unsigned char *buf, unsigned char *buf_sec);
	virtual _int64 GetValue64(unsigned char *buf, unsigned char *buf_sec);

	virtual bool ImpossibleValues(_int64 pack_min, _int64 pack_max);
	virtual void ClearStatistics();

protected:
	_int64 min_val;
	_int64 max_code;

	_int64 min_found;				// local statistics, for rough evaluations
	_int64 max_found;
};

class ColumnBinEncoder::EncoderDecimal : public ColumnBinEncoder::EncoderInt
{
public:
	EncoderDecimal(VirtualColumn *vc, bool decodable, bool nulls_possible, bool _descending);

	virtual bool SecondColumn(VirtualColumn *vc);
	virtual void Encode(unsigned char *buf, unsigned char *buf_sec, VirtualColumn *vc, MIIterator &mit, bool update_stats = false);
	virtual void Encode(unsigned char *buf, unsigned char *buf_sec, _int64 v, bool sec_column, bool update_stats);
	virtual bool ImpossibleValues(_int64 pack_min, _int64 pack_max);
protected:
	int scale;
	double multiplier;
	double sec_multiplier;
};

class ColumnBinEncoder::EncoderDate : public ColumnBinEncoder::EncoderInt
{
public:
	EncoderDate(VirtualColumn *vc, bool decodable, bool nulls_possible, bool _descending);

	virtual bool SecondColumn(VirtualColumn *vc);

	virtual void Encode(unsigned char *buf, unsigned char *buf_sec, VirtualColumn *vc, MIIterator &mit, bool update_stats = false);
	virtual void Encode(unsigned char *buf, unsigned char *buf_sec, _int64 v, bool sec_column, bool update_stats);
	virtual _int64 GetValue64(unsigned char *buf, unsigned char *buf_sec);
	virtual bool ImpossibleValues(_int64 pack_min, _int64 pack_max);
	virtual void ClearStatistics();
};

class ColumnBinEncoder::EncoderYear : public ColumnBinEncoder::EncoderInt
{
public:
	EncoderYear(VirtualColumn *vc, bool decodable, bool nulls_possible, bool _descending);

	virtual bool SecondColumn(VirtualColumn *vc);

	virtual void Encode(unsigned char *buf, unsigned char *buf_sec, VirtualColumn *vc, MIIterator &mit, bool update_stats = false);
	virtual void Encode(unsigned char *buf, unsigned char *buf_sec, _int64 v, bool sec_column, bool update_stats);
	virtual _int64 GetValue64(unsigned char *buf, unsigned char *buf_sec);
	virtual bool ImpossibleValues(_int64 pack_min, _int64 pack_max);
	virtual void ClearStatistics();
};

class ColumnBinEncoder::EncoderDouble : public ColumnBinEncoder::ColumnValueEncoder
{
public:
	EncoderDouble(VirtualColumn *vc, bool decodable, bool nulls_possible, bool _descending);

	virtual bool SecondColumn(VirtualColumn *vc);

	virtual void Encode(unsigned char *buf, unsigned char *buf_sec, _int64 v, bool sec_column, bool update_stats);
	virtual void Encode(unsigned char *buf, unsigned char *buf_sec, VirtualColumn *vc, MIIterator& mit, bool update_stats = false);
	virtual void SetNull(unsigned char *buf, unsigned char *buf_sec);
	virtual bool IsNull(unsigned char *buf, unsigned char *buf_sec);
	virtual _int64 GetValue64(unsigned char *buf, unsigned char *buf_sec);

protected:
	_int64 multiplier_vc1;	// 0 if the first vc is double, otherwise it is regarded as int (decimal) which must be multiplied
	_int64 multiplier_vc2;	// 0 if the second vc is double, otherwise it is regarded as int (decimal) which must be multiplied
};

class ColumnBinEncoder::EncoderText : public ColumnBinEncoder::ColumnValueEncoder
{
public:
	EncoderText(VirtualColumn *vc, bool decodable, bool nulls_possible, bool _descending);
	~EncoderText();

	virtual bool SecondColumn(VirtualColumn *vc);

	virtual void Encode(unsigned char *buf, unsigned char *buf_sec, VirtualColumn *vc, MIIterator& mit, bool update_stats = false);
	virtual void SetNull(unsigned char *buf, unsigned char *buf_sec);
	virtual bool IsNull(unsigned char *buf, unsigned char *buf_sec);
	virtual RCBString GetValueT(unsigned char *buf, unsigned char *buf_sec);

	// Encoding:
	// <comparable text><len+1>,   text is padded with zeros
	// null value: all 0, len = 0;    empty string: all 0, len = 1
};

class ColumnBinEncoder::EncoderText_UTF: public ColumnBinEncoder::ColumnValueEncoder
{
public:
	EncoderText_UTF(VirtualColumn *vc, bool decodable, bool nulls_possible, bool _descending);
	~EncoderText_UTF();

	virtual bool SecondColumn(VirtualColumn *vc);

	virtual void Encode(unsigned char *buf, unsigned char *buf_sec, VirtualColumn *vc, MIIterator& mit, bool update_stats = false);
	virtual void SetNull(unsigned char *buf, unsigned char *buf_sec);
	virtual bool IsNull(unsigned char *buf, unsigned char *buf_sec);
	virtual RCBString GetValueT(unsigned char *buf, unsigned char *buf_sec);

private:
	// Encoding:
	// <comparable text><len+1>,   text is padded with zeros
	// null value: all 0, len = 0;    empty string: all 0, len = 1
	DTCollation collation;
};

#endif /* COLUMNBINENCODER_H_ */
