/* Copyright (C)  2005-2008 Infobright Inc.

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License version 2.0 as
 published by the Free  Software Foundation.

 This program is distributed in the hope that  it will be useful, but
 WITHOUT ANY WARRANTY; without even  the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License version 2.0 for more details.

 You should have received a  copy of the GNU General Public License
 version 2.0  along with this  program; if not, write to the Free
 Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
 02111-1307 USA  */

#include <iostream>
#include <assert.h>
#include <fcntl.h>
#ifdef __GNUC__
#include <sys/stat.h>
#endif

#include "Filter.h"
#include "shared/loader/Loader.h"
#include "system/IBFile.h"
#include "common/bhassert.h"
#include "tools.h"

using namespace std;
using namespace bh;

const int Filter::bitBlockSize = BIT_BLOCK_SIZE;
const int Filter::compressedBitBlockSize = COMPRESSED_BIT_BLOCK_SIZE;
const uchar Filter::file_format_v1 = 1;

Filter::Filter(_int64 no_obj, bool all_ones)
	: iterator_b(-1), iterator_n(-1), no_blocks(0),
	no_of_bits_in_last_block(0), save_path_prefix(0), blocks(0),
	block_status(0), bit_block_pool(0), block_object_pool(0), delayed_stats(NULL)
{
#ifdef FUNCTIONS_EXECUTION_TIMES
	FETOperator fet("Filter::Block::");
#endif
	delayed_stats = NULL;
	no_blocks = bh::common::NoObj2NoPacks(no_obj);
	no_of_bits_in_last_block = (no_obj & 0xFFFF) == 0 ? 0x10000 : (no_obj & 0xFFFF);
	if(no_obj == 0)
		no_of_bits_in_last_block = 0;

	save_path_prefix = 0;
	blocks = 0;

	//if(no_blocks > 0)
	Construct(all_ones);
}

Filter::Filter(const Filter& filter) :
	iterator_b(-1), iterator_n(-1), no_blocks(0), no_of_bits_in_last_block(0), save_path_prefix(0), blocks(0),
			block_status(0), bit_block_pool(0), block_object_pool(0), delayed_stats(NULL)
{
#ifdef FUNCTIONS_EXECUTION_TIMES
	FETOperator fet("Filter::Block::");
#endif
	Filter& tmp_filter = const_cast<Filter&> (filter);
	no_blocks = filter.NoBlocks();
	no_of_bits_in_last_block = tmp_filter.NoAddBits();
	save_path_prefix = 0;
	blocks = new Block *[no_blocks];
	ConstructPool();

	for(int i = 0; i < no_blocks; i++) {
		if(tmp_filter.GetBlock(i))
			blocks[i] = block_object_pool->construct(*(tmp_filter.GetBlock(i)), this); //TODO: JB check copying to use pool
		else
			blocks[i] = NULL;
	}
	block_status = new uchar[no_blocks];
	memcpy(block_status, filter.block_status, no_blocks);
	save_path_prefix = 0;
	Rewind();
}

Filter::Filter(const char* fname) :
	iterator_b(-1), iterator_n(-1), no_blocks(0), no_of_bits_in_last_block(0), save_path_prefix(0), blocks(0),
			block_status(0), bit_block_pool(0), block_object_pool(0), delayed_stats(NULL)
{
	IBFile f;
	int i;

	f.OpenReadOnly(fname);

	Header header;

	f.ReadExact(&header, sizeof(Header));
	if(header.v != file_format_v1)
		BHERROR("unknown Filter file format");

	no_blocks = header.nob;
	no_of_bits_in_last_block = header.bil;

	//if(no_blocks > 0)
	Construct();

	f.ReadExact(block_status, no_blocks);

	for(i = 0; i < no_blocks; i++)
		if(block_status[i] == FB_MIXED)
			blocks[i] = block_object_pool->construct(this, &f);
	f.Close();
}

void Filter::Construct(bool all_ones)
{
	try {
		if(no_blocks > 0) {
			blocks = new Block *[no_blocks];
			for(int i = 0; i < no_blocks; i++)
				blocks[i] = NULL;
			block_status = new uchar[no_blocks];
			memset(block_status, (all_ones ? (int) FB_FULL : (int) FB_EMPTY), no_blocks); // set the filter to all empty
			//No idea how to create an allocator for the pool below to use BH heap, due to static methods in allocator
		}
		ConstructPool();
		Rewind();
		//FIXME if we put throw statement here (which is handled by the catch statement below) then we have crash
		//			on many sql statements e.g. show tables, select * from t - investigete this
	} catch (...) {
		this->~Filter();
		throw OutOfMemoryRCException();
	}
}

void Filter::ConstructPool()
{
	bit_block_pool = new boost::pool<>(bitBlockSize);
	block_object_pool = new boost::object_pool<Block>();
}

Filter::~Filter()
{
#ifdef FUNCTIONS_EXECUTION_TIMES
	FETOperator fet("Filter::Block::");
#endif
	Remove();

	if(blocks) {
		for(int i = 0; i < no_blocks; i++)
			if(blocks[i])
				block_object_pool->destroy(blocks[i]);
		delete[] blocks;
	}
	delete[] block_status;
	delete[] delayed_stats;

	delete block_object_pool;
	delete bit_block_pool;

	free(save_path_prefix);
}

void Filter::InitDelayedStats()
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(!delayed_stats);
	delayed_stats = new int[no_blocks];
	if(delayed_stats == NULL)
		throw OutOfMemoryRCException();
	memset(delayed_stats, -1, no_blocks * sizeof(int));
}

void Filter::SetDelayed(uint b, ushort pos)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(b < no_blocks);
	if(block_status[b] == FB_MIXED) {
		Set(b, pos);
		return;
	}
	if(block_status[b] == FB_EMPTY) {
		if(delayed_stats == NULL)
			InitDelayedStats();
		if(pos > delayed_stats[b] + 1) {// then we can't delay
			if(delayed_stats[b] >= 0)
				SetBetween(b, 0, b, delayed_stats[b]);
			Set(b, pos);
			delayed_stats[b] = -2; // not to use any longer
		} else if(pos == delayed_stats[b] + 1) {
			delayed_stats[b]++;
		}
		// else do nothing
	}
}

void Filter::SetDelayed(_uint64 n)
{
	uint b = (uint) (n >> 16);
	ushort pos = (ushort) (n & 65535);
	SetDelayed(b, n);
}

void Filter::ResetDelayed(uint b, ushort pos)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(b < no_blocks);
	if(block_status[b] == FB_MIXED) {
		Reset(b, pos);
		return;
	}
	if(block_status[b] == FB_FULL) {
		if(delayed_stats == NULL)
			InitDelayedStats();
		if(pos > delayed_stats[b] + 1) {// then we can;t delay
			if(delayed_stats[b] >= 0)
				ResetBetween(b, 0, b, delayed_stats[b]);
			Reset(b, pos);
			delayed_stats[b] = -2; // not to use any longer
		} else if(pos == delayed_stats[b] + 1) {
			delayed_stats[b]++;
		}
		// else do nothing
	}
}

void Filter::ResetDelayed(_uint64 n)
{
	uint b = (uint) (n >> 16);
	ushort pos = (ushort) (n & 65535);
	ResetDelayed(b, pos);
}

void Filter::CommitPack(uint b)
{
	if(delayed_stats) {
		if(block_status[b] == FB_EMPTY && delayed_stats[b] >= 0) {
			SetBetween(b, 0, b, delayed_stats[b]);
		} else if(block_status[b] == FB_FULL && delayed_stats[b] >= 0)
			ResetBetween(b, 0, b, delayed_stats[b]);
		delayed_stats[b] = -1;
	}
}

void Filter::Commit()
{
	if(delayed_stats) {
		for(int b = 0; b < no_blocks; b++)
			CommitPack(b);
	}
}

void Filter::Set()
{
#ifdef FUNCTIONS_EXECUTION_TIMES
	FETOperator fet("Filter::Block::");
#endif
	memset(block_status, FB_FULL, no_blocks); // set the filter to all full

	for(int b = 0; b < no_blocks; b++) {
		if(blocks[b]) {
			block_object_pool->destroy(blocks[b]);
			blocks[b] = NULL;
		}
		if(delayed_stats)
			delayed_stats[b] = -1;
	}
}

void Filter::SetBlock(int b)
{
#ifdef FUNCTIONS_EXECUTION_TIMES
	FETOperator fet("Filter::Block::");
#endif
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(b < no_blocks);
	block_status[b] = FB_FULL; // set the filter to all full
	if(blocks[b]) {
		block_object_pool->destroy(blocks[b]);
		blocks[b] = NULL;
	}
}

void Filter::Set(int b, ushort n)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(b < no_blocks);
	if(block_status[b] == FB_EMPTY) {
		block_status[b] = FB_MIXED;
		if(b == no_blocks - 1)
			blocks[b] = block_object_pool->construct(this, no_of_bits_in_last_block);
		else
			blocks[b] = block_object_pool->construct(this);
		if(blocks[b] == NULL)
			throw OutOfMemoryRCException();
	}
	if(blocks[b]) {
		bool full = blocks[b]->Set(n);
		if(full)
			SetBlock(b);
	}
}

void Filter::SetBetween(_int64 n1, _int64 n2)
{
	if(n1 == n2)
		Set((int) (n1 >> 16), (ushort) (n1 & 65535));
	else
		SetBetween((int) (n1 >> 16), (ushort) (n1 & 65535), (int) (n2 >> 16), (ushort) (n2 & 65535));
}

void Filter::SetBetween(int b1, ushort n1, int b2, ushort n2)
{
#ifdef FUNCTIONS_EXECUTION_TIMES
	FETOperator fet("Filter::Block::");
#endif
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(b2 < no_blocks);
	if(b1 == b2) {
		if(block_status[b1] == FB_EMPTY) {
			// if full block
			if( (b1 == no_blocks - 1 && n1 == 0 && n2 == no_of_bits_in_last_block - 1) ||
			   (b1 < no_blocks - 1 && n1 == 0 && n2 == 65535) )
				block_status[b1] = FB_FULL;
			else {
				block_status[b1] = FB_MIXED;
				if(b1 == no_blocks - 1)
					blocks[b1] = block_object_pool->construct(this, no_of_bits_in_last_block);
				else
					blocks[b1] = block_object_pool->construct(this);
				if(blocks[b1] == NULL)
					throw OutOfMemoryRCException();
			}
		}
		if(blocks[b1]) {
			bool full = blocks[b1]->Set(n1, n2);
			if(full)
				SetBlock(b1);
		}
	} else {
		if(n1 == 0)
			SetBlock(b1);
		else
			SetBetween(b1, n1, b1, 65535); // note that b1 is never the last block
		for(int i = b1 + 1; i < b2; i++)
			SetBlock(i);
		SetBetween(b2, 0, b2, n2);
	}
}

void Filter::Reset()
{
#ifdef FUNCTIONS_EXECUTION_TIMES
	FETOperator fet("Filter::Block::");
#endif
	memset(block_status, FB_EMPTY, no_blocks); // set the filter to all empty
	for(int b = 0; b < no_blocks; b++) {
		if(blocks[b]) {
			block_object_pool->destroy(blocks[b]);
			blocks[b] = NULL;
		}
		if(delayed_stats)
			delayed_stats[b] = -1;
	}
}

void Filter::ResetBlock(int b)
{
#ifdef FUNCTIONS_EXECUTION_TIMES
	FETOperator fet("Filter::Block::");
#endif
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(b < no_blocks);
	block_status[b] = FB_EMPTY;
	if(blocks[b]) {
		block_object_pool->destroy(blocks[b]);
		blocks[b] = NULL;
	}
}

void Filter::Reset(int b, ushort n)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(b < no_blocks);
	if(block_status[b] == FB_FULL) {
		block_status[b] = FB_MIXED;
		if(b == no_blocks - 1)
			blocks[b] = block_object_pool->construct(this, no_of_bits_in_last_block, true); // set as full, then reset a part of it
		else
			blocks[b] = block_object_pool->construct(this, 65536, true);
		if(blocks[b] == NULL)
			throw OutOfMemoryRCException();
	}
	if(blocks[b]) {
		bool empty = blocks[b]->Reset(n);
		if(empty)
			ResetBlock(b);
	}
}

void Filter::ResetBetween(_int64 n1, _int64 n2)
{
	if(n1 == n2)
		Reset((int) (n1 >> 16), (ushort) (n1 & 65535));
	else
		ResetBetween((int) (n1 >> 16), (ushort) (n1 & 65535), (int) (n2 >> 16), (ushort) (n2 & 65535));
}

void Filter::ResetBetween(int b1, ushort n1, int b2, ushort n2)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(b2 <= no_blocks);
	if(b1 == b2) {
		if(block_status[b1] == FB_FULL) {
			// if full block
			if( (b1 == no_blocks - 1 && n1 == 0 && n2 == no_of_bits_in_last_block - 1) ||
			   (b1 < no_blocks - 1 && n1 == 0 && n2 == 65535) )
				block_status[b1] = FB_EMPTY;
			else {
				block_status[b1] = FB_MIXED;
				if(b1 == no_blocks - 1)
					blocks[b1] = block_object_pool->construct(this, no_of_bits_in_last_block, true); // set as full, then reset a part of it
				else
					blocks[b1] = block_object_pool->construct(this, 65536, true);
				if(blocks[b1] == NULL)
					throw OutOfMemoryRCException();
			}
		}
		if(blocks[b1]) {
			bool empty = blocks[b1]->Reset(n1, n2);
			if(empty)
				ResetBlock(b1);
		}
	} else {
		if(n1 == 0)
			ResetBlock(b1);
		else
			ResetBetween(b1, n1, b1, 65535); // note that b1 is never the last block
		for(int i = b1 + 1; i < b2; i++)
			ResetBlock(i);
		ResetBetween(b2, 0, b2, n2);
	}
}

void Filter::Reset(Filter &f2)
{
	int mb = min( f2.NoBlocks(), NoBlocks() );
	for(int b = 0; b < mb; b++) {
		if(f2.block_status[b] == FB_FULL)
			ResetBlock(b);
		else if(f2.block_status[b] != FB_EMPTY) // else no change
		{
			if(block_status[b] == FB_FULL) {
				blocks[b] = block_object_pool->construct(*(f2.GetBlock(b)), this);
				blocks[b]->Not(); // always nontrivial
				block_status[b] = FB_MIXED;
			} else if(blocks[b]) {
				bool empty = blocks[b]->Reset(*(f2.GetBlock(b)));
				if(empty)
					ResetBlock(b);
			}
		}
	}
}

bool Filter::Get(int b, ushort n)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(b < no_blocks);
	if(block_status[b] == FB_EMPTY)
		return false;
	if(block_status[b] == FB_FULL)
		return true;
	return blocks[b]->Get(n);
}

bool Filter::IsEmpty()
{
	for(int i = 0; i < no_blocks; i++)
		if(block_status[i] != FB_EMPTY)
			return false;
	return true;
}

bool Filter::IsEmpty(int b)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(b < no_blocks);
	return (block_status[b] == FB_EMPTY);
}

bool Filter::IsEmptyBetween(_int64 n1, _int64 n2)	// true if there are only 0 between n1 and n2, inclusively
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT((n1 >= 0) && (n1 <= n2));
	if(n1 == n2)
		return !Get(n1);
	int b1 = (n1 >> 16);
	int b2 = (n2 >> 16);
	ushort nn1 = (n1 & 65535);
	ushort nn2 = (n2 & 65535);
	if(b1 == b2) {
		if(block_status[b1] == FB_FULL)
			return false;
		if(block_status[b1] == FB_EMPTY)
			return true;
		return blocks[b1]->IsEmptyBetween(nn1, nn2);
	} else {
		int full_pack_start = (nn1 == 0? b1 : b1 + 1);
		int full_pack_stop  = b2 - 1;
		if(nn2 == 65535 || (b2 == no_blocks - 1 && nn2 == no_of_bits_in_last_block - 1))
			full_pack_stop	= b2;
		for(int i = full_pack_start; i <= full_pack_stop; i++)
			if(block_status[i] != FB_EMPTY)
				return false;
		if(b1 != full_pack_start) {
			if(block_status[b1] == FB_FULL)
				return false;
			if( block_status[b1] != FB_EMPTY &&
				!blocks[b1]->IsEmptyBetween(nn1, 65535))  // note that b1 is never the last block
				return false;
		}
		if(b2 != full_pack_stop) {
			if(block_status[b2] == FB_FULL)
				return false;
			if( block_status[b2] != FB_EMPTY &&
				!blocks[b2]->IsEmptyBetween(0, nn2))
				return false;
		}
	}
	return true;
}

bool Filter::IsFullBetween(_int64 n1, _int64 n2)	// true if there are only 1 between n1 and n2, inclusively
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT((n1 >= 0) && (n1 <= n2));
	if(n1 == n2)
		return Get(n1);
	int b1 = (n1 >> 16);
	int b2 = (n2 >> 16);
	ushort nn1 = (n1 & 65535);
	ushort nn2 = (n2 & 65535);
	if(b1 == b2) {
		if(block_status[b1] == FB_FULL)
			return true;
		if(block_status[b1] == FB_EMPTY)
			return false;
		return blocks[b1]->IsFullBetween(nn1, nn2);
	} else {
		int full_pack_start = (nn1 == 0? b1 : b1 + 1);
		int full_pack_stop  = b2 - 1;
		if(nn2 == 65535 || (b2 == no_blocks - 1 && nn2 == no_of_bits_in_last_block - 1))
			full_pack_stop	= b2;
		for(int i = full_pack_start; i <= full_pack_stop; i++)
			if(block_status[i] != FB_FULL)
				return false;
		if(b1 != full_pack_start) {
			if(block_status[b1] == FB_EMPTY)
				return false;
			if( block_status[b1] != FB_FULL &&
				!blocks[b1]->IsFullBetween(nn1, 65535))  // note that b1 is never the last block
				return false;
		}
		if(b2 != full_pack_stop) {
			if(block_status[b2] == FB_EMPTY)
				return false;
			if( block_status[b2] != FB_FULL &&
				!blocks[b2]->IsFullBetween(0, nn2))
				return false;
		}
	}
	return true;
}

bool Filter::IsFull()
{
	for(int i = 0; i < no_blocks; i++)
		if(block_status[i] != FB_FULL)
			return false;
	return true;
}

bool Filter::IsFull(int b)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(b < no_blocks);
	return (block_status[b] == FB_FULL);
}

void Filter::Copy(Filter& f, int block)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(block < no_blocks);
	if(blocks[block]) {
		block_object_pool->destroy(blocks[block]);
		blocks[block] = NULL;
	}
	block_status[block] = f.block_status[block];
	if(f.GetBlock(block))
		blocks[block] = block_object_pool->construct(*(f.GetBlock(block)), this);
}

bool Filter::IsEqual(Filter &sec)
{
	if(no_blocks != sec.no_blocks || no_of_bits_in_last_block != sec.no_of_bits_in_last_block)
		return false;
	for(int b = 0; b < no_blocks; b++) {
		if(block_status[b] != sec.block_status[b])
			return false;
		if(blocks[b] && blocks[b]->IsEqual(*sec.GetBlock(b)) == false)
			return false;
	}
	return true;
}

void Filter::And(Filter &f2)
{
	int mb = min( f2.NoBlocks(), NoBlocks() );
	for(int b = 0; b < mb; b++) {
		if(f2.block_status[b] == FB_EMPTY)
			ResetBlock(b);
		else if(f2.block_status[b] != FB_FULL) // else no change
		{
			if(block_status[b] == FB_FULL) {
				blocks[b] = block_object_pool->construct(*(f2.GetBlock(b)), this);
				block_status[b] = FB_MIXED;
			} else if(blocks[b]) {
				bool empty = blocks[b]->And(*(f2.GetBlock(b)));
				if(empty)
					ResetBlock(b);
			}
		}
	}
}

void Filter::Or(Filter &f2)
{
	int mb = min( f2.NoBlocks(), NoBlocks() );
	for(int b = 0; b < mb; b++) {
		if(f2.block_status[b] == FB_FULL)
			SetBlock(b);
		else if(f2.block_status[b] != FB_EMPTY) // else no change
		{
			if(block_status[b] == FB_EMPTY) {
				blocks[b] = block_object_pool->construct(*(f2.GetBlock(b)), this);
				block_status[b] = FB_MIXED;
			} else if(blocks[b]) {
				bool full = blocks[b]->Or(*(f2.GetBlock(b)));
				if(full)
					SetBlock(b);
			}
		}
	}
}

void Filter::Not()
{
	for(int b = 0; b < no_blocks; b++) {
		if(block_status[b] == FB_FULL)
			block_status[b] = FB_EMPTY;
		else if(block_status[b] == FB_EMPTY)
			block_status[b] = FB_FULL;
		else
			blocks[b]->Not();
	}
}

_int64 Filter::NoOnes() const
{
	if(no_blocks == 0)
		return 0;
	_int64 count = 0;
	for(int b = 0; b < no_blocks - 1; b++) {
		if(block_status[b] == FB_FULL)
			count += 65536;
		else if(blocks[b])
			count += blocks[b]->NoOnes(); // else empty
	}

	if(block_status[no_blocks - 1] == FB_FULL)
		count += no_of_bits_in_last_block;
	else if(blocks[no_blocks - 1])
		count += blocks[no_blocks - 1]->NoOnes();
	return count;
}

uint Filter::NoOnes(int b)
{
	if(no_blocks == 0)
		return 0;
	if(block_status[b] == FB_FULL) {
		if(b == no_blocks - 1)
			return no_of_bits_in_last_block;
		return 65536;
	} else if(blocks[b])
		return blocks[b]->NoOnes();
	return 0;
}

_int64 Filter::NoOnesBetween(_int64 n1, _int64 n2)	// no of 1 between n1 and n2, inclusively
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT((n1 >= 0) && (n1 <= n2));
	if(n1 == n2)
		return (Get(n1) ? 1 : 0);
	int b1 = (n1 >> 16);
	int b2 = (n2 >> 16);
	ushort nn1 = (n1 & 65535);
	ushort nn2 = (n2 & 65535);
	if(b1 == b2) {
		if(block_status[b1] == FB_FULL)
			return n2 - n1 + 1;
		if(block_status[b1] == FB_EMPTY)
			return 0;
		return blocks[b1]->NoOnesBetween(nn1, nn2);
	}
	_int64 counter = 0;
	int full_pack_start = (nn1 == 0? b1 : b1 + 1);
	int full_pack_stop  = b2 - 1;
	if(nn2 == 65535 || (b2 == no_blocks - 1 && nn2 == no_of_bits_in_last_block - 1))
		full_pack_stop	= b2;
	for(int i = full_pack_start; i <= full_pack_stop; i++) {
		if(block_status[i] == FB_MIXED)
			counter += blocks[i]->NoOnes();
		else if(block_status[i] == FB_FULL)
			counter += (i < no_blocks - 1 ? 65536 : no_of_bits_in_last_block);
	}
	if(b1 != full_pack_start) {
		if(block_status[b1] == FB_FULL)
			counter += 65535 - nn1 + 1;
		else if(block_status[b1] != FB_EMPTY)
			counter += blocks[b1]->NoOnesBetween(nn1, 65535);  // note that b1 is never the last block
	}
	if(b2 != full_pack_stop) {
		if(block_status[b2] == FB_FULL)
			counter += nn2 + 1;
		else if(block_status[b2] != FB_EMPTY)
			counter += blocks[b2]->NoOnesBetween(0, nn2);
	}
	return counter;
}

int Filter::DensityWeight()			// = 65537 for empty filter or a filter with only one nonempty block.
{									// Otherwise it is an average number of ones in nonempty blocks.
	_int64 count = 0;
	int nonempty = 0;
	if(no_blocks > 0) {
		for(int b = 0; b < no_blocks - 1; b++) {
			if(block_status[b] == FB_FULL) {
				count += 65536;
				nonempty++;
			} else if(blocks[b]) {
				count += blocks[b]->NoOnes();
				nonempty++;
			}
		}

		if(block_status[no_blocks - 1] == FB_FULL) {
			count += no_of_bits_in_last_block;
			nonempty++;
		} else if(blocks[no_blocks - 1]) {
			count += blocks[no_blocks - 1]->NoOnes();
			nonempty++;
		}
	}
	if(nonempty < 2)
		return 65537;
	return int(count / double(nonempty));
}

void Filter::GetNext1(int &b, ushort &n)
{
	if(iterator_b == -1 || iterator_n == 65535) {
		iterator_n = 0;
		iterator_b++;
	} else
		iterator_n++; // now we are on the first suspected position - go forward if it is not 1

	bool found = false;
	while(!found && iterator_b < no_blocks) {
		if(block_status[iterator_b] == FB_FULL) {
			if(iterator_b == no_blocks - 1 && iterator_n == no_of_bits_in_last_block) // end of filter
				iterator_b++;
			found = true;
		} else if(block_status[iterator_b] == FB_EMPTY) {
			iterator_b++;
			iterator_n = 0;
		} else if(blocks[iterator_b]) {
			found = blocks[iterator_b]->GetNext1(iterator_n);
			if(!found) {
				iterator_b++;
				iterator_n = 0;
			}
		}
	}
	if(iterator_b == no_blocks) {
		iterator_b = -1;
		iterator_n = 0;
	}
	b = iterator_b;
	n = ushort(iterator_n);
}

void Filter::GetNext0(int &b, ushort &n)
{
	if(iterator_b == -1 || iterator_n == 65535) {
		iterator_n = 0;
		iterator_b++;
	} else
		iterator_n++; // now we are on the first suspected position - go forward if it is not 1

	bool found = false;
	while(!found && iterator_b < no_blocks) {
		if(block_status[iterator_b] == FB_EMPTY) {
			if(iterator_b == no_blocks - 1 && iterator_n == no_of_bits_in_last_block) // end of filter
				iterator_b++;
			found = true;
		} else if(block_status[iterator_b] == FB_FULL) {
			iterator_b++;
			iterator_n = 0;
		} else if(blocks[iterator_b]) {
			found = blocks[iterator_b]->GetNext0(iterator_n);
			if(!found) {
				iterator_b++;
				iterator_n = 0;
			}
		}
	}

	if(iterator_b == no_blocks) {
		iterator_b = -1;
		iterator_n = 0;
	}
	b = iterator_b;
	n = ushort(iterator_n);
}

_int64 Filter::GetNext0()
{
	int b = 0;
	ushort n = 0;
	this->GetNext0(b, n);
	if(b < 0)
		return b;
	return (_int64) ((_int64) b << 16) + n;
}

_int64 Filter::GetCurrent() const
{
	if(iterator_b < 0)
		return iterator_b;
	return (_int64) ((_int64) iterator_b << 16) + iterator_n;
}

_int64 Filter::NoObj() const
{
	_int64 res = _int64(no_blocks - (no_of_bits_in_last_block ? 1 : 0)) << 16;
	return res + no_of_bits_in_last_block;
}

void Filter::Rewind(int b)
{
	iterator_b = b;
	iterator_n = -1;
}

void Filter::RewindTo(_int64 n)
{
	if(n == 0) {
		Rewind();
	} else {
		iterator_b = int(n >> 16);
		iterator_n = (int) n & 0xFFFF;
		if(iterator_n > 0)
			iterator_n--;
		else {
			iterator_n = 0xFFFF;
			iterator_b--;
			if(iterator_b < 0)
				iterator_n = -1;
		}
	}
}

_int64 Filter::Find1(_int64 k)
{
	k++; // note that k is increased here (and blocks[b]->Find1 should work on increased values)
	_int64 ret = -1;
	_int64 count1 = 0;
	_int64 b = 0;
	// TODO: optimize for faster search (e.g. remember the last answer; maybe even a tree of partial sums)

	while(count1 < k && b < no_blocks) {
		if(block_status[b] == FB_FULL)
			count1 += (b < no_blocks - 1 ? 65536 : no_of_bits_in_last_block);
		else if(blocks[b])
			count1 += blocks[b]->NoOnes();
		b++;
	}
	b--;
	if(count1 >= k) {
		if(block_status[b] == FB_FULL)
			ret = (k - 1) - (b < no_blocks - 1 ? count1 - 65536 : count1 - no_of_bits_in_last_block) + (b << 16);
		else if(blocks[b])
			ret = blocks[b]->Find1((int) (k - (count1 - blocks[b]->NoOnes()))) + (b << 16);
	}
	return ret;
}

int Filter::NoBlocks() const
{
	return no_blocks;
}

int Filter::NoAddBits() const
{
	return no_of_bits_in_last_block;
}

char* Filter::GetFilePrefix()
{
	return save_path_prefix;
}

void Filter::Remove()
{
	if(save_path_prefix) {
		size_t file_path_ct = strlen(save_path_prefix) + 18;
		char* file_path = new char[file_path_ct];
		uint no_b_files = 1;
		uint no_bd_files = 1;

		sprintf(file_path, "%s.bh_tmp", save_path_prefix);
		RemoveFile(file_path);
		sprintf(file_path, "%sB%02d.bh_tmp", save_path_prefix, no_b_files);
		RemoveFile(file_path);
		sprintf(file_path, "%s%05d.bh_tmp", save_path_prefix, no_bd_files);
		RemoveFile(file_path);
		delete[] file_path;
	}
}

void Filter::Grow(_int64 grow_size, bool value)
{
	if(grow_size == 0)
		return;
	int new_bits_in_last = (no_of_bits_in_last_block + grow_size) & 0xFFFF;
	if(new_bits_in_last == 0)
		new_bits_in_last = 0x10000;
	int bits_added_to_block = grow_size > 0x10000 - no_of_bits_in_last_block ? 0x10000 - no_of_bits_in_last_block
			: grow_size;
	int new_blocks = no_of_bits_in_last_block == 0 ? (grow_size + 0xFFFF) >> 16 : (grow_size - (0x10000
			- no_of_bits_in_last_block) + 0xFFFF) >> 16;

	//add new blocks
	AddNewBlocks(new_blocks, value);

	//grow the previously last block
	if(no_blocks > 0 && bits_added_to_block > 0) {
		if(!block_object_pool)
			ConstructPool();
		if(value) {
			if(block_status[no_blocks - 1] == FB_EMPTY) {
				block_status[no_blocks - 1] = FB_MIXED;
				blocks[no_blocks - 1]
						= block_object_pool->construct(this, new_blocks == 0 ? new_bits_in_last : 0x10000);
				if(blocks[no_blocks - 1] == NULL)
					throw OutOfMemoryRCException();
				blocks[no_blocks - 1]->Set(no_of_bits_in_last_block, no_of_bits_in_last_block + bits_added_to_block - 1);
			} else if(block_status[no_blocks - 1] == FB_MIXED)
				blocks[no_blocks - 1]->GrowBlock(bits_added_to_block, true);
		} else {
			if(block_status[no_blocks - 1] == FB_FULL) {
				block_status[no_blocks - 1] = FB_MIXED;
				blocks[no_blocks - 1] = block_object_pool->construct(this,
						new_blocks == 0 ? new_bits_in_last : 0x10000, true);
				if(blocks[no_blocks - 1] == NULL)
					throw OutOfMemoryRCException();
				blocks[no_blocks - 1]->Reset(no_of_bits_in_last_block, no_of_bits_in_last_block + bits_added_to_block
						- 1);
			} else if(block_status[no_blocks - 1] == FB_MIXED)
				blocks[no_blocks - 1]->GrowBlock(bits_added_to_block, false);
		}
	}
	no_blocks += new_blocks;
	no_of_bits_in_last_block = new_bits_in_last;
}

void Filter::AddNewBlocks(int new_blocks, bool value)
{
	if(new_blocks > 0) {
		//grow arrays
		Block** tmp_blocks = new Block *[no_blocks + new_blocks];
		uchar* tmp_block_status = new uchar[no_blocks + new_blocks];
		memcpy(tmp_blocks, blocks, sizeof(Block*) * no_blocks);
		memcpy(tmp_block_status, block_status, sizeof(uchar) * no_blocks);
		delete[] blocks;
		delete[] block_status;
		blocks = tmp_blocks;
		block_status = tmp_block_status;

		for(int i = no_blocks; i < no_blocks + new_blocks; i++) {
			blocks[i] = NULL;
			if(value)
				block_status[i] = FB_FULL;
			else
				block_status[i] = FB_EMPTY;
		}
	}
}

// save this object so a restore from file is possible
void Filter::Store(const char * fname)
{
	IBFile f;

	string ftmp(fname);
	ftmp += ".tmp";
	f.OpenCreateEmpty(ftmp.c_str());

	Header header;

	header.v = file_format_v1;
	header.nob = no_blocks;
	header.bil = no_of_bits_in_last_block;

	f.WriteExact(&header, sizeof(Header));
	f.WriteExact(block_status, no_blocks);
	for(int i = 0; i < no_blocks; i++)
		if(block_status[i] == FB_MIXED)
			blocks[i]->Store(f);
	f.Close();
	try {
		RenameFile(ftmp.c_str(), fname);
	} catch (...) {
		//put here error handling - delete mask not saved.
		rclog << lock << "Delete mask rename unsuccessful: " << GetErrorMessage(errno) << unlock;
	}
}

// block b is copied and appended
// the original one is made empty

void Filter::MoveBitsToEnd(int block, int no_of_bits_to_move)
{
	MoveBlockToEnd(block);
	if(block == no_blocks -2)
		return;
	if(no_of_bits_to_move < NoObj() % 65536)
		ResetBetween(no_blocks - 1, no_of_bits_to_move, no_blocks - 1, NoObj() % 65536);
	no_of_bits_in_last_block = no_of_bits_to_move;
	if(blocks[no_blocks - 1]) {
		blocks[no_blocks - 1]->ShrinkBlock(no_of_bits_to_move);
		if(blocks[no_blocks - 1]->NoOnes() == no_of_bits_in_last_block) {
			block_object_pool->destroy(blocks[no_blocks - 1]);
			blocks[no_blocks - 1] = NULL;
			block_status[no_blocks - 1] = FB_FULL;
		} else if(blocks[no_blocks - 1]->NoOnes() == 0) {
			block_object_pool->destroy(blocks[no_blocks - 1]);
			blocks[no_blocks - 1] = NULL;
			block_status[no_blocks - 1] = FB_EMPTY;
		}
	}
}

void Filter::MoveBlockToEnd(int b)
{
	if(b != no_blocks - 1) {
		Grow(0x10000 - no_of_bits_in_last_block, false);
		no_of_bits_in_last_block = 0x10000;
	}

	AddNewBlocks(1, false);
	no_blocks++;
	block_status[no_blocks - 1] = block_status[b];
	blocks[no_blocks - 1] = blocks[b];
	blocks[b] = NULL;
	block_status[b] = FB_EMPTY;
}

template<>
Filter::Iterator<Filter::Bit>::Iterator(Filter* owner, _int64 idx, Filter::iterator::TYPE::type_t const& type) :
	_owner(owner), _idx(idx), _type(type)
{
	switch(_type) {
		case (Filter::iterator::TYPE::D_SET): {
			_owner->RewindTo(_idx);
			_idx = _owner->GetNext1();
		}
			break;
		case (Filter::iterator::TYPE::D_CLEAR): {
			_owner->RewindTo(_idx);
			_idx = _owner->GetNext0();
		}
			break;
		case (Filter::iterator::TYPE::D_BOTH): {
			++_idx;
		}
			break;
		default: {
			BHASSERT_WITH_NO_PERFORMANCE_IMPACT( ! "bad iterator type" );
		}
	}
	if(_idx == -1)
		_idx = _owner->NoObj();
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT( ( _idx >= 0 ) && ( _idx <= _owner->NoObj() ) );
}

template<>
Filter::iterator& Filter::iterator::operator ++(void)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT( _idx < _owner->NoObj() );
	switch(_type) {
		case (Filter::iterator::TYPE::D_SET): {
			_owner->RewindTo(_idx + 1);
			_idx = _owner->GetNext1();
		}
			break;
		case (Filter::iterator::TYPE::D_CLEAR): {
			_owner->RewindTo(_idx + 1);
			_idx = _owner->GetNext0();
		}
			break;
		case (Filter::iterator::TYPE::D_BOTH): {
			++_idx;
		}
			break;
		default: {
			BHASSERT_WITH_NO_PERFORMANCE_IMPACT( ! "bad iterator type" );
		}
	}
	if(_idx == -1)
		_idx = _owner->NoObj();
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT( ( _idx >= 0 ) && ( _idx <= _owner->NoObj() ) );
	return (*this);
}

Filter::iterator Filter::begin(Filter::iterator::TYPE::type_t const& type)
{
	return (iterator(this, 0, type));
}

Filter::iterator Filter::end(Filter::iterator::TYPE::type_t const& type)
{
	return (iterator(this, NoObj(), type));
}

Filter::iterator Filter::find(_uint64 pos, Filter::iterator::TYPE::type_t const& type)
{
	return (iterator(this, pos, type));
}

/*
 Filter::const_iterator Filter::begin( Filter::const_iterator::TYPE::type_t const& type ) const {
 return ( const_iterator( this, 0, type ) );
 }

 Filter::const_iterator Filter::end( Filter::const_iterator::TYPE::type_t const& type ) const {
 return ( const_iterator( this, NoObj(), type ) );
 }

 Filter::const_iterator Filter::find( _uint64 pos, Filter::const_iterator::TYPE::type_t const& type ) const {
 return ( const_iterator( this, pos, type ) );
 }
 */

Filter::Bit::Bit(Filter* owner, _uint64 idx) :
	_owner(*owner), _idx(idx)
{
}

_uint64 Filter::Bit::GetIndex(void) const
{
	return (_idx);
}

