/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "GroupDistinctTable.h"

GroupDistinctTable::GroupDistinctTable()
	:
	input_buffer( NULL ), t( NULL ),
	group_bytes( 0 ), value_bytes( 0 ), total_width( 0 ), max_total_size( 0 ),
	no_rows( 0 ), rows_limit( 0 ), no_of_occupied( 0 ),
	initialized( false ),
	use_CRC( false ),
	use_collation(false),
	collation_buffer(NULL),
	col_buf_size(0),
	filter_implementation( false ), f( NULL ),
	group_factor( 0 ),
	value_offset( 0 )
{
	// initial max. buffer size - to be adjusted later
	max_total_size = 128 * 1024 * 1024;						// default settings for less than 2.5 GB reserved
	int mem_scale = TrackableObject::MemorySettingsScale();
	if(mem_scale == 0)		// low settings (not more than 0.5 GB)
		max_total_size = 64 * 1024 * 1024;
	else if(mem_scale == 3)		// high settings (when 2.5 - 5 GB is reserved in 64-bit version)
		max_total_size = 256 * 1024 * 1024;
	else if(mem_scale == 4)		// high settings (when 5 - 10 GB is reserved in 64-bit version)
		max_total_size = 512 * 1024 * 1024;
	else if(mem_scale > 4)		// high settings (when at least 10 GB is reserved in 64-bit version)
		max_total_size = 1024 * 1024 * 1024;
}

GroupDistinctTable::~GroupDistinctTable()
{
	dealloc(t);
	delete [] input_buffer;
	delete [] collation_buffer;
	delete f;
}

void GroupDistinctTable::Initialize64(	_int64 max_no_groups,
										_int64 min_value, _int64 max_value,
										_int64 max_no_rows,
										_int64 max_bytes)	// numeric initialization
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(!initialized);
	if(max_bytes>0)
		max_total_size = max_bytes;
	// Filter implementation, if possible:
	if( min_value != MINUS_INF_64 && max_value != PLUS_INF_64 &&
		max_no_groups < 1000000000 &&
		(max_value - min_value) > 0 &&				// overflow detection
		(max_value - min_value) < 1000000000 ) {
		_int64 filter_positions = max_no_groups * (max_value - min_value + 1);
		if( filter_positions / 8 < max_total_size &&		// general memory limitation
			filter_positions / 8 < 4*max_no_rows ) {		// compare with a good case hash implementation
			filter_implementation = true;
			group_factor = max_no_groups;
			value_offset = min_value;
			InitializeBuffers(filter_positions);
			initialized = true;
			return;
		}
	}
	// Hash implementation:
	// calculate byte sizes
	max_no_groups += 1;			// 0 is reserved for "empty position"
	group_bytes = 0;
	value_bytes = 0;
	_uint64 m = max_no_groups;
	while(m > 0) {
		m = m>>8;
		group_bytes++;
	}
	if( min_value == MINUS_INF_64 || max_value == PLUS_INF_64 )
		value_bytes = 8;
	else {
		_uint64 whole_scope = 	(min_value < 0 ? _uint64(-min_value) + 1 : 0 ) +
								(max_value < 0 ? 0 : _uint64(max_value) ) + 1;
		while(whole_scope > 0) {
			whole_scope = whole_scope>>8;
			value_bytes++;		// a number of bytes sufficient to distinguish all values
		}
	}
	total_width = value_bytes + group_bytes;
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(group_bytes <= 8);

	/////////// create buffers ////////////////////
	InitializeBuffers(max_no_rows);
	initialized = true;
}

void GroupDistinctTable::InitializeT(	_int64 max_no_groups,
										int max_len,
										DTCollation in_collation,
										_int64 max_no_rows,
										_int64 max_bytes )		// string initialization
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(!initialized);
	input_length = max_len;
	if(max_bytes>0)
		max_total_size = max_bytes;
	// calculate byte sizes
	max_no_groups += 1;			// 0 is reserved for "empty position"
	group_bytes = 0;
	_uint64 m = max_no_groups;
	while(m > 0) {
		m = m>>8;
		group_bytes++;
	}
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(group_bytes <= 8);
	collation = in_collation;
	if(RequiresUTFConversions(collation)) {
		use_collation = true;
		col_buf_size = CollationBufLen(collation, max_len);
		collation_buffer = new unsigned char [col_buf_size];
	}
	if(max_len + 2 > HASH_FUNCTION_BYTE_SIZE || use_collation) {
		value_bytes = HASH_FUNCTION_BYTE_SIZE;
		use_CRC = true;
	}
	else
		value_bytes = max_len + 2;	// 2 for size
	total_width = value_bytes + group_bytes;

	/////////// create buffers ////////////////////
	InitializeBuffers(max_no_rows);
	initialized = true;
}

void GroupDistinctTable::InitializeB(int max_len, _int64 max_no_rows)		// char* initialization
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(!initialized);
	input_length = max_len;
	// calculate byte sizes
	group_bytes = 1;
	if(max_len > HASH_FUNCTION_BYTE_SIZE) {
		value_bytes = HASH_FUNCTION_BYTE_SIZE;
		use_CRC = true;
	}
	else
		value_bytes = max_len;	// 2 for size
	total_width = value_bytes + group_bytes;

	/////////// create buffers ////////////////////
	InitializeBuffers(max_no_rows);
	initialized = true;
}

void GroupDistinctTable::InitializeBuffers(_int64 max_no_rows)		// max_no_rows = 0   =>  limit not known
{
	ConnectionInfo *m_conn = &ConnectionInfoOnTLS.Get();
	if(filter_implementation) {		// special case of filter
		f = new Filter(max_no_rows);
		f->Reset();
		rows_limit = max_no_rows + 1;	// limits do not apply
		rccontrol.lock(m_conn->GetThreadID()) << "GroupDistinctTable initialized as Filter(" << max_no_rows << ")" << unlock;
		return;
	}
	no_rows = max_total_size / total_width;		// memory limitation
	if( max_no_rows != 0 && (max_no_rows + 1) * 1.3 < no_rows)
		no_rows = _int64((max_no_rows + 1) * 1.3);	// make sure that rows_limit>=max_no_groups

	// calculate vertical size (not dividable by 17)
	if( no_rows < 67 )								// too less groups => high collision probability; 67 is prime.
		no_rows = 67;
	if( no_rows % 17 == 0 )
		no_rows++;
	rows_limit = _int64(no_rows * 0.9);		// rows_limit is used to determine whether the table is full

	t = (unsigned char*)alloc(total_width * no_rows, BLOCK_TEMPORARY);
	input_buffer = new unsigned char [ total_width ];
	rccontrol.lock(m_conn->GetThreadID()) << "GroupDistinctTable initialized as Hash(" << no_rows << "), " << group_bytes << "+" << value_bytes << " bytes." << unlock;
	Clear();
}

_int64 GroupDistinctTable::BytesTaken()								// actual size of structures
{
	assert(initialized);
	if(filter_implementation)
		return f->NoObj()/8;
	return total_width * no_rows;
}

GDTResult	GroupDistinctTable::Find(_int64 group, _int64 val)		// numeric values - check if exists
{
	// TODO: date/year/string... encoding with ColumnBinEncoder()
	if(filter_implementation) {
		_int64 pos = group + group_factor*(val - value_offset);
		if(f->Get(pos))
			return GDT_EXISTS;
		return GDT_ADDED;			// "Added" means "found" here.
	}
	group += 1;		// offset; 0 means empty position
	memmove(input_buffer, (unsigned char*)(&group), group_bytes);
	memmove(input_buffer + group_bytes, (unsigned char*)(&val), value_bytes);
	return FindCurrentRow(true);
}

GDTResult GroupDistinctTable::Find(unsigned char *v)
{
	assert(!filter_implementation);
	input_buffer[0] = 1;
	if(use_CRC) {
		memset(input_buffer + group_bytes, 0, HASH_FUNCTION_BYTE_SIZE);
		HashMD5(v, input_length, (input_buffer + group_bytes));
	} else
		memcpy(input_buffer + group_bytes, v, value_bytes);
	return FindCurrentRow(true);
}

GDTResult	GroupDistinctTable::Add(_int64 group, _int64 val)		// numeric values
{
	// TODO: date/year/string... encoding with ColumnBinEncoder()
	if(filter_implementation) {
		_int64 pos = group + group_factor*(val - value_offset);
		if(f->Get(pos))
			return GDT_EXISTS;
		f->Set(pos);
		return GDT_ADDED;
	}
	group += 1;		// offset; 0 means empty position
	memmove(input_buffer, (unsigned char*)(&group), group_bytes);
	memmove(input_buffer + group_bytes, (unsigned char*)(&val), value_bytes);
	return FindCurrentRow();
}

GDTResult	GroupDistinctTable::Add(_int64 group, const RCBString &v)		// string values
{
	assert(!filter_implementation);
	group += 1;		// offset; 0 means empty position
	memmove(input_buffer, (unsigned char*)(&group), group_bytes);
	if(use_CRC) {
		memset(input_buffer + group_bytes, 0, HASH_FUNCTION_BYTE_SIZE);
		if(use_collation) {
			assert(collation_buffer);
			collation.collation->coll->strnxfrm(collation.collation, collation_buffer, col_buf_size, (uchar*)v.Value(), v.len);
			//int coded_len = collation.collation->coll->strnxfrmlen(collation.collation, v.len * collation.collation->mbmaxlen);
			HashMD5(collation_buffer, col_buf_size,(input_buffer + group_bytes));
		} else
			HashMD5((unsigned char*)v.Value(), v.len,(input_buffer + group_bytes));
	} else {
		memset(input_buffer + group_bytes, 0, value_bytes);
		if( v.len == 0 )
			return FindCurrentRow();
		*((unsigned short*)(input_buffer + group_bytes)) = v.len;
		memcpy(input_buffer + group_bytes + 2, v.Value(), v.len);
	}
	return FindCurrentRow();
}

GDTResult	GroupDistinctTable::Add(unsigned char *v)		// arbitrary vector of bytes
{
	assert(!filter_implementation);
	input_buffer[0] = 1;
	if(use_CRC) {
		memset(input_buffer + group_bytes, 0, HASH_FUNCTION_BYTE_SIZE);
		HashMD5(v, input_length, (input_buffer + group_bytes));
	} else
		memcpy(input_buffer + group_bytes, v, value_bytes);
	return FindCurrentRow();
}

GDTResult	GroupDistinctTable::AddFromCache(unsigned char *input_vector)	// input vector from cache
{
	assert(!filter_implementation);
	memcpy(input_buffer, input_vector, total_width);
	return FindCurrentRow();
}

_int64		GroupDistinctTable::GroupNoFromInput()		// decode group number from the current input vector
{
	_int64 group = 0;
	memcpy((unsigned char*)(&group), input_buffer, group_bytes);
	group -= 1;		// offset; 0 means empty position
	return group;
}

_int64		GroupDistinctTable::ValueFromInput()		// decode original value from the current input vector
{
	_int64 val = 0;
	memcpy((unsigned char*)(&val), input_buffer + group_bytes, value_bytes);
	return val;
}

GDTResult	GroupDistinctTable::FindCurrentRow(bool find_only)			// find / insert the current buffer
{
	assert(!filter_implementation);
	unsigned int crc_code = CRC32( input_buffer, total_width );
	_int64 row = crc_code % no_rows;
	_int64 step = 3 + crc_code % 4;
	_int64 local_no_of_checks = 0;
	do {
		unsigned char *p = t + row * total_width;
		if(!RowEmpty(p)) {
			if(memcmp(p, input_buffer, total_width) == 0) {
				// i.e. identical row found
				return GDT_EXISTS;
			}
			local_no_of_checks++;
			row = (row + step + local_no_of_checks) % no_rows;
		} else {
			if(!find_only) {
				memcpy(p, input_buffer, total_width);
				no_of_occupied++;
			}
			return GDT_ADDED;
		}
	} while( local_no_of_checks < 7 );	// search depth
	return GDT_FULL;
}

void GroupDistinctTable::Clear()		// clear the tables
{
	if(filter_implementation) {
		f->Reset();
		return;
	}
	memset(t, 0, total_width * no_rows);
	memset(input_buffer, 0, total_width);
	no_of_occupied = 0;
}
