/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "GroupTable.h"
#include "GroupDistinctTable.h"
#include "core/tools.h"

GroupTable::GroupTable()
{
	max_total_size = 0;
	initialized = false;
	occupied = NULL;
	occupied_table = NULL;
	no_of_occupied = 0;
	input_buffer = NULL;
	column_offset = NULL;
	t = NULL;
	grouping_buf_width = 0;
	total_width = 0;
	operation = NULL;
	distinct = NULL;

	attr_type = NULL;
	size = NULL;
	precision = NULL;

	aggregator = NULL;

	no_attr = 0;
	no_grouping_attr = 0;

	gt_min = NULL;
	gt_max = NULL;

	gdistinct = NULL;
	is_full = false;
	last_conflict = 0;
	search_depth = 0;
	declared_max_no_groups = 0;
}

GroupTable::~GroupTable()
{
	delete occupied;
	delete [] occupied_table;
	dealloc(t);
	delete [] input_buffer;
	delete [] column_offset;
	delete [] distinct;

	delete [] operation;
	delete [] attr_type;
	delete [] size;
	delete [] precision;

	if(aggregator) {
		for(int i=0; i<no_attr; i++) {
			if(aggregator[i])
				delete aggregator[i];
		}
	}
	delete [] aggregator;

	delete [] gt_min;
	delete [] gt_max;

	for(int i=0; i<no_attr; i++) {
		if(gdistinct[i])
			delete gdistinct[i];
	}
	delete [] gdistinct;
}

void GroupTable::AddGroupingColumn(AttributeType type, int size, int precision, DTCollation in_collation)
{
	GroupTable::ColTempDesc desc;
	desc.type = type;
	switch(type) {
		case RC_INT:
		case RC_MEDIUMINT:
		case RC_BYTEINT:
		case RC_SMALLINT:
			size = 4;
			break;				// minimal field is one int (4 byte)
		case RC_STRING:
		case RC_VARCHAR:
		case RC_BIN:
		case RC_BYTE:
		case RC_VARBYTE:
			assert(size>0);
			break;		// size is mandatory!
		default:
			size = 8;
	}
	desc.size = size;
	desc.precision = precision;
	desc.collation = in_collation;
	desc.operation = GT_LIST;		// not used
	desc.distinct = false;			// not used
	grouping_desc.push_back(desc);
}

void GroupTable::AddAggregatedColumn( GT_Aggregation operation, bool distinct,
									  AttributeType type, int size, int precision, DTCollation in_collation)
{
	GroupTable::ColTempDesc desc;
	desc.type = type;						// put defaults first, then check in Initialize() what will be actual result definition
	// Overwriting size in some cases:
	switch(type) {
		case RC_INT:
		case RC_MEDIUMINT:
		case RC_BYTEINT:
		case RC_SMALLINT:
			size = 4;
			break;				// minimal field is one int (4 bytes)
		case RC_STRING:
		case RC_VARCHAR:
		case RC_BIN:
		case RC_BYTE:
		case RC_VARBYTE:
			// left as is
			break;
		// Note: lookup strings will have type RC_STRING or similar, and size 4 (stored as int!)
		default:
			size = 8;
	}
	desc.size = size;
	desc.precision = precision;
	desc.operation = operation;
	desc.distinct = distinct;
	desc.collation = in_collation;
	aggregated_desc.push_back(desc);
}

void GroupTable::Initialize(_int64 max_no_groups)
{
	MEASURE_FET("GroupTable::Initialize(...)");
	declared_max_no_groups = max_no_groups;
	if(initialized)
		return;
	no_grouping_attr = grouping_desc.size();
	no_attr = no_grouping_attr + aggregated_desc.size();
	if(no_attr==0)
		return;

	operation		= new GT_Aggregation [no_attr];
	distinct		= new bool [no_attr];
	attr_type		= new AttributeType [no_attr];
	size			= new int [no_attr];
	precision		= new int [no_attr];
	column_offset 	= new int [no_attr];
	gt_min			= new _int64 [no_attr];
	gt_max			= new _int64 [no_attr];
	gdistinct		= new GroupDistinctTable * [no_attr];
	aggregator		= new Aggregator * [no_attr];

	// rewrite column descriptions (defaults, to be verified)
	int no_columns_with_distinct = 0;
	GroupTable::ColTempDesc desc;
	for(int i = 0; i < no_attr; i++) {
		if(i < no_grouping_attr) {
			desc = grouping_desc[i];
			aggregator[i] = NULL;
		} else {
			desc = aggregated_desc[i - no_grouping_attr];
			//////////////////////// Aggregations: /////////////////////////////////

				    // COUNT(...)
			if(desc.operation == GT_COUNT || desc.operation == GT_COUNT_NOT_NULL) {
				if(desc.max_no_values > 0x7FFFFFFF)
					aggregator[i] = new AggregatorCount64(desc.max_no_values);
				else
					aggregator[i] = new AggregatorCount32(desc.max_no_values);

			} else	// SUM(...)				Note: strings are parsed to double
				if(desc.operation == GT_SUM) {
					if(ATI::IsRealType(desc.type) || ATI::IsStringType(desc.type))
						aggregator[i] = new AggregatorSumD;
					else
						aggregator[i] = new AggregatorSum64;

			} else	// AVG(...)				Note: strings are parsed to double
				if(desc.operation == GT_AVG) {
					if(ATI::IsRealType(desc.type) || ATI::IsStringType(desc.type))
						aggregator[i] = new AggregatorAvgD;
					else
						aggregator[i] = new AggregatorAvg64(desc.precision);

			} else // MIN(...)
				if(desc.operation == GT_MIN) {
					if(ATI::IsStringType(desc.type)) {
						if(RequiresUTFConversions(desc.collation))
							aggregator[i] = new AggregatorMinT_UTF(desc.size, desc.collation);
						else
							aggregator[i] = new AggregatorMinT(desc.size);
					} else if(ATI::IsRealType(desc.type))
						aggregator[i] = new AggregatorMinD;
					else if(desc.min < -(0x7FFFFFFF) || desc.max > 0x7FFFFFFF)
						aggregator[i] = new AggregatorMin64;
					else
						aggregator[i] = new AggregatorMin32;

			} else // MAX(...)
				if(desc.operation == GT_MAX) {
					if(ATI::IsStringType(desc.type)) {
						if(RequiresUTFConversions(desc.collation))
							aggregator[i] = new AggregatorMaxT_UTF(desc.size, desc.collation);
						else
							aggregator[i] = new AggregatorMaxT(desc.size);
					} else if(ATI::IsRealType(desc.type))
						aggregator[i] = new AggregatorMaxD;
					else if(desc.min < -(0x7FFFFFFF) || desc.max > 0x7FFFFFFF)
						aggregator[i] = new AggregatorMax64;
					else
						aggregator[i] = new AggregatorMax32;
			} else	// LIST - just a first value found
				if(desc.operation == GT_LIST) {
					if(ATI::IsStringType(desc.type))
						aggregator[i] = new AggregatorListT(desc.size);
					else if(ATI::IsRealType(desc.type) || (desc.min < -(0x7FFFFFFF) || desc.max > 0x7FFFFFFF))
						aggregator[i] = new AggregatorList64;
					else
						aggregator[i] = new AggregatorList32;

			} else	// VAR_POP(...)
				if(desc.operation == GT_VAR_POP) {
					if(ATI::IsRealType(desc.type) || ATI::IsStringType(desc.type))
						aggregator[i] = new AggregatorVarPopD;
					else
						aggregator[i] = new AggregatorVarPop64(desc.precision);

			} else	// VAR_SAMP(...)
				if(desc.operation == GT_VAR_SAMP) {
					if(ATI::IsRealType(desc.type) || ATI::IsStringType(desc.type))
						aggregator[i] = new AggregatorVarSampD;
					else
						aggregator[i] = new AggregatorVarSamp64(desc.precision);

			} else	// STD_POP(...)
				if(desc.operation == GT_STD_POP) {
					if(ATI::IsRealType(desc.type) || ATI::IsStringType(desc.type))
						aggregator[i] = new AggregatorStdPopD;
					else
						aggregator[i] = new AggregatorStdPop64(desc.precision);

			} else	// STD_SAMP(...)
				if(desc.operation == GT_STD_SAMP) {
					if(ATI::IsRealType(desc.type) || ATI::IsStringType(desc.type))
						aggregator[i] = new AggregatorStdSampD;
					else
						aggregator[i] = new AggregatorStdSamp64(desc.precision);

			} else	// BIT_AND(...)
				if(desc.operation == GT_BIT_AND) {
					aggregator[i] = new AggregatorBitAnd;

			} else	// BIT_Or(...)
				if(desc.operation == GT_BIT_OR) {
					aggregator[i] = new AggregatorBitOr;

			} else	// BIT_XOR(...)
				if(desc.operation == GT_BIT_XOR) {
					aggregator[i] = new AggregatorBitXor;

			}

			if(aggregator[i]->IgnoreDistinct())
				desc.distinct = false;
		}
		attr_type[i]	= desc.type;
		size[i]			= desc.size;
		if(i < no_grouping_attr && RequiresUTFConversions(desc.collation))
			size[i] = CollationBufLen(desc.collation, size[i]);
		assert(size[i] > 0);
		precision[i]	= desc.precision;
		operation[i]	= desc.operation;
		distinct[i]		= desc.distinct;
		gt_min[i]		= NULL_VALUE_64;
		gt_max[i]		= NULL_VALUE_64;
		gdistinct[i]	= NULL;					// for now...
		if(distinct[i])
			no_columns_with_distinct++;
	}
	// calculate column byte sizes
	grouping_buf_width = 0;
	total_width = 0;
	for(int i = 0; i < no_grouping_attr; i++) {
		// round up to 4-byte alignment
		grouping_buf_width = 4 * ( (grouping_buf_width + 3) / 4);				// e.g. 1->4, 12->12, 19->20
		column_offset[i] = grouping_buf_width;
		grouping_buf_width += size[i];
		if(ATI::IsStringType(attr_type[i])) {
			grouping_buf_width += 2;			// 2 bytes for text length on the beginning
		}
	}

	grouping_buf_width = 4 * ( (grouping_buf_width + 3) / 4);				// e.g. 0->0, 1->4, 12->12, 19->20
	total_width = grouping_buf_width;

	for(int i = no_grouping_attr; i < no_attr; i++)	{
		// round up to 4-byte alignment
		total_width = 4 * ( (total_width + 3) / 4);				// e.g. 1->4, 12->12, 19->20
		column_offset[i] = total_width;
		total_width += aggregator[i]->BufferByteSize();
	}
	total_width = 4 * ( (total_width + 3) / 4);

	/////////// create buffers ////////////////////
	// Memory settings
	if(max_total_size == 0) {	// not set yet
		max_total_size = 128 * MBYTE;						// default settings for less than 2.5 GB reserved
		int mem_scale = TrackableObject::MemorySettingsScale();
		if(no_columns_with_distinct>0)
			mem_scale++;		// more memory needed for distinct
		if(mem_scale == 0)		// low settings (not more than 0.5 GB)
			max_total_size = 64 * MBYTE;
		else if(mem_scale == 3)		// high settings (when 2.5 - 5 GB is reserved in 64-bit version)
			max_total_size = 256 * MBYTE;
		else if(mem_scale == 4)		// high settings (when 5 - 10 GB is reserved in 64-bit version)
			max_total_size = 512 * MBYTE;
		else if(mem_scale > 4)		// high settings (when at least 10 GB is reserved in 64-bit version)
			max_total_size = 1024 * MBYTE;
	}

	// calculate sizes
	_int64 primary_total_size = max_total_size;	// size of grouping table, not distinct;
	if(no_columns_with_distinct > 0) {
		// naive algorithm: 1/4 for grouping; 3/4 for distinct tables
		primary_total_size = max_total_size/4;
	}

	// grouping table sizes
	search_depth = 9;								// default search depth
	no_rows = primary_total_size / total_width;		// memory limitation
	double overspace = 1.3;
	if( max_no_groups < 32000 ) {
		overspace = 2;
		search_depth = 20;
	}	// for small groupings (buffer in memory) we may assume more space and deeper search to prevent hash conflicts
	if( (max_no_groups + 1) * overspace < no_rows)
		no_rows = _int64((max_no_groups + 1) * overspace);	// make sure that rows_limit>=max_no_groups

	if(max_no_groups == 1) {		// special case: no grouping, just aggregations
		no_rows = 1;
		rows_limit = 1;
	}
	else {
		// calculate vertical size (not dividable by 17)
		if( no_rows < 67 )								// too less groups => high collision probability; 67 is prime.
			no_rows = 67;
		if( no_rows % 17 == 0 )
			no_rows++;
		rows_limit = _int64(no_rows * 0.9);		// rows_limit is used to determine whether the table is full
	}

	t = (unsigned char*)alloc(total_width * no_rows, BLOCK_TEMPORARY);
	if(grouping_buf_width > 0)
		input_buffer = (unsigned char*)new unsigned int [ grouping_buf_width / 4 ];

	// initialize distinct tables
	_int64 total_size_left = max_total_size - total_width * no_rows;
	for(int i = no_grouping_attr; i < no_attr; i++)
		if(distinct[i]) {
			desc = aggregated_desc[i - no_grouping_attr];
			assert(total_size_left>0);
			gdistinct[i] = new GroupDistinctTable();
			if( ATI::IsStringType(attr_type[i]) )
				gdistinct[i]->InitializeT(  no_rows,
											desc.size,							// original size of text
											desc.collation,
											desc.max_no_values * no_rows,
											total_size_left/no_columns_with_distinct);
			else if(attr_type[i] == RC_DATE)
				gdistinct[i]->Initialize64( no_rows,
											DateSortEncoding(desc.min), DateSortEncoding(desc.max),
											desc.max_no_values * no_rows,
											total_size_left/no_columns_with_distinct);
			else
				gdistinct[i]->Initialize64( no_rows,
											desc.min, desc.max,
											desc.max_no_values * no_rows,
											total_size_left/no_columns_with_distinct);
			total_size_left -= gdistinct[i]->BytesTaken();
			no_columns_with_distinct--;
	}

	// initialize everything
	occupied = new Filter(no_rows);
	occupied_table = new int [no_rows];
	ClearAll();
	initialized = true;
}

//////////////////////////////////////////////////////////////////////////////////////////

void GroupTable::PutGroupingValue(int col, const RCBString &v)
{
	//MEASURE_FET("GroupTable::PutGroupingValue(...)");
	int local_size = size[col] + 2;
	local_size = 4 * ( (local_size + 3) / 4 );
	memset( input_buffer + column_offset[col], 0, local_size);
	if(v.IsNull())
		return;
	if( v.len == 0 ) {
		input_buffer[column_offset[col] + 2] = 1;		// difference between empty string and null: the first one have len=0 and nontrivial character
		return;
	}
	*((unsigned short*)(input_buffer + column_offset[col])) = v.len;
	memcpy( input_buffer + column_offset[col] + 2, v.Value(), v.len );
}

bool GroupTable::FindCurrentRow(_int64 &row)	// a position in the current GroupTable, row==NULL_VALUE_64 if not found and not added
{												// return value: true if already existing, false if put as a new row
	//MEASURE_FET("GroupTable::FindCurrentRow(...)");
	_int64 step = 1;
	int local_no_of_checks = 0;
	if(grouping_buf_width == 0)	{	// special case: no grouping columns
		row = 0;
	} else {
		unsigned int crc_code = CRC32(input_buffer, grouping_buf_width);
		row = crc_code % no_rows;
		step = 3 + crc_code % 4;
	}
	do {
		if(occupied->Get(row)) {
			if(grouping_buf_width == 0 ||
				memcmp( t + row * total_width, input_buffer, grouping_buf_width) == 0) {
				// i.e. identical row found
				last_conflict = 0;
				return true;
			}
			local_no_of_checks++;
			row = (row + step + local_no_of_checks) % no_rows;
		} else {
			if(is_full) {
				row = NULL_VALUE_64;			// when is_full is set, do not add new tuples
				return false;
			}
			if(grouping_buf_width > 0) {
				memcpy(t + row * total_width, input_buffer, grouping_buf_width);
				for(int col = 0; col < no_grouping_attr; col++) {
					if(size[col] == 8)
						PutGroupingValueForStats(col, *(_int64*)(input_buffer + column_offset[col]));
					else if(size[col] == 4)
						PutGroupingValueForStats(col, *(int*)(input_buffer + column_offset[col]));
				}
			}
			occupied->Set(row);
			occupied_table[no_of_occupied] = row;
			no_of_occupied++;
			if(no_of_occupied >= rows_limit || no_of_occupied == declared_max_no_groups)
				is_full = true;
			last_conflict = 0;
			return false;
		}
	} while(local_no_of_checks < search_depth);
	row = NULL_VALUE_64;			// no place found in reasonable depth!
	return false;
}

//////////////////////////////////////////////////////////////////////////////////////////
/*
_int64 GroupTable::GetValue64N(int col, _int64 row)
{
	//MEASURE_FET("GroupTable::GetValue64(...)");
	if( operation[col] == GT_SUM ) {
		if( ATI::IsRealType(attr_type[col]) )
			return *((_int64*)(t + row * total_width + column_offset[col]));		// double value encoded as _int64
		else {
			_int64 res = *((_int64*)(t + row * total_width + column_offset[col]));
			if( res != NULL_VALUE_64 ) {
				double d_res = *((double*)(t + row * total_width + column_offset[col] + 8));
				if( d_res > double(PLUS_INF_64) || d_res < double(MINUS_INF_64)) {	// _int64 overflow
					d_res /= PowOfTen( precision[col] );				// this is a proper result
					throw NotImplementedRCException("Overflow while summing numbers (Group By)");
					throw DataTypeConversionRCException("Overflow: data typed changed", *(_int64*)(&d_res), RC_REAL);
				}
			}
			return res;
		}
	} else if( operation[col] == GT_AVG ) {	// <sum_double><counter_64>
		double res = NULL_VALUE_D;
		_int64 counter = *((_int64*)(t + row * total_width + column_offset[col] + 8));
		if( counter > 0 ) {
			res = *((double*)(t + row * total_width + column_offset[col])) / counter;
			if( precision[col] > 0 )
				res /= PowOfTen( precision[col] );
		}
		return *((_int64*)(&res));
	}
	if( size[col] == 4 ) {		// grouping/list/min/max values
		if( *((int*)(t + row * total_width + column_offset[col])) == NULL_VALUE_32 )
			return NULL_VALUE_64;
		return *((int*)(t + row * total_width + column_offset[col]));
	} else
		return *((_int64*)(t + row * total_width + column_offset[col]));
}

RCBString GroupTable::GetValueTN(int col, _int64 row)
{
	//MEASURE_FET("GroupTable::GetValueT(...)");
	int len = *((unsigned short*)(t + row * total_width + column_offset[col]));
	char *p = (char*)(t + row * total_width + column_offset[col] + 2);
	if(len == 0) {
		if( *p != 0 )						// empty string indicator: len==0 and nontrivial character
			return RCBString("",0,true);	// empty string
		return RCBString();					// null value
	}
	RCBString res( p , len );
	return res;
}
*/
_int64 GroupTable::GetValue64(int col, _int64 row)
{
	unsigned char* p = t + row * total_width + column_offset[col];
	if(col >= no_grouping_attr) {
		return aggregator[col]->GetValue64(p);
	}
	// Grouping column
	if( size[col] == 4 ) {
		if( *((int*)p) == NULL_VALUE_32 )
			return NULL_VALUE_64;
		return *((int*)p);
	} else
		return *((_int64*)p);
}

double GroupTable::GetValueD(int col, _int64 row)
{
	unsigned char* p = t + row * total_width + column_offset[col];
	if(col >= no_grouping_attr) {
		return aggregator[col]->GetValueD(p);
	}
	// Grouping column
	return *((double*)p);
}

RCBString GroupTable::GetValueT(int col, _int64 row)
{
	unsigned char* p = t + row * total_width + column_offset[col];
	if(col >= no_grouping_attr) {
		return aggregator[col]->GetValueT(p);
	}
	// Grouping column
	int len = *((unsigned short*)p);
	char *pstr = (char*)(p + 2);
	if(len == 0) {
		if( *pstr != 0 )					// empty string indicator: len==0 and nontrivial character
			return RCBString("",0,true);	// empty string
		return RCBString();					// null value
	}
	RCBString res(pstr , len);
	return res;
}


void GroupTable::ClearAll()
{
	is_full = false;
	last_conflict = 0;

	memset( t, 0, total_width * no_rows );
	if(grouping_buf_width > 0)
		memset(input_buffer, 0, grouping_buf_width);

	// initialize statistics
	for(int i = 0; i < no_grouping_attr; i++) {
		gt_min[i]		= NULL_VALUE_64;
		gt_max[i]		= NULL_VALUE_64;
	}

	// initialize aggregated values

	for(int i = no_grouping_attr; i < no_attr; i++) {
		aggregator[i]->ResetStatistics();
		for(_int64 row = 0; row < no_rows; row++)
			aggregator[i]->Reset(t + row * total_width + column_offset[i]);
		if(gdistinct[i])
			gdistinct[i]->Clear();
	}
	occupied->Reset();
	no_of_occupied = 0;
}

void GroupTable::ClearUsed()
{
	is_full = false;
	last_conflict = 0;
	if(grouping_buf_width > 0)
		memset(input_buffer, 0, grouping_buf_width);

	FilterOnesIterator it(occupied);
	it.Rewind();
	while(it.IsValid()) {
		if(grouping_buf_width > 0)
			memset(t + *it * total_width, 0, grouping_buf_width);
		for(int i = no_grouping_attr; i < no_attr; i++)
			aggregator[i]->Reset(t + *it * total_width + column_offset[i]);
		++it;
	}
	occupied->Reset();
	no_of_occupied = 0;

	// initialize statistics

	for(int i = 0; i < no_grouping_attr; i++) {		// TODO: unify this mechanism with abstract aggregators
		gt_min[i]		= NULL_VALUE_64;
		gt_max[i]		= NULL_VALUE_64;
	}

	for(int i = no_grouping_attr; i < no_attr; i++) {
		aggregator[i]->ResetStatistics();
		if(gdistinct[i])
			gdistinct[i]->Clear();
	}
}


void GroupTable::ClearDistinct()							// reset the table for distinct values
{
	for(int i = no_grouping_attr; i < no_attr; i++) {
		aggregator[i]->ResetStatistics();
		if(gdistinct[i])
			gdistinct[i]->Clear();
	}
}

//////////////////////////////////////////////////////////////////////////////////////////
void GroupTable::AddCurrentValueToCache(int col, GroupDistinctCache &cache)
{
	cache.SetCurrentValue(gdistinct[col]->InputBuffer(), gdistinct[col]->InputBufferSize());
}

///////////////
bool GroupTable::AggregatePack(int col, _int64 row)	// aggregate based on parameters stored in the aggregator
{
	unsigned char* p = t + row * total_width + column_offset[col];
	return 	aggregator[col]->AggregatePack(p);
}

bool GroupTable::PutAggregatedValue(int col, _int64 row, _int64 factor)			// for agregations which do not need any value
{
	if(aggregator[col]->FactorNeeded() && factor == NULL_VALUE_64)
		throw NotImplementedRCException("Aggregation overflow.");
	unsigned char* p = t + row * total_width + column_offset[col];
	aggregator[col]->PutAggregatedValue(p, factor);
	return true;
}

GDTResult GroupTable::FindDistinctValue(int col, _int64 row, _int64 v)	// for all numerical values
{
	assert(gdistinct[col]);
	if(v == NULL_VALUE_64)		// works also for double
		return GDT_EXISTS;			// null omitted
	if(attr_type[col] == RC_DATE)
		return gdistinct[col]->Find(row, DateSortEncoding(v));
	return gdistinct[col]->Find(row, v);
}

GDTResult GroupTable::AddDistinctValue(int col, _int64 row, _int64 v)	// for all numerical values
{
	assert(gdistinct[col]);
	if(v == NULL_VALUE_64)		// works also for double
		return GDT_EXISTS;			// null omitted
	if(attr_type[col] == RC_DATE)
		return gdistinct[col]->Add(row, DateSortEncoding(v));
	return gdistinct[col]->Add(row, v);
}

bool GroupTable::PutAggregatedValue(int col, _int64 row, _int64 v, _int64 factor)	// for all numerical values
{
	//MEASURE_FET("GroupTable::PutAggregatedValue(...)");
	if(distinct[col]) {
		// Repetition? Return without action.
		assert(gdistinct[col]);
		if(v == NULL_VALUE_64)		// works also for double
			return true;			// null omitted
		GDTResult res;
		if(attr_type[col] == RC_DATE)
			res = gdistinct[col]->Add(row, DateSortEncoding(v));
		else
			res = gdistinct[col]->Add(row, v);
		if(res == GDT_EXISTS)
			return true;	// value found, do not aggregate it again
		if(res == GDT_FULL) {
			if(gdistinct[col]->AlreadyFull())
				is_full = true;	// disable also the main grouping table (if it is a persistent rejection)
			return false;	// value not found in DISTINCT buffer, which is already full
		}
		factor = 1;			// ignore repetitions for distinct
	}
	if(aggregator[col]->FactorNeeded() && factor == NULL_VALUE_64)
		throw NotImplementedRCException("Aggregation overflow.");
	unsigned char* p = t + row * total_width + column_offset[col];
	if(aggregator[col]->IgnoreNulls() && v == NULL_VALUE_64)		// works also for double
		return true;			// null omitted
	aggregator[col]->PutAggregatedValue(p, v, factor);
	return true;
}

bool GroupTable::PutAggregatedValue(int col, _int64 row, const RCBString& v, _int64 factor)	// for all text (not lookup) values
{
	//MEASURE_FET("GroupTable::PutAggregatedValue(...)");
	if(distinct[col]) {
		assert(gdistinct[col]);
		if(v.IsNull())
			return true;
		GDTResult res = gdistinct[col]->Add(row, v);
		if(res == GDT_EXISTS)
			return true;	// value found, do not aggregate it again
		if(res == GDT_FULL) {
			if(gdistinct[col]->AlreadyFull())
				is_full = true;	// disable also the main grouping table (if it is a persistent rejection)
			return false;	// value not found in DISTINCT buffer, which is already full
		}
		factor = 1;			// ignore repetitions for distinct
	}
	if(aggregator[col]->IgnoreNulls() && v.IsNull())
		return true;		// null found (for MIN, MAX, LIST the null value is already initialized)
	if(aggregator[col]->FactorNeeded() && factor == NULL_VALUE_64)
		throw NotImplementedRCException("Aggregation overflow.");
	unsigned char *p = (t + row * total_width + column_offset[col]);
	aggregator[col]->PutAggregatedValue(p, v, factor);
	return true;
}

bool GroupTable::PutCachedValue(int col, GroupDistinctCache &cache)	// for all numerical values
{
	//MEASURE_FET("GroupTable::PutCachedValue(...)");
	assert(distinct[col]);
	GDTResult res = gdistinct[col]->AddFromCache(cache.GetCurrentValue());
	if(res == GDT_EXISTS)
		return true;	// value found, do not aggregate it again
	if(res == GDT_FULL) {
		if(gdistinct[col]->AlreadyFull())
			is_full = true;	// disable also the main grouping table (if it is a persistent rejection)
		return false;	// value not found in DISTINCT buffer, which is already full
	}
	_int64 row = gdistinct[col]->GroupNoFromInput();	// restore group number
	unsigned char* p = t + row * total_width + column_offset[col];
	if(operation[col] == GT_COUNT_NOT_NULL)
		aggregator[col]->PutAggregatedValue(p, 1);			// factor = 1, because we are just after distinct
	else {
		_int64 v = gdistinct[col]->ValueFromInput();
		aggregator[col]->PutAggregatedValue(p, v, 1);
	}
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////

void GroupTable::UpdateAttrStats(int col)	// update the current statistics for a column, if needed
{
	if(!aggregator[col]->StatisticsNeedsUpdate())
		return;
	FilterOnesIterator it(occupied);
	bool stop_updating = false;
	aggregator[col]->ResetStatistics();
	while(!stop_updating && it.IsValid()) {
		stop_updating = aggregator[col]->UpdateStatistics(t + *it * total_width + column_offset[col]);
		++it;
	}
	aggregator[col]->SetStatisticsUpdated();
}

void GroupTable::GetAttrStats(int col, _int64 &min, _int64 &max)	// get the current statistics for a column
{
	min = gt_min[col];
	max = gt_max[col];
}

