/* Copyright (C)  2005-2009 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef _GROUPTABLE_H_
#define _GROUPTABLE_H_

#include "bintools.h"
#include "Filter.h"
#include "GroupDistinctTable.h"
#include "GroupDistinctCache.h"
#include "AggregatorBasic.h"
#include "AggregatorAdvanced.h"


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// GroupTable - a tool for storing values and counters, implementing hash table mechanism
//
//	Usage:
//
//
//
////////////////////////////////////////////////////////////////////////////////////////////////////////
enum GT_Aggregation {GT_LIST,										// GT_LIST - just store the first value
							GT_COUNT, GT_COUNT_NOT_NULL,
							GT_SUM, GT_MIN, GT_MAX, GT_AVG,
							GT_STD_POP, GT_STD_SAMP, GT_VAR_POP, GT_VAR_SAMP,
							GT_BIT_AND, GT_BIT_OR, GT_BIT_XOR,
							GT_GROUP_CONCAT};

class GroupTable : public TrackableObject
{
public:
	GroupTable();
	~GroupTable();

	////////// Group table construction ///////////

	void AddGroupingColumn	(AttributeType type, int b_size, int precision, DTCollation in_collation);		// size in bytes (both for numbers and texts)
	void AddAggregatedColumn(GT_Aggregation operation, bool distinct,
									 AttributeType type, int b_size, int precision, DTCollation in_collation);
	void AggregatedColumnStatistics( int ag_col, _int64 max_no_vals, _int64 min_v = MINUS_INF_64, _int64 max_v = PLUS_INF_64 )
	{
		aggregated_desc[ag_col].max_no_values	= max_no_vals;
		aggregated_desc[ag_col].min				= min_v;
		aggregated_desc[ag_col].max				= max_v;
	}

	void Initialize(_int64 max_no_groups);			// initialize buffers basing on the previously defined grouping/aggregated columns
																// and the best possible upper approximation of number of groups;
																// note: max_no_groups may also be used for "select ... limit n", because we will not produce more groups

	///////// Group table usage //////////////////

	// put values to a temporary buffer (note that it will contain the previous values, which may be reused
	void PutGroupingValue(int col, _int64 v)			// for all numerical values
	{
		if( size[col] == 4 )
			if( v == NULL_VALUE_64 )
				*((int*)(input_buffer + column_offset[col]))	= NULL_VALUE_32;
			else
				*((int*)(input_buffer + column_offset[col]))	= int(v);
		else
			*((_int64*)(input_buffer + column_offset[col]))	= v;
	}
	void PutGroupingValueForStats(int col, _int64 v)			// for all numerical values
	{
		if( v != NULL_VALUE_64 ) {
			if( gt_min[col] == NULL_VALUE_64 || gt_min[col] > v )
				gt_min[col] = v;
			if( gt_max[col] == NULL_VALUE_64 || gt_max[col] < v )
				gt_max[col] = v;
		}
	}

	void PutGroupingValue(int col, const RCBString& v);		// for all text (not lookup) values
	bool FindCurrentRow(_int64 &row);					// a position in the current GroupTable, row==NULL_VALUE_64 if not found
														// return value: true if already existing, false if put as a new row
	GDTResult FindDistinctValue(int col, _int64 row, _int64 v);	// just check whether exists, do not add
	GDTResult AddDistinctValue(int col, _int64 row, _int64 v);	// just add to a list of distincts, not to aggregators

	// columns have common numbering, both grouping and aggregated ones
	// return value: true if value checked in, false if not (possible only on DISTINCT buffer overflow)

	bool PutAggregatedValue(int col, _int64 row, _int64 factor);			// for agregations which do not need any value
	bool PutAggregatedValue(int col, _int64 row, _int64 v, _int64 factor);				// for all numerical values
	bool PutAggregatedValue(int col, _int64 row, const RCBString& v, _int64 factor);	// for all text (not lookup) values
	bool PutCachedValue(int col, GroupDistinctCache &cache);	// for numerics, and only some aggregators

	bool AggregatePack(int col, _int64 row);	// aggregate based on parameters stored in the aggregator

	void AddCurrentValueToCache(int col, GroupDistinctCache &cache);

	///////// Group table output and info //////////////////

	int NoAttr()				{ return no_attr; }				// total number of columns
	int NoGroupingAttr()		{ return no_grouping_attr; }	// number of grouping columns
	bool IsFull()				{ return is_full; }				// no place left?
	bool IsDistinctFull(int col) { return (distinct[col] ? gdistinct[col]->AlreadyFull() : false); }				// no place left in distinct buffer?

	int GetOccupiedRow(int n)		{ assert(n < no_of_occupied); return occupied_table[n]; }
	int GetNoOfOccupiedRows()		{ return no_of_occupied; }

	_int64 GetValue64(int col, _int64 row);			// columns have common numbering
	double GetValueD(int col, _int64 row);
	RCBString GetValueT(int col, _int64 row);

	void ClearAll();								// initialize all
	void ClearUsed();								// initialize only used rows (for next pass)
	void ClearDistinct();							// reset the tables of distinct values, if present

	void GetAttrStats(int col, _int64 &min, _int64 &max);	// get the current statistics for a column
	void UpdateAttrStats(int col);							// update the current statistics for a column, if needed

	AttributeType AttrType(int col)		// note that the type may be different than the type of aggregated column (e.g. COUNT is always integer, AVG is double)
										{ return attr_type[col]; }
	int AttrSize(int col)				{ return size[col]; }
	int AttrPrec(int col)				{ return precision[col]; }
	GT_Aggregation AttrOper(int col)	{ return operation[col]; }
	Aggregator* AttrAggregator(int col)	{ return aggregator[col]; }
	bool AttrDistinct(int col)			{ return distinct[col];  }
	DTCollation GetGroupCollation(int col)	{ return grouping_desc[col].collation; }

	TRACKABLEOBJECT_TYPE TrackableType()	{ return TO_TEMPORARY; }

private:

	Filter*				occupied;
	int 				no_of_occupied;
	int*				occupied_table;

	// content and additional buffers

	unsigned char*		input_buffer;
	int*				column_offset;		// a table of offsets (bytes) of column beginnings wrt. the beginning of a row

	unsigned char*		t;					// common buffer for grouping values and results of aggregations
	int 				grouping_buf_width;	// in bytes
	int 				total_width;		// in bytes

	_int64 				no_rows;			// actual buffer size (vertical)
	_int64 				rows_limit;			// a capacity of one memory buffer
	_int64				search_depth;		// how many steps are we doing before giving up
	_int64				declared_max_no_groups;	// maximal number of groups calculated from KNs etc.

	_int64*				gt_min;				// statistics of grouping columns (numerical) for current GT contents; null values are not counted here
	_int64*				gt_max;				// statistics are updated on all inputs

	// column/operation descriptions

	bool 				initialized;
	struct ColTempDesc					// temporary description - prior to buffers initialization, then sometimes used for complex aggregations
	{
		ColTempDesc()					// defaults:
		{ min = MINUS_INF_64; max = PLUS_INF_64; max_no_values = 0; }

		GT_Aggregation	operation;		// not used for grouping columns
		bool			distinct;		// not used for grouping columns
		AttributeType	type;
		int				size;
		int				precision;
		// optimization statistics, not used for grouping columns:
		_int64			min;
		_int64			max;
		_int64			max_no_values;	// upper approximation of a number of distinct values (without null), 0 - don't know
		DTCollation		collation;
	};
	std::vector<ColTempDesc> grouping_desc;	// input fields description
	std::vector<ColTempDesc> aggregated_desc;

	GT_Aggregation*		operation;		// Note: these tables will be created in Initialize(); using vectors inside grouper may be too slow
	bool*				distinct;		// Note: the description of type/precision is about input (grouped) values,
	AttributeType*		attr_type;		//       i.e. for AVG it may be RC_NUM, although the result of AVG is always double
	int*				precision;
	int*				size;

	//////// a new, abstract aggregators

	Aggregator**		aggregator;		// a table of actual aggregators
	ColumnType**		input_type;
	ColumnType**		outut_type;

	// "distinct" part
	GroupDistinctTable** gdistinct;		// NULL if not used

	// general descriptions

	int					no_attr;
	int					no_grouping_attr;

	bool				is_full;		// true if hash conflicts persisted
	_int64				last_conflict;

	// some memory managing
	_int64		max_total_size;
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#endif
