/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "MIIterator.h"

using namespace std;

MIIterator::MIIterator(MultiIndex* _mind, DimensionVector &dimensions, bool do_rewind) : mind(_mind)
{
	bool success = mind->IteratorLock();
	assert(success);			// Multiindex was already locked for updating!
	no_dims = mind->NoDimensions();
	mind_created_locally = false;
	fi = NULL;
	cur_pos = NULL;
	cur_pack = NULL;
	nulls_found = NULL;
	bigblock_min_size = 0;
	virtual_only = false;
	not_used_only = false;
	Init(dimensions, do_rewind);
}

MIIterator::MIIterator(MultiIndex* _mind, std::set<int> dimensions, bool do_rewind) : mind(_mind)
{
	bool success = mind->IteratorLock();
	assert(success);			// Multiindex was already locked for updating!
	no_dims = mind->NoDimensions();
	mind_created_locally = false;
	fi = NULL;
	cur_pos = NULL;
	cur_pack = NULL;
	nulls_found = NULL;
	bigblock_min_size = 0;
	virtual_only = false;
	not_used_only = false;
	DimensionVector dim2(mind);
	for(std::set<int>::iterator i = dimensions.begin(); i != dimensions.end(); i++)
		dim2[*i] = true;
	Init(dim2, do_rewind);
}

MIIterator::MIIterator(MultiIndex* _mind, bool do_rewind) : mind(_mind)
{
	if(mind == NULL) {
		mind_created_locally = true;
		mind = new MultiIndex();
		mind->AddDimension_cross(1);	// just one row
	} else
		mind_created_locally = false;
	bool success = mind->IteratorLock();
	assert(success);			// Multiindex was already locked for updating!
	no_dims = mind->NoDimensions();
	fi = NULL;
	cur_pos = NULL;
	cur_pack = NULL;
	nulls_found = NULL;
	bigblock_min_size = 0;
	virtual_only = false;
	not_used_only = false;
	DimensionVector dimensions(mind);
	for(int i = 0; i < no_dims; i++)
		dimensions[i] = true;
	Init(dimensions, do_rewind);
}

MIIterator::MIIterator(MultiIndex* _mind, int one_dimension, bool do_not_rewind) : mind(_mind)
{
	bool success = mind->IteratorLock();
	assert(success);			// Multiindex was already locked for updating!
	no_dims = mind->NoDimensions();
	mind_created_locally = false;
	fi = NULL;
	cur_pos = NULL;
	cur_pack = NULL;
	nulls_found = NULL;
	bigblock_min_size = 0;
	virtual_only = false;
	not_used_only = false;
	DimensionVector dimensions(mind);
	if(one_dimension == -1) {
		for(int i = 0; i < no_dims; i++)
			dimensions[i] = true;
	} else
		dimensions[one_dimension] = true;
	Init(dimensions, do_not_rewind);
}

MIIterator::MIIterator(const MIIterator& sec)
	: mind( sec.mind ),
	mind_created_locally(sec.mind_created_locally),
	no_dims( sec.no_dims ),
	cur_pos( no_dims > 0 ? new _int64 [no_dims] : NULL ),
	checked_until( no_dims > 0 ? new _int64 [no_dims] : NULL ),
	cur_pack( no_dims > 0 ? new int [no_dims] : NULL ),
	nulls_found( no_dims > 0 ? new bool [no_dims] : NULL ),
	fi( no_dims > 0 ? new FilterOnesIterator*[no_dims] : NULL ),
	valid( sec.valid ),
	omitted_factor( sec.omitted_factor ),
	no_obj( sec.no_obj ),
	pack_size_left( sec.pack_size_left ),
	bigblock_min_size( sec.bigblock_min_size ),
	dim_mode( sec.dim_mode ),
	dim_order( sec.dim_order ),
	virtual_only (sec.virtual_only),
	not_used_only (sec.not_used_only),
	iterator_m( sec.iterator_m ),
	next_material_pack( sec.next_material_pack ),
	next_pack_started( sec.next_pack_started )
{
	if(sec.mind_created_locally)
		mind = new MultiIndex(*sec.mind);
	bool success = mind->IteratorLock();
	assert(success);			// Multiindex was already locked for updating!
	for(int i = 0; i < no_dims; i++) {
		if(sec.fi[i])
			fi[i] = sec.fi[i]->Copy();
		else
			fi[i] = NULL;
		cur_pos[i] = sec.cur_pos[i];
		cur_pack[i] = sec.cur_pack[i];
		checked_until[i] = sec.checked_until[i];
		nulls_found[i] = sec.nulls_found[i];
	}
	for(int i = 0; i < no_dims; i++)
		if(dim_mode[i] == 2)
			mind->LockForGetIndex(i);
}

void MIIterator::swap( MIIterator& i ) {
	using std::swap;
	swap( mind,               i.mind );
	swap( mind_created_locally, i.mind_created_locally );
	swap( no_dims,            i.no_dims );
	swap( cur_pos,            i.cur_pos );
	swap( checked_until,      i.checked_until);
	swap( cur_pack,           i.cur_pack );
	swap( nulls_found,        i.nulls_found );
	swap( fi,                 i.fi );
	swap( valid,              i.valid );
	swap( omitted_factor,     i.omitted_factor );
	swap( no_obj,             i.no_obj );
	swap( pack_size_left,     i.pack_size_left );
	swap( bigblock_min_size,  i.bigblock_min_size );
	swap( dim_mode,           i.dim_mode );
	swap( dim_order,          i.dim_order );
	swap( virtual_only,		  i.virtual_only );
	swap( not_used_only,	  i.not_used_only );
	swap( iterator_m,         i.iterator_m );
	swap( next_material_pack, i.next_material_pack );
	swap( next_pack_started,  i.next_pack_started );
}

void MIIterator::Init(DimensionVector &dimensions, bool do_rewind)
{
	if(no_dims > 0) {
		BHASSERT(fi == NULL && cur_pos == NULL, "Multiindex Iterator initialized again.");
		fi = new FilterOnesIterator* [no_dims];
		for(int i = 0; i < no_dims; i++)
			fi[i] = NULL;
		cur_pos = new _int64 [no_dims];
		cur_pack = new int [no_dims];
		checked_until = new _int64 [no_dims];
		nulls_found = new bool [no_dims];
		fill(cur_pos, cur_pos + no_dims, NULL_VALUE_64);
		fill(cur_pack, cur_pack + no_dims, -1);
		fill(checked_until, checked_until + no_dims, 0);
		for(int i = 0; i < no_dims; i++)
			nulls_found[i] = false;
	}
	////////// Find boundaries of IndexTable bigblocks
	bigblock_min_size = 100000;	// will be found below
	for(int i = 0; i < no_dims; i++)
		if(dimensions[i] && mind->t[i] && mind->t[i]->BlockShift()<bigblock_min_size)
			bigblock_min_size = mind->t[i]->BlockShift();
	if(bigblock_min_size == 100000)
		bigblock_min_size = mind->material_no_tuples;									// no materialized dimensions to sort
	else
		bigblock_min_size = (_int64(1)<<bigblock_min_size);

	/////////////////////////////////////////////////////////////////
	omitted_factor = 1;
	no_obj = 1;				// special case of no_obj = 0 will be tested at the end
	bool materialized_omitted = false;
	virtual_only = true;
	for(int i = 0; i < no_dims; i++)
		if(dimensions[i] && mind->GetIndTable(i)) {
			virtual_only = false;
			checked_until[i] = 0;
			break;
		}

	not_used_only = true;
	for(int i = 0; i < no_dims; i++) {
		if(dimensions[i] && mind->GetFilter(i) != NULL) {
			dim_mode.push_back(1);		// as filter
			no_obj = SafeMultiplication(no_obj, mind->GetFilter(i)->NoOnes());
			not_used_only = false;
		} else if(!virtual_only && mind->GetIndTable(i)) {
			dim_mode.push_back(2);		// as materialized
			mind->LockForGetIndex(i);
			not_used_only = false;
		} else {
			dim_mode.push_back(0);		// not used
			if(mind->GetFilter(i) != NULL)
				omitted_factor = SafeMultiplication(omitted_factor, mind->GetFilter(i)->NoOnes());
			else
				materialized_omitted = true;
		}
	}
	if(!virtual_only)
		no_obj = SafeMultiplication(no_obj, mind->NoMaterialTuples());
	if(not_used_only)
		no_obj = mind->NoTuples();
	if(omitted_factor != NULL_VALUE_64 && materialized_omitted && virtual_only) 		// all materialized are omitted
		omitted_factor = SafeMultiplication(omitted_factor, mind->NoMaterialTuples());

	vector< pair<int,int> > ordering_filters;
	for(int i = 0; i < no_dims; i++)
		if(dim_mode[i] == 1)				// filters first (note that this order is assumed in operator++)
			ordering_filters.push_back(pair<int,int>(65537 - mind->GetFilter(i)->DensityWeight(), i));
	sort(ordering_filters.begin(), ordering_filters.end());	// order filters starting from the densest one (for pack decompression efficiency)

	for(uint i = 0; i < ordering_filters.size(); i++)	// copy filter dim. numbers from sorter
			dim_order.push_back(ordering_filters[i].second);

	for(int i = 0; i < no_dims; i++)
		if(dim_mode[i] == 2)				// materialized next
			dim_order.push_back(i);
	if(do_rewind) {
		Rewind();
		if(!IsValid())
			no_obj = 0;
	}
}

MIIterator::~MIIterator()
{
	mind->IteratorUnlock();
	for(int i = 0; i < no_dims; i++) {
		delete fi[i];
		if(dim_mode[i] == 2)
			mind->UnlockFromGetIndex(i);
	}
	delete [] fi;
	delete [] cur_pos;
	delete [] cur_pack;
	delete [] checked_until;
	delete [] nulls_found;
	if(mind_created_locally)
		delete mind;
}

void MIIterator::RewindFilterDim(int i)
{
	if(!fi[i])
		fi[i] = new FilterOnesIterator(mind->GetFilter(i));
	else
		fi[i]->Init(mind->GetFilter(i));	// = Rewind()
	if(fi[i]->IsValid()) {				// 0 tuples?
		valid = true;
		cur_pos[i] = **(fi[i]);
	} else {
		valid = false;
	}
}

void MIIterator::Rewind()
{
	valid = false;
	if(mind->index_type == MultiIndex::IT_COUNT) 		// only number of rows may be checked on invalid iterator
		return;
	pack_size_left = 0;
	iterator_m = 0;
	next_material_pack = 0;
	fill(checked_until, checked_until + no_dims, 0);
	next_pack_started = true;
	if(not_used_only) {
		valid = true;
		pack_size_left = mind->NoTuples();
	} else {
		for(int i = 0; i < no_dims; i++) {
			cur_pos[i] = NULL_VALUE_64;
			if(dim_mode[i] == 1) {
				RewindFilterDim(i);
				if(!valid)
					break;
			} else if(dim_mode[i] == 2){
				if(mind->material_no_tuples == 0) {
					valid = false;				// empty material part
					break;
				}
				_uint64 result = mind->t[i]->Get64(0);
				if(result == 0)
					result = NULL_VALUE_64;
				else
					result -= 1;
				cur_pos[i] = result;
				checked_until[i] = 0;
				valid = true;
			} 
		}
	}
	for(uint i = 0; i < (uint)no_dims; i++)
		if(cur_pos[i] != NULL_VALUE_64)
			cur_pack[i] = (cur_pos[i] >> 16);
	InitNextPackrow();
}

MIIterator& MIIterator::operator++()
{
	if(not_used_only) {
		next_pack_started = false;
		--pack_size_left;
		if(pack_size_left == 0)
			valid = false;
		return *this;
	}
	next_pack_started = false;
	int dimndx = 0;
	bool done = false;
	int dim_order_size = dim_order.size();
	// assumption: first go all filters, then all materialized dimensions
	while(!done) {
		int idit = dim_order[dimndx];
		if(dim_mode[idit] == 1) {		// dimensions as filters - use filter iterators
			++(*fi[idit]);
			if(fi[idit]->IsValid()) {
				done = true;
				cur_pos[idit] = **(fi[idit]);			// update dimensions which had changed
			} else {
				fi[idit]->Rewind();
				if(fi[idit]->IsValid())				// we should check it again, because the filer might become empty
					cur_pos[idit] = **(fi[idit]);		// update dimensions which had changed
			}
		}
		else {	// dim_mode == 2  => materialized dimensions, use one common counter
			iterator_m++;
			if(iterator_m >= (_int64)(mind->material_no_tuples)) {
				valid = false;
				break;		// no more tuples at all
			}
			while(dimndx < dim_order_size) {		// fill the rest of dimensions with current materialized position
				idit = dim_order[dimndx];
				_uint64 result = mind->t[idit]->Get64(iterator_m);
				if(result == 0)
					result = NULL_VALUE_64;
				else
					result -= 1;
				cur_pos[idit] = result;
				++dimndx;
			}
			break;
		}
		++dimndx;
		if(!done &&  dimndx >= dim_order_size) {
			valid = false; 		// no more tuples at all
			done = true;
		}
	}
	if(pack_size_left != NULL_VALUE_64) {		// else the packrow is too large to be iterated through
		--pack_size_left;
		if(pack_size_left == 0)
			InitNextPackrow();
	}
	return *this;
}

void MIIterator::InitNextPackrow()		// go forward until the end of current packrow is found, set pack_size_left
{
	// assumption: pack_size_left is accurate only if the method is used at the first row of a packrow.
	if(!valid || not_used_only)
		return;
	pack_size_left = 1;
	next_pack_started = true;
	int dim_order_size = dim_order.size();
	for(int i = 0; i < no_dims; i++)
		nulls_found[i] = false;
	int ditndx = 0;		// iterate through dimensions
	bool done = false;

	next_material_pack = mind->material_no_tuples;		// upper tuple of the current packrow - to be refined below
	if( bigblock_min_size < mind->material_no_tuples &&
		iterator_m < bigblock_min_size * (mind->material_no_tuples / bigblock_min_size)) {
		// this is the first tuple from the next bigblock:
		next_material_pack = bigblock_min_size * (iterator_m / bigblock_min_size + 1);
	}
	// Iterate through the tuples
	// assumption: first go all filters, then all materialized dimensions
	while(!done) {
		int idit = dim_order[ditndx];
		if(dim_mode[idit] == 1) {		// dimensions as filters - use filter iterators
			_int64 loc_pack_size_left = fi[idit]->GetPackSizeLeft();
			pack_size_left = SafeMultiplication(pack_size_left, loc_pack_size_left);
//			if(loc_pack_size_left != fi[*dit].GetTotalSize())		// more than one pack
			if(!fi[idit]->InsideOnePack())		// more than one pack - faster version
				done = true;
		}
		else {	// dim_mode == 2  => materialized dimensions, use one common counter
			_int64 loc_iterator = checked_until[idit];
			assert(checked_until[idit] >= iterator_m);
			if(checked_until[idit] == iterator_m ) {
				int loc_pack = -1;
				while(loc_iterator < next_material_pack && 	// find the first non-NULL row (NULL row is when Get64() = 0)
					  mind->t[idit]->Get64(loc_iterator) == 0) {
					++loc_iterator;
					nulls_found[idit] = true;
				}
				if(loc_iterator < next_material_pack) {
					loc_pack = (mind->t[idit]->Get64(loc_iterator) - 1) >> 16;
					_uint64 ndx;
					while(loc_iterator < next_material_pack && 		// find the first non-NULL row from another pack
							((ndx = mind->t[idit]->Get64(loc_iterator)) == 0 ||
							 ((ndx - 1) >> 16) == loc_pack)) {
						if(ndx == 0)
							nulls_found[idit] = true;
						++loc_iterator;
					}
				}
				checked_until[idit] = loc_iterator;
			}
			if(next_material_pack > loc_iterator)
				next_material_pack = loc_iterator;
		}
		++ditndx;
		if(!done && ditndx >= dim_order_size) {		// no more dimensions
			// take into account the data pack size for the current materialized packrow, if any
			if(mind->material_no_tuples > 0 && !virtual_only)
				pack_size_left = SafeMultiplication(pack_size_left, next_material_pack - iterator_m);
			done = true;
		}
	}
	ditndx = 0;
	// Update pack numbers
	while(ditndx < dim_order_size) {
		int idit = dim_order[ditndx];
		if(dim_mode[idit] == 1) 		// dimensions as filters - use filter iterators
			cur_pack[idit] = (*(*(fi[idit])) >> 16);		// update dimensions which had changed
		else {		// dim_mode == 2
			_int64 loc_iterator = iterator_m;
			while(loc_iterator < next_material_pack && mind->t[idit]->Get64(loc_iterator) == 0)
				++loc_iterator;		// the first non-null record
			if(loc_iterator < next_material_pack)
				cur_pack[idit] = ((mind->t[idit]->Get64(loc_iterator) - 1) >> 16);
			else
				cur_pack[idit] = -1;
		}
		++ditndx;
	}
}

void MIIterator::NextPackrow()
{
	if(!valid || not_used_only)
		return;
	vector<int>::iterator dit = dim_order.begin();		// iterate through dimensions
	bool done = false;
	// Iterate through the tuples
	// assumption: first go all filters, then all materialized dimensions
	while(!done) {
		if(dim_mode[*dit] == 1) {		// dimensions as filters - use filter iterators
			fi[*dit]->NextPack();
			pack_size_left = SafeMultiplication(pack_size_left, fi[*dit]->GetPackSizeLeft());
			if(fi[*dit]->IsValid())
				done = true;
			else
				fi[*dit]->Rewind();
			if(fi[*dit]->IsValid())
				cur_pos[*dit] = **(fi[*dit]);			// update dimensions which had changed
			else {
				valid = false;
				break;
			}
		} else {	// dim_mode == 2  => materialized dimensions, use one common counter
			iterator_m = next_material_pack;
			if(iterator_m >= (_int64)(mind->material_no_tuples)) {
				valid = false;
				break;		// no more tuples at all
			}
			while(dit != dim_order.end()) {		// fill the rest of dimensions with current materialized position
				_uint64 result = mind->t[*dit]->Get64(iterator_m);
				if(result == 0)
					result = NULL_VALUE_64;
				else
					result -= 1;
				cur_pos[*dit] = result;
				++dit;
			}
			break;
		}
		++dit;
		if(!done &&  dit == dim_order.end()) {
			valid = false; 		// no more tuples at all
			done = true;
		}
	}
	InitNextPackrow();		// ignored if invalid
}

bool MIIterator::WholePack(int dim) const
{
	// the only possibility: just one virtual dimension, all other must be omitted
	return (dim_order.size() == 1 && mind->f[dim] && mind->f[dim]->IsFull(cur_pack[dim]));
}

///////////////////////////////////////////////////////////////////////////////////////////////

void MIDummyIterator::Combine(const MIIterator& sec)		// copy position from the second iterator (only the positions used)
{
	for(int i = 0; i < no_dims; i++)
		if(sec.DimUsed(i)) {
			cur_pos[i] = sec[i];
			cur_pack[i] = sec.GetCurPackrow(i);
	}
}

void MIDummyIterator::Set(int dim, _int64 val)			// set a position manually
{
	assert(dim >= 0 && dim < no_dims);
	cur_pos[dim] = val;
	cur_pack[dim] = (val >> 16);
}

void MIDummyIterator::SetPack(int dim, int p)			// set a position manually
{
	assert(dim >= 0 && dim < no_dims);
	cur_pos[dim] = (_int64(p) << 16);
	cur_pack[dim] = p;
}
