/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef MIITERATOR_H_INCLUDED
#define MIITERATOR_H_INCLUDED

#include "MultiIndex.h"

/*! \brief Used for iteration on chosen dimensions in a MultiIndex.
 * Usage example:
 * MIIterator it(mind,...);
 * it.Rewind();
 * while(it.IsValid()) {
 *   ...
 *   ++it;
 * }
 *
 */
class MIIterator
{
public:
	/*! create an iterator for a MultiIndex and set to the \e mind begin
	 *
	 * \param mind the MultiIndex for which the iterator is created
	 * Example:
	 * 		DimensionVector dimensions(mind);
	 *		for(...)
	 *			dimensions[...we are interested in...] = true;
	 *		MIIterator mit(mind, dimensions);
	 *
	 */
	MIIterator(MultiIndex* mind, bool do_rewind = true);							// note: mind=NULL will create a local multiindex - one dim., one row
	MIIterator(MultiIndex* mind, int one_dimension, bool do_rewind = true);		// -1 is an equivalent of all dimensions
	MIIterator(MultiIndex* mind, DimensionVector &dimensions, bool do_rewind = true);
	MIIterator(MultiIndex* mind, std::set<int> dimensions, bool do_rewind = true);

	MIIterator(const MIIterator&);
	~MIIterator();

	/*! Get the array of row numbers for each dimension in the MultiIndex for the current
	 * position of the iterator.
	 *
	 * \return Returns a pointer to a member array of row numbers with its length equal to the number of dimensions
	 * in the associated MultiIndex. Only the elements corresponding to dimensions used for MIIterator
	 * creation are set by the iterator, the rest is left as NULL_VALUE_64.
	 * Cannot be used at the end of MultiIndex (check IsValid() first).
	 */
	const _int64* operator*() const
	{ BHASSERT_WITH_NO_PERFORMANCE_IMPACT(valid); return cur_pos; }

	/*! Get the current value of iterator for the given dimension.
	 *
	 * \return Returns a row number in the associated MultiIndex.
	 * Only the elements corresponding to dimensions used for MIIterator
	 * creation are set by the iterator, the rest is left as NULL_VALUE_64.
	 * Cannot be used at the end of MultiIndex (check IsValid() first).
	 */
	const _int64 operator[](int n) const
	{ BHASSERT_WITH_NO_PERFORMANCE_IMPACT(valid && n >= 0 && n < no_dims); return cur_pos[n]; }

	//! move to the next row position within an associated MultiIndex
	MIIterator& operator++();

	//! set the iterator to the begin of the associated MultiIndex
	void Rewind();
	/*! True if the next call to operator* will return a valid result. False e.g. if end has been reached.
	 */
	bool IsValid() const					{ return valid; }

	/*! Get the number of rows omitted in unused dimensions on each step of iterator. Used e.g. as a factor in sum(...), count(...).
	 *
	 * \return Returns either a number of omitted rows,
	 * or NULL_VALUE_64 if the number is too big (> 2^63).
	 * \pre The result is correct for an initial state of multiindex (no further changes apply).
	 */
	_int64 Factor() const					{ return omitted_factor; }
	/*! Get the number of rows the iterator will iterate through.
	 *
	 * \return Returns either a number of rows to iterate,
	 * or NULL_VALUE_64 if the number is too big (> 2^63).
	 * \pre The result is correct for an initial state of multiindex (no further changes apply).
	 */
	_int64 NoTuples() const					{ return no_obj; }
	//! How many rows (including the current one) are ahead of the current position in the current packrow
	_int64 GetPackSizeLeft() const			{ return pack_size_left; }

	/*! True, if there were any null object numbers (NOT null values!) possibly found in the current packrow for a given dimension
	 * \warning True may mean that the null value is possible, but does not guarantee it!
	*/
	bool NullsPossibleInPack(int dim) const		{ return nulls_found[dim]; }

	bool NullsPossibleInPack() const
	{
		for(int i = 0; i < no_dims; i++) {
			if(nulls_found[i]) return true;
		}
		return false;
	}

	/*! True, if the current packrow contain exactly the whole pack for a dimension (no repetitions, no null objects)
	*/
	bool WholePack(int dim) const;

	/*!  Go to the beginning of the next packrow
	 * If the end is reached, IsValid() becomes false.
	 */
	void NextPackrow();

	/*! Get pack numbers for each dimension in the associated MultiIndex for the current position.
	 * \return Returns a pointer to a member array of pack numbers with its length equal to the number of dimensions
	 * in the associated MultiIndex. Only the elements corresponding to dimensions used for MIIterator
	 * creation are set by the iterator.
	 */
	const int* GetCurPackrow() const		{ return cur_pack; }
	const int GetCurPackrow(int dim) const	{ return cur_pack[dim]; }

	const int GetCurInpack(int dim) const
	{ BHASSERT_WITH_NO_PERFORMANCE_IMPACT(valid && dim >= 0 && dim < no_dims); return int(cur_pos[dim] & 0x000000000000FFFFLL); }

	//! True, if the current row is the first one in the packrow
	bool PackrowStarted()					{ return next_pack_started; }
	void swap(MIIterator&);

	std::vector<int> &DimOrder()			{ return dim_order; }		// a list of dimensions the iterator is using

	bool DimUsed(int dim) const				{ return (dim_mode[dim] != 0); }	// is the dimension used in iterator?

	int NoDimensions() const				{return no_dims;}

protected:		////////////////////////////////////////////////////////////////////////////////

	///////////////////////////////////////////////////////////////////////////////
	MultiIndex*		mind;			// a multiindex the iterator is created for
	bool 			mind_created_locally;	// true, if the multiindex should be deleted in descructor
	int				no_dims;		// number of dimensions in the associated multiindex

	_int64*		cur_pos;		// a buffer storing the current position
	_int64*		checked_until;		// for each dimension the index IndexTable has been scanned to

	int*			cur_pack;		// a buffer storing the current pack number
	bool*			nulls_found;		// true if there were any null objects found in the current packrow
	FilterOnesIterator**	fi;			// a vector of iterators for dimensions (will be not used for non-filter dim.)
	bool			valid;			// false if we are already out of multiindex

	_int64			omitted_factor;	// how many tuples in omitted dimensions are skipped at each iterator step
									// NULL_VALUE_64 means more than 2^63
	_int64			no_obj;			// how many steps (tuples) will the iterator perform
									// NULL_VALUE_64 means more than 2^63
	_int64			pack_size_left;	// how many steps (tuples) will the iterator
									// perform inside the current packrow, incl. the current one
	_int64			bigblock_min_size;	// optimization tip: how many rows is stored in one IndexTable buffer

	////// iterative stuff
	std::vector<int>			dim_mode;	// 0 - not used, 1 - filer, 2 - materialized
	std::vector<int>			dim_order;	// an order of iterating dimensions
	bool				virtual_only;		// all used dimensions are filters
	bool				not_used_only;	    // all dimensions are unused
	_int64				iterator_m;	// iterator through the materialized part of multiindex (only one possible for now)
	_int64				next_material_pack;	// the beginning (iterator_m position) of the next packrow in materialized part
	bool				next_pack_started;	// true if the current row is the first one inside the packrow
private:
	void Init(DimensionVector &dimensions, bool do_rewind = true);
	void InitNextPackrow();		// find boundaries of the current packrow,
								// assumption: pack_size_left is properly updated only if the method is used at the first row of a packrow.
	virtual void RewindFilterDim(int i);	//Initialize iterator on a filter

};

/*! \brief Used for storing position of another iterator and for combining positions from several iterators.
 *
 */
 class MIDummyIterator : public MIIterator
{
public:
	MIDummyIterator(MultiIndex* mind) : MIIterator(mind) 		{ valid = true; }
	MIDummyIterator(const MIIterator& sec) : MIIterator(sec) 	{ valid = true; }
	MIIterator& operator++()		{ assert(0); }		// should not be used for dummy iterator!

	void Combine(const MIIterator& sec);	// copy position from the second iterator (only the positions used)
	void Set(int dim, _int64 val);			// set a position manually
	void SetPack(int dim, int p);			// set a position manually
};

//! Iterate using a provided sequence of datapacks
//! To be specified later
//class MIOrderedIterator : public MIIterator
//{
//public:
//	/* Create a copy of \e mit, pointing to the first row in associated multiindex,
//	 * with pack traversal ordering given by \e pack_numbers
//	 * \param pack_numbers - desired order of rowpack traversal, to be specified later.
//	 *
//	 */
//	MIOrderedIterator(MIIterator& mit, int* pack_numbers);
//
//
//};

//! Used to iterate on a packrow
class MIPackIterator: public MIIterator
{
public:

	/*! Create the pack iterator on a packrow pointed by \e it
	 *  The created pack iterator will point to the same tuple as \e it (not necessarily the first in the packrow)
	 */
	MIPackIterator(MIIterator& it);

	/*! Get the list row numbers for each dimension in the MultiIndex for the current
	 * position of the iterator.
	 *
	 * \return Returns a pointer to a member array of row numbers with its length equal to the number of dimensions
	 * in the associated MultiIndex. Only the elements corresponding to dimensions used for MIIterator
	 * creation are set by the iterator.
	 * If end of the current packrow is reached, then returned row numbers are equal to -1
	 */
	const _int64* operator*() const;

	//! set the iterator to the begin of the current packrow
	void Rewind();

	/*! Go to the beginning of the next packrow (advance the current packrow)
	 *  \return  Returns a pointer to a member array of row numbers with its length equal to the number of dimensions
	 * in the associated MultiIndex. Only the elements corresponding to dimensions used for MIIterator
	 * creation are set by the iterator.
	 * If end is reached, then returned row numbers are equal to -1
	 */
	const _int64* NextPack();

};

#endif /*MIITERATOR_H_INCLUDED*/
