/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef MIUPDATINGITERATOR_H_INCLUDED
#define MIUPDATINGITERATOR_H_INCLUDED

#include "MIIterator.h"

/*! \brief Used for iteration on chosen dimensions in a MultiIndex and for
 *
 * Usage example:
 *
 * MIUpdatingIterator it(mind,...);
 * it.Rewind();
 * while(it.IsValid()) {
 *   ...
 *   if(...condition not met...)
 *       it.DeleteCurrent();			// deletion may be delayed - do not access multiindex by other iterators!
 *   ...
 *   ++it;
 * }
 * it.Commit();		// must be done to ensure
 *
 */
class MIUpdatingIterator : public MIIterator
{
public:
	MIUpdatingIterator(MultiIndex* mind);
	MIUpdatingIterator(MultiIndex* mind, int one_dimension);
	MIUpdatingIterator(MultiIndex* mind, DimensionVector &dimensions);
	MIUpdatingIterator(MultiIndex* mind, std::set<int> dimensions);

	MIUpdatingIterator(const MIUpdatingIterator&);		// Heavy object! Do not copy it.
	~MIUpdatingIterator();

	/*! Exclude the current iterator position from the original multiindex.
	 *  Will be physically deleted not later than on Commit(),
	 *  but earlier deletion is also possible.
	 */
	void ResetCurrent();

	/*! Exclude the current packrow (defined by GetCurPackrow()) from the original multiindex.
	 *  Will be physically deleted not later than on Commit(), but usually immediately.
	 */
	void ResetCurrentPack();

	/*! Confirm deletion of all rows excluded in the original multiindex.
	 *  Must be used at the end of operations, otherwise some deletions may be lost.
	 */
	void Commit();

	/*! A number of tuples still existing after all deletions performed.
	 *  This method uses CommitDelayed, so it may be slow.
	 */
	_int64 UpdatedNoTuples();

	// These methods are overloaded to add new functionality:

	void Rewind();
	MIUpdatingIterator& operator++();
	void NextPackrow();

private:
	void PrepareUpdater();

	bool changed;							// true if anything was reset; false on the beginning or after Commit()
	int one_dim;							// a number (or -1) of dimension in case of one dimensional iterator
	Filter *one_dim_filter;					// external pointer; if not null, then only this dimension od
	Filter *multi_dim_filter;				// internal filter of this->NoTuples() bits; a delete mask for multiindex
	_int64 multi_filter_pos;				// current position of multidimensional filter (0..no_obj-1)
	_int64 multi_filter_pack_start;			// a position of the current rowpack start in multidimensional filter (0..no_obj-1)
};

#endif /* MIUPDATINGITERATOR_H_INCLUDED */

