/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef MULTIINDEX_H_INCLUDED
#define MULTIINDEX_H_INCLUDED

#include "bintools.h"
#include "Filter.h"
#include "CQTerm.h"
#include "system/CacheableItem.h"
#include "system/MemoryManagement/TrackableObject.h"
#include <vector>

//#include "system/ConnectionInfo.h"

class MultiSorter;
class MIIterator;
class MultiIndex;

class DimensionVector
{
public:
	DimensionVector();
	DimensionVector(int no_dims);
	DimensionVector(MultiIndex *);
	DimensionVector(const DimensionVector &sec);
	~DimensionVector();
	DimensionVector &operator=(const DimensionVector &sec);

	bool &operator[](int i)					{ assert(i >= 0 && i < size); return v[i]; }

	void Clean();							// set all dimensions to false
	void SetAll();							// set all dimensions to true (present)
	bool Intersects(DimensionVector &sec);	// true if any common dimension present
	bool Includes(DimensionVector &sec);	// true if all dimensions from sec are present in *this
	void Minus(DimensionVector &sec);		// exclude from *this all dimensions present in sec
	void Plus(DimensionVector &sec);		// include in *this all dimensions present in sec

	int NoDimsUsed();						// return a number of present dimensions
	int Size()								{ return size; }	// return a number of all dimensions (present or not)

private:
	int size;
	bool *v;
};

class IndexTable : private CacheableItem, public TrackableObject
{
public:
	IndexTable(_int64 _size,int _bytes_per_value,int mem_modifier);
	// size - number of objects, bytes_... = 2, 4 or 8
	// mem_modifier - default 0, may also be +1,2... or -1,2..., when positive => IndexTable will take more memory for a buffer
	IndexTable(IndexTable &sec);
	~IndexTable();

	void   Set64(_uint64 n,_uint64 val);
	inline _uint64 Get64(_uint64 n)
			{
				BHASSERT_WITH_NO_PERFORMANCE_IMPACT( IsLocked() );
				assert(n<size);	// warning C4018: '<' : signed/unsigned mismatch
				int b = int(n>>block_shift);
				if(b!=cur_block) LoadBlock(b);
				_uint64 ndx = n & block_mask;
				_uint64 res;
				if(bytes_per_value==8)		res= ((_uint64*)buf)[ndx];
				else if(bytes_per_value==4)	res= ((unsigned int*)buf)[ndx];
				else 						res= ((unsigned short*)buf)[ndx];
				return res;
			}

	int BytesPerValue()			{ return bytes_per_value; }
	_uint64 N()					{ return size; }			// note: this is the upper size, the table can be used partially!
	int BlockShift()			{ return block_shift; } 	// block = int( tuple >> block_shift )

	void ExpandTo(_int64 new_size);

	////////////////////////// Memory management  ////////////////////////
	virtual int CurrentSize();
	virtual int Compress();		// the first stage of conserving memory: delete all redundant data (retrievable from the memory)
	virtual int Collapse();		// the last stage of conserving memory: delete all data, must be reloaded from disk on next use
	void CopyLocksToIT(int d) {
		for(int i = 0; i <= d; ++i)
			Lock();
	}

	//TrackableObject functionality
	TRACKABLEOBJECT_TYPE TrackableType(){return TO_INDEXTABLE;}

private:
	void LoadBlock(int b);

	unsigned char *buf;				// polymorphic: unsigned short, unsigned int or _int64

	size_t max_buffer_size_in_bytes;
	int bytes_per_value;
	int max_block_used;
	size_t buffer_size_in_bytes;
	int block_shift;
	_uint64 block_mask;
	_uint64 size;

	int  cur_block;
	bool block_changed;
	ConnectionInfo& m_conn;		// external pointer
};

///////////////////////////////////////////////////////////////////////////////////
class MINewContents;					// a class to store new (future) contents of multiindex
class MIRoughSorter;

class MultiIndex
{
public:
	MultiIndex();
	MultiIndex(const MultiIndex &s);
	~MultiIndex();

	void Copy(MultiIndex &s);			// copy the object s
	void Canibalize(MultiIndex &s);		// copy the object s by moving pointers, then reset s (faster than Copy())

	void Clear();						// clear the multiindex completely (like destructor + Multiindex())

	// Construction of the index

	// max_value - upper limit of indexes of newly added objects (e.g. number of all objects in a table to be joined)
	void AddDimension_constant(_uint64 const_val, _uint64 size);	// set a constant value in the new column (does not increase the number of tuples); 0 means null, k means the object k-1
	void AddDimension_cross(_uint64 size);		// calculate a cross product of the previous value of index and the full table (trivial filter) of 'size' objects
	void AddDimension_cross(const Filter& new_f);		// calculate a cross product of the previous value of index and the new set of objects (defined as Filter, to be copied inside)
	void AddDimension_cross(MultiIndex &new_f);	// calculate a cross product of the previous value of index and the other one

	//////////// retrieve information

	int NoDimensions() const { return no_dimensions; }	// number of dimensions
	_uint64 NoTuples() const {
		if(!no_tuples_too_big) return no_tuples;
		throw OutOfMemoryRCException("Too many tuples.");
		return 0; }			// number of all tuples

	bool ZeroTuples()		{ return ( !no_tuples_too_big && no_tuples == 0 ); }
	bool TooManyTuples()	{ return no_tuples_too_big; }
	_uint64 NoMaterialTuples()	{ return material_no_tuples; }		// number of all tuples in materialized part

	void SetFilter(Filter *f_in, int dim, bool &nulls_found);  // set 1 in the given filter on every position present in this MultiIndex
	bool SetFilterUnique(Filter *f_in, int dim);  // set 1 on every position present in this MultiIndex; return false if any repetition or null found
	void LockForGetIndex(int dim);				// must precede GetIndex(...)
	void UnlockFromGetIndex(int dim);

	Filter* GetFilter(int dim) const		// Get the pointer to a filter attached to a dimension. NOTE: will be NULL in case of materialized MultiIndex!
		{ assert(dim<no_dimensions && dim>=0); return f[dim]; }
	IndexTable* GetIndTable(int dim) const		// Get the pointer to a index table attached to a dimension. NOTE: will be NULL in case of virtual MultiIndex!
	{ assert(dim<no_dimensions && dim>=0); return t[dim]; }
	bool NullsExist(int dim);				// return true if there exist any 0 value (always false for virtual dimensions)
	bool MarkInvolvedMaterialClasses(DimensionVector &v);	// if any materialized dimension is marked, then mark the rest of this class. Return true if anything new marked.
	void MarkMaterialClasses(DimensionVector &v);
	bool ContainsMaterialClasses(DimensionVector &v);

	//////////// iterated access to the index (optimized, to be used especially in case of virtual indexes)

	void LockAllForUse();				// must precede Rewind(), Next(), GetCurrent(int)
	void UnlockAllFromUse();

	/////////////////////////////////////////////////////////////////////////////////

	_uint64 DimSize(int dim);			// the size of one dimension: material_no_tuples for materialized, NoOnes for virtual
	_uint64 OrigSize(int dim)			{ return dim_size[dim]; }
										// the maximal size of one dimension (e.g. the size of a table, the maximal index possible)

	bool IteratorLock()					// register a normal iterator; false: already locked for updating
	{
		if(iterator_lock > -1)
			iterator_lock++;
		return (iterator_lock > -1);
	}
	bool IteratorUpdatingLock()			// register an updating iterator; false: already locked
	{
		if(iterator_lock == 0) {
			iterator_lock = -1;
			return true;
		}
		return false;
	}
	void IteratorUnlock()
	{
		if(iterator_lock > 0)
			iterator_lock--;
		else
			iterator_lock = 0;
	}

	/////////// operations on the index

	void MarkTuplesPresent(::Filter &f, MultiIndex &sec);	// f - a filter (empty by assumption) equal in size with sec iterated by default MIIterator
															// result: mark in f all tuples of sec which are present in this.
	void Or(MultiIndex &s);		// enlarge the multiindex by adding all tuples from the second one (excluding repetitions)
	void And(MultiIndex &s);	// limit the multiindex by excluding all tuples from the second one
	void Not();					// negation: a tuple is present in F <=> it is absent in F.Not()
	void MIFilterAnd(MIIterator &mit, Filter &fd);	// limit the MultiIndex by excluding all tuples which are not present in fd, in order given by mit

	bool CanBeDistinct(int dim)	const	{ return can_be_distinct[dim]; }	// true if ( distinct(orig. column) => distinct( result ) ), false if we cannot guarantee this
	bool IsUsedInOutput(int dim)		{ return used_in_output[dim]; }	// true if the dimension is used in output columns
	void SetUsedInOutput(int dim)		{ used_in_output[dim] = true; }
	void ResetUsedInOutput(int dim)		{ used_in_output[dim] = false; }
	void ForgetSoon(int dim)			{ forget_now[dim] = true; }		// delete this dimension as soon as possible (e.g. after the next joining)
	bool IsForgotten(int dim)			{ return forget_now[dim]; }
	void DontForget();					// Revoke current optimizations (dims may be still needed)

	void Materialize();		// materialize the index by translating f -> t, filters are no longer needed and may be deleted
	void Materialize(DimensionVector &dims_to_materialize);	// materialize the listed dimensions by translating f -> t
	void Empty(int dim_to_make_empty = -1);	// make an index empty (delete all tuples) with the same dimensions
												// if parameter is set, then do not delete any virtual filter except this one
	void MakeDimensionNull(int dim);	// transform the given dimension into a table of null objects
	void UpdateNoTuples();		// recalculate the number of tuples
	bool IsCountOnly()					{ return (index_type == IT_COUNT); }
	void MakeCountOnly(_int64 mat_tuples = -1);	// set the number of tuples, deleting all existing dimensions; default: leave the current no. of tuples
	void MakeCountOnly(_int64 mat_tuples, DimensionVector &dims_to_materialize);
								// recalculate the number of tuples, assuming mat_tuples is the new material_no_tuples and the dimensions from the list are deleted
	void SortSelfRoughly();		// sort the tuples using objects' pack numbers as keys, locally in bigblocks

	////////////////////////// Manual operations on index  ////////////////////////

	void AddNewMaterializedTuple(DimensionVector &dims, const _int64 *row);

	ConnectionInfo& ConnInfo() const { return *m_conn; }

	///////////////////////////////////////////////////////////////////////////////////

	ConnectionInfo* m_conn;

	friend class MINewContents;
	friend class MIIterator;
	friend class MIRoughSorter;

private:
	enum IndexType { IT_VIRTUAL, IT_MIXED, IT_ORDERED, IT_COUNT }	index_type;
				// index types:
				// IT_VIRTUAL - all dimensions are defined as filters (tuples' order is undefined)
				// IT_MIXED   - at least one dimension is defined as a filter (tuples' order is undefined)
				// IT_ORDERED - the index is fully materialized and may be ordered
				// IT_COUNT   - only no_tuples is proper, all other data in filter are invalid; cannot use this index any longer (except for reading NoTuples())
	void AddDimension();				// declare a new dimension (internal)
	void AddTuple(_uint64 *v);			// add a tuple (given as a vector of objects numbers); not too effective

	//////
	void CheckIfVirtualCanBeDistinct();	// updates can_be_distinct table in case of virtual multiindex

	int no_dimensions;
	_int64 *dim_size;		// the size of a dimension
	_uint64 no_tuples;		// actual number of tuples (also in case of virtual index); should be updated in any change of index
	_uint64 material_no_tuples;	// current number of materialized tuples available in FreeBinTables
	bool	no_tuples_too_big;	// this flag is set if a virtual number of tuples exceeds 2^64
	std::vector<bool> can_be_distinct;	// true if the dimension contain only one copy of original rows, false if we cannot guarantee this
	std::vector<bool> used_in_output;	// true if given dimension is used for generation of output columns
	std::vector<bool> forget_now;		// true if the dimension is not used_in_output, and should be discarded at the first occassion

	// Index data.
	// Note: if f[i] is not null, then the cross product (natural ordering) is to be calculated
	// Otherwise the corresponding t[i] should not be null, defining actual values of tuples
	Filter			**f;
	IndexTable		**t;
	_uint64 mat_max_no_tuples;	// maximal number of materialized tuples available without rebuilding (size of FreeBinTables)

	// Iterator.
	_int64 iterator_m;			// the current position in the materialized part of index

	// Some technical functions
	_uint64 RoundUpTuples(_uint64 n);		// return a number slightly larger (or not less) than n; used for declaring a buffer with some free places for future use
	void MultiplyNoTuples( _uint64 factor );	// the same as "no_tuples*=factor", but set no_tuples_too_big whenever needed

	typedef std::vector<int> noop_locks_t;
	noop_locks_t _noOpLocks;

	int iterator_lock;			// 0 - unlocked, >0 - normal iterator exists, -1 - updating iterator exists
};

/////////////////////////////////////////////////////////////////////////////////////////////

class MINewContents					// a class to store new (future) contents of multiindex
{
public:
	MINewContents(MultiIndex *m);
	~MINewContents();

	void SetDimension(int dim);			// mark this dimension as to be involved
	void SetDimension(DimensionVector &dims);	// mark all these dimensions as to be involved
	void Init(_int64 initial_size);		// initialize temporary structures (set approximate size)
	void Commit(_int64 joined_tuples);	// commit changes to multiindex - must be called at the end, or changes will be lost
	void CommitCountOnly(_int64 joined_tuples);

	void SetNewTableValue(int dim, _int64 val);			// add a value (NULL_VALUE_64 is a null object index); if the index is larger than the current size, enlarge table automatically

private:
	void InitTnew(int dim, _int64 initial_size);		// create t_new of proper size
	void DisableOptimized();							// switch from optimized mode to normal mode

	MultiIndex		*mind;
	IndexTable		**t_new;			// new tables of some size
	_int64			*new_cur_used;		// position of the last index set by SetNewTableValue (plus 1)
	bool			*dim_involved;		// true for dimensions which are about to have new values
	bool			all_virtual;		// all involved dimensions were virtual

	// for optimized case
	int				optimized_dim_stay;	// the dimension which may remain a filter after join
	Filter 			*f_opt;				// new filter for optimized case
};

class MIRoughSorter
{
public:
	MIRoughSorter(MultiIndex *_m);
	~MIRoughSorter();

	void SortMultiIndexRoughly(_int64 limit = -1);	// sort multiindex wrt. packrows; if limit != -1, then sort up to limit rows

private:
	void SwitchMaterial(_uint64 tuple1, _uint64 tuple2);	// one transposition of the whole tuples (for sorting)

	void RoughQSort(_int64 start_tuple, _int64 stop_tuple);
	int RoughCompare( _int64 tuple1, _int64 tuple2);
	bool SortingNeededRC(_int64 bigblock_min_size);		// true if the current MultiIndex should be sorted (RCTable version)
	bool SortingNeededTT(_int64 bigblock_min_size);		// true if the current MultiIndex should be sorted (TempTable version)

	MultiIndex *mind;
	int rough_comp_bit_offset;
	int no_rough_cols_to_sort;
	int *rough_cols_to_sort;
};

#endif /* not MULTIINDEX_H_INCLUDED */

