/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef PACKORDERER_H_
#define PACKORDERER_H_

#include "common/CommonDefinitions.h"
#include <vector>
#include "vc/VirtualColumn.h"

/*!
 *
 */

class PackOrderer
{
public:
	enum OrderType {RangeSimilarity, MinAsc, MinDesc, MaxAsc, MaxDesc, NotSpecified};
	enum State {INIT_VAL = -1, END = -2};

	//! Uninitialized object, use Init() before usage
	PackOrderer() : otype(NotSpecified) {}

	/*!
	 * Create an orderer for given column,
	 * \param vc datapacks of this column will be ordered
	 * \param r_filter use only pack not eliminated by this rough filter
	 * \param order how to order datapacks
	 */
	PackOrderer(VirtualColumn* vc, RSValue* r_filter, OrderType order);

	~PackOrderer() {};

	//! Initialized Orderer constructed with a default constructor,
	//! ignored if used on a initialized orderer
	//! Must use Rewind* just after Init(), Init() sets Current = -1
	//! \return true if successful, otherwise false
	bool Init(VirtualColumn* vc, OrderType order, RSValue* r_filter = NULL);

	/*!
	 * Reset the iterator, so it will start from the first pack in the given sort order
	 */
	void Rewind() { 	
		if(packs.size() == 0)
			curndx = END;
		curndx = prevndx = INIT_VAL;
	}

	/*!
	 * Reset the iterator to the position associated with given datapack from the given column
	 */
	void RewindToMatch(VirtualColumn* vc, MIIterator& mit);

	/*!
	 * Reset the iterator, so it will start from the current position
	 */
	void RewindToCurrent();

	/*!
	 * the current datapack number in the ordered sequence
	 * \return pack number or -1 if end of sequence reached
	 */
	int Current() {return curndx < 0 ? -1 : packs[curndx].second;}

	/*!
	 * Advance to the next position in the ordered sequence, return the next datapack number
	 */
	PackOrderer& operator++();

	//! Is the end of sequence reached?
	bool IsValid() {return curndx >= 0;}

	//! Is the current pack the last one?
	bool IsLast() {return curndx == packs.size() - 1;}

	uint NoPacks() {return packs.size();}

	bool Initialized() {return otype != NotSpecified;}

	OrderType GetOrderType() {return otype;}

	struct OrderingInfo {
		OrderingInfo(VirtualColumn* vc,ColOperation op) : vc(vc), op(op) {}
		VirtualColumn* vc;
		ColOperation op; //distinguishes aggregations only, may need extensions for other operations
	};
	static void ChoosePackOrderer(PackOrderer& po, const vector<OrderingInfo>&);

private:
	enum MinmaxType {MMT_Fixed, MMT_Float, MMT_Double, MMT_String};

	typedef union MMTU {
		_int64 i;
		double d;
		float f;
		char c[8];
		MMTU(_int64 i) :i(i) {}
	} MMTU;

	typedef std::pair<MMTU,int> PackPair;

	class CompIntLess {
	public:
		inline bool operator() (const PackPair& v1, const PackPair& v2) const { return v1.first.i < v2.first.i; }
	};

	class CompIntGreater {
	public:
		inline bool operator() (const PackPair& v1, const PackPair& v2) const { return v1.first.i > v2.first.i; }
	};

	static CompIntLess comp_int_less;	//implemented comparison - sorting on packs[n].first.i ascending
	static CompIntGreater comp_int_greater;	//implemented comparison - sorting on packs[n].first.i desc

	vector<PackPair> packs;
	int curndx;			//Current() == packs[curndx]
	int prevndx;		// i = o.Current(); ++o; i == prevndx;
	bool lastly_left;	//if the last ++ went to the left
	MinmaxType mmtype;
	OrderType otype;
};


#endif /* PACKORDERER_H_ */
