/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "core/TempTable.h"
#include "ParametrizedFilter.h"
#include "vc/VirtualColumn.h"

using namespace std;

ParameterizedFilter::ParameterizedFilter(FilterType filter_type)
: table(NULL), table_alias(NULL_VALUE_32), tree(NULL), filter_type(filter_type)
{
	m_index = new MultiIndex();
}

ParameterizedFilter::ParameterizedFilter(const ParameterizedFilter &pf)
{
	parameters = pf.parameters;
	this->params_set = pf.params_set;
	descriptors = pf.descriptors;
	aliases = pf.aliases;
	parametrized_desc = pf.parametrized_desc;
	join_types = pf.join_types;
	log_connectors = pf.log_connectors;
	if(pf.m_index)
		m_index = new MultiIndex(*pf.m_index);
	else
		m_index = NULL;				// possible e.g. for a temporary data sources
	table = pf.table;
	table_alias = pf.table_alias;
	filter_type = pf.filter_type;
	for(size_t i = 0; i < pf.filters.size(); i++)
		filters.push_back(new ParameterizedFilter(*pf.filters[i]));
	tree = NULL;
	if(pf.tree)
		tree = new DescTree(*pf.tree);
}

ParameterizedFilter::ParameterizedFilter(TempTable *t, int alias, FilterType filter_type)
: filter_type(filter_type)
{
	m_index = new MultiIndex(*t->GetMultiIndexP());
	join_types = t->GetJoinTypes();
	aliases = t->GetAliases();
	table = t;
	table_alias = alias;
	tree = NULL;
}

ParameterizedFilter::~ParameterizedFilter(void)
{
	for(size_t i = 0; i < filters.size(); i++)
		delete filters[i];
	delete m_index;
	delete tree;
}

ParameterizedFilter &ParameterizedFilter::operator=(const ParameterizedFilter &pf)
{
	if(this != &pf) {
		parameters = pf.parameters;
		params_set = pf.params_set;
		descriptors = pf.descriptors;
		aliases = pf.aliases;
		parametrized_desc = pf.parametrized_desc;
		join_types = pf.join_types;
		log_connectors = pf.log_connectors;
		delete m_index;
		m_index = new MultiIndex(*pf.m_index);
		table = pf.table;
		table_alias = pf.table_alias;
		filter_type = pf.filter_type;
		for(size_t i = 0; i < pf.filters.size(); i++)
			filters.push_back(new ParameterizedFilter(*pf.filters[i]));
		delete tree;
		tree = NULL;
		if(pf.tree)
			tree = new DescTree(*pf.tree);
	}
	return *this;
}

// returns no. of parameters including parameters of all filters
uint ParameterizedFilter::NoParameters()
{
	uint no_params = (uint)parameters.size();
	for(uint i = 0; i < filters.size(); i++)
		no_params += filters[i]->NoParameters();
	return no_params;
}

uint ParameterizedFilter::NoUnsetParameters()
{
	uint no_params = (uint)(parameters.size() - params_set.size());
	for(uint i = 0; i < filters.size(); i++)
		no_params += filters[i]->NoUnsetParameters();
	return no_params;
}

bool ParameterizedFilter::IsParameterSet(uint i)
{
	if(i < parameters.size()) {
		for(size_t j = 0 ; j < params_set.size(); j++)
			if(params_set[j] == i)
				return true;
		return false;
	}
	i -= (uint)parameters.size();
	for(uint j = 0; j < filters.size(); j++) {
		if(i < filters[j]->NoParameters())
			return filters[j]->IsParameterSet(i);
		i -= filters[j]->NoParameters();
	}
	assert(0);
	return false;
}

CQTerm ParameterizedFilter::GetParameter(uint i)
{
	assert(i<NoParameters());
	if(i < parameters.size())
		return parameters[i];
	i -= (uint)parameters.size();
	for(uint j = 0; j < filters.size(); j++) {
		if(i < filters[j]->NoParameters())
			return filters[j]->GetParameter(i);
		i -= filters[j]->NoParameters();
	}
	assert(0);
	return CQTerm();
}

void ParameterizedFilter::ClearParamsSet()
{
	params_set.clear();
	for(uint j = 0; j < filters.size(); j++)
		filters[j]->ClearParamsSet();
}

bool ParameterizedFilter::ExtractParameters(CQTerm& e)
{
	if(!e.vc)
		return false;
    bool has_parameter = false;
    if (e.vc->GetParams().size() > 0 )
    	return true;
    else
    	return false;
}

void ParameterizedFilter::AddDescriptor(CQTerm e1, Operator pr, CQTerm e2, CQTerm e3, TempTable* t, int outer_dim, bool pAND)
{
	assert(pAND);	// NOTE: OR may be not implemented on some lower levels! Only valid in HAVING tree.
	bool is_parameterized = false;
	Descriptor desc(t);
	desc.op = pr;
	desc.lop = pAND ? O_AND : O_OR;
	desc.outer_dim = outer_dim;
	is_parameterized = ExtractParameters(e1) || ExtractParameters(e2) || ExtractParameters(e3);

	desc.attr = e1;
	desc.val1 = e2;
	desc.val2 = e3;
	desc.CalculateJoinType();

	if(is_parameterized)
		parametrized_desc.push_back(desc);
	else
		descriptors.push_back(desc);
	return;
}

bool ParameterizedFilter::IsParametrized()
{
	if(parametrized_desc.size() > 0)
		return true;
	for(uint i = 0; i < filters.size(); i++)
		if(filters[i]->IsParametrized())
			return true;
	return false;
}

void ParameterizedFilter::ProcessParameters()
{
	descriptors.clear();
	for(uint i = 0; i < parametrized_desc.size(); i++)
		descriptors.push_back(parametrized_desc[i]);
	for(uint i = 0; i < filters.size(); i++)
		filters[i]->ProcessParameters();
	parametrized_desc.clear();
}

void ParameterizedFilter::Not()
{ // negation of a filter
	if(true)
		m_index->Not();
	else {
		throw NotImplementedRCException("NOT is not implemented in this case.");
		assert(0);	// TODO: nontrivial Not
	}
}

