/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

//////////////////////////////////////////////////////////////////////////////////////////////////
// This is a part of RCBase implementation concerned with the join conditions execution
//////////////////////////////////////////////////////////////////////////////////////////////////

#ifndef __BH_COMMUNITY__
#	include "enterprise/local.h"
#else /* not __BH_COMMUNITY__ */
#	include "community/local.h"
#endif /* __BH_COMMUNITY__ */
#include "Query.h"
#include "Joiner.h"
#include "JoinerHash.h"
#include "OuterJoiner.h"
#include "vc/VirtualColumn.h"
#include "RoughJoinWatcher.h"
#include "TempTable.h"

using namespace std;

void Query::UpdateJoinCondition(MultiIndex& mind, // multi-index to be updated
								RoughMultiIndex& rmind,
								TempTable *table,	// pointer to TempTable corresponding to mind
								vector<Descriptor>& desc, // join descriptors (concerned with the same pair of tables) used for updating
								bool count_only_now,	//	true if materialization is done for counting tuples only
								_int64 limit_now
								)
{
	// return value: number of join conditions executed
	// Calculate joins (i.e. any condition using attributes from two dimensions)
	// as well as other conditions (incl. one-dim) flagged as "outer join"
	int all_conditions_used = 0;

	DimensionVector all_involved_dims(&mind);
	for (int i = 0; i < desc.size(); i++)
		desc[i].DimensionUsed(all_involved_dims);
	bool local_count_only_now = count_only_now;

	/////////////////// Outer join part ////////////////////
	bool is_outer = desc[0].IsOuter();
	auto_ptr<OuterJoiner> outer_joiner;			// may be not used if the join is inner
	if(is_outer) {
		local_count_only_now = false;			// we cannot omit the result calculation for outer join
		outer_joiner = auto_ptr<OuterJoiner>(new OuterJoiner(mind, desc, count_only_now));
	}

	// Apply conditions
	JoinAlgType alg_type_for_display = JTYPE_NONE;

	while(desc.size() > 0) {
		int conditions_used = 0;
		JoinAlgType join_alg = JTYPE_NONE;
		TwoDimensionalJoiner::JoinFailure join_result = TwoDimensionalJoiner::NOT_FAILED;
//		RoughJoinWatcher rough_join(mind, desc);
		join_alg = TwoDimensionalJoiner::ChooseJoinAlgorithm(mind, this, desc);

		/////////////////// Joining itself  ////////////////////
		do {
			TwoDimsJoinerAutoPtr joiner = TwoDimensionalJoiner::CreateJoiner(join_alg, mind, rmind,
																			 local_count_only_now,
																			 limit_now, table);
			conditions_used = joiner->ExecuteJoinConditions(desc);
			join_result = joiner->WhyFailed();

			if(join_result != TwoDimensionalJoiner::NOT_FAILED)
				join_alg = TwoDimensionalJoiner::ChooseJoinAlgorithm(join_result, desc.size());
		} while(join_result != TwoDimensionalJoiner::NOT_FAILED);

		if(conditions_used > 0) {		// join actually done
			MIRoughSorter rsorter(&mind);
			rsorter.SortMultiIndexRoughly(desc.size() == conditions_used ? limit_now : -1);
//			rough_join.UpdatePackPack(mind);
		}
		if(alg_type_for_display == JTYPE_NONE)
			alg_type_for_display = join_alg;
		else if(alg_type_for_display != join_alg)
			alg_type_for_display = JTYPE_MIXED;
		all_conditions_used += conditions_used;
		for(int i = 0; i < conditions_used; i++)
			desc.erase(desc.begin());			// erase the first condition (already used)
	}
	if(is_outer)
		outer_joiner->AddOuterTuples(mind);	// conclude outer joining
	mind.UpdateNoTuples();

	/////////////////// display results (the last alg.) ///////////
	DisplayJoinResults(mind, all_involved_dims, alg_type_for_display, is_outer, all_conditions_used);
}

void Query::DisplayJoinResults(MultiIndex& mind, DimensionVector &all_involved_dims, JoinAlgType join_performed,
							bool is_outer, int conditions_used)
{
	if(rccontrol.isOn()) {
		MIIterator mit(&mind, all_involved_dims);
		_int64 tuples_after_join = mit.NoTuples();

		char buf[30];
		if(conditions_used > 1)
			sprintf(buf, "%d cond. ", conditions_used);
		else
			strcpy(buf, "");
		if(is_outer)
			strcat( buf, "outer ");
		else
			strcat( buf, "inner ");

		char buf_dims[500];
		buf_dims[0] = '\0';
		bool first = true;
		for(int i = 0; i < mind.NoDimensions(); i++)
			if(all_involved_dims[i]) {
				if(first) {
					if(all_involved_dims.NoDimsUsed() == 1)
						sprintf(buf_dims, "...-%d", i);
					else
						sprintf(buf_dims, "%d", i);
					first = false;
				} else
					sprintf(buf_dims, "%s-%d", buf_dims, i);
			}

		rccontrol.lock(m_conn->GetThreadID())
			<< "Tuples after " << buf << "join " << buf_dims
			<<	(join_performed == JTYPE_SORT ?		" [sort]: " :
				(join_performed == JTYPE_MIXED ?	" [mix.]: " :
				(join_performed == JTYPE_HASH ?		" [hash]: " :
				(join_performed == JTYPE_GENERAL ?	" [loop]: " :
													" [????]: " ))))
			<< tuples_after_join
			<< unlock;
	}
}
