/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include <boost/shared_ptr.hpp>

#include "common/defs.h"
#ifndef __BH_COMMUNITY__
#	include "enterprise/local.h"
#else /* not __BH_COMMUNITY__ */
#	include "community/local.h"
#endif /* __BH_COMMUNITY__ */
#include "Query.h"
#include "ValueSet.h"
#include "vc/VirtualColumn.h"
#include "ParametrizedFilter.h"
#include "TempTable.h"
#include "RoughMultiIndex.h"
#include "MIIterator.h"
#include "ConditionEncoder.h"
#include "Sorter3.h"

using namespace std;

void ResolveIndSubquery(CQTerm &e, Operator op = O_EQ);

void Query::UpdateParamFilter(ParameterizedFilter *filter, bool count_only, _int64 limit, bool display_now)
{
	UpdateMultiIndex(*filter->m_index, filter->descriptors, filter->table, count_only, limit);
	for(size_t i = 0; i < filter->filters.size(); i++) {
		UpdateParamFilter(filter->filters[i], false, -1, display_now);
		if(filter->log_connectors[i] == O_AND)
			filter->m_index->And(*filter->filters[i]->m_index);
		else
			filter->m_index->Or(*filter->filters[i]->m_index);
	}
}

bool Query::TryToMerge(Descriptor &d1, Descriptor &d2)   // true, if d2 is no longer needed
{										// Assumption: d1 is earlier than d2 (important for outer joins)
	if(d1 == d2)
		return true;
	if(d1.attr.vc == d2.attr.vc && d1.IsInner() && d2.IsInner()) {
		// IS_NULL and anything based on the same column => FALSE
		// NOT_NULL and anything based on the same column => NOT_NULL is not needed
		// Exceptions:
		//		null NOT IN {empty set}
		//		null < ALL {empty set} etc.
		if( (d1.op == O_IS_NULL && d2.op != O_IS_NULL && !IsSetOperator(d2.op)) ||
			(d2.op == O_IS_NULL && d1.op != O_IS_NULL && !IsSetOperator(d1.op))) {
			d1.op = O_FALSE;
			d1.CalculateJoinType();
			return true;
		}
		if(d1.op == O_NOT_NULL && !IsSetOperator(d2.op)) {
			d1 = d2;
			return true;
		}
		if(d2.op == O_NOT_NULL && !IsSetOperator(d1.op))
			return true;
	}
	// If a condition is repeated both on outer join list and after WHERE, then the first one is not needed
	//    t1 LEFT JOIN t2 ON (a=b AND c=5) WHERE a=b   =>   t1 LEFT JOIN t2 ON c=5 WHERE a=b
	if(d1.EqualExceptOuter(d2)) {
		if(d1.IsInner() && d2.IsOuter())		// delete d2
			return true;
		if(d1.IsOuter() && d2.IsInner()) {		// delete d1
			d1 = d2;
			return true;
		}
	}
	// Content-based merging
	if(d1.attr.vc == d2.attr.vc) {
		if(d1.attr.vc->TryToMerge(d1, d2))
			return true;
	}
	return false;
}

void Query::DelayWhereConditions(vector<Descriptor> &desc)
{
	for(uint i = 0; i < desc.size(); i++)
		if(desc[i].outer_dim == -2) {
			if(desc[i].NullMayBeTrue())
				desc[i].delayed = true;		// delayed, i.e. must be calculated after nulls occur
			desc[i].outer_dim = -1;
		}
}

void Query::SyntacticalDescriptorListPreprocessing( vector<Descriptor> &desc, // descriptors (one-dimensional and joins) used for updating
													MultiIndex &mind,	// mind is used only to determine possible nulls
													TempTable* table)
{
	// outer joins preprocessing (delaying conditions etc.)
	bool outer_join_found = false;
	DimensionVector all_outer(&mind);
	for(uint i = 0; i < desc.size(); i++)
		if(!desc[i].done && desc[i].IsOuter()) {
			outer_join_found = true;
			all_outer[desc[i].outer_dim] = true;
		}

	if(outer_join_found) {
		for(uint i = 0; i < desc.size(); i++)
			if(!desc[i].done && desc[i].IsInner()) {
				DimensionVector inner(&mind);
				desc[i].DimensionUsed(inner);
				if(all_outer.Intersects(inner)) {
					if(desc[i].NullMayBeTrue()) {
						desc[i].delayed = true;		// delayed, i.e. must be calculated after nulls occur
					} else {
						// e.g. t1 LEFT JOIN t2 ON a1=a2 WHERE t2.c = 5   - in such cases all nulls are excluded anyway,
						//                                                  so it is an equivalent of inner join
						all_outer.Minus(inner);		// set all involved dimensions as no longer outer
					}
				}
			}
		for(uint i = 0; i < desc.size(); i++)
			if(!desc[i].done && desc[i].IsOuter() && !all_outer[desc[i].outer_dim]) {
				desc[i].outer_dim = -1;			// change outer joins to inner (these identified above)
			}
	}

	// descriptor preparation (normalization etc.)
	for(uint i = 0; i < desc.size(); i++) {		// Note that desc.size() may enlarge when joining with BETWEEN occur
		assert(desc[i].done == false);	// If not false, check it carefully.
//		desc[i].Simplify();
		if (desc[i].IsTrue()) {
			if(desc[i].IsInner())
				desc[i].done = true;
			continue;
		} else if(desc[i].IsFalse())
			continue;

		if(desc[i].IsDelayed())
			continue;

		if(desc[i].IsType_Join() && (desc[i].op == O_BETWEEN || desc[i].op == O_NOT_BETWEEN) ) {
			if(desc[i].lop != O_AND /* || desc[i].op == O_NOT_BETWEEN*/)
				throw NotImplementedRCException("This kind of join condition with OR is not implemented.");
		}

		// normalization of descriptor of type 1 between a and b
		if(desc[i].GetJoinType() != DT_COMPLEX_JOIN && !desc[i].IsType_Join() && desc[i].attr.vc && desc[i].attr.vc->IsConst() && desc[i].op == O_BETWEEN) {
			swap(desc[i].attr.vc, desc[i].val1.vc);
			desc[i].op = O_LESS_EQ;
			// now, the second part
			Descriptor dd(table);
			dd.attr = desc[i].val2;
			desc[i].val2.vc = NULL;
			dd.op = O_MORE_EQ;
			dd.val1 = desc[i].val1;
			dd.done = false;
			dd.outer_dim = desc[i].outer_dim;
			dd.CalculateJoinType();
			desc[i].CalculateJoinType();
			desc.push_back(dd);
		} else if(desc[i].IsType_JoinSimple() && desc[i].op == O_BETWEEN) {
			// normalization of descriptor of type a between 1 and b
			desc[i].op = O_MORE_EQ;
			Descriptor dd(table);
			dd.attr = desc[i].attr;
			dd.val1 = desc[i].val2;
			desc[i].val2.vc = NULL;
			dd.op = O_LESS_EQ;
			dd.done= false;
			dd.outer_dim = desc[i].outer_dim;
			dd.CalculateJoinType();
			desc[i].CalculateJoinType();
			desc.push_back(dd);
		}
		desc[i].CoerceColumnTypes();
		desc[i].Simplify();
	}

	// join descriptor merging (before normalization, because we may then add a condition for another column)
	vector<Descriptor> added_cond;
	for(uint i = 0; i < desc.size(); i++)				// t1.x == t2.y && t2.y == 5   =>   t1.x == 5
		if(!desc[i].done &&	desc[i].IsType_JoinSimple() &&
			desc[i].IsInner() && desc[i].op == O_EQ) {
			// desc[i] is a joining (eq.) condition
			for(uint j = 0; j < desc.size(); j++)
				if( i != j && !desc[j].done &&
					desc[j].IsType_AttrVal() && desc[j].IsInner()) {
					// desc[j] is a second condition, non-join
					if(desc[j].attr.vc == desc[i].attr.vc) {
						// the same table and column
						if(desc[j].op == O_EQ) {	// t2.y == t1.x && t2.y == 5  change to  t1.x == 5 && t2.y == 5
							desc[i].attr = desc[i].val1;
							desc[i].val1 = desc[j].val1;
							desc[i].CalculateJoinType();
							break;
						}
						else {	// t2.y == t1.x && t2.y > 5  change to  t2.y == t1.x && t2.y > 5 && t1.x > 5
							Descriptor dd(table);
							dd.attr = desc[i].val1;
							dd.op = desc[j].op;
							dd.val1 = desc[j].val1;
							dd.val2 = desc[j].val2;
							dd.CalculateJoinType();
							added_cond.push_back(dd);
						}
					}
					if(desc[j].attr.vc == desc[i].val1.vc) { //the same as above for val1
						// the same table and column
						if(desc[j].op == O_EQ) {	// t1.x == t2.y && t2.y == 5  change to  t1.x == 5 && t2.y == 5
							desc[i].val1 = desc[j].val1;
							desc[i].CalculateJoinType();
							break;
						}
						else {	// t1.x == t2.y && t2.y > 5  change to  t1.x == t2.y && t2.y > 5 && t1.x > 5
							Descriptor dd(table);
							dd.attr = desc[i].attr;
							dd.op = desc[j].op;
							dd.val1 = desc[j].val1;
							dd.val2 = desc[j].val2;
							dd.CalculateJoinType();
							added_cond.push_back(dd);
						}
					}
				}
		}
	if(added_cond.size() > 0)
			rccontrol.lock(m_conn->GetThreadID()) << "Adding " << int(added_cond.size()) << " conditions..." << unlock;
	for(uint i = 0; i < added_cond.size(); i++)
		desc.push_back(added_cond[i]);

	// attribute-based transformation (normalization) of descriptors "attr-operator-value"
	for(uint i = 0; i < desc.size(); i++) {
		if (!desc[i].attr.vc)
			continue; //probably descriptor got simplified
		int loc_dim = desc[i].attr.vc->GetDim();
		bool additional_nulls_possible = loc_dim > -1 ? (mind.GetFilter(loc_dim) == NULL) : false;
		if(desc[i].IsOuter())
			additional_nulls_possible = true;
		ConditionEncoder::Encode(desc[i], *this, additional_nulls_possible);

		if (desc[i].IsTrue()) {			// again, because something might be simplified
			if(desc[i].IsInner())
				desc[i].done = true;
			continue;
		}
	}

	// descriptor merging
	for(uint i = 0; i < desc.size(); i++) {
		if(desc[i].done || desc[i].IsDelayed())
			continue;
		for(uint jj = i + 1; jj < desc.size(); jj++) {
			if(desc[jj].done || desc[jj].IsDelayed() || desc[jj].outer_dim != desc[i].outer_dim)
				continue;
			if(TryToMerge(desc[i], desc[jj])) {
				rccontrol.lock(m_conn->GetThreadID()) << "Merging conditions..." << unlock;
				desc[jj].done = true;
			}
		}
	}
}

void Query::DescriptorListOrdering(vector<Descriptor> &desc)
{
	// descriptor evaluating
	// evaluation weight: higher value means potentially larger result (i.e. to be calculated as late as possible)
	if(desc.size() > 1) {	// else all evaluations remain 0 (default)
		for(uint i = 0; i < desc.size(); i++) {
			if(desc[i].done || desc[i].IsDelayed())
				desc[i].evaluation = 100000;				// to the end of queue
			else
				desc[i].evaluation = EvaluateConditionNonJoinWeight(desc[i]);	// joins are evaluated separately
		}
	}
	// descriptor ordering by evaluation weight
	for(uint k = 0; k < desc.size(); k++) {
		for(uint i = 0; i < desc.size() - 1; i++)	{
			int j = i + 1;			// bubble sort
			if( desc[i].evaluation > desc[j].evaluation &&
				desc[j].IsInner() && desc[i].IsInner())	// do not change outer join order
			{
				desc[i].swap(desc[j]);
			}
		}
	}
}

void Query::DescriptorJoinOrdering( vector<Descriptor> &desc, MultiIndex &mind )
{
	// calculate join weights
	for(uint i = 0; i < desc.size(); i++) {
		if(!desc[i].done && desc[i].IsType_JoinSimple())
			desc[i].evaluation = EvaluateConditionJoinWeight(desc[i], mind);
	}

	// descriptor ordering by evaluation weight - again
	for( uint k = 0; k < desc.size(); k++ )	{
		for( uint i = 0; i < desc.size() - 1; i++ )		{
			int j = i + 1;			// bubble sort
			if( desc[i].evaluation > desc[j].evaluation &&
				desc[j].IsInner() && desc[i].IsInner())	// do not change outer join order
			{
				desc[i].swap(desc[j]);
			}
		}
	}
	// Linearize joins (i.e. reorder conditions to join to already materialized part of multiindex, if possible)
	vector<int> join_desc;
	vector<int> output_order;
	vector<DimensionVector> dims;
	for(uint i = 0; i < desc.size(); i++)
		if(!desc[i].done && desc[i].IsType_Join()) {
			join_desc.push_back(i);
			DimensionVector loc_dim(&mind);
			desc[i].DimensionUsed(loc_dim);
			dims.push_back(loc_dim);
		}

	int max_dim = 0;
	for(uint i = 0; i < desc.size(); i++)
		if(!desc[i].done && desc[i].IsType_JoinSimple())
		{
			int dim1 = desc[i].attr.vc->GetDim();
			int dim2 = desc[i].val1.vc->GetDim();
			if( dim1 > max_dim )
				max_dim = dim1;
			if( dim2 > max_dim )
				max_dim = dim2;
		}

	if(max_dim > 0)
	{
		int *already_material = new int [max_dim+1];
		for( int i = 0; i <= max_dim; i++ )
			already_material[i] = 0;
		bool not_first=false;
		uint j;
		for(uint i = 0; i < desc.size(); i++)
			if(!desc[i].done && desc[i].IsType_JoinSimple())
		{
			int dim1 = desc[i].attr.vc->GetDim();
			int dim2 = desc[i].val1.vc->GetDim();
			if(not_first && already_material[dim1] == 0 && already_material[ dim2 ] == 0)
			{
				bool linearization_found = false;
				if(desc[i].IsInner())			// do not change ordering of outer joins
					for(j = i + 1; j < desc.size(); j++) {
						if( desc[j].done || desc[j].IsType_JoinSimple()==false )
							continue;
						if(desc[j].IsOuter()) 	// do not change ordering of outer joins
							break;
						if( already_material[ desc[j].attr.vc->GetDim() ] || already_material[ desc[j].val1.vc->GetDim() ] ) {
							linearization_found = true;
							break;
						}
					}
				// j is the first condition having any common table with any of the previous one (<i)
				if( linearization_found )		// better one found - change needed
				{
					desc[i].swap(desc[j]);
					dim1 = desc[i].attr.vc->GetDim();
					dim2 = desc[i].val1.vc->GetDim();
				}
			}
			already_material[ dim1 ] = 1;
			already_material[ dim2 ] = 1;
			not_first = true;
		}
		delete [] already_material;
	}
}

void Query::ApplyDescriptor(MultiIndex &mind, RoughMultiIndex &rmind, TempTable *table, Descriptor &desc, int desc_number)
// desc_number = -1 => switch off the rough part
{
	if(desc.op == O_TRUE) {
		desc.done = true;
		return;
	}
	if(desc.op == O_FALSE) {
		mind.Empty();
		desc.done = true;
		return;
	}
	set<int> dims = desc.GetDimensions();
	if (dims.size() == 0 && !desc.IsDeterministic())
		for(int i = 0; i< mind.NoDimensions(); i++)
			dims.insert(i);
	int one_dim = -1;
	RSValue* rf = NULL;
	if(dims.size() == 1) {
		one_dim = *dims.begin();
		rf = rmind.GetLocalDescFilter(one_dim, desc_number, true);	// "true" here means that we demand an existing local rough filter
	}
	RSValue cur_roughval;
	MIUpdatingIterator mit(&mind, dims);
	while(mit.IsValid()) {
		if(rf && mit.GetCurPackrow(one_dim) >= 0)
			cur_roughval = rf[mit.GetCurPackrow(one_dim)];
		else
			cur_roughval = RS_SOME;
		if(cur_roughval == RS_NONE) {
			mit.ResetCurrentPack();
			mit.NextPackrow();
		} else if(cur_roughval == RS_ALL)
			mit.NextPackrow();
		else  // RS_SOME or RS_UNKNOWN
			desc.EvaluatePack(mit);
		if(m_conn->killed())
			throw KilledRCException();
	}
	mit.Commit();
	if(desc.attr.vc)
		desc.attr.vc->UnlockSourcePacks();
	if(desc.val1.vc)
		desc.val1.vc->UnlockSourcePacks();
	if(desc.val2.vc)
		desc.val2.vc->UnlockSourcePacks();
	desc.done = true;
	if(one_dim != -1 && mind.GetFilter(one_dim)) {		// update global rough part
		bool empty_found = false;
		Filter *f = mind.GetFilter(one_dim);
		for(int p = 0; p < rmind.NoPacks(one_dim); p++)
			if(f->IsEmpty(p)) {
				rmind.SetPackStatus(one_dim, p, RS_NONE);
				empty_found = true;
			}
	}
	return;
}

void Query::UpdateMultiIndex(MultiIndex &mind, // multi-index to be updated
							 vector<Descriptor> &desc, // descriptors (one-dimensional and joins) used for updating
							 TempTable *table,	// pointer to TempTable corresponding to mind
							 bool count_only,	//	true if materialization is done for counting tuples only
							 _int64 limit )
{
/*
	if(rccontrol.isOn()) {	// Uncomment it to get an input descriptor list
		rccontrol.lock(m_conn->GetThreadID()) << "TEMPORARY: raw descriptor list:" << unlock;
		for(uint i=0; i<desc.size(); i++) {
			char buf[1000];
			strcpy(buf, " ");
			desc[i].ToString(buf, 1000);
			if(desc[i].done)
				rccontrol.lock(m_conn->GetThreadID()) << "Done: " << buf << unlock;
			else if(desc[i].IsDelayed())
				rccontrol.lock(m_conn->GetThreadID()) << "Delayed: " << buf << unlock;
			else
				rccontrol.lock(m_conn->GetThreadID()) << "Cnd(" << i << "):  " << buf << unlock;
		}
	}
*/
	////////////////////////////////////////////////////////////////////////

	if(desc.size() < 1)
		return;
	DelayWhereConditions(desc);
	SyntacticalDescriptorListPreprocessing(desc, mind, table);

	bool empty_cannot_grow = true;			// if false (e.g. outer joins), then do not optimize empty multiindex as empty result
	for(uint i = 0; i < desc.size(); i++)
		if(desc[i].IsOuter())
			empty_cannot_grow = false;

	// special cases
	bool nonempty = true;

	DescriptorListOrdering(desc);

	///// descriptor display
	if(rccontrol.isOn()) {
		rccontrol.lock(m_conn->GetThreadID()) << "Initial execution plan (non-join):" << unlock;
		for(uint i=0; i<desc.size(); i++) if(!desc[i].done && !desc[i].IsType_Join() && desc[i].IsInner()) {
			char buf[1000];
			strcpy(buf, " ");
			desc[i].ToString(buf, 1000);
			if(desc[i].IsDelayed())
				rccontrol.lock(m_conn->GetThreadID()) << "Delayed: " << buf << " \t(" << int(desc[i].evaluation*100)/100.0 << ")" << unlock;
			else
				rccontrol.lock(m_conn->GetThreadID()) << "Cnd(" << i << "):  " << buf << " \t(" << int(desc[i].evaluation*100)/100.0 << ")" << unlock;
		}
	}

	///// end now if the multiindex is empty
	if(mind.ZeroTuples() && empty_cannot_grow) {
		mind.Empty();
		return;
	}

	///////////////// Prepare execution - rough set part
	for(uint i = 0; i < desc.size(); i++)
		if(!desc[i].done && desc[i].IsInner())	{
			if(desc[i].IsTrue()) {
				desc[i].done = true;
				continue;
			} else if(desc[i].IsFalse()) {
				if(desc[i].attr.vc) {
					desc[i].done = true;
					DimensionVector dims(&mind);
					desc[i].attr.vc->MarkUsedDims(dims);
					for(int cur_dim = 0; cur_dim < mind.NoDimensions(); cur_dim++) {
						if(dims[cur_dim])
							if(mind.GetFilter(cur_dim))
								mind.GetFilter(cur_dim)->Reset();
					}
					if(empty_cannot_grow) {
						mind.Empty();
						return;
					}
				}
			}
		}
	vector<_int64> packs_for_rough;
	for(uint i = 0; i < (uint)mind.NoDimensions(); i++)
		packs_for_rough.push_back(bh::common::NoObj2NoPacks(mind.OrigSize(i)));
	RoughMultiIndex rmind(packs_for_rough);

	nonempty = RoughUpdateMultiIndex(rmind, mind, desc);	// calculate all rough conditions, then exclude RS_NONE from mind
	if(!nonempty && empty_cannot_grow) {
		mind.Empty();		// nonempty==false if the whole result is empty (outer joins considered)
		return;
	}

	///////////////// Apply all one-dimensional filters (after where, i.e. without outer joins)
	int last_desc_dim = -1;
	int cur_dim = -1;
	for(uint i = 0; i < desc.size(); i++)
		if(!desc[i].done && desc[i].IsInner() && !desc[i].IsType_Join() && !desc[i].IsDelayed()) {
			if(desc[i].attr.vc) {
				cur_dim = desc[i].attr.vc->GetDim();
			}
			if(last_desc_dim != -1 && cur_dim != -1 && last_desc_dim != cur_dim ) {
				// Make all possible projections to other dimensions
				RoughMakeProjections(rmind, mind, desc, cur_dim,false);
			}

			ApplyDescriptor(mind, rmind, table, desc[i], i);
			if (!desc[i].attr.vc)
				continue; //probably desc got simplified and is true or false
			if(cur_dim >= 0 && mind.GetFilter(cur_dim) && mind.GetFilter(cur_dim)->IsEmpty() && empty_cannot_grow) {
				mind.Empty();
				return;
			}
			last_desc_dim = cur_dim;

		}
	rmind.UpdateReducedDimension();
	mind.UpdateNoTuples();

	/////////////////////////////////////////////////////////////////////////////////////
	// Some displays

	if(rccontrol.isOn()) {
		int pack_full = 0, pack_some = 0, pack_all = 0;
		rccontrol.lock(m_conn->GetThreadID()) << "Packrows after exact evaluation (WHERE):" << unlock;
		for(uint i = 0; i < (uint)mind.NoDimensions(); i++) if(mind.GetFilter(i)) {
			Filter *f = mind.GetFilter(i);
			pack_full = 0;
			pack_some = 0;
			pack_all = bh::common::NoObj2NoPacks(mind.OrigSize(i));
			for(int b = 0; b < pack_all; b++) {
				if(f->IsFull(b))
					pack_full++;
				else if(!f->IsEmpty(b))
					pack_some++;
			}
			rccontrol.lock(m_conn->GetThreadID()) << "(t" << i << "): " << pack_all << " all packrows, " << pack_full+pack_some << " to open (including " << pack_full << " full)" << unlock;
		}
	}

	/////////////////////////////////////////////////////////////////////////////////////
	DescriptorJoinOrdering(desc, mind);

	///// descriptor display for joins
	if(rccontrol.isOn()) {
		bool first_time = true;
		for(uint i = 0; i < desc.size(); i++) if(!desc[i].done && (desc[i].IsType_Join() || desc[i].IsOuter())) {
			if(first_time) {
				rccontrol.lock(m_conn->GetThreadID()) << "Join execution plan:" << unlock;
				first_time = false;
			}
			char buf[1000];
			strcpy(buf," ");
			desc[i].ToString(buf, 1000);
			if(desc[i].IsDelayed())
				rccontrol.lock(m_conn->GetThreadID()) << "Delayed: " << buf << " \t(" << int(desc[i].evaluation*100)/100.0 << ")" << unlock;
			else
				rccontrol.lock(m_conn->GetThreadID()) << "Cnd(" << i << "):  " << buf << " \t(" << int(desc[i].evaluation*100)/100.0 << ")" << unlock;
		}
	}

	//////// Calculate other types of conditions, e.g. joins (i.e. conditions using attributes from two dimensions)
	int  no_of_join_conditions = 0;		// count also one-dimensional outer join conditions
	bool delayed_conditions_exist = false;
	for(uint i = 0; i < desc.size(); i++)
		if(!desc[i].done) {
			 if(!desc[i].IsDelayed())
				no_of_join_conditions++;
			 else
				delayed_conditions_exist = true;
		}

	for(uint i = 0; i < desc.size(); i++) {
		if(mind.ZeroTuples() && empty_cannot_grow) {
			// set all following descriptors to done
			for(uint j = i; j < desc.size(); j++)
				desc[j].done = true;
			break;
		}
		if(!desc[i].done && !desc[i].IsDelayed()) {
			//////////////// Merging join conditions ///////////////////
			vector<Descriptor> join_desc;
			join_desc.push_back(desc[i]);
			desc[i].done = true;		// setting in advance, but we already copied the descriptor to be processed
			no_of_join_conditions--;
			DimensionVector dims1(&mind);
			desc[i].DimensionUsed(dims1);
			mind.MarkInvolvedMaterialClasses(dims1);
			int cur_outer_dim = join_desc[0].outer_dim;
			for(uint jj = i + 1; jj < desc.size(); jj++)
				if(!desc[jj].done && !desc[jj].IsDelayed()) {
					DimensionVector dims2(&mind);
					desc[jj].DimensionUsed(dims2);
					if(desc[jj].outer_dim == cur_outer_dim &&
							(cur_outer_dim != -1 || dims1.Includes(dims2))) {
						// can be executed together if all dimensions of the other condition are present in the base one
						// or in case of outer join
						join_desc.push_back(desc[jj]);
						desc[jj].done = true;
						no_of_join_conditions--;
					}
				}

			// Optimization: count_only is true => do not materialize multiindex (just counts tuples). WARNING: in this case cannot use multiindex for any operations other than NoTuples().
			bool count_only_now = false;
			_int64 limit_now = -1;
			if(no_of_join_conditions == 0 && // optimization used only for the last group of conditions
				!delayed_conditions_exist ) {
				if(count_only)
					count_only_now = true;
				limit_now = limit;
			}

			// Check whether all dimensions are really used - except outer joins
			if(!join_desc[0].IsOuter()) {
				for(int dim = 0; dim < mind.NoDimensions(); dim++) if(!mind.IsUsedInOutput(dim)) {
					bool dim_still_needed = false;
					for(uint jj = 0; jj < desc.size(); jj++)
						if(jj!=i && !desc[jj].done && desc[jj].DimensionUsed(dim)) {
							dim_still_needed = true;
							break;
						}
					if(!dim_still_needed) {
//						rccontrol.lock(m_conn->GetThreadID()) << "Forget soon: " << dim << unlock;
						mind.ForgetSoon(dim);
					}
				}
			}
			UpdateJoinCondition(mind, rmind, table, join_desc, count_only_now, limit_now);
		}
	}

	/////////////////////////////////////////////////////////////////////////////////////
	// Execute all delayed conditions
	for(uint i = 0; i < desc.size(); i++) {
		if(!desc[i].done) {
			rccontrol.lock(m_conn->GetThreadID()) << "Executing delayed Cnd(" << i << ")" << unlock;
			ApplyDescriptor(mind, rmind, table, desc[i], i);
		}
	}
	mind.UpdateNoTuples();
}
