/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include <boost/algorithm/string.hpp>
#include "Query.h"
#include "compilation_tools.h"
#include "shared/core/RCEngine.h"
#include "shared/common/DataFormat.h"
#include "system/LargeBuffer.h"

using namespace std;
using namespace boost;

bool RCEngine::IsBHRoute(THD * thd, TABLE_LIST* table_list, SELECT_LEX* selects_list, int &in_case_of_failure_can_go_to_mysql, int with_insert)
{
	in_case_of_failure_can_go_to_mysql = true;

	if(!table_list)
		return false;
	if(with_insert)
		table_list = table_list->next_global; //we skip one
	for(TABLE_LIST *tl = table_list; tl; tl = tl->next_global) //SZN:we go through tables
	{
		if(!tl->derived && !tl->view && !IsBHTable(tl->table))
			return false;//In this list we have
		//all views, derived tables and their sources, so anyway we walk through all the source tables
		//even though we seem to reject the control of views
	}

	//then we check the parameter of file format.
	//if it is MYSQL_format AND we write to a file, it is a MYSQL route.
	int is_dump;
	const char * file = GetFilename(selects_list, is_dump);

	if (file) {//it writes to a file
		longlong param = 0;
		string s_res;
		if(!get_parameter(thd, BH_DATAFORMAT, param, s_res)) {

			if(boost::iequals(boost::trim_copy(s_res), "MYSQL"))
				return false;

			if(!DataFormat::GetDataFormat(s_res)) {  //parameter is UNKNOWN VALUE
				my_message(ER_SYNTAX_ERROR, "Brighthouse specific error: Unknown value of BH_DATAFORMAT parameter", MYF(0));
				return false;
			} else
				in_case_of_failure_can_go_to_mysql = false;   //in case of failure
							//it cannot go to MYSQL - it writes to a file,
							//but the file format is not MYSQL
		} else {   //param not set - we assume it is MYSQL
			push_warning(thd, MYSQL_ERROR::WARN_LEVEL_NOTE, ER_UNKNOWN_ERROR,
				"BH_DATAFORMAT parameter is not set, 'MYSQL' value assumed, query executed by MySQL engine.");
			return false;
		}
	}

	return true;
}

bool RCEngine::IsBHTable(TABLE* table)
{
	return table && table->s->db_type() == rcbase_hton;	// table->db_type is always NULL
}

const char* RCEngine::GetFilename(SELECT_LEX* selects_list, int &is_dumpfile)
{
	//if the function returns a filename <> NULL
	//additionally is_dumpfile indicates whether it was 'select into OUTFILE' or maybe 'select into DUMPFILE'
	//if the function returns NULL it was a regular 'select'
	//don't look into is_dumpfile in this case
	if (selects_list->parent_lex->exchange)
	{
		is_dumpfile = selects_list->parent_lex->exchange->dumpfile;
		return selects_list->parent_lex->exchange->file_name;
	}
	return 0;
}

auto_ptr<IOParameters> RCEngine::CreateIOParameters(const string& path)
{
	if(path.empty())
		return auto_ptr<IOParameters>(new IOParameters());

	string data_dir;
	string data_path;

	if(IsAbsolutePath(path)) {
		data_dir = "";
		data_path = path;
	} else {
		data_dir = infobright_data_dir;
		char db_name[256] = {0}, tab_name[256] = {0};
		GetNames(path.c_str(), db_name, tab_name, 256);
		data_path += db_name;
		data_path += '/';
		data_path += tab_name;
	}

	return auto_ptr<IOParameters>(new IOParameters(data_dir, data_path));
}

auto_ptr<IOParameters> RCEngine::CreateIOParameters(THD *thd, TABLE *table)
{
	if(table == NULL)
		return CreateIOParameters("");

	return CreateIOParameters(GetTablePath(thd, table));
}

string RCEngine::GetTablePath(THD *thd, TABLE *table)
{
	BHASSERT(table != NULL, "table pointer shouldn't be null here");
	return table->s->normalized_path.str;
}

//  compute time zone difference between client and server
void RCEngine::ComputeTimeZoneDiffInMinutes(THD* thd, short& sign, short& minutes)
{
	if(stricmp(thd->variables.time_zone->get_name()->ptr(), "System") == 0) {
		// System time zone
		sign = 1;
		minutes = NULL_VALUE_SH;
		return;
	}
	TIME client_zone, utc;
	utc.year = 1970;
	utc.month = 1;
	utc.day = 1;
	utc.hour = 0;
	utc.minute = 0;
	utc.second = 0;
	utc.second_part = 0;
	utc.neg = 0;
	utc.time_type = MYSQL_TIMESTAMP_DATETIME;
	thd->variables.time_zone->gmt_sec_to_TIME(&client_zone, (my_time_t)0); // check if client time zone is set
	longlong secs;
	long msecs;
	sign = 1;
	minutes = 0;
	if(calc_time_diff(&utc, &client_zone, 1, &secs, &msecs))
		sign = -1;
	minutes = (short)(secs/60);
}

BHError RCEngine::GetIOParameters(IOParameters& io_params, THD& thd, sql_exchange& ex, TABLE* table)
{
	String *field_term = ex.field_term, *escaped = ex.escaped;
	String *enclosed = ex.enclosed;

	int io_mode = -1;
	int io_timeout = 600;
	char name[FN_REFLEN];
	char *tdb = 0;
	if(table) {
		tdb = table->s->db.str ? table->s->db.str : thd.db;
	} else
		tdb = thd.db;

	io_params = IOParameters(*RCEngine::CreateIOParameters(&thd, table));
	short sign, minutes;
	ComputeTimeZoneDiffInMinutes(&thd, sign, minutes);
	io_params.SetTimeZone(sign, minutes);

#ifdef DONT_ALLOW_FULL_LOAD_DATA_PATHS
	ex->file_name+=dirname_length(ex->file_name);
#endif

	longlong param = 0;
	string s_res;
	if(DataFormat::GetNoFormats() > 1) {
		if(!get_parameter(&thd, BH_DATAFORMAT, param, s_res)) {
			DataFormatPtr df = DataFormat::GetDataFormat(s_res);
			if(df)
				io_mode = df->GetId();
			else
				return BHError(BHERROR_WRONG_PARAMETER, "Unknown value of BH_DATAFORMAT parameter.");
		} else {
//			vector<string> names;
//			DataFormat::GetFormatsNames(names);
//			string message("Data format for import not set. Please set @bh_dataformat to one of the following supported formats: ");
//			for(int i = 0, size = names.size(); i < size; i++)
//				message += "'" + to_lower_copy(names[i]) + "', ";
//			message += "'" + names[names.size() - 1] + "mysql";
//			return BHError(BHERROR_WRONG_PARAMETER, message);
			BHERROR("BH_DATAFORMAT parameter not set!");
		}
	} else
		io_mode = DataFormat::GetDataFormat(0)->GetId();



	if(!get_parameter(&thd, BH_NULL, param, s_res))
		io_params.SetNullsStr(s_res);

	if(LargeBuffer::IsPipe(ex.file_name))
	{
		strcpy(name, ex.file_name);
		if(!get_parameter(&thd, BH_PIPEMODE, param, s_res))
			io_params.SetOutput(io_mode, ex.file_name, (int)param); // BH_PIPEMODE == 0 - pipe work as a client otherwise as a server
		else
			io_params.SetOutput(io_mode, ex.file_name);
		if(!get_parameter(&thd, BH_TIMEOUT, param, s_res) && param != -1)
			io_params.SetParameter(TIMEOUT, (int)(param));
	}
	else
	{		
		if(!dirname_length(ex.file_name)) {
			strxnmov(name, FN_REFLEN - 1, mysql_real_data_home, tdb, NullS);
			(void)fn_format(name, ex.file_name, name, "", MY_RELATIVE_PATH | MY_UNPACK_FILENAME);
		} else
			(void)fn_format(name, ex.file_name, mysql_real_data_home, "", MY_RELATIVE_PATH | MY_UNPACK_FILENAME);
		io_params.SetOutput(io_mode, name);
	}
	
	

	if(escaped != 0 && escaped->alloced_length() != 0)
		io_params.SetEscapeCharacter(*escaped->c_ptr());
	if(field_term != 0 && !(*field_term->c_ptr() == '\t' && field_term->alloced_length() == 0))
		io_params.SetParameter(DELIMITER, *field_term->c_ptr());
	if(enclosed != 0 && enclosed->length()) {
		if(enclosed->length() == 4 && stricmp(enclosed->ptr(), "NULL") == 0)
			io_params.SetParameter(STRING_QUALIFIER, '\0');
		else
			io_params.SetParameter(STRING_QUALIFIER, *enclosed->ptr());
	}
	return BHERROR_SUCCESS;
}

#ifdef __GNUC__
string RCEngine::GetServiceStartPath(string service_name)
{
	return "";
}
#else

string RCEngine::GetServiceStartPath(string service_name)
{
	LPQUERY_SERVICE_CONFIG lpqscBuf;
	DWORD dwBytesNeeded;
	// Open a handle to the service.
	SC_HANDLE schSCManager = ::OpenSCManager (NULL, NULL, SC_MANAGER_ENUMERATE_SERVICE);
	SC_HANDLE schService = OpenService(schSCManager, TEXT(service_name.c_str()), SERVICE_QUERY_CONFIG);
	if (schService == NULL)
		return "";

	lpqscBuf = (LPQUERY_SERVICE_CONFIG)LocalAlloc(LPTR, 4096);
	if (lpqscBuf == NULL)
		return "";

	if (! QueryServiceConfig(schService, lpqscBuf, 4096, &dwBytesNeeded) )
	{
		LocalFree(lpqscBuf);
		return "";
	}

	string start_path = lpqscBuf->lpBinaryPathName;
	LocalFree(lpqscBuf);
	return start_path;
}
#endif

