/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef _CORE_RSIMAP_H_
#define _CORE_RSIMAP_H_

#include "../common/CommonDefinitions.h"
#include "../types/RCDataTypes.h"
#include "RSI_Framework.h"

/////////////////////////////////////////////////////////////////////////////
// RSIndex_CMap    - histograms for text packs
//
// Usage:
//
//	RSIndex_CMap h;
//	h.Load(fhandle);
//	if(h.NoObj()<current_no_obj) then only a part of it may be safely used, i.e.
//		only between 0 and (h.NoObj()>>16 - 1) packs are up to date.
//
//	result = h.IsValue( value,value,pack,min_pack,max_pack);	// check if a value (or interval) is present in a pack
//
// Updating:
//
//	h.ClearPack(pack);		// clear all packs which are changed
//	h.Update(new_no_obj);	// register new number of objects
//	for(all changed packs)
//		h.PutValue(value,pack,min_pack,max_pack);
//	h.Save(fhandle);
//

class RSIndex_CMap : public RSIndex
{
public:
	RSIndex_CMap(int no_pos = 64);				// create an empty index
	~RSIndex_CMap();

	void	Clear();				// make the index empty

	_int64	NoObj()		{ return no_obj; }
	bool		UpToDate(_int64 cur_no_obj, uint pack);		// true, if this pack is up to date (i.e. either cur_no_obj==no_obj, or pack is not the last pack in the index)
	_int64	NoPositions()		{ return no_positions; }

	// creation, update

	void	Create(_int64 _no_obj,int no_pos = 64);			// create a new histogram based on _no_obj objects;
	void	Update(_int64 _new_no_obj);		// enlarge buffers for the new number of objects;
										// note that all of the last pack must be analyzed by PutValue(),
										// because pack min and max might change. ClearPack may be needed.
	void	ClearPack(uint pack);			// reset the information about the specified pack
										// and prepare it for update (i.e. analyzing again all values)
	void	SetPack(uint pack);			// set the information about the specified pack
										// can be used for fakes
	void	PutValue(RCBString& v, uint pack); // set information that value v does exist in this pack

	// reading histogram information
	// Note: this function is thread-safe, i.e. it only reads the data
	RSValue IsValue(RCBString min_v, RCBString max_v, uint pack);
					// Results:		RS_NONE - there is no objects having values between min_v and max_v (including)
					//				RS_SOME - some objects from this pack do have values between min_v and max_v
					//				RS_ALL	- all objects from this pack do have values between min_v and max_v

	RSValue	IsLike(RCBString pattern, uint pack);

	bool	Intersection(int pack, RSIndex_CMap *sec, int pack2);
					//	Return true if there is any common value possible for two packs from two different columns (histograms).

	int	NoOnes(uint pack, uint pos);	// a number of ones in the pack on a given position 0..no_positions
	bool	IsSet(uint pack, unsigned char c, uint pos);
	bool	IsAnySet(uint pack, unsigned char first, unsigned char last, uint pos); // true, if there is at least one 1
					// in [first, last]

	// Loading/saving

	int		Load(IBFile* frs_index, int current_read_loc);		// fd - file descriptor, open for reading, loc - 0 or 1
	int		LoadLastOnly(IBFile* frs_index, int current_read_loc); // fd - file descriptor, open for reading
	int		Save(IBFile* frs_index, int current_save_loc);		// fd - file descriptor, open for writing

	void	Display(uint pack, unsigned char c);
	void	CopyAppendKN(int pack);		//append a copy of given KN to KN table, extending the number of objects
	void	AppendKN(int pack, RSIndex_CMap*);	//append a given KN to KN table, extending the number of objects

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
private:
	void 	AppendKNs(int no_new_packs);	//append new KNs
	void	Set(uint pack, unsigned char c, uint pos);
	void	UnSet(uint pack, unsigned char c, uint pos);
	_int64	no_obj;				// timeliness signature: a number of objects for which the index is up-to-date
	uint	no_pack;			// number of packs
	uint	start_pack;			// number of the first pack stored in memory
	uint	end_pack;			// number of the last pack stored in memory
	uint	no_pack_declared;	// current size of the buffer (usually more than used, to avoid too frequent reallocations)
	uint	no_positions;		// number of string positions to describe one pack (default: 64 positions)
	uint	*t;			// buffer for values
};

#endif   //_CORE_RSIMAP_H_

