/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "RoughMultiIndex.h"

using namespace std;

#define MAX_P2P_SIZE 200000000				// do not create larger P2P objects

RoughMultiIndex::RoughMultiIndex(vector<_int64> packs)
{
	no_dims = int(packs.size());
	no_packs = new int [no_dims];
	no_empty_packs = new int [no_dims];
	rf = new RSValue * [no_dims];
	rpp = new BinMatrix * [no_dims * no_dims];
	left_used = new bool [no_dims * no_dims];
	local_desc = new vector<RFDesc*> [no_dims];
	for(int d = 0; d < no_dims; d++) {
		no_packs[d] = packs[d];
		no_empty_packs[d] = 0;
		if(no_packs[d] > 0)
			rf[d] = new RSValue [no_packs[d]];
		else
			rf[d] = NULL;
		for(int p = 0; p < no_packs[d]; p++)
			rf[d][p] = RS_UNKNOWN;
		for(int d2 = 0; d2 < no_dims; d2++) {
			rpp[d * no_dims + d2] = NULL;
			left_used[d * no_dims + d2] = false;
		}
	}
	cur_rpp = NULL;
	cur_rpp_switched = false;
}

RoughMultiIndex::~RoughMultiIndex()
{
	for(int d = 0; d < no_dims; d++) {
		if(rf[d])
			delete [] rf[d];
		if(rpp)
			for(int d2 = 0; d2 < no_dims; d2++)
				delete rpp[d * no_dims + d2];
		for(int i = 0; i < local_desc[d].size(); i++)
			delete local_desc[d][i];
	}
	delete [] local_desc;
	delete [] rf;
	delete [] rpp;
	delete [] left_used;
	delete [] no_packs;
}

double RoughMultiIndex::GetP2PDensity(int dim1, int dim2)
{
	if(dim1>dim2) {
		int pp = dim1;
		dim1 = dim2;
		dim2 = pp;
	}
	if(rpp[dim1 * no_dims + dim2] == NULL)
		return -1;
	return rpp[dim1 * no_dims + dim2]->Density();
}

bool RoughMultiIndex::UseP2P(int dim1, int dim2)		// for speed: predefine current P2P
{
	if(_int64(no_packs[dim1]) * no_packs[dim2] > MAX_P2P_SIZE) {
		cur_rpp = NULL;
		return false;
	}
	cur_rpp_switched = false;
	if(dim1 > dim2) {
		int pp = dim1;
		dim1 = dim2;
		dim2 = pp;
		cur_rpp_switched = true;
	}
	if(rpp[dim1 * no_dims + dim2] == NULL) {
		rpp[dim1 * no_dims + dim2] = new BinMatrix(no_packs[dim1], no_packs[dim2], true);
	}
	cur_rpp = rpp[dim1 * no_dims + dim2];
	return true;
}

RSValue RoughMultiIndex::GetCurrentP2PStatus(int pack1, int pack2)
{
	if(!cur_rpp)
		return RS_SOME;
	if(cur_rpp_switched) {
		if(cur_rpp->Get(pack2, pack1))
			return RS_SOME;
	} else {
		if(cur_rpp->Get(pack1, pack2))
			return RS_SOME;
	}
	return RS_NONE;
}

void RoughMultiIndex::SetCurrentP2PStatus(int pack1, int pack2, RSValue v)
{
	if(!cur_rpp)
		return;
	if(cur_rpp_switched) {
		bool prev = cur_rpp->Get(pack2, pack1);
		cur_rpp->Set(pack2, pack1, prev && (v != RS_NONE));
	}
	else {
		bool prev = cur_rpp->Get(pack1, pack2);
		cur_rpp->Set(pack1, pack2, prev && (v != RS_NONE));
	}
}

void RoughMultiIndex::SetP2PLeftJoin(int dim1, int dim2)	// notify that these two dimensions are involved in a left join
{
	// NOTE: if there is a left join (dim1, dim2), then there also is a left join (dim3, dim2) for any dim3<dim2
	// So dim1 is not used here.
	for(int i = 0; i < dim2; i++)
		left_used[ i * no_dims + dim2 ] = true;
}

void RoughMultiIndex::RecalculateP2P()			// make all projections of one dimension through p2p to all other, recursive until any change made
{
	bool *dim_changed = new bool [no_dims];
	bool any_change = false;
	for(int i = 0; i < no_dims; i++)
		dim_changed[i] = true;				// indicates which dimension is to be recalculated again
	do {
		for(int d1 = 0; d1 < no_dims; d1++)		// projection of dimension d1 into d2
			if(dim_changed[d1]) {
				for(int d2 = 0; d2 < no_dims; d2++) {
					bool p2p_exists = UseP2P(d1, d2);
					if(d1 != d2 && p2p_exists &&
							left_used[d2*no_dims+d1] == false &&		// do not make projections for any direction in left joins
							left_used[d1*no_dims+d2] == false)
						for(int pack2 = 0; pack2 < no_packs[d2]; pack2++)
							if(rf[d2][pack2] != RS_NONE) {							// not RS_NONE, but maybe should be RS_NONE?
								bool row_empty = true;
								for(int pack1=0;pack1<no_packs[d1];pack1++) {
									if(rf[d1][pack1]!=RS_NONE && GetCurrentP2PStatus(pack1, pack2)!=RS_NONE) {
										row_empty = false;
										break;
									}
								}
								if(row_empty) {
									dim_changed[d2] = true;
									rf[d2][pack2] = RS_NONE;
								}
							}
				}
				dim_changed[d1] = false;
			}
		//////////////////////////
		any_change = false;
		for(int i = 0; i < no_dims; i++)
			if(dim_changed[i]) {
				any_change = true;
				break;
			}
	} while(any_change);
	delete [] dim_changed;
}

void RoughMultiIndex::AndP2P(int dim1,int dim2,BinMatrix *p2p,bool transpose)		// exclude all pairs of packs which are not present in the given p2p (e.g. RSI_PackPack)
{
	if(dim1>dim2)
	{
		int pp = dim1;
		dim1 = dim2;
		dim2 = pp;
		transpose = !transpose;
	}
	if(rpp[dim1*no_dims+dim2]==NULL)
	{
		rpp[dim1*no_dims+dim2] = new BinMatrix(no_packs[dim1],no_packs[dim2],true);
	}
	if(transpose)
	{
		for(int p1=0;p1<no_packs[dim1];p1++)
			for(int p2=0;p2<no_packs[dim2];p2++)
				if(p2p->Get(p2,p1)==false)						// transposition and conjunction
					rpp[dim1*no_dims+dim2]->Set(p1,p2,false);
	}
	else
	{
		rpp[dim1*no_dims+dim2]->And(*p2p);
	}
}

RSValue *RoughMultiIndex::GetLocalDescFilter(int dim, int desc_num, bool read_only)
{
	if(desc_num < 0)
		return NULL;
	int j;
	for(j = 0; j < local_desc[dim].size(); j++) {
		if((local_desc[dim])[j]->desc_num == desc_num)
			return (local_desc[dim])[j]->desc_rf;
	}
	if(read_only)
		return rf[dim];		// default: the global filter

	// still here? Now j is the first unused descriptor slot, and we should prepare a new table.
	local_desc[dim].push_back(new RFDesc(no_packs[dim], desc_num));
	for(int p = 0; p < no_packs[dim]; p++) {
		if(rf[dim][p] == RS_NONE)								// check global dimension filter
			(local_desc[dim])[j]->desc_rf[p] = RS_NONE;
		else
			(local_desc[dim])[j]->desc_rf[p] = RS_UNKNOWN;
	}
	return (local_desc[dim])[j]->desc_rf;
}

bool RoughMultiIndex::UpdateGlobalRoughFilter(int dim,int desc_num)
{
	bool any_nonempty = false;
	RSValue *loc_rs = GetLocalDescFilter(dim,desc_num,true);
	for(int p=0;p<no_packs[dim];p++) {
		if(rf[dim][p] == RS_UNKNOWN)			// not known yet - get the first information available
		{
			rf[dim][p] = loc_rs[p];
			if(rf[dim][p] != RS_NONE)
				any_nonempty = true;
		}
		else if(loc_rs[p] == RS_NONE)
			rf[dim][p] = RS_NONE;
		else if(rf[dim][p] != RS_NONE)			// else no change
		{
			any_nonempty = true;
			if(loc_rs[p] != RS_ALL || rf[dim][p] != RS_ALL)			// else rf[..] remains RS_ALL
				rf[dim][p] = RS_SOME;
		}
	}
	return any_nonempty;
}

void RoughMultiIndex::UpdateGlobalRoughFilter(int dim, Filter *loc_f)	// if the filter is nontrivial, then copy pack status
{
	if(!loc_f)
		return;
	for(int p = 0; p < no_packs[dim]; p++) {
		if(loc_f->IsEmpty(p))
			rf[dim][p] = RS_NONE;
	}
}

void RoughMultiIndex::UpdateLocalRoughFilters(int dim)						// make projection from global filters to all local ones for given dimensions
{
		for(int p = 0; p < no_packs[dim]; p++) {
			if(rf[dim][p] == RS_NONE) {
				for(int i = 0; i < local_desc[dim].size(); i++)
					(local_desc[dim])[i]->desc_rf[p] = RS_NONE;
			}
		}
}

int RoughMultiIndex::FindReducedDimension()
{
	for(int d = 0; d < no_dims; d++) {
		int omitted = 0;
		for(int i = 0; i < no_packs[d]; i++)
			if(rf[d][i] == RS_NONE)
				omitted++;
		if(no_empty_packs[d] < omitted)
			return d;
	}
	return -1;
}

void RoughMultiIndex::UpdateReducedDimension()
{
	for(int i = 0; i< no_dims; i++) {
		UpdateReducedDimension(i);
	}
}

void RoughMultiIndex::UpdateReducedDimension(int d)
{
	int omitted = 0;
	for(int i = 0; i < no_packs[d]; i++)
		if(rf[d][i] == RS_NONE)
			omitted++;
	no_empty_packs[d] = omitted;
}

vector<int> RoughMultiIndex::GetReducedDimensions()
{
	vector<int> dims;
	for(int d = 0; d < no_dims; d++) {
		int omitted = 0;
		for(int i = 0; i < no_packs[d]; i++)
			if(rf[d][i] == RS_NONE)
				omitted++;
		if(no_empty_packs[d] < omitted)
			dims.push_back(d);
	}
	return dims;

}

////////////////////////////////////////////////////

RoughMultiIndex::RFDesc::RFDesc(int packs,int d)
{
	desc_num = d;
	desc_rf = new RSValue [packs];
}

RoughMultiIndex::RFDesc::~RFDesc()
{
	delete [] desc_rf;
}

