/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "system/RCSystem.h"
#include "ftree.h"

//using namespace std;

/////////////////////////////////////////////////////////////////////

FTree::FTree()
{
	mem=NULL;
	len=NULL;
	index=NULL;
	reverse_index=NULL;
	sorted_to=0;
	value_size=0;
	last_code=-1;
	dic_size=0;
	total_dic_size=0;
	changed=0;
	compressed_size=0;
	comp_mode=-1;
}

FTree::~FTree()
{
	dealloc(mem);
	dealloc(len);
	dealloc(index);
	dealloc(reverse_index);
}

RCBString FTree::GetRealValue(int v)
{
	if(v >= 0 && v < dic_size)
		return RCBString((mem+v*value_size), len[v]);
	return RCBString();
}

char *FTree::GetBuffer(int v)
{
	if(v >= 0 && v < dic_size)
		return (mem+v*value_size);
	return NULL;
}

int FTree::GetEncodedValue(RCBString str)
{
	if(mem==NULL) return -1;
	assert(value_size>0);
	int local_last_code = last_code;	// for multithread safety

	if( local_last_code>-1 &&
		str.len==len[local_last_code] &&
		memcmp( str.val, (mem+local_last_code*value_size),str.len)==0 ) return local_last_code;

	int current_pos=0;
	if(sorted_to>5)		// optimization by sorting (note that the sorting may be partial)
	{
		int step=sorted_to;
		int direction=1;	// 1 - go up, -1 - go down
		int last_direction=0;
		int last_step=0;
		int res;
		while(!(direction!=last_direction && last_step==1))
		// if changing seek direction and step==1 => no such value in dict.
		{
			last_step=step;
			last_direction=direction;
			step/=2;					// bisection search
			if(step<1) step=1;
			if(direction==1)
				current_pos+=step;
			else
				current_pos-=step;
			res = Compare( mem+value_size*index[current_pos], len[index[current_pos]], str.val, str.len);
			if(res==0)
			{
				last_code=index[current_pos];
				return index[current_pos];
			}
			if(res > 0)
			{
				direction = -1;
				if(current_pos==0) break;				// not found
			}
			else
			{
				direction =  1;
				if(current_pos==sorted_to-1) break;		// not found
			}
		}
		current_pos = sorted_to;	// searched up to this place
	}
	// linear search on the rest of data
	for(;current_pos<dic_size;current_pos++)
	{
		if(str.len==len[current_pos] && memcmp(str.val,(mem+current_pos*value_size),str.len)==0)
		{
			last_code=current_pos;
			return current_pos;
		}
	}
	return -1;	// value not found
}

void FTree::Init(int width)
{
	value_size=width;
	dic_size=0;
	total_dic_size=10;	// minimal dictionary size
	dealloc(mem);
	mem = 0;
	if(total_dic_size*value_size*sizeof(char))
		mem = (char*)alloc(total_dic_size*value_size*sizeof(char), BLOCK_TEMPORARY);
	if(!mem) rclog << lock << "Error: out of memory (" << total_dic_size*value_size << " bytes failed). (11)" << unlock;
	memset(mem,0,total_dic_size*value_size);
	last_code=-1;
	changed=0;
}

int FTree::Add(RCBString str)
{
	assert(value_size>0);
	int i;
	int code=GetEncodedValue(str);
	if(code>-1) return code;				// Found? Do not add.
	changed=1;
	if(total_dic_size<dic_size+1)			// Enlarge table, if required
	{
		size_t mem_size = (total_dic_size+10) * value_size * sizeof(char);
		size_t len_size =(total_dic_size+10) * sizeof(ushort);
		if(mem_size)
		{
			mem = (char*)rc_realloc(mem, mem_size, BLOCK_TEMPORARY);
			memset(mem+total_dic_size*value_size,0,10*value_size);
		}
		else
			mem = 0;

		//unsigned short *temp_len = (unsigned short*)alloc((total_dic_size+10)*sizeof(unsigned short), BLOCK_TEMPORARY);
		if(len_size)
		{
			len = (unsigned short*)rc_realloc(len, len_size, BLOCK_TEMPORARY);
			for(i=total_dic_size;i<total_dic_size+10;i++)
				len[i]=0;
		}
		else
			len = 0;
		total_dic_size+=10;
	}
	dic_size++;
	int a = int((total_dic_size) * value_size * sizeof(char));
	if(str.len>0) memcpy(mem+((dic_size-1)*value_size),str.val,str.len);
	len[dic_size-1]=str.len;
	if( sorted_to<(dic_size-30) || (dic_size>5 && index==NULL) )		// sort only occasionally (every 30 new values)
	{
		dealloc(index);
		index = 0;
		if(dic_size*sizeof(int))
			index = (int*)alloc(dic_size*sizeof(int), BLOCK_TEMPORARY);
		for(int i=0;i<dic_size;i++) index[i]=i;
		QuickSort(0,dic_size-1);		// sort and set sorted_to=dic_size
		sorted_to=dic_size;
	}

	return dic_size-1;
}

int FTree::MaxValueSize(int start, int end)				// max. value size for an interval of codes
{
	int max_size = 0;
	for(int i = start; i <= end; i++)
		if(max_size < len[i])
			max_size = len[i];
	return max_size;
}


int FTree::GetSortPosition(int v)		// get the sorting reverse index of value v (the lowest string v return 0 etc.)
{
	if(dic_size==0) return 0;
	if(sorted_to<dic_size || index==NULL)
	{
		dealloc(index);
		dealloc(reverse_index);
		reverse_index = NULL;
		index = 0;
		if(dic_size*sizeof(int))
			index = (int*)alloc(dic_size*sizeof(int), BLOCK_TEMPORARY);
		for(int i=0;i<dic_size;i++) index[i]=i;
		if(dic_size>1)
			QuickSort(0,dic_size-1);
		sorted_to=dic_size;
	}
	if(reverse_index==NULL)
	{
		if(dic_size*sizeof(int))
			reverse_index = (int*)alloc(dic_size*sizeof(int), BLOCK_TEMPORARY);
		for(int i=0;i<dic_size;i++) reverse_index[index[i]]=i;
	}
	return reverse_index[v];
}

int FTree::GetSortValue(int v)
{
	if(dic_size==0) return 0;
	if(sorted_to<dic_size || index==NULL)
	{
		dealloc(index);
		dealloc(reverse_index);
		reverse_index = NULL;
		index = 0;
		if(dic_size*sizeof(int))
			index = (int*)alloc(dic_size*sizeof(int), BLOCK_TEMPORARY);
		for(int i=0;i<dic_size;i++) index[i]=i;
		if(dic_size>1)
			QuickSort(0,dic_size-1);
		sorted_to=dic_size;
	}
	assert(v >= 0 && v < dic_size);
	return index[v];
}

void FTree::QuickSort(int s1,int s2)
{
	if(s2-s1==1)
	{
		if( Compare(mem+value_size*index[s1], len[index[s1]], mem+value_size*index[s2], len[index[s1]]) == 1 )
		{
			int ppp=index[s1];
			index[s1]=index[s2];
			index[s2]=ppp;
		}
	}
	else
	{
		int i = s1;		// QuickSort here
		int j = s2;
		int mid = index[(s1+s2)/2];
		do
		{
			while(Compare(mem+value_size*index[i], len[index[i]], mem+value_size*mid, len[mid]) < 0) i++;
			while(Compare(mem+value_size*index[j], len[index[j]], mem+value_size*mid, len[mid]) > 0) j--;
			if(i<=j)
			{
				int ppp=index[i];
				index[i]=index[j];
				index[j]=ppp;
				i++;
				j--;
			}
		}
		while(i<=j);
		if(s1<j) QuickSort(s1,j);
		if(i<s2) QuickSort(i,s2);
	}
}

 int FTree::Compare(char *val1, int len1, char *val2, int len2)	// compare two strings, return -1 if val1>val2, 0 if equal, 1 if val1<val2
{
	int len = (len1<len2 ? len1 : len2);
	int result = memcmp(val1, val2, len);
	if(result>0) return 1;
	if(result<0) return -1;
	if(len1>len2) return 1;
	if(len1<len2) return -1;
	return 0;
}

int FTree::CountOfUniqueValues()
{
	return dic_size;
}

/////////////////

int FTree::ByteSize()
{
	if(changed==0 && compressed_size>0) return compressed_size;
	// header format: <format><dic_size><value_size><compressed_size><comp_mode><...len...>
	int header_size=12+2*dic_size;
	comp_mode=-1;
	if(dic_size==0) return header_size;

	int len_sum=0;
	for(int i=0;i<dic_size;i++) len_sum+=len[i];

	// format 0:
	compressed_size=header_size+len_sum;
	return compressed_size;
}

//////////////////
//  Save formats:
//
//  <format><dic_size><value_size><compressed_size><comp_mode> <len>...<len> <data>
//
//  where:
//		format=0:
//			<data>= <val1><val2>...<val_n>		- flat file; <val_i> takes exactly len[i] bytes
//

void FTree::SaveData(unsigned char *&buf)
{
	*buf=0;				// current format
	buf++;
	*((int*)buf)=dic_size;
	buf+=4;
	*((unsigned short int*)buf)=value_size;
	buf+=2;
	*((int*)buf)=compressed_size;
	buf+=4;
	*((char*)buf)=char(comp_mode);
	buf++;
	if(dic_size==0) return;
	for(int i=0;i<dic_size;i++)
	{
		*((unsigned short int*)buf)=len[i];
		buf+=2;
	}
	for(int i=0;i<dic_size;i++)
	{
		if(len[i]>0) memcpy(buf,mem+i*value_size,len[i]);
		buf+=len[i];
	}
}

void FTree::Init(unsigned char *&buf)
{
	int format=(*buf);
	buf++;
	if(format!=0)
	{
		rclog << lock << "Error: save mode " << format << " not implemented (FTree)" << unlock;
		return;
	}
	dic_size=*((int*)buf);
	buf+=4;
	total_dic_size=dic_size+10;
	value_size=*((unsigned short int*)buf);
	buf+=2;
	compressed_size=*((int*)buf);
	buf+=4;
	comp_mode=*((char*)buf);
	buf++;

	int i;

	dealloc(len);
	len = 0;
	if(total_dic_size*sizeof(ushort))
		len = (ushort*)alloc(total_dic_size*sizeof(ushort), BLOCK_TEMPORARY);
	if(!len) rclog << lock << "Error: out of memory (" << total_dic_size+10 << " bytes failed). (13a)" << unlock;
	for(i=0;i<dic_size;i++)
	{
		len[i]=*((unsigned short int*)buf);
		buf+=2;
	}
	for(i=dic_size;i<total_dic_size;i++) len[i]=0;
	dealloc(mem);
	mem = 0;
	if(total_dic_size*value_size*sizeof(char))
		mem = (char*)alloc(total_dic_size*value_size*sizeof(char), BLOCK_TEMPORARY);
	if(!mem) rclog << lock << "Error: out of memory (" << total_dic_size*value_size << " bytes failed). (13)" << unlock;
	memset(mem,0,total_dic_size*value_size);
	for(i=0;i<dic_size;i++)
	{
		if(len[i]>0) memcpy(mem+i*value_size,buf,len[i]);
		buf+=len[i];
	}

	if(dic_size>5)
	{
		dealloc(index);
		index = 0;
		if(dic_size*sizeof(int))
			index = (int*)alloc(dic_size*sizeof(int), BLOCK_TEMPORARY);
		for(int i=0;i<dic_size;i++) index[i]=i;
		QuickSort(0,dic_size-1);		// sort and set sorted_to=dic_size
		sorted_to=dic_size;
	}

	last_code=-1;
	changed=0;
}

