/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include <algorithm>
#include "system/RCException.h"
#include "core/RCAttrTypeInfo.h"
#include "DataSource.h"

using namespace std;

namespace na {

//***********************************  DSTable  ***************************************//

void DSTable::Open()
{
	if(current) throw InternalRCException("DSTable::Open() invoked twice before Close()");
	if(packlock >= 0) throw InternalRCException("");
	total = tab->NoObj();
}

bool DSTable::GetNext()
{
	if(current >= total) return false;

	// lock packs that will be read now
	int64 pack = tab->PackIndex(current);
	if(packlock != pack) {
		BHASSERT_WITH_NO_PERFORMANCE_IMPACT(packlock == -1);
		//for each(int c in columns)
		//	tab->LockPackForUse(c, (uint)pack);
		long lCount = columns.size();
		for( long idx=0; idx < lCount;idx++)
			tab->LockPackForUse(columns[idx], (uint)pack, ConnectionInfoOnTLS.Get());		// TODO: should use ConnectionInfo for statistics.
		packlock = pack;
	}

	// read row
	for(size_t i = 0; i < vals.size(); i++) {
		BHASSERT_WITH_NO_PERFORMANCE_IMPACT(tab->AttrType(columns[i]));
		if(tab->IsNull(current, columns[i]))	vals[i]->SetNull();
		else									((VInteger*)vals[i])->Set(tab->GetTable64(current, columns[i]));
	}
	current++;

	// should we unlock the pack before reading next row?
	if(packlock != tab->PackIndex(current)) {
		//for each(int c in columns)
		//	tab->UnlockPackFromUse(c, (uint)packlock);
		long lCount = columns.size();
		for( long idx=0; idx < lCount;idx++)
			tab->UnlockPackFromUse(columns[idx], (uint)packlock);
		packlock = -1;
	}

	return true;
}

void DSTable::Close()
{
	if(packlock < 0) return;
	//for each(int c in columns)
	//	tab->UnlockPackFromUse(c, (uint)packlock);
	long lCount = columns.size();
	for( long idx=0; idx < lCount;idx++)
		tab->UnlockPackFromUse(columns[idx], (uint)packlock);

	packlock = -1;
	current = 0;
}

void DSTable::SetBuffers(BufOfVars *bufs)
{
	if(!bufs) {
		BHERROR("'bufs' should not be null");
		return;
	}

	// which columns are required by the next iterator
	BufOfVars::iterator it;
	for(it = bufs->begin(); it != bufs->end(); it++)
		if(it->first.srctab == this) {
			columns.push_back(it->first.index);
			vals.push_back(CreateBuf(it->first.index));
		}
		else BHERROR("should be 'it->first.srctab == this'");

	// map columns to output fields
	Variable cd;
	cd.srctab = this;
	cd.bhtab = tab;
	for(size_t i = 0; i < columns.size(); i++) {
		cd.index = columns[i];
		(*bufs)[cd] = vals[i];
	}
}

AValue* DSTable::CreateBuf(int col)
{
	if(col >= tab->NoAttrs()) {
		BHERROR("should be 'col < tab->NoAttrs()'");
	}
	AttributeType at = tab->AttrType(col);
	if(!ATI::IsIntegerType(at))	throw UnsupportedDataTypeRCException();

	return new VInteger;
}


//***********************************  DSSelect  ***************************************//

void DSSelect::GetResultBufs(std::vector<AValue*>& bufs)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(bufs.empty());
	bufs.resize(expr.size());
	for(size_t i = 0; i < expr.size(); i++)
		bufs[i] = expr[i]->GetBuffer();
}

void DSSelect::SetBuffers(BufOfVars *bufs_out)
{
	if(bufs_out != NULL) {
		BHERROR("'bufs_out' not equals NULL");
	} // will be used in the future, with 'FROM' subqueries

	SetOfVars vars;
	BufOfVars bufs;
	for(size_t i = 0; i < expr.size(); i++)
		expr[i]->GetVars(&vars);

	SetOfVars::iterator it;
	for(it = vars.begin(); it != vars.end(); it++)
		bufs.insert(pair<Variable,AValue*>(*it, NULL));

	src->SetBuffers(&bufs);

	for(size_t i = 0; i < expr.size(); i++)
		expr[i]->SetBufs(&bufs);
}


}	// namespace na

