/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "vc/VirtualColumn.h"
#include "system/Configuration.h"
#include "shared/core/RCEngine.h"
#include "core/compilation_tools.h"
#include "common/mysql_gate.h"
#include "ha_rcengine.h"

namespace
{
bool AtLeastOneIBTableInvolved(LEX *lex)
{
	for(TABLE_LIST *table_list = lex->query_tables; table_list; table_list = table_list->next_global) {
		TABLE* table = table_list->table;
		if(RCEngine::IsBHTable(table))
			return TRUE;
	}
	return FALSE;
}

bool ForbiddenMySQLQueryPath(LEX *lex)
{
	if(lex->sql_command == SQLCOM_SELECT &&
		Configuration::GetProperty(Configuration::AllowMySQLQueryPath) == 0 &&
		AtLeastOneIBTableInvolved(lex))
		return TRUE;
	return FALSE;
}
}

void BH_UpdateAndStoreColumnComment(TABLE* table, int field_id, Field* source_field, int source_field_id, CHARSET_INFO *cs)
{
	try {
		rceng->UpdateAndStoreColumnComment(table, field_id, source_field, source_field_id, cs);
	}
	catch (std::exception& e) {
		my_message(BHERROR_UNKNOWN, e.what(), MYF(0));
		rclog << lock << "Error: " << e.what() << unlock;
	} catch (...) {
		my_message(BHERROR_UNKNOWN, "An unknown system exception error caught.", MYF(0));
		rclog << lock << "Error: An unknown system exception error caught." << unlock;
	}
}

int BH_HandleSelect(THD *thd, LEX *lex, select_result *&result, ulong setup_tables_done_option, int& res, int& optimize_after_bh, int& bh_free_join, int with_insert)
{
	int ret = RCBASE_QUERY_ROUTE;
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT( MysqlExpression::get_instance_count() == 0 );
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT( VirtualColumn::get_instance_count() == 0 );
	try {
		// handle_select_ret is introduced here because in case of some exceptions (e.g. thrown from ForbiddenMySQLQueryPath)
		// we want to return RCBASE_QUERY_ROUTE
		int handle_select_ret =
			rceng->HandleSelect(thd, lex, result, setup_tables_done_option, res, optimize_after_bh, bh_free_join, with_insert);
		if(handle_select_ret == RETURN_QUERY_TO_MYSQL_ROUTE && ForbiddenMySQLQueryPath(lex)) {
			my_message(BHERROR_UNKNOWN,
					"The query includes syntax that is not supported by the Infobright Optimizer. \
Either restructure the query with supported syntax, or enable the MySQL Query Path in the brighthouse.ini file \
to execute the query with reduced performance.", MYF(0));
			//rceng->Rollback(thd, true, true);
			handle_select_ret = RCBASE_QUERY_ROUTE;
		}
		ret = handle_select_ret;
	}
	catch (std::exception& e) {
		my_message(BHERROR_UNKNOWN, e.what(), MYF(0));
		rclog << lock << "Error: " << e.what() << unlock;
	} catch (...) {
		my_message(BHERROR_UNKNOWN, "An unknown system exception error caught.", MYF(0));
		rclog << lock << "Error: An unknown system exception error caught." << unlock;
	}
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT( VirtualColumn::get_instance_count() == 0 );
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT( MysqlExpression::get_instance_count() == 0 );
	return ret;
}

int BH_LoadData(THD *thd, sql_exchange *ex, TABLE_LIST *table_list,	char* errmsg, int len, int &errcode)
{
	BHError bh_error;
	int ret = BH_LD_Failed;
	try	{
		ret = rceng->RunLoader(thd, ex, table_list, bh_error);
		if(ret == BH_LD_Failed)
			rclog << lock << "Error: "<< bh_error.Message().c_str() << unlock;
	} catch (std::exception& e)	{
		bh_error = BHError(BHERROR_UNKNOWN, e.what());
	} catch (...) {
		bh_error = BHError(BHERROR_UNKNOWN, "An unknown system exception error caught.");
		rclog << lock << "Error: "<< bh_error.Message().c_str() << unlock;
	}
	strncpy(errmsg, bh_error.Message().c_str(), len - 1);
	errmsg[len-1] = '\0';
	errcode = (int)bh_error.ErrorCode();
	return ret;
}

void BH_TerminateLoadBH(THD* thd)
{
	try {
		rceng->TerminateLoader(thd);
	} catch (std::exception& e) {
		my_message(BHERROR_UNKNOWN, e.what(), MYF(0));
		rclog << lock << "Error: " << e.what() << unlock;
	} catch (...) {
		my_message(BHERROR_UNKNOWN, "An unknown system exception error caught.", MYF(0));
		rclog << lock << "Error: An unknown system exception error caught." << unlock;
	}
}

