/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include <boost/algorithm/string.hpp>

#ifndef __BH_COMMUNITY__
#	include "enterprise/local.h"
#	include "enterprise/loader/LoaderFactory.h"
#else /* not __BH_COMMUNITY__ */
#	include "community/local.h"
#	include "community/loader/LoaderFactory.h"
#endif /* __BH_COMMUNITY__ */
#include "RCEngine.h"
#include "system/FileOut.h"
#include "system/StdOut.h"
#include "system/MemoryManager.h"
#include "system/MemoryManagement/Initializer.h"
#include "system/BHToolkit.h"
#include "shared/loader/Loader.h"
#include "core/tools.h"
#include "core/TempTable.h"
#include "system/RCException.h"
#include "system/IBLicense.h"
#include "core/Transaction.h"

#ifdef test
#undef test
#endif

using namespace std;
using namespace boost;

#ifdef PROFILE_LOCK_WAITING
LockProfiler lock_profiler;
#endif

int RCEngine::Init(const char* tab_ext, uint engine_slot)
{
	m_tab_ext = tab_ext;
	m_engine_slot = engine_slot;
	BHASSERT( infobright_data_dir.length(), "data dir not set" );
	BHASSERT( process_type != ProcessType::INVALID, "process type not set" );
	m_log_output = new FileOut(infobright_data_dir + "brighthouse.log");
	rclog.addOutput(m_log_output);
	m_control_output = ( ( process_type != ProcessType::EMBEDDED ) || !log_error_file_ptr || !log_error_file_ptr[0] ) ? static_cast<ChannelOut*>( new StdOut() ) : static_cast<ChannelOut*>( new FileOut( log_error_file_ptr ) );
	rccontrol.addOutput(m_control_output);

	try {
		Configuration::LoadSettings();
	}
	catch (const Configuration::Exception& e) {
		sql_print_error("Unable to load configuration");
		sql_print_error(e.what(), 0);
		rclog << lock << e.what() << " Configuration::Exception" << __FILE__ << __LINE__ << unlock;
		return 1;
	}
    ConfigureRCControl();
    if(Configuration::GetProperty(Configuration::InternalMessages)){
        m_dev_output = new FileOut(infobright_data_dir + "development.log");
        rcdev.addOutput(m_dev_output);
        rcdev.setOn();
    }else{
        rcdev.setOff();
    }
    DefaultPushDown = Configuration::GetProperty(Configuration::PushDown);
#ifdef __BH_EVALUATION__
	try {
		IBLicense* license = IBLicense::GetLicense();
                license->SetProductDir(infobright_home_dir);
                license->SetDataDir(infobright_data_dir);
		string file_name = Configuration::GetProperty(Configuration::LicenseFile);
		if (!license->SetKeyFile(file_name)) return 1;
                license->Load();
		if (license->IsExpired()) {
                    sql_print_error(EXPIRATION_MESSAGE);
                    rclog << lock << EXPIRATION_MESSAGE << unlock;
   		    return 1;
		}
	}
	catch (RCException& e) {
            sql_print_error("Error: License verification failed!");
            sql_print_error(e.what(), 0);
            rclog << lock << e.what() << unlock;
	    return 1;
	}
#endif
    srand(unsigned(time(NULL)));
    size_t main_size = size_t(Configuration::GetProperty(Configuration::ServerMainHeapSize)) << 20;
    size_t compressed_size = size_t(Configuration::GetProperty(Configuration::ServerCompressedHeapSize)) << 20;
    m_loader_main_heapsize = size_t(Configuration::GetProperty(Configuration::LoaderMainHeapSize)) << 20;
    m_loader_compressed_heapsize = 128 << 20;
    string hugefiledir = Configuration::GetProperty(Configuration::HugefileDir);
    if(hugefiledir.empty())
        MemoryManagerInitializer::Instance(compressed_size, main_size, compressed_size / 10, main_size / 10);

    else
        MemoryManagerInitializer::Instance(compressed_size, main_size, compressed_size / 10, main_size / 10, hugefiledir);

		string p = infobright_data_dir + "brighthouse.seq";
    ifstream seq_file(p.c_str());
    if(! DoesFileExist(infobright_data_dir + "brighthouse.seq")) {
		m_tab_sequencer = 0;
		ofstream seq_file(p.c_str());
		if(seq_file)
			seq_file << m_tab_sequencer;
		if(!seq_file) {
			m_tab_sequencer = -1;
			rclog << lock << "Brighthouse: Fatal error: can not create or write the table sequencer file" << unlock;
		}
		seq_file.close();
	} else {
		ifstream seq_file(p.c_str());
		if(seq_file)
			seq_file >> m_tab_sequencer;
		if(!seq_file) {
			m_tab_sequencer = -1;
			rclog << lock << "Brighthouse: Fatal error: can not read the sequence number from the table sequencer file"
					<< unlock;
		}
		seq_file.close();
	}

    if(InitRSIManager(infobright_data_dir, true) != 0)
    	return 1;
	InitCachingLevel();
	//m_globalCache = new GlobalDataCache;

	try {
		string cachefolder_path = Configuration::GetProperty(Configuration::CacheFolder);
		if (SetUpCacheFolder(cachefolder_path) != 0)
			return 1;
		RemoveTemporaryFiles(cachefolder_path);
	}
	catch (const Configuration::Exception&) {
		sql_print_error("'CacheFolder' not defined.");
		rclog << lock << "Error: 'CacheFolder' not defined." << unlock;
		return 1;
	}

#	ifdef FUNCTIONS_EXECUTION_TIMES
	fet = new FunctionsExecutionTimes();
#	endif
	try {
		use_charset =  Configuration::GetProperty(Configuration::UseCharset);
	} catch(...) {
		use_charset = true;
	}

	rclog << lock << "Brighthouse engine started. " << unlock;
	rccontrol.lock(0) << "Brighthouse engine started, ";
	if(rsi_manager)
		rccontrol << "KNs: ok";
	else
		rccontrol << "KNs: none";
	rccontrol << unlock;
	return 0;
}

RCEngine::~RCEngine()
{

	delete rsi_manager;
	rsi_manager=NULL;

//	for(map<string,map<string, boost::shared_ptr<RCTable> > >::const_iterator db_iter = m_tables.begin(); db_iter!=m_tables.end(); db_iter++)
//		for(map<string,RCTable*>::const_iterator tab_iter = db_iter->second.begin(); tab_iter!=db_iter->second.end(); tab_iter++)
//			tab_iter->second.reset();
/**************************************************
#ifdef __GNUC__
	pthread_mutex_destroy(&m_sync_tables);
	pthread_mutex_destroy(&m_sync_threadparams);
#else
	DeleteCriticalSection(&m_sync_tables);
	DeleteCriticalSection(&m_sync_threadparams);
#endif
*****************************************************/

	m_tables.clear();
	if(rccontrol.isOn())
		MemoryManagerInitializer::deinit(true);
	else
		MemoryManagerInitializer::deinit(false);

#ifdef FUNCTIONS_EXECUTION_TIMES
	fet->PrintToRcdev();
	//delete fet;
	//fet = NULL;
#endif
	if (m_control_output) {
		m_control_output->close();
		delete m_control_output;
		m_control_output = NULL;
	}
	rclog << lock << "Brighthouse engine shutdown." << unlock;

	if(m_log_output) {
		m_log_output->close();
		delete m_log_output;
		m_log_output = NULL;
	}
	if(m_dev_output) {
		m_dev_output->close();
		delete m_dev_output;
		m_dev_output = NULL;
	}
}


BHErrorCode RCEngine::CreateTable(const char *table, TABLE *from)
{
	if(table == 0 || from == 0)
		return BHERROR_UNKNOWN;

	if(m_tab_sequencer == -1) {
		rclog << lock << "Brighthouse: Fatal error: can not create tables, the sequencer file did not work" << unlock;
		return BHERROR_UNKNOWN;
	};
	if(m_tab_sequencer == INT_MAX32) {
		rclog << lock
				<< "Brighthouse: Can not create more tables in this version please ask the support team for an update"
				<< unlock;
		return BHERROR_UNKNOWN;
	};
	string p = infobright_data_dir + "brighthouse.seq";
	ofstream seq_file(p.c_str());
	if(seq_file)
		seq_file << (m_tab_sequencer + 1);
	if(!seq_file) {
		m_tab_sequencer = -1;
		rclog << lock << "Brighthouse: Fatal error: can not write the table sequencer file" << unlock;
		seq_file.close();
		return BHERROR_UNKNOWN;
	}
	seq_file.close();
	size_t path_len = strlen(table) + strlen(m_tab_ext) + 1;
	char* tab_path = new char[path_len];

	/*strcpy_s(tab_path, path_len, table);
	 strcat_s(tab_path, path_len, m_tab_ext);*/
	strcpy(tab_path, table);
	strcat(tab_path, m_tab_ext);

	char db_name[256] = { 0 }, tab_name[256] = { 0 };
	GetNames(table, db_name, tab_name, 256);

	vector<DTCollation> charsets;
	if(from->s) {
		for(int i = 0; i < from->s->fields; i++) {
			const Field_str* fstr = dynamic_cast<const Field_str*>(from->s->field[i]);
			if(fstr)
				charsets.push_back(DTCollation(fstr->charset(), fstr->derivation()));
			else
				charsets.push_back(DTCollation());
		}
	}

	shared_ptr<RCTable> tab = boost::shared_ptr<RCTable> (new RCTable(tab_path, charsets, tab_name, m_tab_sequencer++, 0, true));
	delete[] tab_path;

	int fid = 0;
	Field* field = from->field[fid];
	try {
		while(field != 0) {
			AttributeTypeInfo ati = GetAttrTypeInfo(*field);
			tab->AddAttribute(const_cast<char*> (field->field_name), ati, ati.Precision(), ati.Scale(), ati.Params(), ati.GetCollation());
			field = from->field[++fid];
		}
	} catch (UnsupportedDataTypeRCException e) {
		return BHERROR_UNSUPPORTED_DATATYPE;
	} catch (...) {
		return BHERROR_UNKNOWN;
	}
	tab->Save();
	tab->CommitSaveSession(0, 0); // TO DO: provide transaction id

	//pthread_mutex_lock(&m_sync_tables);
    tables_mutex.Lock();

	m_tables[db_name][tab_name] = tab;

	//pthread_mutex_unlock(&m_sync_tables);
    tables_mutex.Unlock();

	return BHERROR_SUCCESS;
	// TO DO: logging ?
}

void RCEngine::GetNames(const char* table_path, char* db_buf, char* tab_buf, size_t buf_size)
{
	size_t path_len = strlen(table_path);
	unsigned last = -1, last_but_one = -1;
	for(unsigned i = 0; i < path_len; i++) {
		if(table_path[i]=='/' || table_path[i]=='\\') {
			last_but_one = last;
			last = i+1;
		}
	}
	if(last_but_one==-1 || last-last_but_one > buf_size-1 || path_len-last>buf_size-2)
		return;
	/*
	 strncpy_s(db_buf, buf_size, &table_path[last_but_one], last-last_but_one-1);
	 strncpy_s(tab_buf, buf_size, &table_path[last], path_len-last); */
	std::strncpy(db_buf, table_path + last_but_one, last-last_but_one-1);
	std::strncpy(tab_buf, table_path + last, path_len-last);
}

ATI RCEngine::GetAttrTypeInfo(const Field& field) throw(UnsupportedDataTypeRCException)
{
	bool is_lookup = false;
	if(has_lookup(field.comment) == 0)
		is_lookup = true;

	bool for_insert = false;
	string c = string(field.comment.str, field.comment.length);
	int i;
	if ((i=c.find("for_insert")) != string::npos)
		for_insert = true;

	AttributeType attrt = RC_INT;
	enum_field_types eft = field.real_type();
	switch(field.type()) {
		case MYSQL_TYPE_SHORT:
		case MYSQL_TYPE_TINY:
		case MYSQL_TYPE_INT24:
		case MYSQL_TYPE_LONG:
		case MYSQL_TYPE_FLOAT:
		case MYSQL_TYPE_DOUBLE:
		case MYSQL_TYPE_LONGLONG:
			if(field.flags & UNSIGNED_FLAG)
				throw UnsupportedDataTypeRCException("UNSIGNED data types are not supported.");
		case MYSQL_TYPE_YEAR:
		case MYSQL_TYPE_TIMESTAMP:
		case MYSQL_TYPE_DATETIME:
		case MYSQL_TYPE_DATE:
		case MYSQL_TYPE_NEWDATE:
			return ATI(RCEngine::GetCorrespondingType(field), field.null_bit == 0, (ushort)field.field_length, 0,
					is_lookup, for_insert);
		case MYSQL_TYPE_TIME:
			return ATI(RC_TIME, field.null_bit == 0, 0, 0, is_lookup, for_insert);
			/*case MYSQL_TYPE_YEAR :
			 return ATI(RC_YEAR, field.null_bit == 0, (ushort)field.field_length, 0, is_lookup);*/
		case MYSQL_TYPE_STRING:
		case MYSQL_TYPE_VARCHAR: {
			if(field.field_length > 65535)
				throw UnsupportedDataTypeRCException();
			if(const Field_str* fstr = dynamic_cast<const Field_string*>(&field))
			{
				if(fstr->charset() != &my_charset_bin)
					return ATI(RC_STRING, field.null_bit == 0, (ushort)field.field_length, 0, is_lookup, for_insert, DTCollation(fstr->charset(), fstr->derivation()));
				return ATI(RC_BYTE, field.null_bit == 0, (ushort)field.field_length, 0, is_lookup, for_insert, DTCollation(fstr->charset(), fstr->derivation()));
			}
			else if(const Field_str* fvstr = dynamic_cast<const Field_varstring*>(&field))
			{
				if(fvstr->charset() != &my_charset_bin)
					return ATI(RC_VARCHAR, field.null_bit == 0, (ushort)field.field_length, 0, is_lookup, for_insert, DTCollation(fvstr->charset(), fvstr->derivation()));
				return ATI(RC_VARBYTE, field.null_bit == 0, (ushort)field.field_length, 0, is_lookup, for_insert, DTCollation(fvstr->charset(), fvstr->derivation()));

			}
			throw UnsupportedDataTypeRCException();
		}
		case MYSQL_TYPE_NEWDECIMAL :
		{
			if(field.flags & UNSIGNED_FLAG)
			throw UnsupportedDataTypeRCException("UNSIGNED data types are not supported.");
			const Field_new_decimal* fnd = ((const Field_new_decimal*)&field);
			if(/*fnd->precision > 0 && */fnd->precision <= 18 /*&& fnd->dec >= 0*/&& fnd->dec <= fnd->precision)
			return ATI(RC_NUM, field.null_bit == 0, fnd->precision, fnd->dec, is_lookup, for_insert);
			throw UnsupportedDataTypeRCException();
		}
		case MYSQL_TYPE_BLOB :
		if(const Field_str* fstr = dynamic_cast<const Field_str*>(&field))
		{
			if(const Field_blob* fblo = dynamic_cast<const Field_blob*>(fstr))
			{
				if(fblo->charset() != &my_charset_bin)
				{//TINYTEXT, MEDIUMTEXT, TEXT, LONGTEXT
					if(field.field_length> 65535)
					throw UnsupportedDataTypeRCException();
					return ATI(RC_VARCHAR, field.null_bit == 0, (ushort)field.field_length, 0, is_lookup, for_insert);
				}
				else
				{
					switch(field.field_length)
					{
						if(field.field_length> 65535)
						throw UnsupportedDataTypeRCException();
						case 255 :
						case 65535 :
						//TINYBLOB, BLOB
						return ATI(RC_VARBYTE, field.null_bit == 0, (ushort)field.field_length, 0, is_lookup, for_insert);
						case 16777215 :
						case 4294967295 :
						//MEDIUMBLOB, LONGBLOB
						return ATI(RC_BIN, field.null_bit == 0, (ushort)field.field_length, 0, is_lookup, for_insert);
						default:
						throw UnsupportedDataTypeRCException();
					}
				}
			}
		}
		default :
		//MYSQL_TYPE_LONG_BLOB
		//MYSQL_TYPE_TINY_BLOB
		//MYSQL_TYPE_MEDIUM_BLOB
		//MYSQL_TYPE_NULL
		//MYSQL_TYPE_BIT
		//MYSQL_TYPE_GEOMETRY
		//MYSQL_TYPE_ENUM
		//MYSQL_TYPE_SET
		//MYSQL_TYPE_VAR_STRING
		throw UnsupportedDataTypeRCException();
	}
	throw;
}

void RCEngine::Rollback(THD* thd, bool all, bool force_error_message)
{
	IBGuard guard(tables_mutex);
	TablesListAutoPtr tables_list = rceng->GetListOfModifiedTables(*thd);
	BHASSERT(tables_list.get(), "Unable to obtain a list of tables to commit.");

	if(tables_list->size()) {

		//if(force_error_message || thd->query_error || (!all && (thd->options & OPTION_NOT_AUTOCOMMIT))) {
		// GA make sure query_error is not required for ga migration
		if(force_error_message || (!all && (thd->options & OPTION_NOT_AUTOCOMMIT))) {
			string message("The Infobright storage engine has encountered an unexpected error. The current transaction has been rolled back.");
			rclog << lock << message.c_str() << unlock;
			message += " For details on the error please see the brighthouse.log file in the " + infobright_data_dir + " directory.";
			my_message(ER_XA_RBROLLBACK, message.c_str(), MYF(0));
		}

		TablesList::iterator iter = tables_list->begin();
		TablesList::iterator end = tables_list->end();

		for(; iter != end; ++iter) {
			DropPendingChanges(thd, *iter, NULL);
			RefreshTable((*iter).c_str(), 0);
			SynchronizedDataCache::GlobalDataCache().ReleasePacks(GetTable((*iter).c_str(), NULL)->GetID());
			table_lock_manager.ReleaseLocks(*thd, *iter);
		}

		if(rsi_manager)
			rsi_manager->UpdateDef();

		if(all && thd->lock)
			thd->lock = 0;
		thd->server_status &= ~SERVER_STATUS_IN_TRANS;
	}
	rceng->GetThreadParams(*thd)->EndTransaction();
}

int has_lookup(const LEX_STRING &comment)
{
	char *str = comment.str;
	uint length = comment.length;
	if(length == 0)
		return 1;

	char lookup_str[] = "LOOKUP";
	uint const lookup_len = (const uint) strlen(lookup_str);

	uint idx = 0;
	while((str[0] == ' ' || str[0] == '\t' || str[0] == '\n') && idx < length) {
		str++;
		idx++;
	}
	if(length - idx < lookup_len)
		return 1;

	uint const buf_size = 128;
	char buf[buf_size] = { 0 };
	//strncpy_s(buf, 128, str, lookup_len);
	std::strncpy(buf, str, lookup_len);
	if(_stricmp(buf, "LOOKUP") == 0)
		return 0;
	else
		return 1;

}

void RCEngine::RenameTable(const char* from, const char* to)
{
	char db_name[256] = { 0 }, tab_name[256] = { 0 };
	GetNames(from, db_name, tab_name, 256);
	//pthread_mutex_lock(&m_sync_tables);
    tables_mutex.Lock();
	RCTablePtr tab = m_tables[db_name][tab_name];
	if(tab) {
		m_tables[db_name].erase(tab_name);
		table_lock_manager.UnRegisterTable(from);
	}
    tables_mutex.Unlock();

	if(!tab) {
		string tab_path = from;
		tab_path += m_tab_ext;

		try {
			tab = RCTablePtr(new RCTable(tab_path.c_str(), vector<DTCollation>(), -1, 0, true, false));	// Version: cannot drop inconsistent tables
//			tab = RCTablePtr(new RCTable(tab_path.c_str(), -1, 0, true, true));		// Version: can drop inconsistent tables
		} catch (NoTableFolderRCException&) {
			return;
		} catch (DatabaseRCException&) {
			throw;
		}
	}

	tab->Rollback(0);
	string f(from);
	f = f + string(".bht");
	string t(to);
	t = t + string(".bht");

	RenameFile(f.c_str(), t.c_str());
}

void RCEngine::DropTable(const char *table)
{
	char db_name[256] = { 0 }, tab_name[256] = { 0 };
	GetNames(table, db_name, tab_name, 256);
	//pthread_mutex_lock(&m_sync_tables);
    tables_mutex.Lock();
	RCTablePtr tab = m_tables[db_name][tab_name];
	if(tab) {
		m_tables[db_name].erase(tab_name);
		table_lock_manager.UnRegisterTable(table);
	}
    tables_mutex.Unlock();

	if(!tab) {
		string tab_path = table;
		tab_path += m_tab_ext;

		try {
			tab = RCTablePtr(new RCTable(tab_path.c_str(), vector<DTCollation>(), -1, 0, true, false));	// Version: cannot drop inconsistent tables
//			tab = RCTablePtr(new RCTable(tab_path.c_str(), -1, 0, true, true));		// Version: can drop inconsistent tables
		} catch (NoTableFolderRCException&) {
			return;
		} catch (DatabaseRCException&) {
			throw;
		}
	}
	SynchronizedDataCache::GlobalDataCache().ReleasePacks(tab->GetID());
	tab->Rollback(0);
	tab->Drop();

	//ReleaseAllLocks(table);
	// TO DO: logging ?
}

RCTablePtr RCEngine::GetTableShared(const string& table_path, struct st_table_share* table_share, bool locked) throw(DatabaseRCException)
{
	char db_name[256] = { 0 }, tab_name[256] = { 0 };
	GetNames(table_path.c_str(), db_name, tab_name, 256);
	//pthread_mutex_lock(&m_sync_tables);
    tables_mutex.Lock();
	RCTablePtr tab = m_tables[db_name][tab_name];
    tables_mutex.Unlock();
	//pthread_mutex_unlock(&m_sync_tables);
	if(!tab) {
		vector<DTCollation> charsets;
		if(table_share) {
			for(int i = 0; i < table_share->fields; i++) {
				const Field_str* fstr = dynamic_cast<const Field_str*>(table_share->field[i]);
				if(fstr)
					charsets.push_back(DTCollation(fstr->charset(), fstr->derivation()));
				else
					charsets.push_back(DTCollation());
			}
		}

		string file_path = table_path + m_tab_ext;
		try {
			tab = RCTablePtr(new RCTable(file_path.c_str(), charsets, -1, 0, true));
		} catch (DatabaseRCException&) {
			throw;
		}
		//pthread_mutex_lock(&m_sync_tables);
        tables_mutex.Lock();
		m_tables[db_name][tab_name] = tab;
        tables_mutex.Unlock();
		//pthread_mutex_unlock(&m_sync_tables);
		// rollback if there was no commit
		if(tab->GetSaveSessionId() != 0xFFFFFFFF) {
			if(!locked)
				tables_mutex.Lock();//pthread_mutex_lock(&m_sync_tables);
			if(tab->GetSaveSessionId() != 0xFFFFFFFF) {
				tab->Rollback();
				rclog << lock << "WARNING: Rollback after uncommitted transaction on table " << db_name << '.'
				<< tab_name << unlock;
			}
			if(!locked)
				tables_mutex.Unlock();//pthread_mutex_unlock(&m_sync_tables);
		}
	}
	return tab;
}

RCTable* RCEngine::GetTable(const string& table_path, struct st_table_share* table_share, bool locked) throw(DatabaseRCException)
{
	return GetTableShared(table_path, table_share, locked).get();
}

std::auto_ptr<RCTable> RCEngine::GetCurrentTableState(const char* table_path, struct st_table_share* table_share) throw(DatabaseRCException)
{
	char db_name[256] = { 0 };
	char tab_name[256] = { 0 };
	GetNames(table_path, db_name, tab_name, 256);
	//pthread_mutex_lock(&m_sync_tables);
    tables_mutex.Lock();
	std::auto_ptr<RCTable> tab;
	string tab_path = table_path;
	tab_path += m_tab_ext;
	try {
		vector<DTCollation> charsets;
		if(table_share) {
			for(int i = 0; i < table_share->fields; i++) {
				const Field_str* fstr = dynamic_cast<const Field_str*>(table_share->field[i]);
				if(fstr)
					charsets.push_back(DTCollation(fstr->charset(), fstr->derivation()));
				else
					charsets.push_back(DTCollation());
			}
		}

		tab = std::auto_ptr<RCTable>(new RCTable(tab_path.c_str(), charsets, -1, 0, true));
	} catch (DatabaseRCException&) {
		//pthread_mutex_unlock(&m_sync_tables);
        tables_mutex.Unlock();
		throw;
	}
	//pthread_mutex_unlock(&m_sync_tables);
    tables_mutex.Unlock();
	return tab;
}

void RCEngine::OpenTable(const char* table_path, struct st_table_share* table_share) throw(DatabaseRCException)
{
	//pthread_mutex_lock(&m_sync_tables);
    tables_mutex.Lock();
	try {
		GetTable(table_path, table_share, true);
	} catch (DatabaseRCException&) {
		//pthread_mutex_unlock(&m_sync_tables);
        tables_mutex.Unlock();
		throw ;
	}
	//pthread_mutex_unlock(&m_sync_tables);
    tables_mutex.Unlock();
}

time_t RCEngine::GetTableCreateTime(const char* table_path, struct st_table_share* table_share) throw(DatabaseRCException)
{
	time_t res = 0;
	//pthread_mutex_lock(&m_sync_tables);
    tables_mutex.Lock();
	try {
		res = GetTable(table_path, table_share, true)->GetCreateTime();
	} catch (DatabaseRCException&) {
		//pthread_mutex_unlock(&m_sync_tables);
        tables_mutex.Unlock();
		throw ;
	}
	//pthread_mutex_unlock(&m_sync_tables);
    tables_mutex.Unlock();
	return res;
}

time_t RCEngine::GetTableUpdateTime(const char* table_path, struct st_table_share* table_share) throw(DatabaseRCException)
{
	time_t res = 0;
	//pthread_mutex_lock(&m_sync_tables);
    tables_mutex.Lock();
	try {
		res = GetTable(table_path, table_share, true)->GetUpdateTime();
	} catch (DatabaseRCException&) {
		//pthread_mutex_unlock(&m_sync_tables);
        tables_mutex.Unlock();
		throw ;
	}
	//pthread_mutex_unlock(&m_sync_tables);
    tables_mutex.Unlock();
	return res;
}

void RCEngine::GetTableAttributes(RCTable& tab, AttrInfo*& attributes, uint& no_attr)
{
	attributes = tab.GetAttributes();
	no_attr = tab.NoAttrs();
}

void RCEngine::GetTableAttributes(const char* table_path, struct st_table_share* table_share, AttrInfo*& attributes, uint& no_attr)
	throw(DatabaseRCException)
{
	RCTable* tab;
	//pthread_mutex_lock(&m_sync_tables);
    tables_mutex.Lock();
	try {
		tab = GetTable(table_path, table_share, true);
	} catch (DatabaseRCException&) {
		//pthread_mutex_unlock(&m_sync_tables);
        tables_mutex.Unlock();
		throw ;
	}
	GetTableAttributes(*tab, attributes, no_attr);
	//pthread_mutex_unlock(&m_sync_tables);
    tables_mutex.Unlock();
	rccontrol << "Table " << table_path << " (" << tab->GetID() << ") accessed by MySQL engine." << endl;
}

void RCEngine::GetCurrentTableStateAttributes(const char* table_path, struct st_table_share* table_share, AttrInfo*& attributes, uint& no_attr)
	throw(DatabaseRCException)
{
	std::auto_ptr<RCTable> tab;
	//pthread_mutex_lock(&m_sync_tables);
    tables_mutex.Lock();
	try {
		tab = GetCurrentTableState(table_path, table_share);
	} catch (DatabaseRCException&) {
		//pthread_mutex_unlock(&m_sync_tables);
        tables_mutex.Unlock();
		throw ;
	}
	GetTableAttributes(*tab, attributes, no_attr);
	//pthread_mutex_unlock(&m_sync_tables);
    tables_mutex.Unlock();
	rccontrol << "Table " << table_path << " (" << tab->GetID() << ") accessed by MySQL engine." << endl;
}

_uint64 RCEngine::GetTableNoObj(const char* table_path, struct st_table_share* table_share) throw(DatabaseRCException)
{
	_uint64 res = 0;
	//pthread_mutex_lock(&m_sync_tables);
    tables_mutex.Lock();
	try {
		res = GetTable(table_path, table_share, true)->NoValues();
	} catch (DatabaseRCException&) {
		//pthread_mutex_unlock(&m_sync_tables);
        tables_mutex.Unlock();
		throw ;
	}
	//pthread_mutex_unlock(&m_sync_tables);
    tables_mutex.Unlock();
	return res;
}

void RCEngine::GetTableIterator(const char* table_path, JustATable::RecordIterator& iter_begin,
		JustATable::RecordIterator& iter_end, JustATable*& table,
		JustATable::RcTableLockPolicy::attr_uses_t const& attrUses, THD* thd, ConnectionInfo *conn) throw(DatabaseRCException)
{
	//pthread_mutex_lock(&m_sync_tables);
    tables_mutex.Lock();
	try {
		RefreshTable(table_path, GetThreadParams(*thd)->GetTransactionID(), false);
		table = GetTable(table_path, NULL, true);
	} catch (DatabaseRCException&) {
		//pthread_mutex_unlock(&m_sync_tables);
        tables_mutex.Unlock();
		throw ;
	}

	iter_begin = table->begin( conn, attrUses );
	iter_end = table->end( conn );
	//pthread_mutex_unlock(&m_sync_tables);
    tables_mutex.Unlock();
}

bool parameter_equals(THD* thd, enum bh_var_name vn, longlong value)
{
	longlong param = 0;
	string s_res;

	if(!get_parameter(thd, vn, param, s_res))
		if(param == value)
			return true;

	return false;
}

#ifdef __WIN__
#define snprintf _snprintf
#endif

void RCEngine::UpdateAndStoreColumnComment(TABLE* table, int field_id, Field* source_field, int source_field_id,
		CHARSET_INFO *cs)
{
	if(source_field->orig_table->s->db_type() == rcbase_hton) //do not use table (cont. default values)
	{
		char buf_size[256] = { 0 };
		char buf_ratio[256] = { 0 };
		uint buf_size_count = 0;
		uint buf_ratio_count = 0;
		_int64 sum_c = 0, sum_u = 0;
		AttrInfo* attr_info;
		uint no_attr;
		if(table->in_use && table_lock_manager.HasAnyLockForTable(*table->in_use, source_field->orig_table->s->path.str)) {
			GetCurrentTableStateAttributes(source_field->orig_table->s->path.str, source_field->orig_table->s, attr_info, no_attr);
		} else
			GetTableAttributes(source_field->orig_table->s->path.str, source_field->orig_table->s, attr_info, no_attr);

		bool is_unique = false;
		if(source_field_id < (int)no_attr) {
			is_unique = attr_info[source_field_id].actually_unique;
			sum_c += attr_info[source_field_id].comp_size;
			sum_u += attr_info[source_field_id].uncomp_size;
		}
		delete [] attr_info;

		double d_comp = int(sum_c/104857.6)/10.0; // 1 MB = 2^20 bytes
		if(d_comp < 0.1)
			d_comp = 0.1;
		double ratio = (sum_c > 0 ? sum_u / double(sum_c) : 0);
		if(ratio>1000)
			ratio = 999.99;

		buf_size_count = snprintf(buf_size, 256, "Size[MB]: %.1f", d_comp);
		if(is_unique)
			buf_ratio_count = snprintf(buf_ratio, 256, "; Ratio: %.2f; unique", ratio);
		else
			buf_ratio_count = snprintf(buf_ratio, 256, "; Ratio: %.2f", ratio);

		uint len = source_field->comment.length + buf_size_count + buf_ratio_count + 3;
		//	char* full_comment = new char(len);  !!!! DO NOT USE NEW !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!11
		char* full_comment = (char*)my_malloc(len, MYF(0));
		for(uint i = 0; i < len; i++)
			full_comment[i]=' ';
		full_comment[len-1]=0;

		char* pos = full_comment + source_field->comment.length;
		if(source_field->comment.length) {
			memcpy(full_comment, source_field->comment.str, source_field->comment.length);
			*pos++ = ';';
			*pos++ = ' ';
		}

		//memcpy_s(pos, buf_size_count, buf_size, buf_size_count);
		memcpy(pos, buf_size, buf_size_count);
		pos += buf_size_count;
		//memcpy_s(pos, buf_ratio_count, buf_ratio, buf_ratio_count);
		memcpy(pos, buf_ratio, buf_ratio_count);
		pos[buf_ratio_count] = 0;

		table->field[field_id]->store(full_comment, len-1, cs);
		//	delete [] full_comment;	!!!! DO NOT USE DELETE !!!!!!!!!!!!!!!!!!!!!!!!!!!!!
		my_free(full_comment, MYF(0));
		//	table->field[field_id]->store("abcde", 5, cs);
	} else {
		table->field[field_id]->store(source_field->comment.str, source_field->comment.length, cs);
	}
}

ConnectionInfo* RCEngine::GetThreadParams(const THD& thd)
{

	IBGuard guard(threadparams_mutex);
	if(thd.ha_data[m_engine_slot].ha_ptr == NULL)
		CeateThreadParams(const_cast<THD&>(thd));
	return (ConnectionInfo*)thd.ha_data[m_engine_slot].ha_ptr;

}

void RCEngine::CeateThreadParams(THD& thd)
{

	IBGuard guard(threadparams_mutex);
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(thd.ha_data[m_engine_slot].ha_ptr == 0);
	ConnectionInfo *ci = new ConnectionInfo(thd);
	ci->GetNewTransaction(*this);
	thd.ha_data[m_engine_slot].ha_ptr = ci;
	ConnectionInfoOnTLS.Set(*static_cast<ConnectionInfo*>(thd.ha_data[m_engine_slot].ha_ptr));
}

void RCEngine::ClearThreadParams(THD& thd)
{

	IBGuard guard(threadparams_mutex);
	ConnectionInfoOnTLS.Release();
	delete (ConnectionInfo*)thd.ha_data[m_engine_slot].ha_ptr;
	thd.ha_data[m_engine_slot].ha_ptr = NULL;

}

//bool RCEngine::HasAnyLockForTable(const THD& thd, const string& table_path)
//{
//	IBGuard guard(threadparams_mutex);
//	ThreadParamsMap::iterator iter	= m_ThreadParamsMap.begin();
//	ThreadParamsMap::iterator end	= m_ThreadParamsMap.end();
//
//	for(; iter != end; ++iter) {
//		if(thd.thread_id != iter->second->GetThreadID() && iter->second->HasAnyLockForTable(table_path))
//			return true;
//	}
//	return false;
//}

//void RCEngine::ReleaseTableLock(const THD& thd, const TableLockPtr table_lock)
//{
//	IBGuard guard(threadparams_mutex);
//	ThreadParamsMap::iterator iter	= m_ThreadParamsMap.begin();
//	ThreadParamsMap::iterator end	= m_ThreadParamsMap.end();
//
//	for(; iter != end; ++iter) {
//		if(thd.thread_id == iter->second->GetThreadID())
//			iter->second->ReleaseTableLock(table_lock);
//	}
//}
//
//void RCEngine::ReleaseAllWriteLocks(const THD& thd, const string& table_path)
//{
//	IBGuard guard(threadparams_mutex);
//	GetThreadParams(thd)->ReleaseAllWriteLocks(table_path);
//}

#ifdef USE_NEW_ENGINE
int RCEngine::Convert(int& is_null, Field* field, void* out_ptr, AValue* value)
{
	if(value->IsNull())
	{
		is_null = TRUE;
		return 1;
	}
	VInteger* vi = dynamic_cast<VInteger*>(value);

	switch(field->type())
	{
		case MYSQL_TYPE_SHORT:
		*(short*)field->ptr = vi->Get();
		break;
		case MYSQL_TYPE_TINY:
		*(char*)field->ptr = vi->Get();
		break;
		case MYSQL_TYPE_LONG:
		*(int*)field->ptr = vi->Get();
		break;
		case MYSQL_TYPE_LONGLONG:
		*(int64*)out_ptr = vi->Get();
	}
	return 1;
}
#endif

int RCEngine::SetUpCacheFolder(const string& cachefolder_path)
{
	const char* cache_path = cachefolder_path.c_str();
	char* buf = new char[strlen(cache_path)+1];
	for(uint i = 0; i==0 || cache_path[i-1]!=0; i++) {
		if(cache_path[i]=='\\' || (cache_path[i]=='/' && i > 0) || cache_path[i]==0) {
			buf[i] = 0;
            if (!DoesFileExist(buf)) {
                try {
                	string cache_folder_path;
                	if(IsAbsolutePath(string(buf)))
                		cache_folder_path = buf;
                	else
                		cache_folder_path = infobright_data_dir + buf;

#ifdef __WIN__  // WINPORT - deal with "c:\" path substring
                    if ((i > 0) && (buf[i-1] != ':')) CreateDir( cache_folder_path );
#else

                    CreateDir( cache_folder_path );
#endif
                } catch(DatabaseRCException& e) {
                    sql_print_error("Brighthouse: Can not create folder %s.", buf);
                    rclog << lock << e.what() << " DatabaseRCException" << unlock;
                    delete[] buf;
                    return 1;
                }
            }
		}
		buf[i] = cache_path[i];
	}
	delete [] buf;
    if (! IsReadWriteAllowed(cache_path)) {
		sql_print_error("Brighthouse: Can not access cache folder %s.", cache_path);
		rclog << lock << "Error: Can not access cache folder " << cache_path << '.' << unlock;
		return 1;
	}
	return 0;
}

void RCEngine::RemoveTemporaryFiles(string cachefolder_path)
{
	ClearDirectory(cachefolder_path.c_str());
    /**************************************
	string shell_command = "rm -f ";
	shell_command += cachefolder_path;
	shell_command += "/*.bh_tmp";
	system(shell_command.c_str());
    **************************************/
}


bool RCEngine::IsRouteMySqlLoader(THD* thd)
{
	longlong load_route = -1;
	string s_res;

	if(get_parameter(thd, BH_DATAFORMAT, load_route, s_res) || iequals(trim_copy(s_res), "MYSQL"))
		return true;
	return false;
}

void RCEngine::TerminateLoader(THD *thd)
{
	BHASSERT(thd != 0, "'thd' should not be null");
	ConnectionInfo* ci = (ConnectionInfo*) thd->ha_data[m_engine_slot].ha_ptr;
#ifdef __WIN__
#define PROCESS_ID   pi.hProcess
#else
#define PROCESS_ID   pi
#endif
	if((ci != 0) && (ci->PROCESS_ID > 0)) {
        IBProcess ib_process(ci->PROCESS_ID);
        ib_process.Terminate();
        int return_code, is_finished, is_terminated;
        ib_process.Wait(is_finished, is_terminated, return_code, 1);
	}
}


int get_parameter(THD* thd, enum bh_var_name vn, longlong &result, string& s_result)
{
	user_var_entry* m_entry;
	const uint var_size = 1024;
	char var_data[var_size];

	switch(vn) {
		case BH_TIMEOUT:
			strcpy(var_data, "BH_TIMEOUT");
			break;
		case BH_DATAFORMAT:
			strcpy(var_data, "BH_DATAFORMAT");
			break;
		case BH_PIPEMODE:
			strcpy(var_data, "BH_PIPEMODE");
			break;
		case BH_NULL:
			strcpy(var_data, "BH_NULL");
			break;
		default:
			return 1;
	}

	if(m_entry = (user_var_entry*) hash_search(&thd->user_vars, (uchar*) var_data, (uint)strlen(var_data))) {
		if(m_entry->type == INT_RESULT) {
			if(vn != BH_TIMEOUT) {
				result = -1;
				return 0;
			} else {
				result = *(longlong*) m_entry->value;
				return 0;
			}
		} else if(m_entry->type == STRING_RESULT) {
			result = -1;
			if(m_entry->value == NULL)
				return 1;
			strcpy(var_data, m_entry->value);
			if(vn == BH_DATAFORMAT) {
				s_result = var_data;
			} else if(vn == BH_PIPEMODE) {
				if(stricmp(var_data, "SERVER") == 0)
					result = 1;
				if(stricmp(var_data, "CLIENT") == 0)
					result = 0;
			} else if(vn == BH_NULL) {
				s_result = var_data;
			}
			return 0;
		} else {
			result = 0;
			return 2;
		}
	} else
		return 1;
}

void RCEngine::VerifyTablesIntegrity(const TablesList& table_list) throw (DatabaseRCException)
{
	TablesList::const_iterator iter = table_list.begin();
	TablesList::const_iterator end = table_list.begin();

	for(; iter != end; ++iter) {
		RCTable* tab = GetTable(*iter, NULL);
		if(tab == NULL)
			continue;

		if(tab->Verify() != 1) {
			string err_msg = "RCEngine::Commit  Data integrity is broken in table ";
			err_msg += *iter;
			throw DatabaseRCException(err_msg.c_str());
		}
	}
}

TablesListAutoPtr RCEngine::GetListOfModifiedTables(const THD& thd)
{
	IBGuard guard(threadparams_mutex);
	return GetThreadParams(thd)->GetListOfModifiedTables();
}

int InitRSIManager(string datadir_path, bool with_updater, bool force)
{
	if(rsi_manager == NULL) {
		char rsi_path[2048];
		int index_level = 99; // 0 to switch off RSIndex
		try {
			index_level = Configuration::GetProperty(Configuration::KNLevel);
		} catch (...) {
			index_level = 99;
		}
		if(index_level > 0 || force) {
			try {
				strcpy(rsi_path, Configuration::GetProperty(Configuration::KNFolder).c_str());
			} catch (...) {
				strcpy(rsi_path, "BH_RSI_Repository");
			}
			try {
				bool dir_exists = DoesFileExist(rsi_path);
				if(MakeDirs(rsi_path) < 0) {
					rclog << lock << "WARNING: KN directory not present/creation failed" << unlock;
					return -1;
				} else {
					if(dir_exists) {
						if(with_updater && RunRSIUpdater(datadir_path, rsi_path) != 0)
								return -1;
					} else {
						string metadata_file_path = MakeDirectoryPathOsSpecific(rsi_path);
						metadata_file_path += RSI_METADATAFILE ;
						try {
							//ofstream vf(infobright_data_dir + metadata_file_path.c_str());
							//vf.close();
							RSI_Manager::CreateMetadataFile(metadata_file_path, RSI_Manager::GetRSIVersionCode((RSI_MANAGER_VERSION)));
						} catch(std::exception& e) {
							rclog << lock << "WARNING: Metadata file creation failed" << unlock;
							return -1;
						}
					}
					rsi_manager = new RSI_Manager(rsi_path, index_level);
				}
			} catch (...) {
				// just run without RSI...
			}
		} else {
			rccontrol << lock << "KNLevel set to 0, no Knowledge Grid in use." << unlock;
		}
	}
	return 0;
}

int RunRSIUpdater(const string& datadir_path, const string& kn_folder)
{
	char updater_app[1048] = "";
// GA
	size_t len = 0;
	dirname_part(updater_app, my_progname, &len);
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(updater_app[len] == 0);

// 	dirname_part(updater_app, my_progname);


	strcat(updater_app, "updater");

	vector<string> params;
	params.push_back(string("--datadir=\"") + datadir_path + "\"");
	if(IsAbsolutePath(kn_folder))
		params.push_back(string("--knfolder=\"") + kn_folder + "\"");
	else
		params.push_back(string("--knfolder=\"") + datadir_path + kn_folder + "\"");
	params.push_back(string("--migrate-to=\"") + RSI_MANAGER_VERSION + "\"");
	params.push_back(string("--log-file=\"updater.log\""));

	IBProcess ibpr;
	int is_finished, is_terminated, return_code = -128;

	try {
		ibpr.Start(updater_app, params);
		pid_t pi = (pid_t)ibpr.GetProcess();
		ibpr.Wait(is_finished, is_terminated, return_code);
	} catch(...) {
		//rclog << lock << "Brighthouse: Fatal error: Update process failed." << unlock;
		return_code = -128;
	}

	if(return_code == -128)
		rclog << lock << "Brighthouse: Fatal error: Unable to start IB Updater (" << updater_app << ")." << unlock;
	else if(return_code < 0) {
		rclog << lock <<	string("Brighthouse: Fatal error: Update process failed.") +
							string(" For details on the error please see the updater.log file in the " + datadir_path + " directory.")
		<< unlock;
	}
	return min(0, return_code);
}

void InitCachingLevel()
{
	try {
		CachingLevel = Configuration::GetProperty(Configuration::CachingLevel);
		if(CachingLevel == 2)
			CachingLevel = 1;
	} catch (...) {
	}
}
