/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef _RCENGINE_H_
#define _RCENGINE_H_

#ifndef RCENGINE_H
#define RCENGINE_H

#include <boost/function.hpp>
#include <boost/shared_array.hpp>

#ifndef __BH_COMMUNITY__
#	include "enterprise/local_fwd.h"
#include "enterprise/loader/LoaderFactory.h"
#else /* not __BH_COMMUNITY__ */
#	include "community/local_fwd.h"
#include "community/loader/LoaderFactory.h"
#endif /* __BH_COMMUNITY__ */
#include "system/RCSystem.h"
#include "core/TempTable.h"
#include "system/RCException.h"
#include "system/IOParameters.h"
#include "system/ConnectionInfo.h"
#include "core/Query.h"
#include "shared/loader/Loader.h"
#include "core/DataCache.h"

#ifdef __GNUC__
#include <ext/hash_map>
#ifndef stdext
#define stdext __gnu_cxx
#endif
#else
#include <hash_map>
#endif

#ifdef USE_NEW_ENGINE
class AValue;
namespace na {
	class DataSource;
}
#endif

#define BHLoaderAppName "bhloader"

enum BHEngineReturnValues {LD_Successed = 100, LD_Failed = 101, LD_Continue = 102};

#ifdef __WIN__
void unlock_locked_tables(THD *thd);
#endif

extern handlerton* rcbase_hton;

//////////////////////////////////////////////////////////////////////////////////////////
///////////////////  RCEngine:   ///////////////////////////////////////////////////////////
//
// A class representing RCBase engine
//
// Assumption:	only one RCEngine per MySQL server will be created,
//
class select_bh_export;
class RCDataType;
class Field;
struct AttrInfo;

enum TABLE_LOCK_ENUM {READ_LOCK = 0, WRITE_LOCK};

class RCEngine
{
	friend class BrighthouseHandler;
	friend class TableLockManager;
public:
	RCEngine() : m_dev_output(NULL) {}
	~RCEngine();

	static std::string GetServiceStartPath(std::string service_name);
	int Init(const char* tab_ext, uint engine_slot);


	BHErrorCode CreateTable(const char *table, TABLE *from);
	void DropTable(const char *table);
	void RenameTable(const char* from, const char* to);

	void OpenTable(const char* table_path, struct st_table_share *table_share) throw(DatabaseRCException);
	time_t GetTableCreateTime(const char* table_path, struct st_table_share* table_share) throw(DatabaseRCException);
	time_t GetTableUpdateTime(const char* table_path, struct st_table_share* table_share) throw(DatabaseRCException);
	void GetTableAttributes(RCTable& tab, AttrInfo*& attributes, uint& no_attr);
	void GetTableAttributes(const char* table_path, struct st_table_share* table_share, AttrInfo*& attributes, uint& no_attr)
		throw(DatabaseRCException);
	void GetCurrentTableStateAttributes(const char* table_path, struct st_table_share* table_share, AttrInfo*& attributes, uint& no_attr)
		throw(DatabaseRCException);
	_uint64 GetTableNoObj(const char* table_path, struct st_table_share* table_share) throw(DatabaseRCException);
	void UpdateAndStoreColumnComment(TABLE* table, int field_id, Field* source_field, int source_field_id,
			CHARSET_INFO *cs);
	void GetTableIterator(const char* table_path, JustATable::RecordIterator& iter_begin,
			JustATable::RecordIterator& iter_end, JustATable*& table, const JustATable::RcTableLockPolicy::attr_uses_t& , THD *thd, ConnectionInfo *conn) throw (DatabaseRCException);
    RCTable *GetTable(const std::string& table_path, struct st_table_share* table_share, bool locked = false) throw (DatabaseRCException);
    RCTablePtr GetTableShared(const std::string& table_path, struct st_table_share* table_share, bool locked = false) throw (DatabaseRCException);
    BHEngineReturnValues RunLoader(THD *thd, sql_exchange *ex, TABLE_LIST *table_list, BHError & bherror);
    BHEngineReturnValues ExternalLoad(THD *thd, sql_exchange *ex, TABLE_LIST *table_list, BHError & bherror);
    void TerminateLoader(THD *thd);
    void Commit(THD *thd, bool all);
    void Rollback(THD *thd, bool all, bool force_error_message = false);
    ConnectionInfo* GetThreadParams(const THD& thd);
    void ClearThreadParams(THD& thd);
    int HandleSelect(THD *thd, LEX *lex, select_result *& result_output, ulong setup_tables_done_option, int & res, int & optimize_after_bh, int & bh_free_join, int with_insert = false);
    static AttributeType GetCorrespondingType(const Field & field);
    static int Convert(int & is_null, Field *field, void *out_ptr, RCItem *rcitem);
    static bool IsBHTable(TABLE *table);
    int WriteRow(THD *thd, TABLE *table, Field **field, uchar *buf,RCTable *&);
    int UpdateRow(THD *thd, TABLE *table, _uint64 row_id, Field **field,RCTable *&);
    int DeleteRow(THD *thd, TABLE *table, _uint64 row_id, RCTable *&);
    int DeleteAllRows(THD *thd, TABLE *table, RCTable *&);
	BHError PrivateLoad(THD* thd, const string& table_path, struct st_table_share* table_share, bool in_parallel = false);
    //int DropColumn(THD* thd, TABLE* table, Field** fields, byte* buf, RCTable *&tab);

	static AttributeTypeInfo GetAttrTypeInfo(const Field & field) throw (UnsupportedDataTypeRCException);
	
	SynchronizedDataCache* getGlobalDataCache() const { return &SynchronizedDataCache::GlobalDataCache(); }

private:
	void CeateThreadParams(THD& thd);
    int SetUpCacheFolder(const std::string & cachefolder_path);
    void RemoveTemporaryFiles(std::string cachefolder_path);
    ChannelOut *m_log_output;
    ChannelOut *m_control_output;
    ChannelOut *m_dev_output;
    int m_tab_sequencer;
    const char *m_tab_ext;
	DataCache *m_globalCache;

    uint m_engine_slot;
    std::map<std::string,std::map<std::string,boost::shared_ptr<RCTable> > > m_tables;
    typedef std::map<ulong,ConnectionInfo*> ThreadParamsMap;
    ThreadParamsMap m_ThreadParamsMap;
    size_t m_loader_main_heapsize;
    size_t m_loader_compressed_heapsize;
    IBMutex tables_mutex;
    IBMutex threadparams_mutex;
    static void GetNames(const char *table_path, char *db_buf, char *tab_buf, size_t buf_size);
    RCTablePtr RefreshTable(const char *table_path, uint session_id = 0, bool force = true) throw (DatabaseRCException);
    std::auto_ptr<RCTable> GetCurrentTableState(const char* table_path, struct st_table_share* table_share) throw (DatabaseRCException);
    void ApplyPendingChanges(THD* thd, const std::string& table_path, struct st_table_share* table_share);
    void DropPendingChanges(THD* thd, const std::string& table_path, struct st_table_share* table_share);
    bool RefreshTableApplyPendingChanges(RCTable *tab);
    int Execute(THD *thd, LEX *lex, select_result *result_output, SELECT_LEX_UNIT *unit_for_union = NULL);
    int ExecuteOld(THD *thd, LEX *lex, select_result *result_output, SELECT_LEX_UNIT *unit_for_union = NULL);
    void VerifyTablesIntegrity(const TablesList & table_list) throw (DatabaseRCException);
    TablesListAutoPtr GetListOfModifiedTables(const THD & thd);
	BHError PrivateLoadInternal(THD* thd, RCTable* tab, const string& table_path, struct st_table_share* table_share, bool refresh = true);

private:
    static void SendResults(JustATable & results, THD *thd, select_result *res, List<Item> & fields, ConnectionInfo *conn);
    static bool AreConvertible(RCItem *rcitem, Field *field);
    static bool AreConvertible(RCItem *rcitem, enum_field_types my_type, uint length = 0);
    static int Convert(int & is_null, my_decimal *value, RCItem *rcitem);
    static int Convert(int & is_null, int64 & value, RCItem *rcitem);
    static int Convert(int & is_null, double & value, RCItem *rcitem);
    static int Convert(int & is_null, String *value, RCItem *rcitem, enum_field_types f_type);
    static bool IsBHRoute(THD *thd, TABLE_LIST *table_list, SELECT_LEX *selects_list, int & in_case_of_failure_can_go_to_mysql, int with_insert);
    static const char *GetFilename(SELECT_LEX *selects_list, int & is_dumpfile);
    static void ExportResults(THD *thd, TempTable & results, select_bh_export *res, List<Item> & fields, const char *filename, int is_dumpfile);
    static BHError GetIOParameters(IOParameters & io_params, THD & thd, sql_exchange & ex, TABLE *table = 0);
    static std::auto_ptr<IOParameters> CreateIOParameters(const std::string & path);
    static std::auto_ptr<IOParameters> CreateIOParameters(THD *thd, TABLE *table);
    static std::string GetTablePath(THD *thd, TABLE *table);
    static void ComputeTimeZoneDiffInMinutes(THD *thd, short  & sign, short  & minutes);

public:
	bool IsRouteMySqlLoader(THD* thd);
	static DWORD WINAPI insert_th_func(void* f) {
			(*(boost::function0<void>*)(f))();
			return NULL;
		}
};

enum bh_var_name
{
	BH_TIMEOUT,
	BH_DATAFORMAT,
	BH_PIPEMODE,
	BH_NULL
};

int get_parameter(THD* thd, enum bh_var_name vn, longlong &result, std::string &s_result);
bool parameter_equals(THD* thd, enum bh_var_name vn, longlong value);
int has_lookup(const LEX_STRING &comment);

int		InitRSIManager(std::string datadir_path = "", bool with_updater = false, bool force = false);
int 	RunRSIUpdater(const std::string& datadir_path, const std::string& kn_folder);
void 	InitCachingLevel();

//extern RCEngine* rceng;// RCBase engine used by MySQL

#endif

#endif // RCENGINE_H
