/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef _RCTABLE_H_
#define _RCTABLE_H_

#include <boost/shared_ptr.hpp>
#include <boost/utility.hpp>
#include <boost/enable_shared_from_this.hpp>

#include "common/CommonDefinitions.h"
#include "core/JustATable.h"
#include "core/RCAttr.h"
#include "system/ib_system.h"

class BufferForInsert;
class IndexTable;
class Updater;
class RSIndex_PP;

#ifdef USE_NEW_ENGINE
#include "types/AValue.h"
#endif

struct AttrInfo
{
	const char* name;
	const char* desc;
	AttributeType type;
	int size;
	int precision;
	bool no_nulls;
	bool lookup;
	bool actually_unique;
	//bool unique;
	//bool primary;
	unsigned int current_trans;
	_int64 uncomp_size;
	_int64 comp_size;
	int column_size;
};

class RCTableImpl : public JustATable, public boost::noncopyable
{
protected:
	RCTableImpl() : name(NULL), db_name(NULL), desc(NULL), path(NULL), path_no_ext(NULL), charsets(), a(NULL) { }
public:
	RCTableImpl(const char* path, vector<DTCollation> charsets, char* nm, int t_n, unsigned int s_id = 0, int use_path_only = 0);
	RCTableImpl(const char* path, vector<DTCollation> charsets, int t_n, int current_state, int use_path_only = 0, bool force_open = false, bool for_loader = false);
	virtual ~RCTableImpl();

	/////////////////////// Synchronization //////////////////////////////////////
	void LockTable() { table_impl_mutex.Lock(); /*pthread_mutex_lock(&synchr);*/ }
	void UnlockTable() { table_impl_mutex.Unlock(); /*pthread_mutex_unlock(&synchr);*/ }

	int waiting_for_drop;

	// WriteSync nodrop;

	void Refresh(int current_state); // inter-process synchronization

	void LockPackInfoForUse(); // lock attribute data against memory manager
	void UnlockPackInfoFromUse(); // return attribute data to memory manager
	void LockPackForUse(unsigned attr, unsigned pack_no, ConnectionInfo& conn);
	void UnlockPackFromUse(unsigned attr, unsigned pack_no);

	///////////////////////////// Table definition & information //////////////////////////////
	void AddAttribute(char* nm, AttributeType a_type, int a_field_size, int a_dec_places = 0, unsigned int param = 0, DTCollation collation = DTCollation());

	virtual void Save(); // save table definition and data
	void SaveHeader(); // save table definition
	void Drop(); // drop this table

	const char		*Name() const 			{ return name; }
	const char		*Description() const 	{ return desc; }
	int 			GetID() const 			{ return tab_num; }
	TType 			TableType() const 		{ return RC_TABLE; }
	std::string 	GetPath() const 		{ return std::string(path); }

	uint NoAttrs() { return no_attr; }
	uint NoDisplaybleAttrs() { return NoAttrs(); }
	AttrInfo* GetAttributes(); // attributes info
	int AttrNum(const char *a_name); // return the number of an attribute (given by name) or -1 if attribute not found
	const char *AttrName(int n_a);
	AttributeType AttrType(int n_a);
	std::vector<ATI> GetATIs();
	int GetAttrSize(int n_a);
	int GetFieldSize(int n_a);
	int GetInternalSize(int n_a);
	NullMode GetAttrNullMode(int n_a);
	int GetAttrScale(int n_a);
	bool IsLookup(int n_a);
	const ColumnType& GetColumnType(int a) ;
	const PhysicalColumn* GetColumn(int col_no) { return a[col_no]; }
	RCAttr* GetAttr(int n_a);
	time_t GetCreateTime() const { return m_create_time; }
	time_t GetUpdateTime() const { return m_update_time; }

	//////////////////////////// Transaction management ////////////////////////
	int Verify(); // verify table integrity (number of objects in each attribute)
	bool LockedBy(unsigned int currentTransaction);
	unsigned int GetSaveSessionId();
	unsigned int GetLastSaveSessionId();
	virtual unsigned int CommitSaveSession(unsigned int sess_id, int new_state);
	virtual void Rollback(unsigned int s_id = 0);
	bool IsRLCPossible();
	void RollbackLastCommitted(uint s_id, int current_state);
	bool Committed();

	//////////////////////////// Data load & export ////////////////////////
	void SetWriteMode(unsigned int s_id) { SetMode(1, s_id); } // reopen all attributes information in the write mode
//	std::string GetErrorMsg() { return error_msg; }
	void LoadDirtyData(unsigned int session_id);

	/////////////////////// Data access & information //////////////////////////
	virtual _uint64 NoObj();
	virtual _uint64 NoValues() { return NoObj(); }
	virtual _uint64 NoNulls(int attr);

	RCBString GetTable_S(_uint64 obj, int attr, bool bin_as_hex = true);
//	int GetTable_T(_uint64 obj, int attr, char *val_buf); // value from table in string form
	void GetTable_B(_uint64 obj, int attr, int &size, char *val_buf); // value from table in string form
	int GetTable32(_uint64 obj, int attr); // value from table in 1-level numerical form
	_int64 GetTable64(_uint64 obj, int attr); // value from table in 1-level numerical form
	//int GetTableOrderedLookup(_uint64 obj, int attr, ConnectionInfo *conn); // value from table as an ordered position
																			// in the dictionary
//	double GetTable_N(_uint64 obj, int attr);
	bool IsNull(_uint64 obj, int attr); // return true if the value of attr. is null
	RCValueObject GetValue(_uint64 obj, int attr, ConnectionInfo *conn = NULL);
	int GetLength(_uint64 obj, int attr, ConnectionInfo *conn);

#ifdef USE_NEW_ENGINE
	void GetAValue(_uint64 obj, int attr, AValue& value, ConnectionInfo *conn);
	//AValuePtr		GetMinAValuePtr(int attr, ConnectionInfo *conn);
	//AValuePtr		GetMaxAValuePtr(int attr, ConnectionInfo *conn);
	void GetMinMaxBounds(int attr, VInteger& min, VInteger& max, ConnectionInfo *conn);
#endif

	RCBString DecodeValue_S(_int64 code, int attr); // original 0-level text value for a given 1-level code,
													// return the number of characters
	//_int64 DecodeOrderedLookup(_int64 code, int attr); // lookup code on the given ordered position in dictionary
	_int64 EncodeValue_T(char *val, int attr); // 1-level code value for a given 0-level (text) value
	_int64 EncodeValue_T(RCBString& val, int attr, bool sort_order); // if sort_order is set, then a number in
																	 // sortable order is returned

	/////////////////////// Query execution ///////////////////////////////////

	_int64 RoughMin(int n_a, Filter *f = NULL);		// for numerical: best rough approximation of min for a given filter (or global min if filter is NULL)
	_int64 RoughMax(int n_a, Filter *f = NULL);		// for numerical: best rough approximation of max for a given filter (or global max if filter is NULL)
	ushort MaxStringSize(int n_a, Filter *f = NULL);
	RSValue AttrIsDistinct(int n_a); // three-valued answer
	void SetDistinct(int n_a, RSValue v);

	/////////////////////////// Rough set part ////////////////////////////////////////////

	// provide the most probable approximation of number of objects matching the condition
	_uint64 ApproxAnswerSize(int attr, Descriptor &d);

	RSIndex_PP* LoadRSI_PP(RCTableImpl *t2, int a1, int a2, bool &flipped);
	void ReleaseIndex(RSIndex *rsi);
	/*
	void UpdateRSI_PP(Sorter2 &s1, Sorter2 &s2, RCTableImplImpl *sec, int attr1, int attr2, Filter *f1, Filter *f2,
			bool nulls_happen);
	void UpdateRSI_PP(IndexTable *t1_t, IndexTable *t2_t, RCTableImpl *sec, int attr1, int attr2, Filter *f1, Filter *f2);
	bool RSI_PP_Updatable(RCTableImpl *sec, int attr1, int attr2); // true if RSI_PP does not exist or is outdated
	*/

	void DisplayRSI();
	vector<DTCollation> GetCharsets() { return charsets; }
	void SetCharsets(vector<DTCollation> _charsets) { charsets = _charsets; } 
protected:
	char* name;
	char* db_name;
	char* desc; // table description (comments)
	int tab_num; // Index (unique) of the table in database. Used e.g. for determining folder name, identifying RSI etc.
	char* path;
	char* path_no_ext;

	vector<DTCollation> charsets; // charset definition for each column

	int no_attr;
	RCAttr** a;

	//unsigned char *a_mode;		// one-byte attribute parameters; 0 for deleted attribute, !=0 otherwise

    IBMutex  table_impl_mutex;
	//pthread_mutex_t synchr;

	int m_state;
	bool check_integrity;			// if false, then do not check no. of obj. integrity in open

	time_t m_create_time, m_update_time;

	int conn_mode;
	unsigned int session;
//	std::string error_msg;

	// concatenate db path 'a_path' with folder name of the table (\D00...) and store in 'path'
	void SetPath(const char *a_path, int use_path_only = 0);
	virtual void LoadAttribute(int attr_no);

	void SetMode(int conn_mode, unsigned int s_id); // reopen all attributes information in the new mode (see above)

	/////////////////////// Housekeeping //////////////////////////////////////
	void CleanUp();
	virtual RecordIterator do_find( ConnectionInfo *conn, _uint64 pos, RcTableLockPolicy::attr_uses_t const& );
	virtual RecordIterator do_find( ConnectionInfo *conn, _uint64 pos, Filter const*, RcTableLockPolicy::attr_uses_t const& );
	virtual RecordIterator do_begin( ConnectionInfo *conn, RcTableLockPolicy::attr_uses_t const& );
	virtual RecordIterator do_begin( ConnectionInfo *conn, Filter const& filter, RcTableLockPolicy::attr_uses_t const& );
	virtual RecordIterator do_end( ConnectionInfo *conn );

public:
	int 		DeleteRow(_uint64 row_id);
	const char* GetPath() { return path; }
	const char* GetPathNoExt() { return path_no_ext; }
	const char*	GetDBName() { return db_name; }


};

#endif

