/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "DataParser.h"
#include "types/ValueParser.h"
#include "core/tools.h"


extern my_time_t sec_since_epoch_TIME(MYSQL_TIME *t);

#ifdef __WIN__
#define gmtime_r(timearg, tmarg) gmtime_s((tmarg),(timearg))
#endif /* __WIN__ */

using namespace std;

DataParser::DataParser(vector<RCAttrLoad*> attrs, Buffer& buffer, const IOParameters& iop)
	:	attrs(attrs), buffer(&buffer), error_value(-1, -1)
{
	if((buffer.BufStatus() != 1 && buffer.BufStatus() != 4 && buffer.BufStatus() != 5))
		throw FileRCException("Unable to read data. Wrong file or pipe name.");
	edf = iop.GetEDF();
	iop.GetTimeZone(sign, minute);
	values_ptr = new char* [65536];
	rows_ptr = new char* [65536];
	no_attr = attrs.size();
	no_prepared = 0;
	buf_start = buf_end = buf_ptr = 0;
	cur_attr = -1;
	row_header_byte_len = 0;
	objs_sizes = 0;
	objs_sizes = new uint * [no_attr];
	for(int i = 0; i < no_attr; i++)
		objs_sizes[i] = new uint [65536];

	buf_ptr = buf_start = this->buffer->Buf(0);
	buf_end = buf_start + this->buffer->BufSize();
	this->num_values64 = 0;

	loaderStartTime = RCDateTime::GetCurrent().GetInt64();

	InitAttributesTypeInfo();

	value_sizes = 0;
	value_sizes = new uint * [no_attr];
	for(int i = 0; i < no_attr; i++)
		if (atis[i].Type() == RC_STRING)
			value_sizes[i] = new uint [65536];
		else
			value_sizes[i] = objs_sizes[i];
}

DataParser::~DataParser()
{
	delete [] values_ptr;
	delete [] rows_ptr;
	delete [] num_values64;
	for(int i = 0; i < no_attr; i++) {
		delete [] objs_sizes[i];
		if(atis[i].Type() == RC_STRING)
			delete [] value_sizes[i];
	}
	delete [] value_sizes;
	delete [] objs_sizes;
}

void DataParser::InitAttributesTypeInfo()
{
	for(int i = 0; i < no_attr; i++) {
		atis.push_back(AttributeTypeInfo(attrs[i]->TypeName(), attrs[i]->Type().GetNullsMode() == NO_NULLS,
			attrs[i]->Type().GetPrecision(), attrs[i]->Type().GetScale()));
	}
}

int DataParser::Prepare(int no_rows)
{
	no_prepared = 0;
	cur_attr = - 1;

	if(GetRowSize(buf_ptr, 0) == -1)
		return 0;

	int row_size = GetRowSize(buf_ptr, 0);
	while((row_size >= 0) && (no_prepared < no_rows)) {
		rows_ptr[no_prepared] = buf_ptr;
		values_ptr[no_prepared] = buf_ptr + row_header_byte_len;
		row_sizes[no_prepared] = row_size;
		no_prepared++;
		buf_ptr += row_size;
		if(no_prepared < no_rows)
			row_size = GetRowSize(buf_ptr, no_prepared);
	};
	return no_prepared;
}

void DataParser::PrepareNextCol()
{
	cur_attr++;
	cur_attr_type = atis[cur_attr].Type();

	PrepareValuesPointers();
	PrepareObjsSizes();
	PrepareNulls();
	FormatSpecificProcessing();

	if(!CheckData())
		throw FormatRCException(BHERROR_DATA_ERROR, error_value.first, error_value.second);

	if(attrs[cur_attr]->PackType() == PackN)
		PrepareNumValues();
}

uint DataParser::GetObjSize(int ono)
{
	return objs_sizes_ptr[ono];
}

uint DataParser::GetValueSize(int ono)
{
	return value_sizes[cur_attr][ono];
}

void DataParser::PrepareNumValuesForLookup()
{
	int size = 0;
	for(int i = 0; i < no_prepared; i++) {
		if(!IsNull(i)) {
			size = GetValueSize(i);
			num_values64[i] = attrs[cur_attr]->EncodeValue_T(RCBString(size ? values_ptr[i] : 0, size), 1);
		}
	}
}

bool IsTimeStampZero(MYSQL_TIME& t)
{
	return t.day == 0 && t.hour == 0 && t.minute == 0 && t.month == 0 && t.second == 0 && t.second_part == 0 && t.year == 0;
}

/*! \brief Converts UTC/GMT time given in seconds since beginning of the EPOCHE to TIME representation
 * \param tmp
 * \param t
 * \return void
 */
void GMTSec2GMTTime(MYSQL_TIME* tmp, my_time_t t)
{
	time_t tmp_t = (time_t)t;
	struct tm tmp_tm;
	gmtime_r(&tmp_t, &tmp_tm);
	localtime_to_TIME(tmp, &tmp_tm);
	tmp->time_type= MYSQL_TIMESTAMP_DATETIME;
}

void DataParser::PrepareNumValues()
{

	BHASSERT(attrs[cur_attr]->PackType() == PackN, "should be 'attrs[cur_attr]->PackType() == PackN'");
	if(!num_values64) {
		num_values64 = new _int64 [MAX_NO_OBJ];
	}

	if(attrs[cur_attr]->Type().IsLookup()) {
		PrepareNumValuesForLookup();
	} else {
		int os = 0;
		RCBString rcs(ZERO_LENGTH_STRING, 0);
		ParsingFunction parsing_function = value_parser->GetParsingFuntion(atis[cur_attr]);

		for(int i = 0; i < no_prepared; i++) {
			if(!IsNull(i)) {
				os = GetObjSize(i);
				TemporalValueReplacement<char>(values_ptr[i][os], 0);
				rcs.val = values_ptr[i];
				rcs.len = os;

				if(parsing_function(rcs, num_values64[i]) == BHRC_FAILD)
					throw FormatRCException(BHERROR_DATA_ERROR, i + 1, cur_attr + 1);

			} else if(cur_attr_type == RC_TIMESTAMP && attrs[cur_attr]->Type().GetNullsMode() == NO_NULLS) {
				num_values64[i] = loaderStartTime;
				nulls[i] = false;
			}
			if(cur_attr_type == RC_TIMESTAMP) {
				RCDateTime dt(num_values64[i], RC_TIMESTAMP);
				if(dt.IsNull())
					continue;
				TIME t;
				t.year = dt.Year();
				t.month = dt.Month();
				t.day = dt.Day();
				t.hour = dt.Hour();
				t.minute = dt.Minute();
				t.second = dt.Second();
				t.second_part = 0;
				if(t.year == 0 || t.month == 0 || t.day == 0)
					continue;
				time_t secs = 0;
				if(minute == NULL_VALUE_SH) {
					// System time zone
					my_bool myb;
					long tz;
					// time in system time zone is converted into UTC and expressed as seconds since EPOCHE
					secs = my_system_gmt_sec(&t, &tz, &myb);
				} else {
					// time in client time zone is converted into UTC and expressed as seconds since EPOCHE
					secs = sec_since_epoch_TIME(&t) + sign * minute * 60;
				}
				// UTC seconds converted to UTC struct tm
				struct tm utc_t;
				gmtime_r(&secs, &utc_t);
				//GMTSec2GMTTime(&t, secs);
				// UTC time stored on server
				num_values64[i] = RCDateTime((utc_t.tm_year + 1900) % 10000, utc_t.tm_mon + 1, utc_t.tm_mday, utc_t.tm_hour, utc_t.tm_min, utc_t.tm_sec, RC_TIMESTAMP).GetInt64();
			}
		}
	}

}

bool DataParser::IsLastColumn()
{
	return (cur_attr == (no_attr - 1));
}

bool DataParser::IsNull(int ono)
{
	return nulls[ono] == 1;
}

bool DataParser::CheckPreparedValuesSizes()
{
	if(ATI::IsStringType(cur_attr_type)) {
		for(int i = 0; i < no_prepared; i++) {
			if(!IsNull(i) && GetValueSize(i) > (uint)attrs[cur_attr]->Type().GetPrecision()) {
				error_value.first = i + 1;
				error_value.second = cur_attr + 1;
				return false;
			}
		}
	}
	return true;
}

bool DataParser::CheckData()
{
	return CheckPreparedValuesSizes() && FormatSpecificDataCheck();
}

char* DataParser::GetValue(int ono)
{
	if(attrs[cur_attr]->PackType() == PackN)
		return (char*)&num_values64[ono];
	return values_ptr[ono];
}

