/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include <vector>
#include <set>
#include <algorithm>
#include <boost/bind.hpp>
#include <boost/lambda/lambda.hpp>

#include "BHToolkit.h"
#include "ConnectionInfo.h"
#include "Channel.h"
#include "RCSystem.h"
#include "system/ib_system.h"
#include "core/Transaction.h"
#include "shared/core/RCEngine.h"

using namespace std;
using namespace boost;
namespace bl = boost::lambda;

IBThreadStorage<ConnectionInfo> ConnectionInfoOnTLS;

bool TableLock::operator==(const TableLock& table_lock) const
{
	return	table_path == table_lock.table_path &&
			lock_type == table_lock.lock_type &&
			thd.thread_id == table_lock.thd.thread_id &&
			(granted_in_lock_tables == table_lock.granted_in_lock_tables ||
			 granted_in_lock_tables == GRANTED_IN_DOES_NOT_MATTER ||
			 table_lock.granted_in_lock_tables == GRANTED_IN_DOES_NOT_MATTER);
}

ConnectionInfo::ConnectionInfo(THD& thd)
	:	current_transaction(NULL),thd(thd)
{
	memset(&pi, 0, sizeof(pi));
	transaction_id = GenerateTransactionNumber();
	display_attr_stats = false;
	display_lock = rccontrol.isOn() ? 0 : 1;
}

ConnectionInfo::~ConnectionInfo()
{
	Transaction *t;
	while(current_transaction) {
		t = current_transaction;
		current_transaction = t->getParent();
		delete t;
		
	}
}

TablesListAutoPtr ConnectionInfo::GetListOfModifiedTables() const
{
	IBGuard guard(modified_tables_guard);
	//TODO: <Michal> Do we really need to copy this TablesList here?
	return TablesListAutoPtr(new TablesList(modified_tables));
}

void ConnectionInfo::AddModifiedTable(const std::string& table_path)
{
	IBGuard guard(modified_tables_guard);
	modified_tables.insert(table_path);
}

void ConnectionInfo::RemoveTable(const std::string& table_path)
{
	IBGuard guard(modified_tables_guard);
	modified_tables.erase(table_path);
}

void ConnectionInfo::SuspendDisplay()
{
	if(++display_lock == 1) 
		rccontrol.setOff();
}
void ConnectionInfo::ResumeDisplay()
{
	if(--display_lock == 0) 
		rccontrol.setOn();
}

void ConnectionInfo::ResetDisplay()
{
	display_lock = rccontrol.isOn() ? 0 : 1;
	ConfigureRCControl();
}

void ConnectionInfo::EndTransaction()
{
	IBGuard guard(modified_tables_guard);
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(modified_tables.size() == 0);
	transaction_id = GenerateTransactionNumber();
	BHASSERT(current_transaction != NULL, "Transaction is NULL");
	Transaction *prev = current_transaction;
	current_transaction = current_transaction->getParent();
	delete prev;
	if(!current_transaction)
		GetNewTransaction(*rceng);
}

TableLockWeakPtr IBTableShared::AcquireWriteLock(const THD& thd)
{
	//int sleep_time = 0;
	while(!thd.killed && !thd.some_tables_deleted /*&& sleep_time < 20000*/) {
		{
			IBGuard guard(lock_request_sync);

			if(HasWriteLockForThread(thd))
				return write_lock;
			else if(!HasWriteLock() && !HasReadLockForOtherThread(thd))
				return CreateWriteLockForThread(thd);
		}
		IBThread::SleepNow(300);
		//sleep_time += 300;
	}
	return TableLockPtr();
}

TableLockWeakPtr IBTableShared::AcquireReadLock(const THD& thd)
{
	//int sleep_time = 0;
	while(!thd.killed && !thd.some_tables_deleted /*&& sleep_time < 20000*/) {
		{
			IBGuard guard(lock_request_sync);
			if(HasWriteLockForThread(thd) || !HasWriteLock())
				return CreateReadLockForThread(thd);
		}
		IBThread::SleepNow(300);
		//sleep_time += 300;
	}
	return TableLockPtr();
}

void IBTableShared::ReleaseLock(TableLockWeakPtr lock)
{
	IBGuard guard(lock_request_sync);
	if(lock.expired())
		return;
	TableLockPtr tlock = lock.lock();
	if(tlock->IsWriteLock()) {
		BHASSERT_WITH_NO_PERFORMANCE_IMPACT(write_lock.get() == tlock.get());
		write_lock.reset();
	} else {
		BHASSERT_WITH_NO_PERFORMANCE_IMPACT(find_if(read_lock_list.begin(), read_lock_list.end(), bl::_1 == tlock) != read_lock_list.end());
		read_lock_list.erase(find_if(read_lock_list.begin(), read_lock_list.end(), bl::_1 == tlock));
	}
}

void IBTableShared::ReleaseLocks(const THD& thd)
{
	IBGuard guard(lock_request_sync);
	if(HasWriteLockForThread(thd))
		write_lock.reset();

	read_lock_list.erase(
		remove_if(read_lock_list.begin(), read_lock_list.end(),	*bl::_1 == TableLock(table_path, thd, F_RDLCK)),
		read_lock_list.end()
	);
}

bool IBTableShared::HasWriteLockForThread(const THD& thd) const
{
	return write_lock && write_lock->Thd().thread_id == thd.thread_id;
}

bool IBTableShared::HasWriteLock() const
{
	return write_lock.get() != 0;
}

bool IBTableShared::HasReadLock() const
{
	return read_lock_list.size() != 0;
}

bool IBTableShared::HasAnyLockForThread(const THD& thd)
{
	IBGuard guard(lock_request_sync);
	return HasWriteLockForThread(thd) || HasReadLockForThread(thd);
}

bool IBTableShared::HasReadLockForThread(const THD& thd) const
{
	return find_if(read_lock_list.begin(), read_lock_list.end(),
			*bl::_1 == TableLock(table_path, thd, F_RDLCK)) !=  read_lock_list.end();
}

bool IBTableShared::HasReadLockForOtherThread(const THD& thd) const
{
	TablesLocksList::const_iterator iter = read_lock_list.begin();
	TablesLocksList::const_iterator end  = read_lock_list.end();
	for(; iter != end; ++iter)
		if((*iter)->IsReadLock() && (*iter)->Thd().thread_id != thd.thread_id)
			return true;
	return false;
}

TableLockWeakPtr IBTableShared::CreateReadLockForThread(const THD& thd)
{
	GrantedInLockTables in_lock_tables =
		((thd.options & OPTION_TABLE_LOCK) != 0) ? GRANTED_IN_LOCK_TABLES : GRANTED_IN_REGULAR_WAY;
	TableLockPtr table_lock(new TableLock(table_path, thd, F_RDLCK, in_lock_tables));
	read_lock_list.push_back(table_lock);
	return TableLockWeakPtr(table_lock);
}

TableLockWeakPtr IBTableShared::CreateWriteLockForThread(const THD& thd)
{
	if(!HasWriteLockForThread(thd)) {
		GrantedInLockTables in_lock_tables =
					((thd.options & OPTION_TABLE_LOCK) != 0) ? GRANTED_IN_LOCK_TABLES : GRANTED_IN_REGULAR_WAY;
		write_lock = TableLockPtr(new TableLock(table_path, thd, F_WRLCK, in_lock_tables));
	}
	return TableLockWeakPtr(write_lock);
}



#include "system/ConnectionInfo.h"
#include "shared/core/RCEngine.h"

Transaction * 
ConnectionInfo::GetNewTransaction( const RCEngine &engine) {
	current_transaction = new Transaction(current_transaction,const_cast<RCEngine&>(engine), engine.getGlobalDataCache());
	return current_transaction;
}

Transaction * 
ConnectionInfo::GetTransaction( ) { return current_transaction; }

void TableLockManager::RegisterTable(const std::string& table_path)
{
	IBGuard guard(mutex);
	registered_tables.insert(make_pair(table_path, IBTableSharedPtr(new IBTableShared(table_path))));
}

void TableLockManager::UnRegisterTable(const std::string& table_path)
{
	IBGuard guard(mutex);
	std::map<std::string, IBTableSharedPtr>::iterator it = registered_tables.find(table_path);
	if(it != registered_tables.end() && it->second->HasWriteLock())
		rceng->GetThreadParams(it->second->WriteLockOwner())->RemoveTable(table_path);
	registered_tables.erase(table_path);
}

TableLockWeakPtr TableLockManager::AcquireWriteLock(const THD& thd, const std::string& table_path)
{
	TableLockWeakPtr lock = GetTableShare(table_path)->AcquireWriteLock(thd);
	if(!lock.expired())
		rceng->GetThreadParams(thd)->AddModifiedTable(table_path);
	return lock;
}

TableLockWeakPtr TableLockManager::AcquireReadLock(const THD& thd, const std::string& table_path)
{
	return GetTableShare(table_path)->AcquireReadLock(thd);
}

void TableLockManager::ReleaseLocks(const THD& thd, const std::string& table_path)
{
	GetTableShare(table_path)->ReleaseLocks(thd);
	rceng->GetThreadParams(thd)->RemoveTable(table_path);
}

void TableLockManager::ReleaseLocks(const THD& thd)
{
	IBGuard guard(mutex);
	std::map<std::string, IBTableSharedPtr>::iterator iter = registered_tables.begin();
	std::map<std::string, IBTableSharedPtr>::iterator end =	registered_tables.end();
	for(; iter != end; ++iter)
		iter->second->ReleaseLocks(thd);
}

void TableLockManager::ReleaseLockIfNeeded(TableLockWeakPtr table_lock)
{
	if(!table_lock.expired()) {
		TableLockPtr tl(table_lock.lock());
		if(tl->IsReadLock())
			GetTableShare(table_lock.lock()->TablePath())->ReleaseLock(table_lock);
		else if(!(tl->Thd().options & OPTION_NOT_AUTOCOMMIT)) {
			rceng->GetThreadParams(tl->Thd())->RemoveTable(tl->TablePath());
			GetTableShare(table_lock.lock()->TablePath())->ReleaseLock(table_lock);
		}
	}
}

bool TableLockManager::HasAnyLockForTable(const THD& thd, const std::string& table_path)
{
	return GetTableShare(table_path)->HasAnyLockForThread(thd);
}

IBTableSharedPtr TableLockManager::GetTableShare(const std::string& table_path)
{
	IBGuard guard(mutex);
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(registered_tables.find(table_path) != registered_tables.end());
	return registered_tables[table_path];
}
