/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef _SYSTEM_CONNECTIONINFO_H_
#define _SYSTEM_CONNECTIONINFO_H_

#include "common/CommonDefinitions.h"
#include "system/ib_system.h"

#include <set>
#include <map>
#include <vector>

#include <boost/shared_ptr.hpp>
#include <boost/utility.hpp>
#include <boost/bind.hpp>
class THD;
class Channel;
class Transaction;
class RCEngine;

#ifndef __WIN__
typedef pid_t process_info;
#else
typedef PROCESS_INFORMATION process_info;
#endif


//TODO: <Michal> move TableLock to separated file

enum GrantedInLockTables
{
	GRANTED_IN_LOCK_TABLES,
	GRANTED_IN_REGULAR_WAY,
	GRANTED_IN_DOES_NOT_MATTER
};

class TableLock
{
public:
	TableLock(const std::string& table_path, const THD& thd, int lock_type, GrantedInLockTables in_lock_tables = GRANTED_IN_DOES_NOT_MATTER)
		:	table_path(table_path), lock_type(lock_type), granted_in_lock_tables(in_lock_tables), thd(thd)
	{}

	bool operator==(const TableLock& table_lock) const;

	bool IsWriteLock() const	{ return lock_type == F_WRLCK; }
	bool IsReadLock() const		{ return lock_type == F_RDLCK; }

	const THD& Thd() const		{ return thd; }

	const std::string& TablePath() const { return table_path; }

private:
	std::string table_path;
	int lock_type;
	GrantedInLockTables granted_in_lock_tables;
	const THD& thd;
};

typedef boost::shared_ptr<TableLock> TableLockPtr;
typedef std::vector<TableLockPtr> TablesLocksList;
typedef std::set<std::string> TablesList;
typedef std::auto_ptr<TablesList > TablesListAutoPtr;

class ConnectionInfo : public boost::noncopyable
{
public:
	ConnectionInfo(THD& thd);
	~ConnectionInfo();
	
	ulong	GetThreadID() const			{ return thd.thread_id; }
	ulong	GetTransactionID() const	{ return transaction_id; }
	THD&	Thd() const 				{ return thd; }
	bool 	killed() const 				{ return (thd.killed != 0); }

	TablesListAutoPtr	GetListOfModifiedTables() const;
	void 				AddModifiedTable(const std::string& table_path);
	void 				RemoveTable(const std::string& table_path);

	/** Notify the end of transaction - switch transaction number
	 */
	void				EndTransaction();
	
	Transaction * GetNewTransaction( const RCEngine &engine); 
	Transaction * GetTransaction( );


	// Statistical stuff
	void	ResetStatistics()						{ packs_loaded=0; }
	void	NotifyPackLoad()						{ packs_loaded++; }
	_int64	GetPackCount()							{ return packs_loaded; }

	void	SetDisplayAttrStats(bool v = true)		{ display_attr_stats = v; }
	bool	DisplayAttrStats()						{ return display_attr_stats; }
	void SuspendDisplay();
	void ResumeDisplay() ;
	void ResetDisplay() ;


private:
	int		display_lock;			// if >0 disable messages e.g. in subqueries
	//ulong m_thd_id;
	uint  transaction_id;
	Transaction *current_transaction;
	
	TablesList modified_tables;
	//const int volatile& m_killed;

	_int64		packs_loaded;			// for query statistics
	bool		display_attr_stats;		// if set, then statistics on attributes should be displayed at the end of query
	THD& 		thd;
	mutable IBMutex modified_tables_guard;
public:
	process_info pi;
};

class IBTableShared
{
public:
	IBTableShared(const std::string& table_path)
		:	table_path(table_path)
	{
	}

	TableLockWeakPtr 	AcquireWriteLock(const THD& thd);
	TableLockWeakPtr	AcquireReadLock(const THD& thd);
	void				ReleaseLock(TableLockWeakPtr lock);
	void 				ReleaseLocks(const THD& thd);
	bool 				HasAnyLockForThread(const THD& thd);

	bool 				HasWriteLock() const;
	const THD&			WriteLockOwner() const { return write_lock->Thd(); }

private:
	bool 				HasWriteLockForThread(const THD& thd) const;

	bool 				HasReadLock() const;
	bool 				HasReadLockForThread(const THD& thd) const;
	bool 				HasReadLockForOtherThread(const THD& thd) const;

	/** Create Table READ LOCK.
	 * 	\param thd Thread which ask for a LOCK
	 *  \return weak pointer to new TableLock
	 */
	TableLockWeakPtr	CreateReadLockForThread(const THD& thd);

	/** Create Table READ LOCK.
	 *  \param thd Thread which ask for a LOCK
	 *  \return weak pointer to new TableLock
	 */
	TableLockWeakPtr	CreateWriteLockForThread(const THD& thd);

private:
	TablesLocksList read_lock_list;
	TableLockPtr 	write_lock;
	std::string 	table_path;
	IBMutex 		lock_request_sync;
};

typedef boost::shared_ptr<IBTableShared> IBTableSharedPtr;

class TableLockManager
{
public:

	void RegisterTable(const std::string& table_path);
	void UnRegisterTable(const std::string& table_path);

	TableLockWeakPtr AcquireWriteLock(const THD& thd, const std::string& table_path);
	TableLockWeakPtr AcquireReadLock(const THD& thd, const std::string& table_path);


	/** Release all if
	 * 	it is READ LOCK
	 *	it is WRITE LOCK and AUTOCOMMIT is turned off
	 *  \param table_lock table lock
	 */
	void ReleaseLockIfNeeded(TableLockWeakPtr table_lock);

	/** Release all LOCKS owned by a given thread on a given table
	 *  \param thd thread object
	 * 	\param table_path table identifier
	 */
	void ReleaseLocks(const THD& thd, const std::string& table_path);
	void ReleaseLocks(const THD& thd);

	bool HasAnyLockForTable(const THD& thd, const std::string& table_path);

private:
	IBTableSharedPtr GetTableShare(const std::string& table_path);

private:

	std::map<std::string, IBTableSharedPtr> registered_tables;
	IBMutex mutex;
};


#endif //_SYSTEM_CONNECTIONINFO_H_

