/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include <algorithm>

#include "RCSystem.h"
#include "LargeBuffer.h"
#include "core/WinTools.h"
#include "common/bhassert.h"
#include "core/tools.h"
#include "system/IOParameters.h"

#ifdef _MSC_VER
#include <windows.h>
#endif


using namespace std;
/////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////  LargeBuffer   //////////////////////////////////////////////////////////////

static int const D_OVERRUN_GUARDIAN = -66;

LargeBuffer::LargeBuffer(int numOfBufs, int requestedSize)
{
	this->numOfBufs = numOfBufs;
	bufs= new char*[numOfBufs];
	usageOfBufs = new int[numOfBufs];
	failed = false;

	size = requestedSize;

	if (bufs == NULL || usageOfBufs == NULL)
		rclog << lock << "Error: out of memory small malloc failed (42a)" << unlock;

	for (int i = 0; i< numOfBufs; i++) {
		bufs[i] = (char*)alloc((size + 2)*sizeof(char), BLOCK_TEMPORARY);				// additional 2 bytes for safety and terminating '\0';
		if(!bufs[i])
			rclog << lock << "Error: out of memory (" << size + 2 << " bytes failed). (42)" << unlock;
		usageOfBufs[i] = 0;
		bufs[i][size]='\0';		//maybe unnecessary
		bufs[i][size+1]=D_OVERRUN_GUARDIAN; //maybe unnecessary
	}

	currBufNo = 0;
	buf = bufs[currBufNo];
	currBuf2No = 1;
	buf2= bufs[currBuf2No];

	buf_status=0;
	buf_used=0;
	buf_incomplete=0;
	bytes_in_read_thread_buffer = 0;
	params.path = 0;
	//bufFlushThread = INVALID_THREAD;
	read_thread_buffer = 0;
	//reading_thread = 0;

	//pthread_mutexattr_t mattr;
	//pthread_mutexattr_init(&mattr);
	//pthread_mutexattr_settype(&mattr, PTHREAD_MUTEX_RECURSIVE);
	//pthread_mutex_init(&synch_read_thread, &mattr);
	//pthread_mutex_init(&bufUsageMutex, NULL);
	//pthread_cond_init(&isFreeBuf, NULL);
}

LargeBuffer::~LargeBuffer()
{
	if(read_thread.IsRunning()) {
		read_mutex.Lock();
		stop_reading_thread = true;
		read_mutex.Unlock();
		read_thread.Join();
	}

	if(flush_thread.IsRunning())
        flush_thread.Join();
	if(curfile.get())
		curfile->Close();
	for (int i = 0; i< numOfBufs; i++)
		dealloc(bufs[i]);
	delete [] bufs;
	delete [] usageOfBufs;
	delete [] params.path;
	if(read_thread_buffer)
		dealloc(read_thread_buffer);
}

bool LargeBuffer::IsPipe(const char *name)
{
	if (strncasecmp(name, "\\\\", 2)!=0) return false;
	const char *rest = strstr(name+2, "\\");
	if ((!rest) || (rest == name+2)) return false;   // backslash not found or found at the beginning of name+2
	if (strncasecmp(rest, "\\pipe\\", 6)!=0) return false;

	return true;
}

bool LargeBuffer::BufOpen(const IOParameters& iop, BufOpenType mode)
{
	const char* fname = iop.Path();
	bool result = false;
	
	if(buf_status != 0) {		
		return 0;
	}
	BufClose(); 

	try {
		curfile = IBStream::CreateOpenIBStream(iop, mode);
		if(!curfile.get() || !curfile->IsOpen()) {
			BufClose();
			return false;
		}		
	} catch ( ... ) {
		buf_status = 0;
		return false;
	}

	if(mode == READ) {
		buf_used = Read(curfile.get(), buf, size);
		if(buf_used == -1) {
			curfile->Close();
			buf_incomplete = 0;
			buf_status = 0;
			return false;
		} else if(buf_used == size) {
			StartReadingThread();
			buf_incomplete = 1;
		} else {
			curfile->Close();
			buf_incomplete = 0;
		}
		buf_status = 1;
		result = true;
	} else if(mode == OVERWRITE || mode == APPEND) {		
			if(mode == OVERWRITE) {			
				buf_used = 0;
				buf_status = 2;
				result = true;
			} else {			
				buf_used = 0;
				buf_status = 3;
				result = true;
			}		
	} else
		rclog << lock << "BufOpen warning: mode not implemented!" << unlock;
	return result;
}

bool LargeBuffer::StartReadingThread()
{
	if(BufferingLevel > 0)
	{
		if(!read_thread_buffer)
			read_thread_buffer  = (char*)alloc((size+1)*sizeof(char), BLOCK_TEMPORARY);
		bytes_in_read_thread_buffer = 0;

		THREADVAR threadId;
		//reading_thread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)ReadThread, this,0, &threadId);
		stop_reading_thread = false;
        read_thread.Start(ReadThread, (void*) this);
		threadId = read_thread.GetThreadID(); //pthread_create(&reading_thread, NULL, ReadThread, (void*) this);
		//if (threadId)
		//	return false;
		return true;
	}
	return false;
}

struct FlushThreadParams
{
	FlushThreadParams(IBStream *_file, char* _buf, int _len, bool& _failed)
       : file(_file), buf(_buf), buf_used(_len), failed(&_failed) {}
	IBStream *file;
    char* buf;
    int buf_used;
	bool* failed;
};

void LargeBuffer::BufFlush()		// save the data to the file (in write mode)
{
	if(buf_used>size)
		rclog << lock << "LargeBuffer error: Buffer overrun (Flush)" << unlock;
	if(flush_thread.IsRunning()) {
		//WaitForSingleObject(bufFlushThread, INFINITE);
        flush_thread.Join();
		//pthread_join(bufFlushThread, NULL);
		//bufFlushThread = INVALID_THREAD;
	}
	if(failed) {
		rclog << lock << "ERROR: Write operation to file or pipe failed." << unlock;	
		throw FileRCException("Write operation to file or pipe failed.");
	}
	THREADVAR threadId;
	//bufFlushThread = CreateThread(0, 0, (LPTHREAD_START_ROUTINE)BufFlushThread, &pars, 0, &threadId);
    FlushThreadParams *ftp = new FlushThreadParams(curfile.get(), buf, buf_used, failed);
	//threadId = pthread_create(&bufFlushThread, NULL, BufFlushThread, (void*)ftp);
    flush_thread.Start(BufFlushThread, (void*)ftp);
    threadId = flush_thread.GetThreadID();
	UseNextBuf();
}

#ifdef __GNUC__
void *LargeBuffer::BufFlushThread(void* _pars)
#else
DWORD WINAPI LargeBuffer::BufFlushThread(void* _pars)
#endif
{
    FlushThreadParams *ftp = (FlushThreadParams*) _pars;	
	IBStream *file = ftp->file;
	char* buf_ptr = ftp->buf;
	int len = ftp->buf_used;
	bool* failed = ftp->failed;
    delete ftp;

	if(file->IsOpen())
	{
		int start = 0, end = 0;
		while(end != len)
		{
			start = end;
			end += WRITE_TO_PIPE_NO_BYTES;
			if(end > len)
				end = len;			
			try {
				file->WriteExact(buf_ptr+start, end-start);
			} catch(DatabaseRCException& e)
			{				
				*failed = true;
				break;
			}
		}
	}
	return 0;
}

void LargeBuffer::BufClose()		// close the buffer; warning: does not flush data on disk (write mode)
{	
	if(flush_thread.IsRunning())
        flush_thread.Join();

	if(read_thread.IsRunning()) {
		read_mutex.Lock();
		stop_reading_thread = true;
		read_mutex.Unlock();
		read_thread.Join();
	}

	if(curfile.get())
		curfile->Close();

	buf_status=0;
	buf_used=0;
	buf_incomplete=0;
	if(buf[size+1]!=D_OVERRUN_GUARDIAN) {
		rclog << lock << "INTERNAL WARNING: buffer overrun detected in LargeBuffer::BufClose." << unlock;
		assert(0);
	}
}

void LargeBuffer::FlushAndClose()
{
	BufFlush();
	BufClose();
}

void LargeBuffer::BufDeclare(int bytes)
{
	if(bytes>size) rclog << lock << "LargeBuffer error: cannot initialize memory!" << unlock;
	buf_used = bytes;
}

char* LargeBuffer::BufAppend(unsigned int len)
{
	if((int)len>size)
	{
		rclog << lock << "Error: LargeBuffer buffer overrun (BufAppend)" << unlock;
		return NULL;
	}
	int buf_pos = buf_used;
	if(size > (int)(buf_used + len))
		buf_used += len;
	else
	{
		BufFlush();
		buf_status = 3;
		buf_pos = 0;
		buf_used = len;
	}
	return this->Buf(buf_pos);	// NOTE: will point to the first undeclared byte in case of len=0.
								// Fortunately we always assume one additional byte for '\0' to be insertable at the end of system buffer
}

int LargeBuffer::BufFetch(int unused_bytes, int to_read)
{
	if(to_read == 0)
		to_read = size;
	if(buf_status == 1)
	{
		//pthread_mutex_lock(&synch_read_thread);
        IBGuard guard(read_mutex);

		if(bytes_in_read_thread_buffer == -1)
		{
			curfile->Close();
			buf_incomplete = 0;
			//pthread_mutex_unlock(&synch_read_thread);
			throw FileRCException("Unable to read from the input file.");
		}

		if(!curfile->IsOpen() && bytes_in_read_thread_buffer == 0)
		{
			//pthread_mutex_unlock(&synch_read_thread);
			return 0;
		}

		for(int i = 0; i < unused_bytes; i++)
			buf2[i] = *(buf + ((buf_used - unused_bytes) + i));
		to_read -= unused_bytes;
		int to_read_from_th_buf = min(to_read, bytes_in_read_thread_buffer);

		memcpy(buf2 + unused_bytes, read_thread_buffer, to_read_from_th_buf);

		for(int i = 0; i < bytes_in_read_thread_buffer - to_read_from_th_buf; i++)
			read_thread_buffer[i] = read_thread_buffer[i+to_read_from_th_buf];

		bytes_in_read_thread_buffer -= to_read_from_th_buf;
		unused_bytes += to_read_from_th_buf;
		to_read -= to_read_from_th_buf;

		buf_used = unused_bytes;
		if(curfile->IsOpen())
		{
			ushort no_steps = 0;
			int r = 0;
			while(no_steps < 15 && to_read > 0)
			{
				int errno_before = errno;
				try {
				r = curfile->Read(buf2+unused_bytes, to_read);		// Read to buffer and set '\0' on the end of file
				} catch(DatabaseRCException& e)
				{
					r = -1;
				}
				if(r != -1)
				{
					buf_used += r;
					if(r == 0)
					{
						curfile->Close();
						buf_incomplete = 0;
						break;
					}
					else
					{
						buf_incomplete = 1;
					}
					if(r != to_read)
					{
						unused_bytes += r;
						to_read -= r;
					}
					else
						break;
				}
				else
				{
					no_steps++;
					rclog << lock << "Reading from the input file error: " << DisplayError().c_str() << unlock;
					IBThread::SleepNow(500 + no_steps*500);
				}

			}
			if(r == -1)
			{
				curfile->Close();
				buf_incomplete = 0;
				//pthread_mutex_unlock(&synch_read_thread);
				throw FileRCException("Unable to read from the input file.");
			}
		}
		buf_status=1;
		//pthread_mutex_unlock(&synch_read_thread);
	}
	UseNextBuf();
	return buf_used;
}

int LargeBuffer::Read(IBStream *file, char* buffer, int bytes_to_read)
{
	int requested_size = bytes_to_read;
	int no_attempts = 0;
	int bytes_read = 0;
	int read_b = -1;
	bool do_stop = false;
	bytes_read = 0;
	while(no_attempts < 16 && bytes_read != requested_size)
	{
		no_attempts++;
		do_stop = false;
		while(!do_stop)
		{
			try {
			read_b = file->Read(buffer + bytes_read , min(bytes_to_read, requested_size - bytes_read));
			} catch(DatabaseRCException& e)
			{
				read_b = -1;
			}
			if(read_b == -1)
				do_stop = true;
			else if(read_b == 0)
				return bytes_read;
			else
				bytes_read += read_b;
		}
		bytes_to_read /= 2;
	}
	return read_b != -1 ? bytes_read : read_b;
}

#ifdef __GNUC__
void *LargeBuffer::ReadThread(void* _lb)
#else
DWORD WINAPI LargeBuffer::ReadThread(void* _lb)
#endif
{
	int to_read = 0;
	LargeBuffer* lb = (LargeBuffer*)_lb;
	int to_read_in_one_loop = lb->size / BufferingLevel;
	bool do_sleep = false;
	bool do_stop = false;

	int no_read_bytes = -1;

	while(!do_stop) {
		no_read_bytes = 0;
        lb->read_mutex.Lock();
		if(lb->buf_incomplete != 1 || !lb->curfile->IsOpen() || lb->stop_reading_thread) {
            lb->read_mutex.Unlock();
			break;
		}

		if(lb->bytes_in_read_thread_buffer != lb->size) {
			to_read = to_read_in_one_loop;
			if(lb->size - lb->bytes_in_read_thread_buffer < to_read_in_one_loop)
				to_read = lb->size - lb->bytes_in_read_thread_buffer;
			ushort no_steps = 0;
			while(no_steps < 15  && to_read > 0) {
				int errno_before = errno;
				try {
				no_read_bytes = lb->curfile->Read(lb->read_thread_buffer + lb->bytes_in_read_thread_buffer, to_read);
				} catch(DatabaseRCException& e) {
					no_read_bytes = -1;
				}				
				if(no_read_bytes == -1) {
					no_steps++;
					rclog << lock << "Reading from the input file error: " << DisplayError().c_str() << unlock;
					IBThread::SleepNow(500 + no_steps*500);
					do_stop = true;
				} else {
					do_stop = false;
					lb->bytes_in_read_thread_buffer += no_read_bytes;
					if(no_read_bytes != 0)
						lb->buf_incomplete = 1;
					else {
						lb->curfile->Close();
						lb->buf_incomplete = 0;
						//LeaveCriticalSection(&lb->synch_read_thread);
						do_stop = true;
					}
					break;
				}
			}

			if(no_read_bytes == -1) {
				lb->bytes_in_read_thread_buffer = -1;
				do_stop = true;
			}
		}

		if(lb->bytes_in_read_thread_buffer == lb->size)
			do_sleep = true;
		else
			do_sleep = false;
        lb->read_mutex.Unlock();
		//pthread_mutex_unlock(&lb->synch_read_thread);
		if(do_sleep)
			IBThread::SleepNow(20);
		if(no_read_bytes != -1)
			IBThread::SleepNow(20);
	}
	return 0;
}


void LargeBuffer::UseNextBuf()
{
	int tmp = currBuf2No;
	currBuf2No = FindUnusedBuf();
	buf2 = bufs[currBuf2No];		//to be loaded with data
	currBufNo = tmp;
	buf =  bufs[currBufNo];			//to be parsed
}

int LargeBuffer::FindUnusedBuf()
{
	//pthread_mutex_lock(&bufUsageMutex);
    buffer_cond.Lock();
	while (true) {
		for (int i = 0; i< numOfBufs; i++) {
			if (usageOfBufs[i] == 0 && i != currBuf2No) { //not used and not just loaded
                buffer_cond.Unlock();
				//pthread_mutex_unlock(&bufUsageMutex);
				return i;
			}
		}
        buffer_cond.Wait();
		//pthread_cond_wait(&isFreeBuf, & bufUsageMutex);
	}

}

void LargeBuffer::MarkUseOf(int bufNo)
{
	BHASSERT(bufNo < numOfBufs, " marking use of too big bufNo");
	//pthread_mutex_lock(&bufUsageMutex);
    buffer_cond.Lock();
	usageOfBufs[bufNo]++;
    buffer_cond.Unlock();
	//pthread_mutex_unlock(&bufUsageMutex);
}

void LargeBuffer::ReleaseFromUse(int bufNo)
{
	BHASSERT(bufNo < numOfBufs,"releasing too big bufNo");
	//pthread_mutex_lock(&bufUsageMutex);
    buffer_cond.Lock();
	usageOfBufs[bufNo]--;
	if (usageOfBufs[bufNo] == 0) {
        buffer_cond.Signal();
		//pthread_cond_signal(&isFreeBuf);
    }
    buffer_cond.Unlock();
	//pthread_mutex_unlock(&bufUsageMutex);
}

#ifdef _XMK
#ifdef _MSC_VER
bool createSecurityDescritpor(SECURITY_ATTRIBUTES &sa) 
{
	//returns true on success
	class tmpException {};

	unsigned long dwRes;
	PSID pEveryoneSID = NULL;
	PACL pACL = NULL;
	PSECURITY_DESCRIPTOR pSD = NULL;
	EXPLICIT_ACCESS ea;
	SID_IDENTIFIER_AUTHORITY SIDAuthWorld = SECURITY_WORLD_SID_AUTHORITY;
	HKEY hkSub = NULL;

	try {


		// Create a well-known SID for the Everyone group.
		if(!AllocateAndInitializeSid(&SIDAuthWorld, 1,
			SECURITY_WORLD_RID,
			0, 0, 0, 0, 0, 0, 0,
			&pEveryoneSID)) throw tmpException();

		// Initialize an EXPLICIT_ACCESS structure for an ACE.
		// The ACE will allow Everyone read access to the key.
		ZeroMemory(&ea, sizeof(EXPLICIT_ACCESS));
		ea.grfAccessPermissions = GENERIC_READ | GENERIC_WRITE;
		ea.grfAccessMode = SET_ACCESS;
		ea.grfInheritance= NO_INHERITANCE;
		ea.Trustee.TrusteeForm = TRUSTEE_IS_SID;
		ea.Trustee.TrusteeType = TRUSTEE_IS_COMPUTER;
		ea.Trustee.ptstrName  = (LPTSTR) pEveryoneSID;

		// Create a new ACL that contains the new ACEs.
		dwRes = SetEntriesInAcl(1, &ea, NULL, &pACL);
		if (ERROR_SUCCESS != dwRes) throw tmpException();

		// Initialize a security descriptor.  
		pSD = (PSECURITY_DESCRIPTOR) LocalAlloc(LPTR, 
			SECURITY_DESCRIPTOR_MIN_LENGTH); 
		if (NULL == pSD) throw tmpException();

		if (!InitializeSecurityDescriptor(pSD, SECURITY_DESCRIPTOR_REVISION)) throw tmpException();


		// Add the ACL to the security descriptor. 
		if (!SetSecurityDescriptorDacl(pSD, 
			TRUE,     // bDaclPresent flag   
			pACL, 
			FALSE))   // not a default DACL 
			throw tmpException();

		// Initialize a security attributes structure.
		sa.nLength = sizeof (SECURITY_ATTRIBUTES);
		sa.lpSecurityDescriptor = pSD;
		sa.bInheritHandle = FALSE;

		// Use the security attributes to set the security descriptor 
	}
	catch (tmpException &) {
		if (pEveryoneSID) 
			FreeSid(pEveryoneSID);
		if (pACL) 
			LocalFree(pACL);
		if (pSD) 
			LocalFree(pSD);

		return false;
	}

	return true;
}
#endif
#endif