/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "BasicHeapPolicy.h"

#include "system/RCSystem.h"
#include "core/tools.h"
#include "common/bhassert.h"

using namespace std;

BasicHeap::BasicHeap()
{
	this->heap_frame = NULL;
	is_huge= false;
}

BasicHeap::~BasicHeap()
{
	if(heap_frame) {
#ifdef __GNUC__  // should be #ifdef LINUX but it's left __GNUC__ for meantime
		if (is_huge) {
			munmap(heap_frame, size);
			unlink(hugefilename);
		} else {
			delete[] heap_frame;
		}
#else
		delete[] heap_frame;
#endif //__GNUC__
	}

	heap_frame = NULL;
}

MemoryBlock<MEM_HANDLE_BH>* BasicHeap::CreateHeap(size_t size, string hugedir)
{
	/*
		Allocate buffer of a given size and put corresponding
		memory block into ADDRESS_TREE and LENGTH_TREE
	*/

	try
	{
#ifdef __GNUC__  // should be #ifdef LINUX but it's left __GNUC__ for meantime
		if (!hugedir.empty()) {
			 char pidtext[12];
			 strcpy(hugefilename, hugedir.c_str());
			 sprintf(pidtext, "%d", getpid());
			 strcat(hugefilename, "/bhhuge.");
			 strcat(hugefilename, pidtext);
			 int fd = open(hugefilename, O_CREAT | O_RDWR, 0700);
			 if (fd < 0) {
					m_hs = HEAP_OUT_OF_MEMORY;
					rccontrol << lock << "Memory Manager Error: Unable to create hugepage file: " << hugefilename << unlock;
					return NULL;
			 }
			 // MAP_SHARED to have mmap fail immediately if not enough pages
			 heap_frame = (char*)mmap(NULL, size,PROT_READ | PROT_WRITE, MAP_PRIVATE, fd, 0);
			 if (heap_frame == MAP_FAILED ) {
				 	unlink(hugefilename);
				 	m_hs = HEAP_OUT_OF_MEMORY;
					rccontrol << lock << "Memory Manager Error: hugepage file mmap error: " << GetErrorMessage(errno) << unlock;
					return NULL;
			 }
			 is_huge = true;
		 } else
			 heap_frame = new char[size];
#else
	    heap_frame = new char[size];
#endif //__GNUC__

		if(heap_frame == NULL)
		{
			m_hs = HEAP_OUT_OF_MEMORY;
			rccontrol << lock << "Memory Manager Error: Unable to allocate: " << (unsigned int)size << " bytes" << unlock;
			return NULL;
		}

		MemoryBlock<MEM_HANDLE_BH>* init_blck = new MemoryBlock<MEM_HANDLE_BH>();

		init_blck->setAddress(heap_frame);
		init_blck->setNext(NULL);
		init_blck->setPrev(NULL);
		init_blck->setSize(size);
		init_blck->setType(BLOCK_FREE);
		init_blck->setOwner(NULL);

		LENGTH_TREE.insert(size, init_blck);
		ADDRESS_TREE.insert(heap_frame, init_blck);

		m_hs = HEAP_SUCCESS;

		return init_blck;
	}
	catch(bad_alloc)
	{
		m_hs = HEAP_OUT_OF_MEMORY;
		rccontrol << lock << "Memory Manager Error: Unable to allocate: " << (unsigned int)size << " bytes" << unlock;
		return NULL;
	}
}

MEM_HANDLE_BH BasicHeap::alloc(size_t size, BLOCK_TYPE type, TrackableObject* owner)
{
#ifdef FUNCTIONS_EXECUTION_TIMES
	FETOperator feto("BasicHeap::alloc(...)");
#endif
	BHASSERT(type != BLOCK_FREE && type != BLOCK_FIXED, "should be 'type != BLOCK_FREE && type != BLOCK_FIXED'");

	/*
		Allocation algorithm:

		1. Splay the LENGTH_TREE to get best matching free memory block (or tree node)
		2. Remove the node from the LENGTH_TREE and remove the corresponding address node from the POINTER_TREE
		3. If the memory block is larger than the requested size, split the memory block
			and insert the remainder into the LENGTH_TREE and POINTER_TREE
		4. Return the pointer of the block (minus the heap block header)
	*/

	LENGTH_TREE.splay(size);

	//block will receive first memory block from root's list
	MemoryBlock<MEM_HANDLE_BH>* block = LENGTH_TREE.removeRoot(); //Splay op moved best matching block to root position

	if(!block || block->getSize() < size)
	{
		/*
			There is no free space in the memory frame
		*/
		if(block)
			LENGTH_TREE.insert(block->getSize(), block); // Put root back to the tree
		m_hs = HEAP_OUT_OF_MEMORY;
		return NULL;
	}
	if(ADDRESS_TREE.remove(block->getAddress()) == NULL)
	{
		//Heap corruption occurred!
		rccontrol << lock << "Heap corruption error 1" << unlock;
		BHERROR("Heap corruption error 1");
		delete block;
		m_hs = HEAP_CORRUPTED;
		return NULL;
	}

	block->setType(type);
	block->setOwner(owner);

	if(block->getSize() > size)
	{
		MemoryBlock<MEM_HANDLE_BH>* remainder_block = block->Split(size);
		LENGTH_TREE.insert(remainder_block->getSize(), remainder_block);
		ADDRESS_TREE.insert(remainder_block->getAddress(), remainder_block);
	}

	alloc_map.insert(pair<MEM_HANDLE_BH, MemoryBlock<MEM_HANDLE_BH>* >(block->getAddress(), block));

	m_hs = HEAP_SUCCESS;
	return block->getAddress();
}

void BasicHeap::dealloc(MEM_HANDLE_BH mh)
{
#ifdef FUNCTIONS_EXECUTION_TIMES
	FETOperator feto("BasicHeap::dealloc(...)");
#endif
	/*
		deallocation algorithm:

		1. Get the successor memory block (CURRENT_BLOCK + length of the block)
		2. Splay the POINTER_TREE to check if the successor block is in the free list or not
		3. If the successor is in the free list, merge the current node with it
		4. Check if the left node of the successor node is actually the predecessor node of the current node
		5. If a predecessor node is found, merge the current node with it
		6. Insert the merged node into the POINTER_TREE
		7. Insert the merged node into the LENGTH_TREE
	*/

	m_hs = HEAP_SUCCESS;

	if(mh == (MEM_HANDLE_BH)0)
	{
		return;
	}

	map<MEM_HANDLE_BH, MemoryBlock<MEM_HANDLE_BH>* >::iterator it = alloc_map.find(mh);
	if(it == alloc_map.end())
	{
		rccontrol << lock << "Heap corruption error 2" << unlock;
		m_hs = HEAP_CORRUPTED;
		return;
	}

	MemoryBlock<MEM_HANDLE_BH>* dealloc_block = it->second;
	size_t size = dealloc_block->getSize();

	SplayNode<void*, MemoryBlock<MEM_HANDLE_BH>* >* successor = ADDRESS_TREE.strictSplay((MEM_HANDLE_BH)(size + (size_t)mh));
	if(successor) {
		/*
		Merge successor with dealloc_block
		*/

		BHASSERT(successor->items->size() == 1, "should be 'successor->items->size() == 1'");
		//__assume(successor->items->size() == 1);

		MemoryBlock<MEM_HANDLE_BH>* succ_item = successor->items->front();

		alloc_map.erase(it);
		dealloc_block->Merge(succ_item);
		dealloc_block->setType(BLOCK_FREE);
		dealloc_block->setOwner(NULL);

		SplayNode<void*, MemoryBlock<MEM_HANDLE_BH>* >* pred_node = ADDRESS_TREE.predecessor(successor);
		//if (successor->left)
		if(pred_node) {
			//MemoryBlock<MEM_HANDLE_BH>* succ_item = successor->item; -> successor_block
			//BHASSERT(successor->left->items->size() == 1, "should be 'successor->left->items->size() == 1'");
			BHASSERT(pred_node->items->size() == 1, "should be 'pred_node->items->size() == 1'");
			//__assume(pred_node->items->size() == 1);
			//__assume(successor->left->items->size() == 1);

			//MemoryBlock<MEM_HANDLE_BH>* predecessor = successor->left->items->front();
			MemoryBlock<MEM_HANDLE_BH>* predecessor = pred_node->items->front();

			if((MEM_HANDLE_BH)((size_t)predecessor->getAddress() + predecessor->getSize()) == dealloc_block->getAddress())
			{
				LENGTH_TREE.removeEx(predecessor->getSize(), predecessor);

				ADDRESS_TREE.remove(succ_item->getAddress());
				LENGTH_TREE.removeEx(succ_item->getSize(), succ_item);
				delete succ_item;

				predecessor->Merge(dealloc_block);

				LENGTH_TREE.insert(predecessor->getSize(), predecessor);
				predecessor->setType(BLOCK_FREE);
				predecessor->setOwner(NULL);
				delete dealloc_block;
			}
			else
			{
				ADDRESS_TREE.remove(succ_item->getAddress());
				LENGTH_TREE.removeEx(succ_item->getSize(), succ_item);
				delete succ_item;

				ADDRESS_TREE.insert(dealloc_block->getAddress(), dealloc_block);
				LENGTH_TREE.insert(dealloc_block->getSize(), dealloc_block);
				dealloc_block->setType(BLOCK_FREE);
				dealloc_block->setOwner(NULL);
			}
		}
		else
		{
			//MemoryBlock<MEM_HANDLE_BH>* item = successor->item;

			ADDRESS_TREE.remove(succ_item->getAddress());
			LENGTH_TREE.removeEx(succ_item->getSize(), succ_item);
			delete succ_item;

			ADDRESS_TREE.insert(dealloc_block->getAddress(), dealloc_block);
			LENGTH_TREE.insert(dealloc_block->getSize(), dealloc_block);
			dealloc_block->setType(BLOCK_FREE);
			dealloc_block->setOwner(NULL);
		}
	}
	else
	{
		//successor is not a free block
		alloc_map.erase(it);
		ADDRESS_TREE.insert(dealloc_block->getAddress(), dealloc_block);

		dealloc_block->setType(BLOCK_FREE);
		dealloc_block->setOwner(NULL);

		SplayNode<void*, MemoryBlock<MEM_HANDLE_BH>* >* dealloc_node = ADDRESS_TREE.strictSplay(dealloc_block->getAddress());

		SplayNode<void*, MemoryBlock<MEM_HANDLE_BH>* >* pred_node = ADDRESS_TREE.predecessor(dealloc_node);
		//if(dealloc_node->left)
		if(pred_node)
		{
			//BHASSERT(dealloc_node->left->items->size() == 1, "should be 'dealloc_node->left->items->size() == 1'");
			BHASSERT(pred_node->items->size() == 1, "should be 'pred_node->items->size() == 1'");
			BHASSERT(dealloc_node->items->size() == 1, "should be 'dealloc_node->items->size() == 1'");

			//__assume(pred_node->items->size() == 1);
			//__assume(dealloc_node->items->size() == 1);

			//MemoryBlock<MEM_HANDLE_BH>* predecessor = dealloc_node->left->items->front();
			MemoryBlock<MEM_HANDLE_BH>* predecessor = pred_node->items->front();
			MemoryBlock<MEM_HANDLE_BH>* dealloc_itm = dealloc_node->items->front();

			if((MEM_HANDLE_BH)((size_t)predecessor->getAddress() + predecessor->getSize()) == dealloc_block->getAddress())
			{
				ADDRESS_TREE.remove(dealloc_itm->getAddress());
				LENGTH_TREE.removeEx(predecessor->getSize(), predecessor);
				predecessor->Merge(dealloc_block);

				LENGTH_TREE.insert(predecessor->getSize(), predecessor);
				predecessor->setType(BLOCK_FREE);
				predecessor->setOwner(NULL);

				delete dealloc_block;
			}
			else
				LENGTH_TREE.insert(dealloc_block->getSize(), dealloc_block);
		}
		else
			LENGTH_TREE.insert(dealloc_block->getSize(), dealloc_block);
	}
}

MEM_HANDLE_BH BasicHeap::rc_realloc(MEM_HANDLE_BH mh, size_t size)
{
#ifdef FUNCTIONS_EXECUTION_TIMES
	FETOperator feto("BasicHeap::rc_realloc(...)");
#endif
	/*
		1. Get the successor memory block (CURRENT_BLOCK + length of the block)
		2. Splay the POINTER_TREE to check is the successor block is in the free list or not
		3. If the successor is in the free list merge the current node with it
		4. If there is no successor or it is too small allocate [size] of memory
		5. copy contents
		6. dealloc old memory block
	*/

	if(mh == (MEM_HANDLE_BH)0)
	{
		return NULL;
	}


	map<MEM_HANDLE_BH, MemoryBlock<MEM_HANDLE_BH>* >::iterator it = alloc_map.find(mh);
	if(it == alloc_map.end())
	{
		rccontrol << lock << "Heap corruption error 3" << unlock;
		m_hs = HEAP_CORRUPTED;
		return NULL;
	}

	size_t old_size = it->second->getSize();

	SplayNode<void*, MemoryBlock<MEM_HANDLE_BH>* >* successor = ADDRESS_TREE.strictSplay((MEM_HANDLE_BH)(old_size + (size_t)mh));

	if(successor) {
		BHASSERT(successor->items->size() == 1, "should be 'successor->items->size() == 1'");
		//__assume(successor->items->size() == 1);

		MemoryBlock<MEM_HANDLE_BH>* item = successor->items->front();

		if( item->getSize() + old_size >= size)
		{
			/*
			Merge successor with dealloc_block
			*/
			it->second->Merge(item);

			ADDRESS_TREE.remove(item->getAddress());
			LENGTH_TREE.removeEx(item->getSize(), item);
			delete item;

			MemoryBlock<MEM_HANDLE_BH>* new_node = it->second->Split(size);

			if(new_node)
			{
				ADDRESS_TREE.insert(new_node->getAddress(), new_node);
				LENGTH_TREE.insert(new_node->getSize(), new_node);
			}

			m_hs = HEAP_SUCCESS;
			return mh;
		}
		else
		{

			MEM_HANDLE_BH new_ptr = alloc(size, it->second->getType(), it->second->getOwner());
			if(!new_ptr)
				return NULL;

			memcpy(new_ptr, mh, old_size);
			dealloc(mh);
			return new_ptr;
		}
	}
	else
	{

		MEM_HANDLE_BH new_ptr = alloc(size, it->second->getType(), it->second->getOwner());
		if(!new_ptr)
			return NULL;

		memcpy(new_ptr, mh, old_size);
		dealloc(mh);
		return new_ptr;
	}
}

size_t BasicHeap::getBlockSize(MEM_HANDLE_BH mh)
{
	map<MEM_HANDLE_BH, MemoryBlock<MEM_HANDLE_BH>* >::iterator it = alloc_map.find(mh);
	if(it == alloc_map.end())
	{
		m_hs = HEAP_CORRUPTED;
		return 0;
	}
	m_hs = HEAP_SUCCESS;
	return it->second->getSize();
}

size_t BasicHeap::getMaximalBlock()
{
	return this->LENGTH_TREE.getMostRightKey();
}
/*
template class BasicHeap<void*>;
*/

