/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include <map>

#include "system/linux/hash_set_ext.h"

#include "MemoryHandlingPolicy.h"
#include "TrackableObject.h"
#include "core/RCAttrPack.h"
#include "core/tools.h"
#include "system/RCSystem.h"
#include "system/RCException.h"
#include "common/bhassert.h"
#include "core/DataCache.h"

using namespace std;

int cmp_eval_func(unsigned short cmp, unsigned short free)
{
	return 5 * cmp - free;
}
int uncmp_eval_func(unsigned short uncmp, unsigned short uptodate, unsigned short free)
{
	return free + uptodate + 5 * uncmp;
}

MemoryHandling::MemoryHandling(size_t comp_heap_size, size_t uncomp_heap_size, size_t cmp_rls, size_t uncomp_rls, string hugedir)
{
	comp_heap_release = cmp_rls;
	uncomp_heap_release = uncomp_rls;
	main_heap_MB = int(uncomp_heap_size>>20);				// ServerMainHeapSize, in MB

	this->comp_heap_mem = this->compressed_heap.CreateHeap(comp_heap_size);
	this->uncomp_heap_mem = this->uncompressed_heap.CreateHeap(uncomp_heap_size, hugedir);


	if(comp_heap_mem == NULL || uncomp_heap_mem == NULL)
	{
		throw OutOfMemoryRCException("Memory manager is unable to allocate specified amount of memory");
	}

	this->cmp_eval = &cmp_eval_func;
	this->uncmp_eval = &uncmp_eval_func;
}

MemoryHandling::~MemoryHandling(void)
{

	if(this->comp_heap_mem)
		delete this->comp_heap_mem;
	if(this->uncomp_heap_mem)
		delete this->uncomp_heap_mem;
	//Clear owner maps
	OwnerMemoryMap::iterator it;
	for(it = cmp_blck_mbrshp.begin(); it != cmp_blck_mbrshp.end(); it++)
		delete it->second; //delete hash_set
	for(it = uncmp_blck_mbrshp.begin(); it != uncmp_blck_mbrshp.end(); it++)
		delete it->second;
	OwnerMemoryMapSystem::iterator it2;
	for(it2 = system_heap_map.begin(); it2 != system_heap_map.end(); it2++)
		delete it2->second;
}

void MemoryHandling::ReportLeaks()
{
	unsigned uncomp = 0, comp = 0, temp = 0;
	for (AllocBlckMap::const_iterator it = alloc_blck_type.begin(); it!=alloc_blck_type.end(); it++)
		if (it->second==BLOCK_UNCOMPRESSED) uncomp++;
		else if (it->second==BLOCK_COMPRESSED) comp++;
		else if (it->second==BLOCK_TEMPORARY) temp++;
	if (uncomp > 0)
		rclog << lock << "Warning: " << uncomp << " uncompressed memory block(s) leaked out." << unlock;
	if (comp > 0)
		rclog << lock << "Warning: " << comp << " compressed memory block(s) leaked out." << unlock;
	if (temp > 0)
		rclog << lock << "Warning: " << temp << " temporary memory block(s) leaked out." << unlock;
	unsigned sys_block = 0;
	for (OwnerMemoryMapSystem::const_iterator it = system_heap_map.begin(); it!=system_heap_map.end(); it++)
		sys_block += it->second->size();
	if (sys_block > 0)
		rclog << lock << "Warning: " << sys_block << " system memory block(s) (used when Brighthouse memory area ended up) leaked out." << unlock;
}

size_t MemoryHandling::rc_msize(MEM_HANDLE_MP mh, TrackableObject* owner)
{
    IBGuard guard(ccs.counting_mutex);

	if(mh == NULL)
	{
		return 0;
	}
	AllocBlckMap::iterator it = alloc_blck_type.find(mh);
	if (it != alloc_blck_type.end())
	{
		size_t size = 0;

		switch(it->second)
		{
		case BLOCK_COMPRESSED:

			size = compressed_heap.getBlockSize(mh);

			if((size != 0) && compressed_heap.getHeapStatus() != HEAP_CORRUPTED)
			{
				return size;
			}
			else
			{
				rccontrol << lock << "rc_msize failed" << unlock;
				BHERROR("rc_msize failed");
				return -1;
			}
			break;
		case BLOCK_UNCOMPRESSED:
		case BLOCK_TEMPORARY:

			size = uncompressed_heap.getBlockSize(mh);

			if((size != 0) && uncompressed_heap.getHeapStatus() != HEAP_CORRUPTED)
			{
				return size;
			}
			else
			{
				rccontrol << lock << "rc_msize failed" << unlock;
				BHERROR("rc_msize failed");
				return -1;
			}
			break;
		default:
			rccontrol << lock << "rc_msize failed" << unlock;
			BHERROR("rc_msize failed");
			return -1;
		}
	}
	else
	{
		OwnerMemoryMapSystem::iterator it = system_heap_map.find(owner);
		if(it != system_heap_map.end())
		{
			hash_set_for_system_mem_heaps::iterator hs_it = it->second->find(MEM_HANDLE_MP_WITH_SIZE(mh,0));
			if(hs_it != it->second->end())
				return hs_it->GetSize();
			else
			{
				rccontrol << lock << "Request for a size of a pointer which is not traced" << unlock;
				BHERROR("Request for a size of a pointer which is not traced");
				return -1;
			}
		}
		else
		{
			rccontrol << lock << "Request for a size of a pointer which is not traced" << unlock;
			BHERROR("Request for a size of a pointer which is not traced");
			return -1;
		}
	}
}

MEM_HANDLE_MP MemoryHandling::alloc(size_t size, BLOCK_TYPE type, TrackableObject* owner)
{
    IBGuard guard(ccs.counting_mutex);

#ifdef FUNCTIONS_EXECUTION_TIMES
	FETOperator feto("MemoryHandling::alloc(...)");
#endif
	if(size <= 0)
	{
		return NULL;
	}
	MEM_HANDLE_MP addr;
	OwnerMemoryMap::iterator it;

	switch(type)
	{
	case BLOCK_COMPRESSED:
		addr = this->compressed_heap.alloc(size, type, owner);
		if (this->compressed_heap.getHeapStatus() == HEAP_OUT_OF_MEMORY)
		{
			bool released = ReleaseMemory(comp_heap_release < size ? size : comp_heap_release, RH_COMPRESSED, owner);
			if(!released && !ReleaseMemory(size, RH_COMPRESSED, owner))
			{
				//Last resort procedure - try to allocate memory using system heap
				addr = (MEM_HANDLE_MP)malloc(size);
				if(addr)
				{
					//system_heap_alloc.push_back(addr);

					OwnerMemoryMapSystem::iterator sys_it;
					sys_it = system_heap_map.find(owner);
					if(sys_it != system_heap_map.end())
						sys_it->second->insert(MEM_HANDLE_MP_WITH_SIZE(addr, size));
					else
					{
						pair<OwnerMemoryMapSystem::iterator, bool> res = system_heap_map.insert(make_pair(owner, new hash_set_for_system_mem_heaps));
						res.first->second->insert(MEM_HANDLE_MP_WITH_SIZE(addr, size));
					}
					return addr;
				}

				rclog << lock << "Returned NULL in alloc (system heap): " << __FILE__ << " " << __LINE__ << unlock;
				BHERROR("Returned NULL in alloc (system heap)");
				return NULL; // Unable to free memory and allocate new chunk
			}
			else
				addr = this->compressed_heap.alloc(size, type, owner);
		}

		it = this->cmp_blck_mbrshp.find(owner);
		if(it == cmp_blck_mbrshp.end())
		{
			pair<OwnerMemoryMap::iterator, bool> res = cmp_blck_mbrshp.insert(mbrshp_pair(owner, new stdext::hash_set<MEM_HANDLE_MP> ));
			res.first->second->insert(addr);
		}
		else
			it->second->insert(addr);
		break;
	case BLOCK_TEMPORARY:
		addr = this->uncompressed_heap.alloc(size, type, owner);
		if(this->uncompressed_heap.getHeapStatus() == HEAP_OUT_OF_MEMORY)
		{
			bool released = ReleaseMemory(uncomp_heap_release < size ? size : uncomp_heap_release, RH_UNCOMPRESSED, owner);
			if(!released && !ReleaseMemory(size, RH_UNCOMPRESSED, owner))
			{
				//Last resort procedure - try to allocate memory using system heap
				addr = (MEM_HANDLE_MP)malloc(size);
				if(addr)
				{
					//system_heap_alloc.push_back(addr);

					OwnerMemoryMapSystem::iterator sys_it;
					sys_it = system_heap_map.find(owner);
					if(sys_it != system_heap_map.end())
						sys_it->second->insert(MEM_HANDLE_MP_WITH_SIZE(addr, size));
					else
					{
						pair<OwnerMemoryMapSystem::iterator, bool> res = system_heap_map.insert(make_pair(owner, new hash_set_for_system_mem_heaps));
						res.first->second->insert(MEM_HANDLE_MP_WITH_SIZE(addr, size));
					}
					return addr;
				}

				rclog << lock << "Returned NULL in alloc (system heap): " << __FILE__ << " " << __LINE__ << unlock;
				BHERROR("Returned NULL in alloc (system heap)");
				return NULL; // Unable to free memory and allocate new chunk
			}
			else
			{
				addr = this->uncompressed_heap.alloc(size, type, owner);
				BHASSERT(uncompressed_heap.getHeapStatus() == HEAP_SUCCESS && addr,
						"should be 'uncompressed_heap.getHeapStatus() == HEAP_SUCCESS && addr'");
			}
		}
		break;
	case BLOCK_UNCOMPRESSED:

		addr = this->uncompressed_heap.alloc(size, type, owner);
		if(CachingLevel == 1 && this->uncompressed_heap.getHeapStatus() == HEAP_OUT_OF_MEMORY)
		{
			//cout << "CLEANING 1." << endl;
			ReleaseUnlockedDataPacks();
			addr = this->uncompressed_heap.alloc(size, type, owner);
		}
		if(this->uncompressed_heap.getHeapStatus() == HEAP_OUT_OF_MEMORY)
		{
			//cout << "CLEANING 2." << endl;
			bool released = ReleaseMemory(uncomp_heap_release < size ? size : uncomp_heap_release, RH_UNCOMPRESSED, owner);
			if(!released && !ReleaseMemory(size, RH_UNCOMPRESSED, owner))
			{
				//Last resort procedure - try to allocate memory using system heap
				addr = (MEM_HANDLE_MP)malloc(size);
				if(addr)
				{
					//system_heap_alloc.push_back(addr);

					OwnerMemoryMapSystem::iterator sys_it;
					sys_it = system_heap_map.find(owner);
					if(sys_it != system_heap_map.end())
						sys_it->second->insert(MEM_HANDLE_MP_WITH_SIZE(addr, size));
					else
					{
						pair<OwnerMemoryMapSystem::iterator, bool> res = system_heap_map.insert(make_pair(owner, new hash_set_for_system_mem_heaps));
						res.first->second->insert(MEM_HANDLE_MP_WITH_SIZE(addr, size));
					}
					return addr;
				}

				rclog << lock << "Returned NULL in alloc (system heap): " << __FILE__ << " " << __LINE__ << unlock;
				BHERROR("Returned NULL in alloc (system heap)");
				return NULL; // Unable to free memory and allocate new chunk
			}
			else
			{
				addr = this->uncompressed_heap.alloc(size, type, owner);
				BHASSERT(uncompressed_heap.getHeapStatus() == HEAP_SUCCESS && addr,
						"should be 'uncompressed_heap.getHeapStatus() == HEAP_SUCCESS && addr'");
			}
		}

		it = this->uncmp_blck_mbrshp.find(owner);
		if(it == uncmp_blck_mbrshp.end())
		{
			pair<OwnerMemoryMap::iterator, bool> res = uncmp_blck_mbrshp.insert(mbrshp_pair(owner, new stdext::hash_set<MEM_HANDLE_MP> ));
			res.first->second->insert(addr);
		}
		else
			it->second->insert(addr);

		break;
	default:
		rccontrol << lock << "Returned NULL in alloc: " << __FILE__ << " " << __LINE__ << unlock;
		BHERROR("Returned NULL in alloc");
		return NULL;
	}
	if(addr)
		alloc_blck_type.insert(alloc_blck_pair(addr, type));

	if (addr == NULL) {
		rccontrol << lock << "Returned NULL in alloc: " << __FILE__ << " " << __LINE__ << unlock;
		BHERROR("Returned NULL in alloc");
	}
	return addr;
}

void MemoryHandling::dealloc(MEM_HANDLE_MP mh, TrackableObject* owner)
{
    IBGuard guard(ccs.counting_mutex);

#ifdef FUNCTIONS_EXECUTION_TIMES
	FETOperator feto("MemoryHandling::dealloc(...)");
#endif
	if(mh == NULL)
	{
		return;
	}
	AllocBlckMap::iterator it = alloc_blck_type.find(mh);

	stdext::hash_set<MEM_HANDLE_MP>::iterator hs_it;
	OwnerMemoryMap::iterator omm_it;

	if(it != alloc_blck_type.end())
	{
		switch( it->second )
		{
		case BLOCK_COMPRESSED:
			compressed_heap.dealloc(mh);
			alloc_blck_type.erase(it);
			omm_it = cmp_blck_mbrshp.find(owner);
			hs_it = omm_it->second->find(mh);
			omm_it->second->erase(hs_it);
			if(omm_it->second->size() == 0)
			{
				delete omm_it->second;
				cmp_blck_mbrshp.erase(omm_it);
			}
			break;
		case BLOCK_TEMPORARY:
			uncompressed_heap.dealloc(mh);
			alloc_blck_type.erase(it);
			break;
		case BLOCK_UNCOMPRESSED:
			uncompressed_heap.dealloc(mh);
			alloc_blck_type.erase(it);
			omm_it = uncmp_blck_mbrshp.find(owner);
			hs_it = omm_it->second->find(mh);
			omm_it->second->erase(hs_it);
			if(omm_it->second->size() == 0)
			{
				delete omm_it->second;
				uncmp_blck_mbrshp.erase(omm_it);
			}
			break;
		default:
			BHERROR("alloc block type not supported");
			break;
		}
	}
	else
	{
		OwnerMemoryMapSystem::iterator sys_it;
		sys_it = system_heap_map.find(owner);
		if(sys_it != system_heap_map.end())
		{
			sys_it->second->erase(MEM_HANDLE_MP_WITH_SIZE(mh,0));
			if(sys_it->second->size() == 0)
			{
				delete sys_it->second;
				system_heap_map.erase(owner);
			}
			free(mh);
		}
		else
		{
			rccontrol << lock << "dealloc tries to free block which is not traced" << unlock;
		}
	}
}


bool MemoryHandling::FindAndErase(vector<MEM_HANDLE_MP_WITH_SIZE>& vec, MEM_HANDLE_MP_WITH_SIZE element) {
	vector<MEM_HANDLE_MP_WITH_SIZE>::iterator toErase;
	toErase = std::find<vector<MEM_HANDLE_MP_WITH_SIZE>::iterator, MEM_HANDLE_MP_WITH_SIZE>(vec.begin(), vec.end(), element);
	if (toErase == vec.end())
		return false;
	vec.erase(toErase);
	return true;
}


MEM_HANDLE_MP MemoryHandling::rc_realloc(MEM_HANDLE_MP mh, size_t size, TrackableObject* owner, BLOCK_TYPE type)
{
    IBGuard guard(ccs.counting_mutex);

#ifdef FUNCTIONS_EXECUTION_TIMES
	FETOperator feto("MemoryHandling::re_alloc(...)");
#endif
	BHASSERT(size != 0, "should be 'size != 0'");
	if (mh == NULL) //treat as alloc
	{
		return alloc(size, type, owner);
	}

	BHASSERT(type != BLOCK_FIXED, "should be 'type != BLOCK_FIXED'");
	AllocBlckMap::iterator it = alloc_blck_type.find(mh);

	OwnerMemoryMap::iterator omm_it;
	stdext::hash_set<MEM_HANDLE_MP>::iterator hs_it;

	if(it != alloc_blck_type.end())
	{
		MEM_HANDLE_MP new_mh;

		switch( it->second )
		{
		case BLOCK_COMPRESSED:
			new_mh = compressed_heap.rc_realloc(mh, size);
			if (compressed_heap.getHeapStatus() == HEAP_OUT_OF_MEMORY)
			{
				bool released = ReleaseMemory(comp_heap_release < size ? size : comp_heap_release, RH_COMPRESSED, owner);
				if(!released && !ReleaseMemory(size, RH_COMPRESSED, owner))
				{
					//Try to allocate block using system heap, copy it and deallocate old one
					MEM_HANDLE_MP new_addr = (MEM_HANDLE_MP)malloc(size);
					if(new_addr)
					{
						size_t old_size = compressed_heap.getBlockSize(mh);
						memcpy(new_addr, mh, old_size);
						//system_heap_alloc.push_back(new_addr);

						OwnerMemoryMapSystem::iterator sys_it;
						sys_it = system_heap_map.find(owner);
						if(sys_it != system_heap_map.end())
							sys_it->second->insert(MEM_HANDLE_MP_WITH_SIZE(new_addr, size));
						else
						{
							pair<OwnerMemoryMapSystem::iterator, bool> res = system_heap_map.insert(make_pair(owner, new hash_set_for_system_mem_heaps));
							res.first->second->insert(MEM_HANDLE_MP_WITH_SIZE(new_addr, size));
						}

						dealloc(mh, owner);
						return new_addr;
					}
					rccontrol << lock << "Returned NULL in rc_realloc (system heap): " << __FILE__ << " " << __LINE__ << unlock;
					BHERROR("Returned NULL in rc_realloc (system heap)");
					return NULL; //Unable to release memory and realloc the memory block
				}
				else
					new_mh = compressed_heap.rc_realloc(mh, size);
			}
			if(new_mh)
			{
				alloc_blck_type.erase(it);
				alloc_blck_type.insert(alloc_blck_pair(new_mh, BLOCK_COMPRESSED));
				omm_it = cmp_blck_mbrshp.find(owner);
				if(new_mh != mh) //delete old address and insert new
				{
					hs_it = omm_it->second->find(mh);
					omm_it->second->erase(hs_it);
					omm_it->second->insert(new_mh);
				}
			}
			break;
		case BLOCK_TEMPORARY:
			new_mh = uncompressed_heap.rc_realloc(mh, size);
			if (uncompressed_heap.getHeapStatus() == HEAP_OUT_OF_MEMORY)
			{
				bool released = ReleaseMemory(uncomp_heap_release < size ? size : uncomp_heap_release, RH_UNCOMPRESSED, owner);
				if(!released && !ReleaseMemory(size, RH_UNCOMPRESSED, owner))
				{
					//Try to allocate block using system heap, copy it and deallocate old one
					MEM_HANDLE_MP new_addr = (MEM_HANDLE_MP)malloc(size);
					if(new_addr)
					{
						size_t old_size = uncompressed_heap.getBlockSize(mh);
						memcpy(new_addr, mh, old_size);
						//system_heap_alloc.push_back(new_addr);
						OwnerMemoryMapSystem::iterator sys_it;
						sys_it = system_heap_map.find(owner);
						if(sys_it != system_heap_map.end())
							sys_it->second->insert(MEM_HANDLE_MP_WITH_SIZE(new_addr, size));
						else
						{
							pair<OwnerMemoryMapSystem::iterator, bool> res = system_heap_map.insert(make_pair(owner, new hash_set_for_system_mem_heaps));
							res.first->second->insert(MEM_HANDLE_MP_WITH_SIZE(new_addr, size));
						}

						dealloc(mh, owner);
						return new_addr;
					}
					rccontrol << lock << "Returned NULL in rc_realloc (system heap): " << __FILE__ << " " << __LINE__ << unlock;
					BHERROR("Returned NULL in rc_realloc (system heap)");
					return NULL; //Unable to release memory and realloc the memory block
				}
				else
					new_mh = uncompressed_heap.rc_realloc(mh, size);
			}
			if (new_mh)
			{
				alloc_blck_type.erase(it);
				alloc_blck_type.insert(alloc_blck_pair(new_mh, BLOCK_TEMPORARY));
			}
			break;
		case BLOCK_UNCOMPRESSED:
			new_mh = uncompressed_heap.rc_realloc(mh, size);
			if (uncompressed_heap.getHeapStatus() == HEAP_OUT_OF_MEMORY)
			{
				bool released = ReleaseMemory(uncomp_heap_release < size ? size : uncomp_heap_release, RH_UNCOMPRESSED, owner);
				if(!released && !ReleaseMemory(size, RH_UNCOMPRESSED, owner))
				{
					//Try to allocate block using system heap, copy it and deallocate old one
					MEM_HANDLE_MP new_addr = (MEM_HANDLE_MP)malloc(size);
					if(new_addr)
					{
						size_t old_size = uncompressed_heap.getBlockSize(mh);
						memcpy(new_addr, mh, old_size);
						//system_heap_alloc.push_back(new_addr);
						OwnerMemoryMapSystem::iterator sys_it;
						sys_it = system_heap_map.find(owner);
						if(sys_it != system_heap_map.end())
							sys_it->second->insert(MEM_HANDLE_MP_WITH_SIZE(new_addr, size));
						else
						{
							pair<OwnerMemoryMapSystem::iterator, bool> res = system_heap_map.insert(make_pair(owner, new hash_set_for_system_mem_heaps));
							res.first->second->insert(MEM_HANDLE_MP_WITH_SIZE(new_addr, size));
						}

						dealloc(mh, owner);
						return new_addr;
					}

					rccontrol << lock << "Returned NULL in rc_realloc (system heap): " << __FILE__ << " " << __LINE__ << unlock;
					BHERROR("Returned NULL in rc_realloc (system heap)");
					return NULL; //Unable to release memory and realloc the memory block
				}
				else
				{
					new_mh = uncompressed_heap.rc_realloc(mh, size);
					BHASSERT(uncompressed_heap.getHeapStatus() == HEAP_SUCCESS && new_mh,
							"should be 'uncompressed_heap.getHeapStatus() == HEAP_SUCCESS && new_mh'");
				}
			}
			if(new_mh)
			{
				alloc_blck_type.erase(it);
				alloc_blck_type.insert(alloc_blck_pair(new_mh, BLOCK_UNCOMPRESSED));
				omm_it = uncmp_blck_mbrshp.find(owner);
				if(new_mh != mh) //delete old address and insert new
				{
					hs_it = omm_it->second->find(mh);
					omm_it->second->erase(hs_it);
					omm_it->second->insert(new_mh);
				}
			}
			break;
		default:
			rccontrol << lock << "Returned NULL in rc_realloc: " << __FILE__ << " " << __LINE__ << unlock;
			BHERROR("Returned NULL in rc_realloc");
			return NULL;
		}
		if(new_mh == NULL) {
			rccontrol << lock << "Returned NULL in rc_realloc: " << __FILE__ << " " << __LINE__ << unlock;
			BHERROR("Returned NULL in rc_realloc");
		}
		return new_mh;
	}
	else
	{
		OwnerMemoryMapSystem::iterator sys_it;
		sys_it = system_heap_map.find(owner);
		if(sys_it == system_heap_map.end())
		{
			rccontrol << lock << "Rc_realloc: address not traced" << __FILE__ << " " << __LINE__ << unlock;
			BHERROR("Rc_realloc: address not traced");
			return NULL;
		}

		hash_set_for_system_mem_heaps::iterator it = sys_it->second->find(MEM_HANDLE_MP_WITH_SIZE(mh, 0));
		if(it != sys_it->second->end())
		{
			//Try to go allocate on BH heaps one again
			MEM_HANDLE_MP new_mh = 0;
			switch(type)
			{
			case BLOCK_COMPRESSED:
			{
				new_mh = compressed_heap.alloc(size, type, owner);
				if (compressed_heap.getHeapStatus() == HEAP_OUT_OF_MEMORY)
				{
					bool released = ReleaseMemory(comp_heap_release < size ? size : comp_heap_release, RH_COMPRESSED, owner);
					if(!released && !ReleaseMemory(size, RH_COMPRESSED, owner))
					{
						break;
					}
					else
						new_mh = compressed_heap.alloc(size, type, owner);
				}
				if(new_mh)
				{
					alloc_blck_type.insert(alloc_blck_pair(new_mh, BLOCK_COMPRESSED));
					omm_it = cmp_blck_mbrshp.find(owner);
					if(omm_it == cmp_blck_mbrshp.end())
					{
						pair<OwnerMemoryMap::iterator, bool> res = cmp_blck_mbrshp.insert(make_pair(owner, new stdext::hash_set<MEM_HANDLE_MP> ));
						res.first->second->insert(new_mh);
					}
					else
						omm_it->second->insert(new_mh);
				}
				memcpy(new_mh, it->GetPointer(), it->GetSize());
				free(it->GetPointer());
				FindAndErase(system_heap_alloc, MEM_HANDLE_MP_WITH_SIZE(it->GetPointer(), 0));
				sys_it->second->erase(MEM_HANDLE_MP_WITH_SIZE(it->GetPointer(), 0));
				if(sys_it->second->size() == 0)
				{
					delete sys_it->second;
					system_heap_map.erase(sys_it);
				}

				return new_mh;
			}
			case BLOCK_TEMPORARY:
				new_mh = uncompressed_heap.alloc(size, type, owner);
				if (uncompressed_heap.getHeapStatus() == HEAP_OUT_OF_MEMORY)
				{
					bool released = ReleaseMemory(uncomp_heap_release < size ? size : uncomp_heap_release, RH_UNCOMPRESSED, owner);
					if(!released && !ReleaseMemory(size, RH_UNCOMPRESSED, owner))
					{
						break;
					}
					else
						new_mh = uncompressed_heap.alloc(size, type, owner);
				}
				if(new_mh)
					alloc_blck_type.insert(alloc_blck_pair(new_mh, BLOCK_TEMPORARY));
				memcpy(new_mh, it->GetPointer(), it->GetSize());
				free(it->GetPointer());
				FindAndErase(system_heap_alloc, MEM_HANDLE_MP_WITH_SIZE(it->GetPointer(), 0));
				sys_it->second->erase(MEM_HANDLE_MP_WITH_SIZE(it->GetPointer(), 0));
				if(sys_it->second->size() == 0)
				{
					delete sys_it->second;
					system_heap_map.erase(sys_it);
				}

				return new_mh;
			case BLOCK_UNCOMPRESSED:
				new_mh = uncompressed_heap.alloc(size, type, owner);
				if (uncompressed_heap.getHeapStatus() == HEAP_OUT_OF_MEMORY)
				{
					bool released = ReleaseMemory(uncomp_heap_release < size ? size : uncomp_heap_release, RH_UNCOMPRESSED, owner);
					if(!released && !ReleaseMemory(size, RH_UNCOMPRESSED, owner))
					{
						break;
					}
					else
						new_mh = uncompressed_heap.alloc(size, type, owner);
				}
				if(new_mh)
				{
					alloc_blck_type.insert(alloc_blck_pair(new_mh, BLOCK_UNCOMPRESSED));
					omm_it = uncmp_blck_mbrshp.find(owner);
					if(omm_it == uncmp_blck_mbrshp.end())
					{
						pair<OwnerMemoryMap::iterator, bool> res = uncmp_blck_mbrshp.insert(make_pair(owner, new stdext::hash_set<MEM_HANDLE_MP> ));
						res.first->second->insert(new_mh);
					}
					else
						omm_it->second->insert(new_mh);
				}
				memcpy(new_mh, it->GetPointer(), it->GetSize());
				free(it->GetPointer());
				FindAndErase(system_heap_alloc, MEM_HANDLE_MP_WITH_SIZE(it->GetPointer(), 0));
				sys_it->second->erase(MEM_HANDLE_MP_WITH_SIZE(it->GetPointer(), 0));
				if(sys_it->second->size() == 0)
				{
					delete sys_it->second;
					system_heap_map.erase(sys_it);
				}
				return new_mh;
			default:
				//No space available - realloc using system heap
				break;
			}

			new_mh = (MEM_HANDLE_MP)realloc(it->GetPointer(), size);
			sys_it->second->erase(it);
			FindAndErase(system_heap_alloc, MEM_HANDLE_MP_WITH_SIZE(mh, 0));
			if(new_mh)
			{
				sys_it->second->insert(MEM_HANDLE_MP_WITH_SIZE(new_mh, size));
//				system_heap_alloc.push_back(new_mh);
				return new_mh;
			}
			else
			{
				rccontrol << lock << "Returned NULL in rc_realloc (system heap): " << __FILE__ << " " << __LINE__ << unlock;
				BHERROR("Returned NULL in rc_realloc (system heap)");
				return NULL;
			}
		}
	}
	rccontrol << lock << "Returned NULL in rc_realloc: " << __FILE__ << " " << __LINE__ << unlock;
	BHERROR("Returned NULL in rc_realloc");
	return NULL;
}

bool MemoryHandling::ReleaseMemory(size_t size, RELEASE_HEAP rh, TrackableObject* untouchable)
{
	/*
		Memory release algorithm:

			Allocate large buffers on one heap and smaller data (compressed) on second one (during allocation each buffer will be marked with its storage type)
			If there is no space for uncompressed data on the first heap
			o	Locate memory block that could be compressed and placed on second heap
			o	Invoke object's StayCompressed function (if compressed data is up to date, then no compression will be done and uncompressed buffer will be released)
			o	If compression is not up to date, compress the uncompressed data and place the buffer on the second heap (if possible), otherwise delete the object.
			o	If the large buffer is marked as collapsible, collapse the object and place the remaining data elsewhere
			If there is no space for compressed buffer
			o	Locate the best memory block on the second heap and delete the owner object (or if possible collapse it).
			Buffers of objects that are currently locked are not being investigated by memory manager while releasing the memory
	*/

	/*
		Naive way:
		Locate best memory region:
		 - free chunks
		 - owners that are up to date
		 - to be compressed buffers (count all owner's buffers, not only those in the region!)

		 minimize evaluation function - A*(free_chunks+up_to_date) + B*to_be_compressed_buf
	*/

	/*
		Start new region when next() is:
			- temporary
			- owner is Locked
			- owner == untouchable
	*/


#ifdef SHOW_MM_RELEASE
	printf("*");
#endif

#ifdef DO_MEMORY_TEST
	//Test cycle
	MemoryBlock<MEM_HANDLE_MP>* blck_checker = uncomp_heap_mem;
	while (blck_checker->getNext())
	{
		if (blck_checker->getType() == BLOCK_FREE && blck_checker->getNext()->getType() == BLOCK_FREE)
			rccontrol << lock << "Error: FREE blocks are not merged!" << unlock;
		blck_checker = blck_checker->getNext();
	}
#endif

	MemoryRegion* mem_reg = new MemoryRegion();
	memset(mem_reg, 0, sizeof(MemoryRegion));

	MemoryBlock<MEM_HANDLE_MP>* last_end = NULL;

	MemoryRegion* best_region = NULL;
	mem_reg->size = 0;

	unsigned int eval_func = 0;
	unsigned int best_eval_func = UINT_MAX;

	OwnerMemoryMap::iterator om_it;

	map<TrackableObject*, unsigned short> obj_counter;
	map<TrackableObject*, unsigned short>::iterator oc_it;
	//int cnt = 0;

	bool found = false;

	if(rh == RH_UNCOMPRESSED)
	{
		mem_reg->start = this->uncomp_heap_mem;
		mem_reg->end = this->uncomp_heap_mem;

		do
		{
			while(mem_reg->end)
			{
				found = true;
				if(mem_reg->end->getOwner() &&  //Owner == NULL when type == FREE
					(mem_reg->end->getOwner()->IsLocked() ||
					mem_reg->end->getOwner() == untouchable ||
					mem_reg->end->getType() == BLOCK_TEMPORARY ||
					mem_reg->end->getOwner()->IsBlocked()) )
				{
					//Reset whole memory block
					//Traverse to first not locked, touchable, not temporary memory block
					if(mem_reg->end->getNext())
					{
						while( mem_reg->end &&
							mem_reg->end->getType() != BLOCK_FREE &&
							(mem_reg->end->getOwner()->IsLocked() ||
							mem_reg->end->getOwner() == untouchable ||
							mem_reg->end->getType() == BLOCK_TEMPORARY ||
							mem_reg->end->getOwner()->IsBlocked()) )
						{
							mem_reg->start = mem_reg->end->getNext();
							mem_reg->end = mem_reg->start;
						}

						if(!mem_reg->end)
						{
							found = false;
							break;
						}
					}
					else
					{
						found = false;
						break;
					}

					last_end = NULL;
					mem_reg->size = 0;

					mem_reg->heapunits.uncmp_units.free_units = 0;
					mem_reg->heapunits.uncmp_units.uncompressed_units = 0;
					mem_reg->heapunits.uncmp_units.uncompressed_up_to_date_units = 0;

					obj_counter.clear();
				}
				if(!found)
					break;

				if(last_end != mem_reg->end)
				{
					mem_reg->size += mem_reg->end->getSize();

					if(mem_reg->end->getOwner()) //FREE blocks are not inserted
					{
						oc_it = obj_counter.find(mem_reg->end->getOwner());
						if(oc_it == obj_counter.end())
							obj_counter.insert(pair<TrackableObject*, unsigned short>(mem_reg->end->getOwner(), 1));
						else
							oc_it->second++;
					}

					switch(mem_reg->end->getType())
					{
					case BLOCK_FREE:
						mem_reg->heapunits.uncmp_units.free_units++;
						break;
					case BLOCK_UNCOMPRESSED:
						if(oc_it == obj_counter.end())
						{
							om_it = uncmp_blck_mbrshp.find(mem_reg->end->getOwner());
							if(om_it != uncmp_blck_mbrshp.end())
							{
								if(mem_reg->end->getOwner()->UpToDate())
									mem_reg->heapunits.uncmp_units.uncompressed_up_to_date_units += (unsigned short)om_it->second->size();
								else
									mem_reg->heapunits.uncmp_units.uncompressed_units += (unsigned short)om_it->second->size();
							}
							else
							{
								return false; //Corruption
							}
						}
						break;
					case BLOCK_TEMPORARY:
						//mem_reg->uncmp_units.temporary_units++;
						//break;
					default:
						return false;
					}
				}
				if(mem_reg->size < size)
					mem_reg->end = mem_reg->end->getNext();
				else
					break;
			}

			if(mem_reg->end && found)
			{
				//Calculate evaluation functions
				/*
					Evaluation function:
					- award for having free units (subtract)
					- penalize for having compressed (not up to date)
				*/

				eval_func = uncmp_eval(mem_reg->heapunits.uncmp_units.uncompressed_units, mem_reg->heapunits.uncmp_units.uncompressed_up_to_date_units, mem_reg->heapunits.uncmp_units.free_units);
				if(eval_func < best_eval_func)
				{
					best_eval_func = eval_func;
					if(!best_region)
						best_region = new MemoryRegion();
					best_region->start = mem_reg->start;
					best_region->end = mem_reg->end;
					best_region->size = mem_reg->size;
				}

				//Erase first block
				//If it removes TrackableObject* permanently, remove its compressed units

				mem_reg->size -= mem_reg->start->getSize();

				if(mem_reg->start->getOwner())
				{
					oc_it = obj_counter.find(mem_reg->start->getOwner());
					if(oc_it != obj_counter.end())
					{
						oc_it->second--;
						if(oc_it->second == 0)
						{
							//Remove units that belong to this TrackableObject
							if(mem_reg->start->getType() == BLOCK_UNCOMPRESSED)
							{
								om_it = uncmp_blck_mbrshp.find(mem_reg->start->getOwner());

								if(om_it != uncmp_blck_mbrshp.end())
								{
									if(mem_reg->start->getOwner()->UpToDate())
									{
										mem_reg->heapunits.uncmp_units.uncompressed_up_to_date_units -= (unsigned short)om_it->second->size();
									}
									else
									{
										mem_reg->heapunits.uncmp_units.uncompressed_units -= (unsigned short)om_it->second->size();
									}
								}

							}
							obj_counter.erase(oc_it);
						}
					}
					else
					{
						BHERROR("'oc_it' equals obj_counter.end()");
						return false;
					}
				}
				else if(mem_reg->start->getType() == BLOCK_FREE)
				{
					mem_reg->heapunits.uncmp_units.free_units--;
				}

				//If start is equal to end -> set start and end as next block
				//last_end = NULL to prevent omitting the insertion of owner pointer into the map
				if(mem_reg->start == mem_reg->end)
				{
					mem_reg->start = mem_reg->end = mem_reg->start->getNext();
					last_end = NULL;
					//mem_reg->size = 0;
					BHASSERT(obj_counter.size() == 0 && mem_reg->size == 0,
							"should be 'obj_counter.size() == 0 && mem_reg->size == 0'");
				}
				else
				{
					mem_reg->start = mem_reg->start->getNext();
					last_end = mem_reg->end;
				}
			}
			else
				break; //Heap end was reached and required memory was not collected
		}while(mem_reg->start->getNext());
		/*
			We assume there is not enough memory so only last block
			cannot have the space we require, thus it is not checked
			separately
		*/

		delete mem_reg;
		if(best_region)
		{
			/*
				invoke stayCompressed on all TrackableObjects owned by the region
				if method return false, delete the object
				Blocking objects that are being released prevents from releasing them in case
				when compressed heap is full and must be investigated
			*/

			//rccontrol << lock << "Releasing: " << best_region->size << " bytes" << unlock;

			stdext::hash_set<TrackableObject*> invoked_obj;
			stdext::hash_set<TrackableObject*>::iterator io_it;

			MemoryBlock<MEM_HANDLE_MP>* iterator = best_region->start;

			//while(best_region->start != best_region->end->getNext())
			while(iterator != best_region->end->getNext())
			{
				//if(best_region->start->getOwner())
				if(iterator->getOwner())
				{
					//best_region->start->getOwner()->SetBlocked(true);
					//invoked_obj.insert(best_region->start->getOwner());
					iterator->getOwner()->SetBlocked(true);
					invoked_obj.insert(iterator->getOwner());
				}
#ifdef _DEBUG
				else
					BHASSERT(iterator->getType() == BLOCK_FREE, "should be 'iterator->getType() == BLOCK_FREE'");
					//The only case here with owner == NULL
					//BHASSERT(best_region->start->getType() == BLOCK_FREE,
					//		"should be 'best_region->start->getType() == BLOCK_FREE'");
#endif

				//best_region->start = best_region->start->getNext();
				iterator = iterator->getNext();
			}
			for(io_it = invoked_obj.begin(); io_it != invoked_obj.end(); io_it++)
			{
				switch((*io_it)->TrackableType())
				{
				case TO_PACK:
					(*io_it)->SetBlocked(true);
					(*io_it)->StayCompressed();
					(*io_it)->SetBlocked(false);

#ifdef MEMORY_LEAK_CHECK
					UncmpMemoryLeaks(*io_it);
#endif
					break;
				case TO_FILTER:
					//printf("f");
					(*io_it)->SetBlocked(true);
					(*io_it)->StayCompressed();
					(*io_it)->SetBlocked(false);

#ifdef MEMORY_LEAK_CHECK
					UncmpMemoryLeaks(*io_it);
#endif
					break;
				case TO_INDEXTABLE:
					//printf("i");
					(*io_it)->SetBlocked(true);
					(*io_it)->Collapse(); // In IndexTable & Sorter Compress == Collapse
					(*io_it)->SetBlocked(false);

#ifdef MEMORY_LEAK_CHECK
					UncmpMemoryLeaks(*io_it);
#endif
					break;
				case TO_SORTER:
					//printf("s");
					(*io_it)->SetBlocked(true);
					(*io_it)->Collapse(); // In IndexTable & Sorter Compress == Collapse
					(*io_it)->SetBlocked(false);

#ifdef MEMORY_LEAK_CHECK
					UncmpMemoryLeaks(*io_it);
#endif
					break;
				//case TO_MULTIFILTER2:
				case TO_RSINDEX:
					BHERROR("RSIndex object cannot be released");
					return false;
				case TO_DEPENDENT:
					//Logical error during integration
					BHERROR("Dependent TrackableObject's memory block was wrongly assigned");
					return false;
				case TO_TEMPORARY:
					//Logical error in algorithm's selection process
					BHERROR("Temporary object cannot be released");
					return false;
				case TO_FTREE:
					//Logical error in algorithm's selection process
					BHERROR("FTree object is not compressible whether collapsible");
					return false;
				case TO_CACHEDBUFFER:
					//Logical error in algorithm's selection process
					BHERROR("CachedBuffer object cannot be released");
					return false;
				default:
					return false;
				}
			}
			//printf("%d bytes freed\n", best_region->size);
			//printf("%d bytes maximal block in tree\n", this->uncompressed_heap.getMaximalBlock());
			//if(best_region->size > this->uncompressed_heap.getMaximalBlock())
			//	rccontrol << lock << "Less memory released than expected" << unlock;

			if (best_region)
				delete best_region;

			//Destroy all packs that have any buffers on system heap
			//Assumption: unlocked and unblocked packs are ready to be destroyed (no temporary buffers)
			stdext::hash_set<TrackableObject*> to_hs;
			for(OwnerMemoryMapSystem::iterator sys_it = system_heap_map.begin(); sys_it != system_heap_map.end(); sys_it++)
				to_hs.insert(sys_it->first);

			stdext::hash_set<TrackableObject*>::iterator to_hs_it ;
			for(stdext::hash_set<TrackableObject*>::iterator to_hs_it = to_hs.begin(); to_hs_it != to_hs.end(); to_hs_it++)
			{
				if ((*to_hs_it)->TrackableType() == TO_PACK)
				{
				AttrPack* pack_ptr = (AttrPack*)(*to_hs_it);
				if (!pack_ptr->IsBlocked() && !pack_ptr->IsLocked() && pack_ptr != untouchable)
					//pack_ptr->Destroy();
					//delete pack_ptr;
					SynchronizedDataCache::GlobalDataCache().ReleasePack(pack_ptr);
				}
			}

			return true;
		}
		else
		{
			return false; // No continuous memory region was found!!!
		}
	}
	else if (rh == RH_COMPRESSED)
	{
		mem_reg->start = this->comp_heap_mem;
		mem_reg->end = this->comp_heap_mem;

		do
		{
			while (mem_reg->end)
			{
				found = true;
				if(mem_reg->end->getOwner() &&  //Owner == NULL when type == FREE
					(mem_reg->end->getOwner()->IsLocked() ||
					mem_reg->end->getOwner() == untouchable ||
					mem_reg->end->getType() == BLOCK_TEMPORARY ||
					mem_reg->end->getOwner()->IsBlocked()) )
				{
					//Reset whole memory block
					//Traverse to first not locked, touchable, not temporary memory block
					if(mem_reg->end->getNext())
					{
						while( mem_reg->end &&
							mem_reg->end->getType() != BLOCK_FREE &&
							(mem_reg->end->getOwner()->IsLocked() ||
							mem_reg->end->getOwner() == untouchable ||
							mem_reg->end->getType() == BLOCK_TEMPORARY ||
							mem_reg->end->getOwner()->IsBlocked()))
						{
							mem_reg->start = mem_reg->end->getNext();
							mem_reg->end = mem_reg->start;
						}

						if(!mem_reg->end)
						{
							found = false;
							break;
						}
					}
					else
					{
						found = false;
						break;
					}

					last_end = NULL;
					mem_reg->size = 0;

					mem_reg->heapunits.cmp_units.free_units = 0;
					mem_reg->heapunits.cmp_units.compressed_units = 0;

					obj_counter.clear();
				}
				if(!found)
					break;

				if(last_end != mem_reg->end)
				{
					mem_reg->size += mem_reg->end->getSize();

					if( mem_reg->end->getOwner()) //FREE blocks are not inserted
					{
						oc_it = obj_counter.find(mem_reg->end->getOwner());
						if(oc_it == obj_counter.end())
							obj_counter.insert(pair<TrackableObject*, unsigned short>(mem_reg->end->getOwner(), 1));
						else
							oc_it->second++;
					}

					switch(mem_reg->end->getType())
					{
					case BLOCK_FREE:
						mem_reg->heapunits.cmp_units.free_units++;
						break;
					case BLOCK_COMPRESSED:
						if(oc_it == obj_counter.end())
						{
							om_it = cmp_blck_mbrshp.find(mem_reg->end->getOwner());
							if(om_it != cmp_blck_mbrshp.end())
							{
								mem_reg->heapunits.cmp_units.compressed_units += (unsigned short)om_it->second->size();
							}
							else
							{
								BHERROR("'om_it' equals obj_counter.end()");
								return false; //Corruption
							}
						}
						break;
					default:
						return false;
					}
				}
				if(mem_reg->size < size)
					mem_reg->end = mem_reg->end->getNext();
				else
					break;
			}
			if(mem_reg->end && found)
			{
				//Calculate evaluation functions
				eval_func = cmp_eval(mem_reg->heapunits.cmp_units.compressed_units, mem_reg->heapunits.cmp_units.free_units);
				if(eval_func < best_eval_func)
				{
					best_eval_func = eval_func;
					if(!best_region)
						best_region = new MemoryRegion();
					best_region->start = mem_reg->start;
					best_region->end = mem_reg->end;
					best_region->size = mem_reg->size;
				}

				//Erase first block
				//If it removes TrackableObject* permanently, remove its compressed units

				mem_reg->size -= mem_reg->start->getSize();

				if(mem_reg->start->getOwner())
				{
					oc_it = obj_counter.find(mem_reg->start->getOwner());
					if(oc_it != obj_counter.end())
					{
						oc_it->second--;
						if(oc_it->second == 0)
						{
							//Remove units that belong to this TrackableObject
							if(mem_reg->start->getType() == BLOCK_COMPRESSED)
							{
								om_it = cmp_blck_mbrshp.find(mem_reg->start->getOwner());

								if(om_it != cmp_blck_mbrshp.end())
								{
									mem_reg->heapunits.cmp_units.compressed_units -= (unsigned short)om_it->second->size();
								}
							}
							obj_counter.erase(oc_it);
						}
					}
					else
					{
						BHERROR("'oc_it' equals obj_counter.end()");
						return false;
					}
				}
				else if (mem_reg->start->getType() != BLOCK_FREE)
				{
					mem_reg->heapunits.cmp_units.free_units--;
				}

				//If start is equal to end -> set start and end as next block
				//last_end = NULL to prevent omitting the insertion of owner pointer into the map
				if(mem_reg->start == mem_reg->end)
				{
					mem_reg->start = mem_reg->end = mem_reg->start->getNext();
					last_end = NULL;
				}
				else
				{
					mem_reg->start = mem_reg->start->getNext();
					last_end = mem_reg->end;
				}
			}
			else
				break; // Heap end was reached and the required size was not collected
		} while(mem_reg->start->getNext());

		delete mem_reg;
		if(best_region)
		{
			/*
				Invoke Collapse on all collapsable objects and delete data packs
			*/

			stdext::hash_set<TrackableObject*> invoked_obj;
			stdext::hash_set<TrackableObject*>::iterator io_it;

			while(best_region->start != best_region->end->getNext())
			{
				if (best_region->start->getOwner())
				{
					best_region->start->getOwner()->SetBlocked(true);
					invoked_obj.insert(best_region->start->getOwner());
				}
#ifdef _DEBUG
				else
					BHASSERT(best_region->start->getType() == BLOCK_FREE,
							"should be 'best_region->start->getType() == BLOCK_FREE'");
#endif
				best_region->start = best_region->start->getNext();
			}
			for(io_it = invoked_obj.begin(); io_it != invoked_obj.end(); io_it++)
			{
				switch((*io_it)->TrackableType())
				{
				case TO_PACK:

					(*io_it)->SetBlocked(true);
					//delete ((AttrPack*)(*io_it));
					SynchronizedDataCache::GlobalDataCache().ReleasePack((AttrPack*)(*io_it));

#ifdef MEMORY_LEAK_CHECK
					UncmpMemoryLeaks(*io_it);
					CmpMemoryLeaks(*io_it);
#endif
					break;
				case TO_FILTER:

					//printf("c_f");
					(*io_it)->SetBlocked(true);
					(*io_it)->Collapse();
					(*io_it)->SetBlocked(false);

#ifdef MEMORY_LEAK_CHECK
					UncmpMemoryLeaks(*io_it);
					CmpMemoryLeaks(*io_it);
#endif
					break;
				case TO_INDEXTABLE:

					//printf("c_i");
					(*io_it)->SetBlocked(true);
					(*io_it)->Collapse();
					(*io_it)->SetBlocked(false);

#ifdef MEMORY_LEAK_CHECK
					UncmpMemoryLeaks(*io_it);
					CmpMemoryLeaks(*io_it);
#endif
					break;
				case TO_SORTER:

					//printf("c_s");
					(*io_it)->SetBlocked(true);
					(*io_it)->Collapse();
					(*io_it)->SetBlocked(false);

#ifdef MEMORY_LEAK_CHECK
					UncmpMemoryLeaks(*io_it);
					CmpMemoryLeaks(*io_it);
#endif
					break;
				//case TO_MULTIFILTER2:
				case TO_RSINDEX:
					BHERROR("RSIndex object cannot be released");
					return false;
				case TO_DEPENDENT:
					//Logical error in code while integrating with memory manager
					BHERROR("Dependent TrackableObject's memory block was wrongly assigned");
					return false;
				case TO_TEMPORARY:
					//Logical error in algorithm's selection process
					BHERROR("Temporary object cannot be released");
					return false;
				case TO_FTREE:
					//Logical error in algorithm's selection process
					BHERROR("FTree object is not compressible whether collapsible");
					return false;
				case TO_CACHEDBUFFER:
					//Logical error in algorithm's selection process
					BHERROR("CachedBuffer object is cannot be released");
				default:
					return false;
				}
			}

			if(best_region)
				delete best_region;

			//Destroy all packs that have any buffers on system heap
			//Assumption: unlocked and unblocked packs are ready to be destroyed (no temporary buffers)
			stdext::hash_set<TrackableObject*> to_hs;
			for(OwnerMemoryMapSystem::iterator sys_it = system_heap_map.begin(); sys_it != system_heap_map.end(); sys_it++)
				to_hs.insert(sys_it->first);

			stdext::hash_set<TrackableObject*>::iterator to_hs_it;
			for(to_hs_it = to_hs.begin(); to_hs_it != to_hs.end(); to_hs_it++)
			{
				if ((*to_hs_it)->TrackableType() == TO_PACK)
				{
					AttrPack* pack_ptr = (AttrPack*)(*to_hs_it);
					if (!pack_ptr->IsBlocked() && !pack_ptr->IsLocked() && pack_ptr != untouchable)
						//delete pack_ptr;
						SynchronizedDataCache::GlobalDataCache().ReleasePack(pack_ptr);
				}
			}

			return true;
		}
		else
		{
			return false;
		}
	}
	return false;
}

void MemoryHandling::ReleaseUnlockedDataPacks()
{
	vector<TrackableObject*> dps;
	OwnerMemoryMap::iterator it = uncmp_blck_mbrshp.begin();
	while(it != uncmp_blck_mbrshp.end())
	{
		if((*it).first->TrackableType() == TO_PACK && (*it).first->IsLocked() == false && (*it).first->IsBlocked() == false )
		{
			dps.push_back((*it).first);
			//delete ((AttrPack*)(*it).first);
			//uncmp_blck_mbrshp.erase(it);
		}
		it++;
	}
	vector<TrackableObject*>::iterator iter = dps.begin();
	for(iter; iter != dps.end(); iter++)
		//delete (AttrPack*)(*iter);
		SynchronizedDataCache::GlobalDataCache().ReleasePack((AttrPack*)*iter);
}

void MemoryHandling::CleanPacks()
{
	MemoryBlock<MEM_HANDLE_MP>* blck = NULL;

	stdext::hash_set<MEM_HANDLE_MP> owner_set;
	//destroy packs
	for(blck = this->uncomp_heap_mem; blck != NULL;)
	{
		if (blck->getOwner() && blck->getOwner()->TrackableType() == TO_PACK)
			owner_set.insert(blck->getOwner());
		blck = blck->getNext();
	}
	stdext::hash_set<MEM_HANDLE_MP>::iterator it = owner_set.begin();
	for(; it != owner_set.end(); it++)
		//((AttrPack*)(*io_it))delete ((AttrPack*)*it);
		SynchronizedDataCache::GlobalDataCache().ReleasePack((AttrPack*)*it);

	owner_set.clear();


	for(blck = this->comp_heap_mem; blck != NULL;)
	{
		if (blck->getOwner() && blck->getOwner()->TrackableType() == TO_PACK)
			owner_set.insert(blck->getOwner());
		blck = blck->getNext();
	}
	for(stdext::hash_set<MEM_HANDLE_MP>::iterator it = owner_set.begin(); it != owner_set.end(); it++)
		//delete ((AttrPack*)*it);
		SynchronizedDataCache::GlobalDataCache().ReleasePack((AttrPack*)*it);

}

void MemoryHandling::CompressPacks()
{
	MemoryBlock<MEM_HANDLE_MP>* blck = NULL;
	stdext::hash_set<MEM_HANDLE_MP> owner_set;

	//destroy packs
	for(blck = this->uncomp_heap_mem; blck != NULL;)
	{
		if (blck->getOwner() && blck->getOwner()->TrackableType() == TO_PACK)
		{
			owner_set.insert(blck->getOwner());
		}
		blck = blck->getNext();
	}
	for(stdext::hash_set<MEM_HANDLE_MP>::iterator it = owner_set.begin(); it != owner_set.end(); it++)
		((AttrPack*)*it)->StayCompressed();
}

void MemoryHandling::UncmpMemoryLeaks(TrackableObject* obj)
{
	OwnerMemoryMap::iterator it = this->uncmp_blck_mbrshp.find(obj);
	if (it != uncmp_blck_mbrshp.end())
	{
		rccontrol << lock << "Memory is leaking in uncompressed part of heap" << unlock;
		BHERROR("Memory is leaking in uncompressed part of heap");
	}
}

void MemoryHandling::CmpMemoryLeaks(TrackableObject* obj)
{
	OwnerMemoryMap::iterator it = this->cmp_blck_mbrshp.find(obj);
	if (it != cmp_blck_mbrshp.end())
	{
		rccontrol << lock << "Memory is leaking in compressed part of heap" << unlock;
		BHERROR("Memory is leaking in compressed part of heap");
	}
}


//template class MemoryHandling<void*, BasicHeap<void*> >;

