/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef _TRACKABLE_OBJECT_DD9E054B_5B06_4cc8_B404_D3C1C1AC934B
#define _TRACKABLE_OBJECT_DD9E054B_5B06_4cc8_B404_D3C1C1AC934B

#include "common/bhassert.h"
#include "MemoryHandlingPolicy.h"

/*
	Classes that will use memory manager should inherit from this class
*/

#ifdef __GNUC__
enum TRACKABLEOBJECT_TYPE
{
	TO_PACK,
	TO_SORTER,
	TO_CACHEDBUFFER,
	TO_FILTER,
	TO_MULTIFILTER2,
	TO_INDEXTABLE,
	TO_RSINDEX,
	TO_TEMPORARY,
	TO_FTREE,
	TO_INITIALIZER,
	TO_DEPENDENT // object is dependent from other one
};
#else
enum TRACKABLEOBJECT_TYPE:char
{
	TO_PACK,
	TO_SORTER,
	TO_CACHEDBUFFER,
	TO_FILTER,
	TO_MULTIFILTER2,
	TO_INDEXTABLE,
	TO_RSINDEX,
	TO_TEMPORARY,
	TO_FTREE,
	TO_INITIALIZER,
	TO_DEPENDENT // object is dependent from other one
};
#endif

//ooh!, class should be named Traceable not Trackable! To be changed later
class TrackableObject
{
    friend class MemoryManagerGuard;

public:

	TrackableObject();
	TrackableObject(size_t, size_t, size_t, size_t, std::string = "");
	virtual ~TrackableObject();

	virtual void StayCompressed()				{BHERROR("method should not be invoked");}
	virtual TRACKABLEOBJECT_TYPE TrackableType() = 0;
	virtual bool UpToDate()						{return false;}
	virtual int Collapse()						{BHERROR("method should not be invoked"); return 0;}

	virtual void Uncompress()					{BHERROR("method should not be invoked");} //Used in Locking mechanism
	virtual bool IsCompressed()				{BHERROR("method should not be invoked"); return 0;} //Used in Locking mechanism

	//void Lock(unsigned short &lock_count);

	//void Unlock(unsigned short &lock_count);

	void Lock();
	void Unlock();

	void CopyLockStatus(TrackableObject &sec)
	{
		m_Locked = sec.m_Locked;
		m_Blocked = sec.m_Blocked;
		m_lock_count = sec.m_lock_count;
	}

//	int MemorySettingsScale()							// Memory scale factors: 0 - <0.5GB, 1 - <1.2GB, 2 - <2.5GB, ..., 99 - higher than all these
//	{
//		return Instance()->MemorySettingsScale();
//	}

	static int MemorySettingsScale()					// Memory scale factors: 0 - <0.5GB, 1 - <1.2GB, 2 - <2.5GB, ..., 99 - higher than all these
	{
		return Instance()->MemorySettingsScale();
	}

	//Lock to be used during object construction
	void ConstructionLock();
	void ConstructionUnlock();

	//Lock the object during destruction to prevent garbage collection
	void DestructionLock();

	bool IsLocked() const	{ return m_Locked; }

	/*
		Blocked is not equal to Locked!
		Blocking is used on a memory management level (its more internal locking)
		Locking is used by RCAttr etc. for packs manipulation purposes
	*/
	bool IsBlocked() const	{ return m_Blocked; }
	short NoLocks() const	{ return m_lock_count; }

	void SetBlocked(bool blocked)
	{
		m_Blocked = blocked;
	}

	void CleanPacks()
	{
		ccs->EnterCountingCriticalSection();
		Instance()->CleanPacks();
		ccs->LeaveCountingCriticalSection();
	}
	void CompressPacks()
	{
		ccs->EnterCountingCriticalSection();
		Instance()->CompressPacks();
		ccs->LeaveCountingCriticalSection();
	}

protected:

	void* alloc(size_t, BLOCK_TYPE);
	void dealloc(void* ptr);
	void* rc_realloc(void* ptr, size_t size, BLOCK_TYPE type = BLOCK_FIXED);
	size_t rc_msize(void* ptr);

	//memory manipulation functions \w TrackableObject parent mechanism
	//void* alloc(size_t, BLOCK_TYPE, TrackableObject* parent_owner);
	//void* rc_realloc(void* ptr, size_t size, TrackableObject*parent_owner, BLOCK_TYPE type = BLOCK_FIXED);
	//void dealloc(void* ptr, TrackableObject* parent_owner);

	void deinitialize(bool detect_leaks)
	{
		if(TrackableType() != TO_INITIALIZER)
		{
			BHERROR("TrackableType() not equals 'TO_INITIALIZER'");
			return;
		}

		if (m_MemHandling) //Used only in MemoryManagerInitializer!
		{
			if (detect_leaks) m_MemHandling->ReportLeaks();
			delete m_MemHandling;
		}
	}

	void LockMemoryManager(){ ccs->EnterCountingCriticalSection();/*EnterCriticalSection(cs);*/}
	void UnlockMemoryManager(){ccs->LeaveCountingCriticalSection();/*LeaveCriticalSection(cs);*/}

protected:

	//static MemoryHandling<void*, BasicHeap<void*> >* m_MemHandling;
	static MemoryHandling* m_MemHandling;

	static MemoryHandling* Instance()
	{
		if(!m_MemHandling)
		{
			//m_MemHandling = new MemoryHandling<void*, BasicHeap<void*> >(COMP_SIZE, UNCOMP_SIZE, COMPRESSED_HEAP_RELEASE, UNCOMPRESSED_HEAP_RELEASE);
			BHERROR("Memory manager is not instantiated");
			return NULL;
		}
		else
			return m_MemHandling;
	}
	static MemoryHandling* Instance(size_t comp_size, size_t uncomp_size, size_t cmp_rls, size_t uncmp_rls, std::string hugedir = "")
	{
		if(!m_MemHandling)
			m_MemHandling = new MemoryHandling(comp_size, uncomp_size, cmp_rls, uncmp_rls, hugedir);
		return m_MemHandling;
	}

	bool m_Locked;
	bool m_Blocked;	//Used to block memory blocks when due to compression process compressed_heap must be released
	short m_lock_count;

	CountingCriticalSection* ccs;
	//pthread_mutex_t m_mutex;
};

class MemoryManagerGuard
{
public:
    MemoryManagerGuard( TrackableObject& to) : css(*to.ccs) { css.EnterCountingCriticalSection(); }
    MemoryManagerGuard() : css(*TrackableObject::m_MemHandling->GetCriticalSection()) { css.EnterCountingCriticalSection(); }
    ~MemoryManagerGuard() { css.LeaveCountingCriticalSection(); }

private:
    CountingCriticalSection& css;

};

#define MEMORY_LOCK_GUARD  					MemoryManagerGuard memoryManagerGuard(*this);
#define MEMORY_LOCK_GUARD_USE_ANYWHERE	MemoryManagerGuard memoryManagerGuard;


#endif

