/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef _MUTEX_H_
#define _MUTEX_H_

#include <pthread.h>
#include <string>
#include "system/RCException.h"

namespace bh
{

	/**
 		\brief Simple mutex wrapper class.

		A simple non-recursive mutex class that wraps a pthread mutex.
	 */
	class Mutex
	{
		public:
			/**
 				\brief A mutex specific exception.

				This object is raised for all errors within the Mutex class.
			 */
			class Exception : public RCException
			{
				public:
					/**
 						\brief Creates an exception object to be raised by the
						mutex.

						\param action is a brief (one to three word)
						description of the mutex operation that triggered
						the exception.
						\param code is the value of the code returned by the
						underlying failed pthread call.
					 */
					Exception(std::string action, int code) 
						:	RCException("Mutex " + action +	" operation failed: " + strerror(code))
					{
					}
			};

			/**
 				\brief A convenience class to lock a mutex.

				So long as it is in scope, this class locks the
				delegate mutex.  This class is primarily used to
				lock a mutex automatically for the duration of
				a single function call or the lifespan of a given
				object.
			 */
			class AutoLock
			{
				public:
					/**
 						\brief Instantiates the auto lock.

						The constructor call immediately locks the delegate
						mutex and only unlocks it when this instance falls
						out of scope.

						\param m is a reference to the delegate mutex.
					 */
					AutoLock(Mutex& m):mutex(m)
					{
						mutex.lock();
					}

					/**
 						Releases the object and unlocks the delegate mutex.
					 */
					~AutoLock()
					{
						mutex.unlock();
					}

				private:
					/// A reference to the managed delegate mutex.
					Mutex& mutex;
			};

			/**
				\brief Instantiates a non-recursive mutex.
				\exception Mutex::Exception if the underlying pthread call fails.
			 */
			Mutex() throw(Mutex::Exception)
			{
				ProcessReturnCode("initialization", pthread_mutex_init(&mutex, 0));
			}

			/**
				\brief Destroys the mutex.
			 */
			~Mutex()
			{
				pthread_mutex_destroy(&mutex);
			}

			/**
				\brief Locks the mutex.
				\exception Mutex::Exception if the underlying pthread call fails.
			 */
			void lock() throw(Mutex::Exception)
			{
				ProcessReturnCode("lock", pthread_mutex_lock(&mutex));
			}

			/**
				\brief Unlocks the mutex.
				\exception Mutex::Exception if the underlying pthread call fails.
			 */
			void unlock() throw(Mutex::Exception)
			{
				ProcessReturnCode("unlock", pthread_mutex_unlock(&mutex));
			}

		protected:
			/**
				\brief Handles a pthread return code.

				This call raises an exception if the code passed in
				represents a standard pthread error condition.

				\param action is a short description of the pthread call.
				\param code is the return value from the pthread call.
				\exception Mutex::Exception If the value of code represents
				a pthread error condition.
			 */
			int ProcessReturnCode(const char *action, int code) throw(Mutex::Exception)
			{
				if( 0 != code ) {
					throw Mutex::Exception(action, code);
				}
				return code;
			}

			/// The managed pthread mutex object.
			pthread_mutex_t mutex;
	};

} // end of bh namespace

#endif // _MUTEX_H_
