/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef _SINGLETON_H_
#define _SINGLETON_H_

#include <memory>
#include "util/Mutex.h"

namespace bh
{

	/**
 		\brief A base singleton pattern object

		Describes a base object for creating singletons.  Though not
		a pure virtual base class, this object cannot be instantiated
		directly.  It must be derived and will in turn be instantiated
		via the child.  See the configuration object in the file
		Configuration.h for an example of how to use this pattern.
	 */
	template <class _Object>
	class Singleton
	{
		public:
			/**
				\brief Cleans up the singleton base object.
			 */
			virtual ~Singleton() {}

			/**
				\brief Fetches a reference to the single instance
				of the class.
			 */
			static _Object& GetInstance()
			{
			  if (0 == singleton.get())
			  {
				Mutex::AutoLock lock(singleton_mutex);

				if (0 == singleton.get()) {
					singleton.reset(new _Object);
				}
			  }

				return *singleton;
			}

		protected:
			// Ensure this singleton class can't be
			// directly instantiated ...must be
			// created as a derived object.
			Singleton() {}
			// Copy protection
			Singleton(const Singleton&) {}
			const Singleton& operator=(const Singleton&);

			/// The singleton instance's guarding mutex
			static Mutex singleton_mutex;

			/// The actual instance of the derived singleton object
			static std::auto_ptr<_Object> singleton;
	};

	template <class _Object> Mutex Singleton<_Object>::singleton_mutex;
	template <class _Object> std::auto_ptr<_Object> Singleton<_Object>::singleton;

} // end of bh namespace

#endif // _SINGLETON_H_
