/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include <assert.h>
#include "core/CompiledQuery.h"
#include "ConstColumn.h"
#include "core/MysqlExpression.h"
#include "core/RCAttr.h"


BHReturnCode ConstColumn::DoGetValueDateTime(MIIterator const&, RCDateTime& dt, AttributeType t) {
	BHReturnCode rc = BHRC_SUCCESS;

	if(t == RC_UNKNOWN)
		t= TypeName();
	if (ATI::IsTxtType(TypeName()))
		rc = RCDateTime::Parse(value.sp, dt, t);
	else if (ATI::IsDateTimeType(TypeName()))
//TODO: not sure where to parse and when to construct - was IsEncoded()
		dt = RCDateTime(value.GetDateTime64(), t);
	else
		rc = RCDateTime::Parse(value.Get64(), dt, t);
	return rc;
}

BHReturnCode ConstColumn::DoGetValueRCNum(MIIterator const& mit, RCNum& rcn) {
	BHReturnCode rc = BHRC_SUCCESS;
	assert(ATI::IsNumericType(TypeName()));
	if(!IsNull(mit))
			rcn = RCNum(value.x, Type().GetScale(), Type().GetScale() == -1, TypeName());
		else
			rcn = RCNum();
	return rc;
}

double ConstColumn::DoGetValueDouble(const MIIterator& mit) {
	assert(ATI::IsNumericType(TypeName()));
	double val = 0;
	if (value.IsNull())
		val = NULL_VALUE_D;
	if (ATI::IsIntegerType(TypeName()))
		val = (double) value.Get64();
	else if(ATI::IsFixedNumericType(TypeName()))
		val = ((double) value.Get64()) / PowOfTen(ct.GetScale());
	else if(ATI::IsRealType(TypeName())) {
		union { double d; _int64 i;} u;
		u.i = value.Get64();
		val = u.d;
	} else
		assert(0 && "conversion to double not implemented");
	return val;
}

void ConstColumn::DoGetValueBin(const MIIterator& mit, int& size, char* buf) {
	if(ATI::IsStringType((TypeName()))) {
		RCBString s;
		value.GetString(s);
		size = s.size();
		memcpy(buf, s.GetDataBytesPointer(), size);
	} else {
		_int64 val = value.Get64();
		size = 8;
		memcpy(buf, &val, size);
	}
}

RCValueObject ConstColumn::DoGetValue(const MIIterator& mit, bool lookup_to_num) {

	if (value.null)
		return RCValueObject();

	if(ATI::IsStringType((TypeName()))) {
		RCBString s;
		value.GetString(s);
		return s;
	}
	if(ATI::IsIntegerType(TypeName()))
		return RCNum(value.Get64(), -1, false, TypeName());
	if(ATI::IsDateTimeType(TypeName()))
		return RCDateTime(value.GetDateTime64() , TypeName());
	if(ATI::IsRealType(TypeName()))
		return RCNum(value.Get64(), 0, true, TypeName());
	if(lookup_to_num || TypeName() == RC_NUM)
		return RCNum((_int64)value.Get64(), Type().GetScale());
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(!"Illegal execution path");
	return RCValueObject();
}

_int64 ConstColumn::DoGetSum(const MIIterator &mit)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(!ATI::IsStringType(TypeName()));
	if(value.IsNull())
		return NULL_VALUE_64;		// note that this is a bit ambiguous: the same is for sum of nulls and for "not implemented"
	if( ATI::IsRealType(TypeName()) ) {
		double res = value.GetDouble() * mit.GetPackSizeLeft();
		return *(_int64 *)&res;
	}
	return (value.Get64() * mit.GetPackSizeLeft());
}

RCBString ConstColumn::DoGetMinString(const MIIterator &mit) {
	RCBString s;
	value.GetString(s);
	return s;
}

RCBString ConstColumn::DoGetMaxString(const MIIterator &mit) {
	RCBString s;
	value.GetString(s);
	return s;
}

_int64 ConstColumn::DoGetApproxDistVals(bool incl_nulls)
{
	return 1;
}

_int64 ConstColumn::GetExactDistVals()
{
	return (value.IsNull() ? 0 : 1);
}


ushort ConstColumn::DoMaxStringSize()		// maximal byte string length in column
{
	return ct.GetPrecision();
}

PackOntologicalStatus ConstColumn::DoGetPackOntologicalStatus(const MIIterator &mit)
{
	if (value.IsNull())
		return NULLS_ONLY;
	return UNIFORM;
}

void ConstColumn::DoEvaluatePack(MIUpdatingIterator& mit, Descriptor& desc)
{
	assert(0); 	// comparison of a const with a const should be simplified earlier
}

RSValue ConstColumn::DoRoughCheck(const MIIterator& mit, Descriptor& d)
{
	BHASSERT(false, "RoughCheck on VirtualColumn representing constant. Descriptor should be evaluated earlier.");
	return RS_SOME; //not implemented
}


#ifdef __WIN__
#define snprintf _snprintf
#endif

char *ConstColumn::ToString(char p_buf[], size_t buf_ct) const
{
	if(value.IsNull() || value.Get64() == NULL_VALUE_64)
		snprintf(p_buf, buf_ct, "<null>");
	else if(value.Get64() == PLUS_INF_64)
		snprintf(p_buf, buf_ct, "+inf");
	else if(value.Get64() == MINUS_INF_64)
		snprintf(p_buf, buf_ct, "-inf");
	else if(ct.IsInt())
		snprintf(p_buf, buf_ct, "%lld", value.Get64());
	else if(ct.IsFixed())
		snprintf(p_buf, buf_ct, "%g", value.Get64() / PowOfTen(ct.GetScale()));
	else if(ct.IsFloat())
		snprintf(p_buf, buf_ct, "%g", value.GetDouble());
	else if(ct.IsString()) {
		RCBString val;
		value.GetString(val);
		snprintf(p_buf, buf_ct - 2, "\"%.*s", (val.len < buf_ct - 4 ? val.len : buf_ct - 4), val.GetDataBytesPointer());
		strcat(p_buf, "\"");
	}
	return p_buf;
}
